/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011  Hohlraum (h0h1r4um@yahoo.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

int
main (int argc, char *argv[])
{
  int i, j, k;
  unsigned char *ip4_flags;
  gchar *ipaddress, *value;
  GtkTextBuffer *textbuffer;

  SPSData *data;
  GtkBuilder *builder;
  GError *error = NULL;
  unsigned int iseed;
  void *tmp;

  // Secure glib
  if (! g_thread_supported()) {
    g_thread_init (NULL);
  }

  gtk_init (&argc, &argv);

  // Create new GtkBuilder object
  builder = gtk_builder_new();
  if (!gtk_builder_add_from_file (builder, "sps.ui", &error)) {
    g_warning ("%s", error->message );
    g_free (error);
    exit (EXIT_FAILURE);
  }

  // Allocate data structure
  data = g_slice_new (SPSData);

  // Get objects from user interface
  data->main_window = GTK_WIDGET (gtk_builder_get_object (builder, "main_window"));
  data->menubar1    = GTK_WIDGET (gtk_builder_get_object (builder, "menubar1"));
  data->view        = GTK_WIDGET (gtk_builder_get_object (builder, "view"));

  // Send page

  // Select Packet Type
  data->radiobutton1 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton1"));  // TCP (IPv4) packet type
  data->radiobutton2 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton2"));  // ICMP (IPv4) packet type
  data->radiobutton3 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton3"));  // UDP (IPv4) packet type
  data->radiobutton4 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton4"));  // All (IPv4) packet types (i.e., cycle)
  data->radiobutton5 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton5"));  // TCP (IPv6) packet type
  data->radiobutton6 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton6"));  // ICMP (IPv6) packet type
  data->radiobutton7 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton7"));  // UDP (IPv6) packet type
  data->radiobutton8 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton8"));  // All (IPv6) packet types (i.e., cycle)
  // Number of Packets to Send
  data->entry18      = GTK_WIDGET (gtk_builder_get_object (builder, "entry18"));              // Number of packets to send
  data->checkbutton11 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton11"));  // Flood
  // Activity Log
  data->textview1    = GTK_WIDGET (gtk_builder_get_object (builder, "textview1"));              // SPS activity log
  // IPv6 over IPv4 (6to4)
  data->checkbutton12 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton12"));  // Tunnel IPv6 over IPv4
  data->entry125      = GTK_WIDGET (gtk_builder_get_object (builder, "entry125"));              // TCP source IPv4 address for 6to4 packets
  data->entry160      = GTK_WIDGET (gtk_builder_get_object (builder, "entry160"));              // ICMP source IPv4 address for 6to4 packets
  data->entry161      = GTK_WIDGET (gtk_builder_get_object (builder, "entry161"));              // UDP source IPv4 address for 6to4 packets
  data->checkbutton13 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton13"));  // TCP source IPv4 "Randomize"
  data->checkbutton26 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton26"));  // ICMP source IPv4 "Randomize"
  data->checkbutton27 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton27"));  // UDP source IPv4 "Randomize"
  data->checkbutton14 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton14"));  // Specify IPv4 ethernet header
  data->entry156     = GTK_WIDGET (gtk_builder_get_object (builder, "entry156"));             // Destination link-layer (MAC) address
  data->entry157     = GTK_WIDGET (gtk_builder_get_object (builder, "entry157"));             // Source link-layer (MAC) address
  data->entry158     = GTK_WIDGET (gtk_builder_get_object (builder, "entry158"));             // Source interface name
  data->entry159     = GTK_WIDGET (gtk_builder_get_object (builder, "entry159"));             // Ethernet type code
  // Send / Stop
  data->button1      = GTK_BUTTON (gtk_builder_get_object (builder, "button1"));              // Send
  data->button5      = GTK_BUTTON (gtk_builder_get_object (builder, "button5"));              // Stop Flood

  // TCP (IPv4) page

  // Ethernet header
  data->checkbutton18 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton18"));  // Specify ethernet header
  data->entry144 = GTK_WIDGET (gtk_builder_get_object (builder, "entry144"));   // Destination link-layer (MAC) address
  data->entry145 = GTK_WIDGET (gtk_builder_get_object (builder, "entry145"));   // Source link-layer (MAC) address
  data->entry150 = GTK_WIDGET (gtk_builder_get_object (builder, "entry150"));   // Source interface name
  data->entry146 = GTK_WIDGET (gtk_builder_get_object (builder, "entry146"));   // Ethernet type code
  // IPv4 header
  data->entry29   = GTK_WIDGET (gtk_builder_get_object (builder, "entry29"));   // IP header length
  data->entry30   = GTK_WIDGET (gtk_builder_get_object (builder, "entry30"));   // Protocol version
  data->entry31   = GTK_WIDGET (gtk_builder_get_object (builder, "entry31"));   // Type of service
  data->entry32   = GTK_WIDGET (gtk_builder_get_object (builder, "entry32"));   // Total length of datagram
  data->entry33   = GTK_WIDGET (gtk_builder_get_object (builder, "entry33"));   // ID sequence number
  data->entry122  = GTK_WIDGET (gtk_builder_get_object (builder, "entry122"));  // Zero
  data->entry37   = GTK_WIDGET (gtk_builder_get_object (builder, "entry37"));   // Do not fragment flag
  data->entry38   = GTK_WIDGET (gtk_builder_get_object (builder, "entry38"));   // More fragments following flag
  data->entry39   = GTK_WIDGET (gtk_builder_get_object (builder, "entry39"));   // Fragmentation offset
  data->entry40   = GTK_WIDGET (gtk_builder_get_object (builder, "entry40"));   // Time-to-live
  data->entry74   = GTK_WIDGET (gtk_builder_get_object (builder, "entry74"));   // Transport layer protocol
  data->entry35   = GTK_WIDGET (gtk_builder_get_object (builder, "entry35"));   // Source IP address
  data->checkbutton5 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton5"));  // Source IP "Randomize" checkbutton
  data->entry36   = GTK_WIDGET (gtk_builder_get_object (builder, "entry36"));   // Destination IP address
  data->entry34   = GTK_WIDGET (gtk_builder_get_object (builder, "entry34"));   // IP checksum
  // TCP header (IPv4)
  data->entry41   = GTK_WIDGET (gtk_builder_get_object (builder, "entry41"));   // Source port number
  data->checkbutton7 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton7"));  // Source port "Randomize" checkbutton
  data->entry42   = GTK_WIDGET (gtk_builder_get_object (builder, "entry42"));   // Destination port number
  data->entry43   = GTK_WIDGET (gtk_builder_get_object (builder, "entry43"));   // Sequence number
  data->entry44   = GTK_WIDGET (gtk_builder_get_object (builder, "entry44"));   // Ack number
  data->entry45   = GTK_WIDGET (gtk_builder_get_object (builder, "entry45"));   // Reserved
  data->entry46   = GTK_WIDGET (gtk_builder_get_object (builder, "entry46"));   // Data offset
  data->entry47   = GTK_WIDGET (gtk_builder_get_object (builder, "entry47"));   // FIN flag
  data->entry48   = GTK_WIDGET (gtk_builder_get_object (builder, "entry48"));   // SYN flag
  data->entry49   = GTK_WIDGET (gtk_builder_get_object (builder, "entry49"));   // RST flag
  data->entry50   = GTK_WIDGET (gtk_builder_get_object (builder, "entry50"));   // PSH flag
  data->entry51   = GTK_WIDGET (gtk_builder_get_object (builder, "entry51"));   // ACK flag
  data->entry52   = GTK_WIDGET (gtk_builder_get_object (builder, "entry52"));   // URG flag
  data->entry53   = GTK_WIDGET (gtk_builder_get_object (builder, "entry53"));   // ECE flag
  data->entry54   = GTK_WIDGET (gtk_builder_get_object (builder, "entry54"));   // CWR flag
  data->entry55   = GTK_WIDGET (gtk_builder_get_object (builder, "entry55"));   // Window size
  data->entry56   = GTK_WIDGET (gtk_builder_get_object (builder, "entry56"));   // TCP Checksum
  data->entry77   = GTK_WIDGET (gtk_builder_get_object (builder, "entry77"));   // Urgent pointer
  data->radiobutton17 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton17"));  // ASCII data entry
  data->radiobutton18 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton18"));  // Hexadecimal data entry
  data->entry162  = GTK_WIDGET (gtk_builder_get_object (builder, "entry162"));  // TCP data from keyboard
  data->button20  = GTK_BUTTON (gtk_builder_get_object (builder, "button20"));  // TCP data from file

  data->button3   = GTK_BUTTON (gtk_builder_get_object (builder, "button3"));   // View packet
  data->button11  = GTK_BUTTON (gtk_builder_get_object (builder, "button11"));  // Restore default values

  // TCP (IPv6) page

  // Ethernet header
  data->entry131 = GTK_WIDGET (gtk_builder_get_object (builder, "entry131"));  // Destination link-layer (MAC) address
  data->entry132 = GTK_WIDGET (gtk_builder_get_object (builder, "entry132"));  // Source link-layer (MAC) address
  data->entry154 = GTK_WIDGET (gtk_builder_get_object (builder, "entry154"));  // Source interface name
  data->entry133 = GTK_WIDGET (gtk_builder_get_object (builder, "entry133"));  // Ethernet type code
  // IPv6 header
  data->entry22  = GTK_WIDGET (gtk_builder_get_object (builder, "entry22"));   // IP version
  data->entry23  = GTK_WIDGET (gtk_builder_get_object (builder, "entry23"));   // Traffic class
  data->entry24  = GTK_WIDGET (gtk_builder_get_object (builder, "entry24"));   // Flow label
  data->entry25 = GTK_WIDGET (gtk_builder_get_object (builder, "entry25"));  // Payload length
  data->entry26 = GTK_WIDGET (gtk_builder_get_object (builder, "entry26"));  // Next header
  data->entry27 = GTK_WIDGET (gtk_builder_get_object (builder, "entry27"));  // Hop limit
  data->entry28 = GTK_WIDGET (gtk_builder_get_object (builder, "entry28"));    // Source IP address
  data->checkbutton2 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton2"));  // Source IP "Randomize" checkbutton
  data->entry78 = GTK_WIDGET (gtk_builder_get_object (builder, "entry78"));  // Destination IP address
  // TCP header (IPv6)
  data->entry79   = GTK_WIDGET (gtk_builder_get_object (builder, "entry79"));   // Source port number
  data->checkbutton4 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton4"));  // Source port "Randomize" checkbutton
  data->entry80   = GTK_WIDGET (gtk_builder_get_object (builder, "entry80"));   // Destination port number
  data->entry81   = GTK_WIDGET (gtk_builder_get_object (builder, "entry81"));   // Sequence number
  data->entry82   = GTK_WIDGET (gtk_builder_get_object (builder, "entry82"));   // Ack number
  data->entry83   = GTK_WIDGET (gtk_builder_get_object (builder, "entry83"));   // Reserved
  data->entry84   = GTK_WIDGET (gtk_builder_get_object (builder, "entry84"));   // Data offset
  data->entry85   = GTK_WIDGET (gtk_builder_get_object (builder, "entry85"));   // FIN flag
  data->entry86   = GTK_WIDGET (gtk_builder_get_object (builder, "entry86"));   // SYN flag
  data->entry87   = GTK_WIDGET (gtk_builder_get_object (builder, "entry87"));   // RST flag
  data->entry88   = GTK_WIDGET (gtk_builder_get_object (builder, "entry88"));   // PSH flag
  data->entry89   = GTK_WIDGET (gtk_builder_get_object (builder, "entry89"));   // ACK flag
  data->entry90   = GTK_WIDGET (gtk_builder_get_object (builder, "entry90"));   // URG flag
  data->entry91   = GTK_WIDGET (gtk_builder_get_object (builder, "entry91"));   // ECE flag
  data->entry92   = GTK_WIDGET (gtk_builder_get_object (builder, "entry92"));   // CWR flag
  data->entry93   = GTK_WIDGET (gtk_builder_get_object (builder, "entry93"));   // Window size
  data->entry94   = GTK_WIDGET (gtk_builder_get_object (builder, "entry94"));   // TCP Checksum
  data->entry95   = GTK_WIDGET (gtk_builder_get_object (builder, "entry95"));   // Urgent pointer
  data->radiobutton19 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton19"));  // ASCII data entry
  data->radiobutton20 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton20"));  // Hexadecimal data entry
  data->entry163  = GTK_WIDGET (gtk_builder_get_object (builder, "entry163"));  // TCP data from keyboard
  data->button22  = GTK_BUTTON (gtk_builder_get_object (builder, "button22"));  // TCP data from file

  data->button6   = GTK_BUTTON (gtk_builder_get_object (builder, "button6"));   // View packet
  data->button16  = GTK_BUTTON (gtk_builder_get_object (builder, "button16"));  // Restore default values

  // ICMP (IPv4) page

  // Ethernet header
  data->checkbutton17 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton17"));  // Specify ethernet header
  data->entry141 = GTK_WIDGET (gtk_builder_get_object (builder, "entry141"));   // Destination link-layer (MAC) address
  data->entry142 = GTK_WIDGET (gtk_builder_get_object (builder, "entry142"));   // Source link-layer (MAC) address
  data->entry151 = GTK_WIDGET (gtk_builder_get_object (builder, "entry151"));   // Source interface name
  data->entry143 = GTK_WIDGET (gtk_builder_get_object (builder, "entry143"));   // Ethernet type code
  // IPv4 header
  data->entry57   = GTK_WIDGET (gtk_builder_get_object (builder, "entry57"));   // IP header length
  data->entry58   = GTK_WIDGET (gtk_builder_get_object (builder, "entry58"));   // Protocol version
  data->entry59   = GTK_WIDGET (gtk_builder_get_object (builder, "entry59"));   // Type of service
  data->entry60   = GTK_WIDGET (gtk_builder_get_object (builder, "entry60"));   // Total length of datagram
  data->entry61   = GTK_WIDGET (gtk_builder_get_object (builder, "entry61"));   // ID sequence number
  data->entry123  = GTK_WIDGET (gtk_builder_get_object (builder, "entry123"));  // Zero
  data->entry65   = GTK_WIDGET (gtk_builder_get_object (builder, "entry65"));   // Do not fragment flag
  data->entry66   = GTK_WIDGET (gtk_builder_get_object (builder, "entry66"));   // More fragments following flag
  data->entry67   = GTK_WIDGET (gtk_builder_get_object (builder, "entry67"));   // Fragmentation offset
  data->entry68   = GTK_WIDGET (gtk_builder_get_object (builder, "entry68"));   // Time-to-live
  data->entry75   = GTK_WIDGET (gtk_builder_get_object (builder, "entry75"));   // Transport layer protocol
  data->entry63   = GTK_WIDGET (gtk_builder_get_object (builder, "entry63"));   // Source IP address
  data->checkbutton9 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton9"));  // Source IP "Randomize" checkbutton
  data->entry64   = GTK_WIDGET (gtk_builder_get_object (builder, "entry64"));   // Destination IP address
  data->entry62   = GTK_WIDGET (gtk_builder_get_object (builder, "entry62"));   // IP checksum
  // ICMP header (IPv4)
  data->entry69   = GTK_WIDGET (gtk_builder_get_object (builder, "entry69"));   // Message type
  data->entry70   = GTK_WIDGET (gtk_builder_get_object (builder, "entry70"));   // Message code
  data->entry71   = GTK_WIDGET (gtk_builder_get_object (builder, "entry71"));   // ICMP checksum
  data->entry72   = GTK_WIDGET (gtk_builder_get_object (builder, "entry72"));   // Identifier
  data->entry73   = GTK_WIDGET (gtk_builder_get_object (builder, "entry73"));   // Sequence number
  data->radiobutton9 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton9"));  // ASCII data entry
  data->radiobutton10 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton10"));  // Hexadecimal data entry
  data->entry126  = GTK_WIDGET (gtk_builder_get_object (builder, "entry126"));    // ICMP data from keyboard
  data->button7      = GTK_BUTTON (gtk_builder_get_object (builder, "button7"));  // ICMP data from file

  data->button2   = GTK_BUTTON (gtk_builder_get_object (builder, "button2"));   // View packet
  data->button12  = GTK_BUTTON (gtk_builder_get_object (builder, "button12"));  // Restore default values

  // ICMP (IPv6) page

  // Ethernet header
  data->entry128 = GTK_WIDGET (gtk_builder_get_object (builder, "entry128"));  // Destination link-layer (MAC) address
  data->entry129 = GTK_WIDGET (gtk_builder_get_object (builder, "entry129"));  // Source link-layer (MAC) address
  data->entry153 = GTK_WIDGET (gtk_builder_get_object (builder, "entry153"));  // Source interface name
  data->entry130 = GTK_WIDGET (gtk_builder_get_object (builder, "entry130"));  // Ethernet type code
  // IPv6 header
  data->entry110  = GTK_WIDGET (gtk_builder_get_object (builder, "entry110"));   // IP version
  data->entry111  = GTK_WIDGET (gtk_builder_get_object (builder, "entry111"));   // Traffic class
  data->entry112  = GTK_WIDGET (gtk_builder_get_object (builder, "entry112"));   // Flow label
  data->entry113 = GTK_WIDGET (gtk_builder_get_object (builder, "entry113"));  // Payload length
  data->entry114 = GTK_WIDGET (gtk_builder_get_object (builder, "entry114"));  // Next header
  data->entry115 = GTK_WIDGET (gtk_builder_get_object (builder, "entry115"));  // Hop limit
  data->entry109 = GTK_WIDGET (gtk_builder_get_object (builder, "entry109"));    // Source IP address
  data->checkbutton10 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton10"));  // Source IP "Randomize" checkbutton
  data->entry116 = GTK_WIDGET (gtk_builder_get_object (builder, "entry116"));  // Destination IP address
  // ICMP header (IPv6)
  data->entry117   = GTK_WIDGET (gtk_builder_get_object (builder, "entry117"));   // Message type
  data->entry118   = GTK_WIDGET (gtk_builder_get_object (builder, "entry118"));   // Message code
  data->entry119   = GTK_WIDGET (gtk_builder_get_object (builder, "entry119"));   // ICMP checksum
  data->entry120   = GTK_WIDGET (gtk_builder_get_object (builder, "entry120"));   // Identifier
  data->entry121   = GTK_WIDGET (gtk_builder_get_object (builder, "entry121"));   // Sequence number
  data->radiobutton11 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton11"));  // ASCII data entry
  data->radiobutton12 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton12"));  // Hexadecimal data entry
  data->entry127   = GTK_WIDGET (gtk_builder_get_object (builder, "entry127"));   // ICMP data from keyboard
  data->button9      = GTK_BUTTON (gtk_builder_get_object (builder, "button9"));  // ICMP data from file

  data->button17  = GTK_BUTTON (gtk_builder_get_object (builder, "button30"));   // View packet
  data->button18  = GTK_BUTTON (gtk_builder_get_object (builder, "button18"));  // Restore default values

  // UDP (IPv4) page

  // Ethernet header
  data->checkbutton19 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton19"));  // Specify ethernet header
  data->entry147 = GTK_WIDGET (gtk_builder_get_object (builder, "entry147"));   // Destination link-layer (MAC) address
  data->entry148 = GTK_WIDGET (gtk_builder_get_object (builder, "entry148"));   // Source link-layer (MAC) address
  data->entry152 = GTK_WIDGET (gtk_builder_get_object (builder, "entry152"));   // Source interface name
  data->entry149 = GTK_WIDGET (gtk_builder_get_object (builder, "entry149"));   // Ethernet type code
  // IPv4 header
  data->entry1   = GTK_WIDGET (gtk_builder_get_object (builder, "entry1"));   // IP header length
  data->entry2   = GTK_WIDGET (gtk_builder_get_object (builder, "entry2"));   // Protocol version
  data->entry3   = GTK_WIDGET (gtk_builder_get_object (builder, "entry3"));   // Type of service
  data->entry4   = GTK_WIDGET (gtk_builder_get_object (builder, "entry4"));   // Total length of datagram
  data->entry5   = GTK_WIDGET (gtk_builder_get_object (builder, "entry5"));   // ID sequence number
  data->entry124  = GTK_WIDGET (gtk_builder_get_object (builder, "entry124"));  // Zero
  data->entry9   = GTK_WIDGET (gtk_builder_get_object (builder, "entry9"));   // Do not fragment flag
  data->entry10  = GTK_WIDGET (gtk_builder_get_object (builder, "entry10"));  // More fragments following flag
  data->entry11  = GTK_WIDGET (gtk_builder_get_object (builder, "entry11"));  // Fragmentation offset
  data->entry12  = GTK_WIDGET (gtk_builder_get_object (builder, "entry12"));  // Time-to-live
  data->entry76  = GTK_WIDGET (gtk_builder_get_object (builder, "entry76"));  // Transport layer protocol
  data->entry7   = GTK_WIDGET (gtk_builder_get_object (builder, "entry7"));   // Source IP address
  data->checkbutton1 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton1"));  // Source IP "Randomize" checkbutton
  data->entry8   = GTK_WIDGET (gtk_builder_get_object (builder, "entry8"));   // Destination IP address
  data->entry6   = GTK_WIDGET (gtk_builder_get_object (builder, "entry6"));   // IP checksum
  // UDP header (IPv4)
  data->entry13   = GTK_WIDGET (gtk_builder_get_object (builder, "entry13"));   // Source port
  data->checkbutton3 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton3"));  // Source IP "Randomize" checkbutton
  data->entry14   = GTK_WIDGET (gtk_builder_get_object (builder, "entry14"));   // Destination port
  data->entry15   = GTK_WIDGET (gtk_builder_get_object (builder, "entry15"));   // Length of UDP datagram
  data->entry16   = GTK_WIDGET (gtk_builder_get_object (builder, "entry16"));   // UDP checksum
  data->radiobutton13 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton13"));  // ASCII data entry
  data->radiobutton14 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton14"));  // Hexadecimal data entry
  data->entry17   = GTK_WIDGET (gtk_builder_get_object (builder, "entry17"));     // UDP data from keyboard
  data->button8      = GTK_BUTTON (gtk_builder_get_object (builder, "button8"));  // UDP data from file

  data->button4   = GTK_BUTTON (gtk_builder_get_object (builder, "button4"));   // View packet
  data->button15  = GTK_BUTTON (gtk_builder_get_object (builder, "button15"));  // Restore default values

  // UDP (IPv6) page

  // Ethernet header
  data->entry134 = GTK_WIDGET (gtk_builder_get_object (builder, "entry134"));  // Destination link-layer (MAC) address
  data->entry135 = GTK_WIDGET (gtk_builder_get_object (builder, "entry135"));  // Source link-layer (MAC) address
  data->entry155 = GTK_WIDGET (gtk_builder_get_object (builder, "entry155"));  // Source interface name
  data->entry136 = GTK_WIDGET (gtk_builder_get_object (builder, "entry136"));  // Ethernet type code
  // IPv6 header
  data->entry97  = GTK_WIDGET (gtk_builder_get_object (builder, "entry97"));   // IP version
  data->entry98  = GTK_WIDGET (gtk_builder_get_object (builder, "entry98"));   // Traffic class
  data->entry99  = GTK_WIDGET (gtk_builder_get_object (builder, "entry99"));   // Flow label
  data->entry100 = GTK_WIDGET (gtk_builder_get_object (builder, "entry100"));  // Payload length
  data->entry101 = GTK_WIDGET (gtk_builder_get_object (builder, "entry101"));  // Next header
  data->entry102 = GTK_WIDGET (gtk_builder_get_object (builder, "entry102"));  // Hop limit
  data->entry96 = GTK_WIDGET (gtk_builder_get_object (builder, "entry96"));    // Source IP address
  data->checkbutton6 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton6"));  // Source IP "Randomize" checkbutton
  data->entry103 = GTK_WIDGET (gtk_builder_get_object (builder, "entry103"));  // Destination IP address
  // UDP header (IPv6)
  data->entry104 = GTK_WIDGET (gtk_builder_get_object (builder, "entry104"));  // Source port
  data->checkbutton8 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton8"));  // Source IP "Randomize" checkbutton
  data->entry105 = GTK_WIDGET (gtk_builder_get_object (builder, "entry105"));  // Destination port
  data->entry106 = GTK_WIDGET (gtk_builder_get_object (builder, "entry106"));  // Length of UDP datagram
  data->entry107 = GTK_WIDGET (gtk_builder_get_object (builder, "entry107"));  // UDP checksum
  data->radiobutton15 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton15"));  // ASCII data entry
  data->radiobutton16 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton16"));  // Hexadecimal data entry
  data->entry108 = GTK_WIDGET (gtk_builder_get_object (builder, "entry108"));     // UDP data from keyboard
  data->button10    = GTK_BUTTON (gtk_builder_get_object (builder, "button10"));  // UDP data from file

  data->button13 = GTK_BUTTON (gtk_builder_get_object (builder, "button13"));  // View packet
  data->button19  = GTK_BUTTON (gtk_builder_get_object (builder, "button19"));  // Restore default values

  // Host/IP page
  data->entry19   = GTK_WIDGET (gtk_builder_get_object (builder, "entry19"));    // Host-to-IP: hostname
  data->textview3 = GTK_WIDGET (gtk_builder_get_object (builder, "textview3"));  // Host-to-IPv4 results
  data->textview6 = GTK_WIDGET (gtk_builder_get_object (builder, "textview6"));  // Host-to-IPv6 results

  data->entry20   = GTK_WIDGET (gtk_builder_get_object (builder, "entry20"));    // IP-to-Host IPv4 address
  data->entry21   = GTK_WIDGET (gtk_builder_get_object (builder, "entry21"));    // IP-to-Host IPv6 address
  data->textview2 = GTK_WIDGET (gtk_builder_get_object (builder, "textview2"));  // IP-to-Host hostname result

  data->textview4 = GTK_WIDGET (gtk_builder_get_object (builder, "textview4"));  // This host hostname
  data->textview5 = GTK_WIDGET (gtk_builder_get_object (builder, "textview5"));  // This host IP address

  data->entry139   = GTK_WIDGET (gtk_builder_get_object (builder, "entry139"));    // This host interface name
  data->textview10 = GTK_WIDGET (gtk_builder_get_object (builder, "textview10"));  // This host interface MAC address

  data->entry140   = GTK_WIDGET (gtk_builder_get_object (builder, "entry140"));    // IP-to-MAC source interface name
  data->spinbutton1 = GTK_WIDGET (gtk_builder_get_object (builder, "spinbutton1"));  // Timeout for ARP/Neighbor discovery
  data->entry137   = GTK_WIDGET (gtk_builder_get_object (builder, "entry137"));    // IP-to-MAC IPv4 address
  data->entry138   = GTK_WIDGET (gtk_builder_get_object (builder, "entry138"));    // IP-to-MAC IPv6 address
  data->textview11 = GTK_WIDGET (gtk_builder_get_object (builder, "textview11"));  // IP-to-MAC results

  // Load/Save page
  // Load
  data->button21 = GTK_BUTTON (gtk_builder_get_object (builder, "button13"));                   // Load file ...
  // Save
  // Send Packet settings
  data->checkbutton25 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton25"));  // Send Packet settings
  // IPv6 over IPv4 (6to4)
  data->checkbutton15 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton15"));  // IPv4 Ethernet header and IPv4 header
  // IPv4
  data->checkbutton16 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton16"));  // TCP packet
  data->checkbutton20 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton20"));  // ICMP packet
  data->checkbutton21 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton21"));  // UDP packet
  // IPv6
  data->checkbutton22 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton22"));  // TCP packet
  data->checkbutton23 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton23"));  // ICMP packet
  data->checkbutton24 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton24"));  // UDP packet
  // Save file
  data->button14 = GTK_BUTTON (gtk_builder_get_object (builder, "button14"));                   // Save file ...

  // Traceroute page (sps.ui)

  // Probe Protocol
  data->radiobutton21 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton21"));  // TCP (IPv4) probe packet
  data->radiobutton22 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton22"));  // ICMP (IPv4) probe packet
  data->radiobutton23 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton23"));  // UDP (IPv4) probe packet
  data->radiobutton26 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton26"));  // TCP (IPv6) probe packet
  data->radiobutton27 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton27"));  // ICMP (IPv6) probe packet
  data->radiobutton28 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton28"));  // UDP (IPv6) probe packet
  data->checkbutton29 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton29"));  // Tunnel IPv6 over IPv4
  // Number of Probes per Hop
  data->spinbutton2 = GTK_WIDGET (gtk_builder_get_object (builder, "spinbutton2"));  // Number of probes per hop
  // Timeout for Reply
  data->spinbutton4 = GTK_WIDGET (gtk_builder_get_object (builder, "spinbutton4"));  // Timeout for reply
  // Maximum Number of Hops
  data->spinbutton3 = GTK_WIDGET (gtk_builder_get_object (builder, "spinbutton3"));  // Timeout for reply
  // Resolve Node Hostnames
  data->checkbutton28 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton28"));  // Yes
  // Edit, Start, Stop buttons
  data->button23 = GTK_BUTTON (gtk_builder_get_object (builder, "button23"));  // Edit packets
  data->button24 = GTK_BUTTON (gtk_builder_get_object (builder, "button24"));  // Start
  data->button25 = GTK_BUTTON (gtk_builder_get_object (builder, "button25"));  // Stop
  // Activity Log
  data->textview7 = GTK_WIDGET (gtk_builder_get_object (builder, "textview7"));  // Traceroute activity log

  // IP / ASN delegation page

  data->checkbutton38 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton38"));  // Flag to show sent and received messages on stdout when files downloading
  data->textview8 = GTK_WIDGET (gtk_builder_get_object (builder, "textview8"));  // Status of ARIN delegation file.
  data->button50 = GTK_BUTTON (gtk_builder_get_object (builder, "button50"));  // Download ARIN delegation file.
  data->textview9 = GTK_WIDGET (gtk_builder_get_object (builder, "textview9"));  // Status of RIPE delegation file.
  data->button51 = GTK_BUTTON (gtk_builder_get_object (builder, "button51"));  // Download RIPE delegation file.
  data->textview12 = GTK_WIDGET (gtk_builder_get_object (builder, "textview12"));  // Status of AFRINIC delegation file.
  data->button52 = GTK_BUTTON (gtk_builder_get_object (builder, "button52"));  // Download AFRINIC delegation file.
  data->textview13 = GTK_WIDGET (gtk_builder_get_object (builder, "textview13"));  // Status of APNIC delegation file.
  data->button53 = GTK_BUTTON (gtk_builder_get_object (builder, "button53"));  // Download APNIC delegation file.
  data->textview14 = GTK_WIDGET (gtk_builder_get_object (builder, "textview14"));  // Status of LACNIC delegation file.
  data->button54 = GTK_BUTTON (gtk_builder_get_object (builder, "button54"));  // Download LACNIC delegation file.
  data->textview15 = GTK_WIDGET (gtk_builder_get_object (builder, "textview15"));  // Status of ISO 3166-1 country code file.
  data->button55 = GTK_BUTTON (gtk_builder_get_object (builder, "button55"));  // Download www.iso.org's ISO 3166-1 country code file.
  data->button56 = GTK_BUTTON (gtk_builder_get_object (builder, "button56"));  // View www.iso.org's ISO 3166-1 country code file.
  data->button57 = GTK_BUTTON (gtk_builder_get_object (builder, "button57"));  // Download all RIR files and www.iso.org's country code list.

  data->entry500 = GTK_WIDGET (gtk_builder_get_object (builder, "entry500"));  // Country name for country name/code search.
  data->entry501 = GTK_WIDGET (gtk_builder_get_object (builder, "entry501"));  // Country code for country name/code search.
  data->entry502 = GTK_WIDGET (gtk_builder_get_object (builder, "entry502"));  // Country name for ASN search.
  data->entry503 = GTK_WIDGET (gtk_builder_get_object (builder, "entry503"));  // Country code for ASN search.
  data->textview16 = GTK_WIDGET (gtk_builder_get_object (builder, "textview16"));  // Results of search for ASN delegations by country.
  data->entry504 = GTK_WIDGET (gtk_builder_get_object (builder, "entry504"));  // ASN for reverse-search for delegation of an ASN.
  data->textview17 = GTK_WIDGET (gtk_builder_get_object (builder, "textview17"));  // Results of reverse-search for delegation of an ASN.
  data->entry505 = GTK_WIDGET (gtk_builder_get_object (builder, "entry505"));  // Country name for IP delegation search.
  data->entry506 = GTK_WIDGET (gtk_builder_get_object (builder, "entry506"));  // Country code for IP delegation search.
  data->textview18 = GTK_WIDGET (gtk_builder_get_object (builder, "textview18"));  // IPv4 address results for IP delegation search.
  data->textview19 = GTK_WIDGET (gtk_builder_get_object (builder, "textview19"));  // IPv6 address results for IP delegation search.
  data->entry507 = GTK_WIDGET (gtk_builder_get_object (builder, "entry507"));  // Country name for IP delegation reverse-search.
  data->entry508 = GTK_WIDGET (gtk_builder_get_object (builder, "entry508"));  // Country code for IP delegation reverse-search.
  data->textview20 = GTK_WIDGET (gtk_builder_get_object (builder, "textview20"));  // IPv6 address results for IP delegation reverse-search.

  // Initialize random number generator
  iseed = (unsigned int) time (NULL);
  srand (iseed);

  // Start the message listener. Checks every 100 ms for error
  // messages which are to be displayed in a pop-up dialog.
  g_timeout_add (100, (GSourceFunc) thread_listen, data);

  // String for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    data->error_text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->error_text' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Allocate memory for TCP payload and find length of payload.

  // Maximum IPv4 TCP payload size = 65535 - IPv4 header (20 bytes) - TCP header (20 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->tcp4_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp4_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv6 TCP payload size = 65535 - IPv6 header (40 bytes) - TCP header (20 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->tcp6_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp6_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv4 traceroute TCP payload size = 65535 - IPv4 header (20 bytes) - TCP header (20 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->tcp4_tr_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp4_tr_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv6 traceroute TCP payload size = 65535 - IPv6 header (40 bytes) - TCP header (20 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->tcp6_tr_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp6_tr_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Allocate memory for ICMP payload and find length of payload.

  // Maximum IPv4 ICMP payload size = 65535 - IPv4 header (20 bytes) - ICMP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->icmp4_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp4_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv6 ICMP payload size = 65535 - IPv6 header (40 bytes) - ICMP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->icmp6_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp6_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv4 traceroute ICMP payload size = 65535 - IPv4 header (20 bytes) - ICMP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->icmp4_tr_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp4_tr_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv6 traceroute ICMP payload size = 65535 - IPv6 header (40 bytes) - ICMP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->icmp6_tr_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp6_tr_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Allocate memory for UDP payload and find length of payload.

  // Maximum IPv4 UDP payload size = 65535 - IPv4 header (20 bytes) - UDP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->udp4_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp4_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv6 UDP payload size = 65535 - IPv6 header (40 bytes) - UDP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->udp6_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp6_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv4 traceroute UDP payload size = 65535 - IPv4 header (20 bytes) - UDP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->udp4_tr_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp4_tr_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Maximum IPv6 traceroute UDP payload size = 65535 - IPv6 header (40 bytes) - UDP header (8 bytes)
  tmp = (guchar *) g_malloc0 ((IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN) * sizeof (guchar));
  if (tmp != NULL) {
    data->udp6_tr_payload = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp6_tr_payload' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Interface names

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->tcp4_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp4_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->icmp4_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp4_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->udp4_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp4_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->tcp6_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp6_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->icmp6_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp6_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->udp6_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp6_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->sixto4_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->sixto4_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->tcp4_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp4_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->icmp4_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp4_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->udp4_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp4_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->tcp6_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tcp6_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->icmp6_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->icmp6_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->udp6_tr_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->udp6_tr_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (255 * sizeof (gchar));
  if (tmp != NULL) {
    data->sixto4_interfacename = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->sixto4_interfacename' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Array of flags indicating which RIR files we have.
  tmp = (gint *) g_malloc0 (6 * sizeof (gint));
  if (tmp != NULL) {
    data->rir_files_have = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->rir_files_have' in main().\n");
    exit (EXIT_FAILURE);
  }

  // IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in main().\n");
    exit (EXIT_FAILURE);
  }

  // String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in main().\n");
    exit (EXIT_FAILURE);
  }

  // SET DEFAULT VALUES FOR ALL PARAMETERS

  // Load/Save page
  // Default to saving all information
   data->settings_save = 1;
   data->sixto4_save = 1;
   data->ipv4_tcp_save = 1;
   data->ipv4_icmp_save = 1;
   data->ipv4_udp_save = 1;
   data->ipv6_tcp_save = 1;
   data->ipv6_icmp_save = 1;
   data->ipv6_udp_save = 1;

  // Host / IP page

  // Fill out hostname and IP address textview fields for this machine.
  this_host (data);

  // Manually set properties of timeout spin button (adjustment object method seems buggy).

  data->timeout = 2;

  gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->spinbutton1), 1.0, 20.0);

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->spinbutton1), 0);

  gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->spinbutton1), 1.0, 0.0);

  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->spinbutton1), TRUE);

  gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->spinbutton1), FALSE);

  gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->spinbutton1), TRUE);

  gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton1), 2.0);

  // Send page

  // Select Packet Type:
  //   IPv4: 1 = TCP, 2 = ICMP, 3 = UDP, 4 = All (cycle)
  //   IPv6: 5 = TCP, 6 = ICMP, 7 = UDP, 8 = All (cycle)
  data->packet_type = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton1), TRUE);

  // Number of Packets to Send
  data->npackets = 1;
  sprintf (value, "%u", data->npackets);
  gtk_entry_set_text (GTK_ENTRY (data->entry18), value);

  // Default packet settings

  tcp4_default (data);  // Default IPv4 TCP settings

  tcp6_default (data);  // Default IPv6 TCP settings

  icmp4_default (data);  // Default IPv4 ICMP settings

  icmp6_default (data);  // Default IPv6 ICMP settings

  udp4_default (data);  // Default IPv4 UDP settings

  udp6_default (data);  // Default IPv6 UDP settings

  tcp4_tr_default (data);  // Default IPv4 TCP traceroute settings

  icmp4_tr_default (data);  // Default IPv4 ICMP traceroute settings

  udp4_tr_default (data);  // Default IPv4 UDP traceroute settings

  tcp6_tr_default (data);  // Default IPv6 TCP traceroute settings

  icmp6_tr_default (data);  // Default IPv6 ICMP traceroute settings

  udp6_tr_default (data);  // Default IPv6 UDP traceroute settings

  // IPv6 over IPv4 (6to4) page

  // IPv6 over IPv4 flag
  data->sixto4_flag = 0;  // Default to pure IPv6 without tunnelling over IPv4

  // Default to user-specified 6to4 ethernet header
  data->sixto4_specify_ether = 1;

  // Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->sixto4ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->sixto4ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry156), value);

  // Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->sixto4ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->sixto4ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry157), value);

  // Ethernet type code (16 bits): default to IPv4
  // http://www.iana.org/assignments/ethernet-numbers
  data->sixto4ethhdr.type_code = htons (ETH_P_IP);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->sixto4ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry159), value);

  // IPv4 header for tunnelling IPv6 over IPv4 (6to4) (Section 3.5 of RFC 4213)

  // IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->sixto4ip4hdr_tcp.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  data->sixto4ip4hdr_icmp.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  data->sixto4ip4hdr_udp.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  data->sixto4ip4hdr_tcp.ip_v = 4u;      // Internet Protocol version (4 bits) - IPv4
  data->sixto4ip4hdr_icmp.ip_v = 4u;     // Internet Protocol version (4 bits) - IPv4
  data->sixto4ip4hdr_udp.ip_v = 4u;      // Internet Protocol version (4 bits) - IPv4
  data->sixto4ip4hdr_tcp.ip_tos = 0;     // Type of service (8 bits)
  data->sixto4ip4hdr_icmp.ip_tos = 0;    // Type of service (8 bits)
  data->sixto4ip4hdr_udp.ip_tos = 0;     // Type of service (8 bits)

  // Total length of datagram (16 bits): IPv6 payload length + IPv4 header + IPv6 header
  data->sixto4ip4hdr_tcp.ip_len = htons (TCP_HDRLEN + IP4_HDRLEN + IP6_HDRLEN);
  data->sixto4ip4hdr_icmp.ip_len = htons (ICMP_HDRLEN + data->icmp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);
  data->sixto4ip4hdr_udp.ip_len = htons (UDP_HDRLEN + data->udp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

  // ID sequence number (16 bits): unused, since single datagram
  data->sixto4ip4hdr_tcp.ip_id = htons (0);
  data->sixto4ip4hdr_icmp.ip_id = htons (0);
  data->sixto4ip4hdr_udp.ip_id = htons (0);

  // Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  ip4_flags[0] = 0;               // Zero (1 bit)
  ip4_flags[1] = 1;               // Do not fragment flag (1 bit)
  ip4_flags[2] = 0;               // More fragments following flag (1 bit)
  ip4_flags[3] = 0;               // Fragmentation offset (13 bits)
  data->sixto4ip4hdr_tcp.ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->sixto4ip4hdr_icmp.ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->sixto4ip4hdr_udp.ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);

  // Time-to-Live (8 bits): use maximum value
  data->sixto4ip4hdr_tcp.ip_ttl = 255u;
  data->sixto4ip4hdr_icmp.ip_ttl = 255u;
  data->sixto4ip4hdr_udp.ip_ttl = 255u;

  // Transport layer protocol (8 bits): 41 for IPv6 (Section 3.5 of RFC 4213)
  data->sixto4ip4hdr_tcp.ip_p = IPPROTO_IPV6;
  data->sixto4ip4hdr_icmp.ip_p = IPPROTO_IPV6;
  data->sixto4ip4hdr_udp.ip_p = IPPROTO_IPV6;

  // Source IPv4 address (32 bits): default to random
  data->ran_tcp6to4_sourceip = 1;
  inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_tcp.ip_src);
  inet_ntop (AF_INET, &(data->sixto4ip4hdr_tcp.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry125), value);

  data->ran_icmp6to4_sourceip = 1;
  inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_icmp.ip_src);
  inet_ntop (AF_INET, &(data->sixto4ip4hdr_icmp.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry160), value);

  data->ran_udp6to4_sourceip = 1;
  inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_udp.ip_src);
  inet_ntop (AF_INET, &(data->sixto4ip4hdr_udp.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry161), value);

  // Destination IPv4 address (32 bits): 6to4 anycast address
  inet_pton (AF_INET, "192.88.99.1", &data->sixto4ip4hdr_tcp.ip_dst);
  inet_pton (AF_INET, "192.88.99.1", &data->sixto4ip4hdr_icmp.ip_dst);
  inet_pton (AF_INET, "192.88.99.1", &data->sixto4ip4hdr_udp.ip_dst);

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->sixto4ip4hdr_tcp.ip_sum = 0;
  data->sixto4ip4hdr_icmp.ip_sum = 0;
  data->sixto4ip4hdr_udp.ip_sum = 0;
  data->sixto4ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
  data->sixto4ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
  data->sixto4ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp, IP4_HDRLEN);

  // Traceroute page (sps.ui)

  // Probe protocol type
  data->packet_type_tr = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton21), TRUE);

  // Number of Probes per Hop: Manually set properties of spin button (adjustment object method seems buggy).

  data->num_probes = 1;

  gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->spinbutton2), 1.0, 20.0);

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->spinbutton2), 0);

  gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->spinbutton2), 1.0, 0.0);

  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->spinbutton2), TRUE);

  gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->spinbutton2), FALSE);

  gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->spinbutton2), TRUE);

  gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton2), data->num_probes);

  // Timeout for Reply (seconds): Manually set properties of spin button for (adjustment object method seems buggy).

  data->timeout_tr = 2;

  gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->spinbutton4), 1.0, 20.0);

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->spinbutton4), 0);

  gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->spinbutton4), 1.0, 0.0);

  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->spinbutton4), TRUE);

  gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->spinbutton4), FALSE);

  gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->spinbutton4), TRUE);

  gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton4), 2.0);

  // Maximum Number of Hops: Manually set properties of spin button (adjustment object method seems buggy).

  data->maxhops = 30;

  gtk_spin_button_set_range (GTK_SPIN_BUTTON (data->spinbutton3), 1.0, 99.0);

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (data->spinbutton3), 0);

  gtk_spin_button_set_increments (GTK_SPIN_BUTTON (data->spinbutton3), 1.0, 0.0);

  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (data->spinbutton3), FALSE);

  gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (data->spinbutton3), FALSE);

  gtk_spin_button_set_snap_to_ticks (GTK_SPIN_BUTTON (data->spinbutton3), TRUE);

  gtk_spin_button_set_value (GTK_SPIN_BUTTON (data->spinbutton3), 30.0);

  // Default to not resolve hostnames of nodes during traceroute.
  data->resolve_tr = 0;

  // Traceroute page (traceroute.ui)

  // Traceroute IPv6 over IPv4 (6to4) page

  // IPv6 over IPv4 flag
  data->sixto4_tr_flag = 0;  // Default to pure IPv6 without tunnelling over IPv4

  // Default to user-specified 6to4 ethernet header
  data->sixto4_tr_specify_ether = 1;

  // Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->sixto4ethhdr_tr.dst_mac, 0, 6);

  // Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->sixto4ethhdr_tr.src_mac, 0, 6);

  // Ethernet type code (16 bits): default to IPv4
  // http://www.iana.org/assignments/ethernet-numbers
  data->sixto4ethhdr_tr.type_code = htons (ETH_P_IP);

  // IPv4 header for tunnelling IPv6 over IPv4 (6to4) (Section 3.5 of RFC 4213)

  // IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->sixto4ip4hdr_tcp_tr.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  data->sixto4ip4hdr_icmp_tr.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  data->sixto4ip4hdr_udp_tr.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  data->sixto4ip4hdr_tcp_tr.ip_v = 4u;      // Internet Protocol version (4 bits) - IPv4
  data->sixto4ip4hdr_icmp_tr.ip_v = 4u;     // Internet Protocol version (4 bits) - IPv4
  data->sixto4ip4hdr_udp_tr.ip_v = 4u;      // Internet Protocol version (4 bits) - IPv4
  data->sixto4ip4hdr_tcp_tr.ip_tos = 0;     // Type of service (8 bits)
  data->sixto4ip4hdr_icmp_tr.ip_tos = 0;    // Type of service (8 bits)
  data->sixto4ip4hdr_udp_tr.ip_tos = 0;     // Type of service (8 bits)

  // Total length of datagram (16 bits): IPv6 payload length + IPv4 header + IPv6 header
  data->sixto4ip4hdr_tcp_tr.ip_len = htons (TCP_HDRLEN + IP4_HDRLEN + IP6_HDRLEN);
  data->sixto4ip4hdr_icmp_tr.ip_len = htons (ICMP_HDRLEN + data->icmp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);
  data->sixto4ip4hdr_udp_tr.ip_len = htons (UDP_HDRLEN + data->udp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

  // ID sequence number (16 bits): unused, since single datagram
  data->sixto4ip4hdr_tcp_tr.ip_id = htons (0);
  data->sixto4ip4hdr_icmp_tr.ip_id = htons (0);
  data->sixto4ip4hdr_udp_tr.ip_id = htons (0);

  // Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  ip4_flags[0] = 0;               // Zero (1 bit)
  ip4_flags[1] = 1;               // Do not fragment flag (1 bit)
  ip4_flags[2] = 0;               // More fragments following flag (1 bit)
  ip4_flags[3] = 0;               // Fragmentation offset (13 bits)
  data->sixto4ip4hdr_tcp_tr.ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->sixto4ip4hdr_icmp_tr.ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);
  data->sixto4ip4hdr_udp_tr.ip_off = htons ((ip4_flags[0] << 15) + (ip4_flags[1] << 14) +
    (ip4_flags[2] << 13) + ip4_flags[3]);

  // Time-to-Live (8 bits): use maximum value
  data->sixto4ip4hdr_tcp_tr.ip_ttl = 255u;
  data->sixto4ip4hdr_icmp_tr.ip_ttl = 255u;
  data->sixto4ip4hdr_udp_tr.ip_ttl = 255u;

  // Transport layer protocol (8 bits): 41 for IPv6 (Section 3.5 of RFC 4213)
  data->sixto4ip4hdr_tcp_tr.ip_p = IPPROTO_IPV6;
  data->sixto4ip4hdr_icmp_tr.ip_p = IPPROTO_IPV6;
  data->sixto4ip4hdr_udp_tr.ip_p = IPPROTO_IPV6;

  // Source IPv4 address (32 bits): default to this host's IPv4 address
  memset (ipaddress, 0, 40);
  inet_pton (AF_INET, thishost_ipv4 (ipaddress), &data->sixto4ip4hdr_tcp_tr.ip_src);
  memset (ipaddress, 0, 40);
  inet_pton (AF_INET, thishost_ipv4 (ipaddress), &data->sixto4ip4hdr_icmp_tr.ip_src);
  memset (ipaddress, 0, 40);
  inet_pton (AF_INET, thishost_ipv4 (ipaddress), &data->sixto4ip4hdr_udp_tr.ip_src);

  // Destination IPv4 address (32 bits): 6to4 anycast address
  inet_pton (AF_INET, "192.88.99.1", &data->sixto4ip4hdr_tcp_tr.ip_dst);
  inet_pton (AF_INET, "192.88.99.1", &data->sixto4ip4hdr_icmp_tr.ip_dst);
  inet_pton (AF_INET, "192.88.99.1", &data->sixto4ip4hdr_udp_tr.ip_dst);

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->sixto4ip4hdr_tcp_tr.ip_sum = 0;
  data->sixto4ip4hdr_icmp_tr.ip_sum = 0;
  data->sixto4ip4hdr_udp_tr.ip_sum = 0;
  data->sixto4ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp_tr, IP4_HDRLEN);
  data->sixto4ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp_tr, IP4_HDRLEN);
  data->sixto4ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp_tr, IP4_HDRLEN);

  // Allocate memory for RIR names, URLs, FTP URLs, directories, and filenames.

  tmp = (gchar **) g_malloc0 (5 * sizeof (gchar *));
  if (tmp != NULL) {
    rir_name = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_name' in main().\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<5; i++) {
    tmp = (gchar *) g_malloc0 (10 * sizeof (gchar));
    if (tmp != NULL) {
      rir_name[i] = tmp;
    } else {
      fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_name' in main().\n");
      exit (EXIT_FAILURE);
    }
  }

  tmp = (gchar **) g_malloc0 (5 * sizeof (gchar *));
  if (tmp != NULL) {
    rir_url = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_url' in main().\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<5; i++) {
    tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
    if (tmp != NULL) {
      rir_url[i] = tmp;
    } else {
      fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_url' in main().\n");
      exit (EXIT_FAILURE);
    }
  }

  tmp = (gchar **) g_malloc0 (5 * sizeof (gchar *));
  if (tmp != NULL) {
    rir_www = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_www' in main().\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<5; i++) {
    tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
    if (tmp != NULL) {
      rir_www[i] = tmp;
    } else {
      fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_www' in main().\n");
      exit (EXIT_FAILURE);
    }
  }

  tmp = (gchar **) g_malloc0 (5 * sizeof (gchar *));
  if (tmp != NULL) {
    rir_dir = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_dir' in main()\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<5; i++) {
    tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
    if (tmp != NULL) {
      rir_dir[i] = tmp;
    } else {
      fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_dir' in main().\n");
      exit (EXIT_FAILURE);
    }
  }

  tmp = (gchar **) g_malloc0 (5 * sizeof (gchar *));
  if (tmp != NULL) {
    rir_file = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_file' in main().\n");
    exit (EXIT_FAILURE);
  }
  for (i=0; i<5; i++) {
    tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
    if (tmp != NULL) {
      rir_file[i] = tmp;
    } else {
      fprintf (stderr, "ERROR: Cannot allocate memory for array 'rir_file' in main().\n");
      exit (EXIT_FAILURE);
    }
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar *));
  if (tmp != NULL) {
    data->cname = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->cname' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (3 * sizeof (gchar));
  if (tmp != NULL) {
    data->ccode = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ccode' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    data->ip4string = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ip4string' in main().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    data->ip6string = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ip6string' in main().\n");
    exit (EXIT_FAILURE);
  }

  // Regional Internet Registry (RIR) names.
  strcpy (rir_name[0], "arin");
  strcpy (rir_name[1], "ripe");
  strcpy (rir_name[2], "afrinic");
  strcpy (rir_name[3], "apnic");
  strcpy (rir_name[4], "lacnic");

  // Regional Internet Registry (RIR) URLs.
  strcpy (rir_www[0], "www.arin.net");
  strcpy (rir_www[1], "www.ripe.net");
  strcpy (rir_www[2], "www.afrinic.net");
  strcpy (rir_www[3], "www.apnic.net");
  strcpy (rir_www[4], "www.lacnic.net");

  // Regional Internet Registry (RIR) FTP URLs for IP delegations.
  strcpy (rir_url[0], "ftp.arin.net");
  strcpy (rir_url[1], "ftp.ripe.net");
  strcpy (rir_url[2], "ftp.afrinic.net");
  strcpy (rir_url[3], "ftp.apnic.net");
  strcpy (rir_url[4], "ftp.lacnic.net");

  // Regional Internet Registry (RIR) IP delegation file locations.
  strcpy (rir_dir[0], "pub/stats/arin/");
  strcpy (rir_dir[1], "ripe/stats/");
  strcpy (rir_dir[2], "pub/stats/afrinic/");
  strcpy (rir_dir[3], "pub/stats/apnic/");
  strcpy (rir_dir[4], "pub/stats/lacnic/");

  // Regional Internet Registry (RIR) IP delegation files.
  strcpy (rir_file[0], "delegated-arin-latest");
  strcpy (rir_file[1], "delegated-ripencc-latest");
  strcpy (rir_file[2], "delegated-afrinic-latest");
  strcpy (rir_file[3], "delegated-apnic-latest");
  strcpy (rir_file[4], "delegated-lacnic-latest");

  // IP / ASN delegation page
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview8));
  if (file_exists (rir_file[0]) == 1) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->rir_files_have[0] = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->rir_files_have[0] = 0;
  }
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview9));
  if (file_exists (rir_file[1]) == 1) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->rir_files_have[1] = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->rir_files_have[1] = 0;
  }
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview12));
  if (file_exists (rir_file[2]) == 1) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->rir_files_have[2] = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->rir_files_have[2] = 0;
  }
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview13));
  if (file_exists (rir_file[3]) == 1) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->rir_files_have[3] = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->rir_files_have[3] = 0;
  }
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview14));
  if (file_exists (rir_file[4]) == 1) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->rir_files_have[4] = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->rir_files_have[4] = 0;
  }
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview15));
  if (file_exists (COUNTRY_FILE) == 1) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->countries_have = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->countries_have = 0;
  }

  // Flag to indicate no files are currently downloading.
  downloading = 0;

  // Start with no messages available to be reported via message_listen().
  message_available = 0;

  // Flag to stop timeout function thread_listen().
  // thread_listen() gets invoked when a file download or traceroute thread is started.
  stop_thread_listen = 0;

  // Load RIR files and country codes into memory.
  if (files_available (data) != 0) {
    read_delegations_and_countries (data);
  }

  // Default to show sent and received messages on stdout as files download.
  show_dl_messages = 1;

  // Connect signals
  gtk_builder_connect_signals (builder, data);

  // Destroy builder, since we don't need it anymore
  g_object_unref (G_OBJECT (builder));

  // Show window. All other widgets are automatically shown by GtkBuilder
  gtk_widget_show (data->main_window);

  // Hand over control to GTK+
  gtk_main ();

  // Free allocated memory.
  g_free (data->error_text);
  g_free (ip4_flags);
  g_free (value);
  g_free (ipaddress);
  g_free (data->tcp4_payload);
  g_free (data->tcp6_payload);
  g_free (data->udp4_payload);
  g_free (data->udp6_payload);
  g_free (data->icmp4_payload);
  g_free (data->icmp6_payload);
  g_free (data->sixto4_tr_interfacename);
  g_free (data->tcp4_interfacename);
  g_free (data->tcp6_interfacename);
  g_free (data->icmp4_interfacename);
  g_free (data->icmp6_interfacename);
  g_free (data->udp4_interfacename);
  g_free (data->udp6_interfacename);
  g_free (data->sixto4_interfacename);
  g_free (data->rir_files_have);
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      for (k=0; k<7; k++) {
        free (data->array[i][j][k]);
      }
    }
  }
  free (data->array);

  for (i=0; i<5; i++) {
    g_free (rir_name[i]);
    g_free (rir_www[i]);
    g_free (rir_url[i]);
    g_free (rir_dir[i]);
    g_free (rir_file[i]);
  }
  g_free (rir_name);
  g_free (rir_www);
  g_free (rir_url);
  g_free (rir_dir);
  g_free (rir_file);
  g_free (data->cname);
  g_free (data->ccode);
  g_free (data->ip4string);
  g_free (data->ip6string);

  for (i=0; i<data->ncountry; i++) {
    for (j=0; j<2; j++) {
      free (data->cc[i][j]);
    }
  }
  free (data->cc);
  g_slice_free (SPSData, data);

  // Remove thread_listen() timeout function.
  stop_thread_listen = 1;
  sleep (1);  // Give it a chance to stop.

  return (EXIT_SUCCESS);
}
