/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011  Hohlraum (h0h1r4um@yahoo.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

gint flood_mode = 0;  // Flood flag.
gint looping = 0;     // Looping flag for sending finite number of packets.

// MUTEXs for the flags.
G_LOCK_DEFINE_STATIC (downloading);  // Obtain a MUTEX for this flag which indicates a download thread is active.
G_LOCK_DEFINE_STATIC (message_available);  // Obtain a MUTEX for this flag which indicates a message from a download or traceroute thread is available to be displayed in a dialog by thread_listen().
G_LOCK_DEFINE_STATIC (show_dl_messages);  // Obtain a MUTEX for this flag which indicates messages from file download threads should be sent to stdout.
G_LOCK_DEFINE_STATIC (tracing);  // Obtain a MUTEX for this flag which indicates a traceroute thread is active.
G_LOCK_DEFINE_STATIC (ip_asn_searching);  // Obtain a MUTEX for this flag which indicates an ASN delegation search or IP address delegation search thread is active.

// Someone has entered a hostname in Host-to-IP converter
gboolean
on_entry19_activate (GtkWidget *entry19, SPSData *data)
{
  const gchar *entry19_text;
  GtkTextBuffer *textbuffer3, *textbuffer6;  // IPv4 textview, IPv6 textview
  GtkTextIter end;
  struct addrinfo hints, *res, *p;
  int status;
  char ipstr [INET6_ADDRSTRLEN];
  gchar *text;
  void *ptr;

  // Array for error messages.
  ptr = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (ptr != NULL) {
    text = ptr;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry19_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry19_text = gtk_entry_get_text (GTK_ENTRY (entry19));

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

  if ((status = getaddrinfo (entry19_text, NULL, &hints, &res)) != 0) {
    sprintf (text, "In callbacks.c: getaddrinfo() failed in on_entry19_activate().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

// Clear previous IP results from textviews
  textbuffer3 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview3));
  gtk_text_buffer_set_text (textbuffer3, "", -1);
  textbuffer6 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview6));
  gtk_text_buffer_set_text (textbuffer6, "", -1);
 
  for (p = res; p != NULL; p = p->ai_next) {

    switch (p->ai_family) {
      case AF_INET:
        ptr = &((struct sockaddr_in *) p->ai_addr)->sin_addr;
        break;
      case AF_INET6:
        ptr = &((struct sockaddr_in6 *) p->ai_addr)->sin6_addr;
        break;
    }
    inet_ntop (p->ai_family, ptr, ipstr, sizeof ipstr);
    sprintf (ipstr, "%s\n", ipstr);

    if (p->ai_family == AF_INET6) {
      textbuffer6 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview6));
      gtk_text_buffer_get_end_iter (textbuffer6, &end);
      gtk_text_buffer_insert (textbuffer6, &end, ipstr, -1);
      gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (data->textview6), &end, 0.0, FALSE, 0, 0);
    } else {
      textbuffer3 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview3));
      gtk_text_buffer_get_end_iter (textbuffer3, &end);
      gtk_text_buffer_insert (textbuffer3, &end, ipstr, -1);
      gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (data->textview3), &end, 0.0, FALSE, 0, 0);
    }

  }

  g_free (text);
  freeaddrinfo (res);

  return (TRUE);
}

// Someone has entered an IPv4 address in IP-to-Hostname converter
gboolean
on_entry20_activate (GtkWidget *entry20, SPSData *data)
{
  int status;
  const gchar *entry20_text;
  GtkTextBuffer *textbuffer2;
  struct sockaddr_in sa; 
  sa.sin_family = AF_INET;
  char hostname[NI_MAXHOST];
  gchar *text;
  void *tmp;

  memset (hostname, 0, NI_MAXHOST);

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry20_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry20_text = gtk_entry_get_text (GTK_ENTRY (entry20));
  textbuffer2 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview2));

// Is it a valid IP address?
  if (is_valid_ip4 (entry20_text) != TRUE) {
    gtk_text_buffer_set_text (textbuffer2, "", -1);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry20_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

  if ((status = inet_pton (AF_INET, entry20_text, &sa.sin_addr)) != 1) {
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry20_activate().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

  if ((status = getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
    sprintf (text, "In callbacks.c: getnameinfo() failed in on_entry20_activate().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

// Clear IPv6 text entry so there's no confusion about which IP belongs to the hostname
  gtk_entry_set_text (GTK_ENTRY (data->entry21), "");

// Show hostname
  gtk_text_buffer_set_text (textbuffer2, hostname, -1);

  g_free (text);
  return (TRUE);
}

// Someone has entered an IPv6 address in IP-to-Hostname converter
gboolean
on_entry21_activate (GtkWidget *entry21, SPSData *data)
{
  int status;
  const gchar *entry21_text;
  GtkTextBuffer *textbuffer2;
  struct sockaddr_in6 sa;
  sa.sin6_family = AF_INET6;
  char hostname[NI_MAXHOST];
  gchar *text;
  void *tmp;

  memset (hostname, 0, NI_MAXHOST);

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry21_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry21_text = gtk_entry_get_text (GTK_ENTRY (entry21));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry21_text, &sa.sin6_addr) < 1) {

// Invalid IP
    gtk_entry_set_text (GTK_ENTRY (data->entry21), "");
    textbuffer2 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview2));
    gtk_text_buffer_set_text (textbuffer2, "", -1);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry21_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);

// Valid IP
  } else {
    if ((status = getnameinfo ((struct sockaddr *)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
      sprintf (text, "In callbacks.c: getnameinfo() failed in on_entry21_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      return (FALSE);
    }

// Clear IPv4 text entry so there's no confusion about which IP belongs to the hostname
    gtk_entry_set_text (GTK_ENTRY (data->entry20), "");

// Show hostname
    textbuffer2 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview2));
    gtk_text_buffer_set_text (textbuffer2, hostname, -1);

    g_free (text);
    return (TRUE);
  }
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry139_activate (GtkWidget *entry139, SPSData *data)
{
  int i, sd, status;
  const gchar *entry139_text;
  GtkTextBuffer *textbuffer10;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry139_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry139_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry139_text = gtk_entry_get_text (GTK_ENTRY (entry139));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor for ioctl() in on_entry139_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry139_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry139_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Print MAC address in textview
  textbuffer10 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview10));
  gtk_text_buffer_set_text (textbuffer10, text, -1);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Interface name for IP to link-layer (MAC) address conversion.
gboolean
on_entry140_activate (GtkWidget *entry140, SPSData *data)
{
  int status;
  const gchar *entry140_text;
  struct sockaddr_ll device;
  gchar *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry140_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry140_text = gtk_entry_get_text (GTK_ENTRY (entry140));

// We attempt to resolve interface index only to see if interface name exists.
// The contents of entry140 will be accessed by arp() or neighbor_discovery().
  if ((device.sll_ifindex = if_nametoindex (entry140_text)) == 0) {
    status = errno;
    sprintf (text, "In callbacks.c: if_nametoindex() failed to get interface index in on_entry140_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

  return (TRUE);
}

// Someone has entered an IPv4 address into IP to link-layer (MAC) address converter
gboolean
on_entry137_activate (GtkWidget *entry137, SPSData *data)
{
  gint i;
  GtkTextBuffer *textbuffer11;
  gchar *text;
  unsigned char *dst_mac;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry137_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    dst_mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'dst_mac' in on_entry137_activate().\n");
    exit (EXIT_FAILURE);
  }

  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));

// Do ARP request
  arp (dst_mac, data);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) dst_mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) dst_mac[5]);

// Print MAC address in textview
  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));
  gtk_text_buffer_set_text (textbuffer11, text, -1);

  g_free (text);
  g_free (dst_mac);

  return (TRUE);
}

// Someone has entered an IPv6 address into IP to link-layer (MAC) address converter
gboolean
on_entry138_activate (GtkWidget *entry138, SPSData *data)
{
  gint i;
  GtkTextBuffer *textbuffer11;
  gchar *text;
  unsigned char *dst_mac;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry138_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    dst_mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'dst_mac' in on_entry138_activate().\n");
    exit (EXIT_FAILURE);
  }

  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));

// Do neighbor discovery
  neighbor_discovery (dst_mac, data);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) dst_mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) dst_mac[5]);

// Print MAC address in textview
  textbuffer11 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview11));
  gtk_text_buffer_set_text (textbuffer11, text, -1);

  g_free (text);
  g_free (dst_mac);

  return (TRUE);
}

// Someone has requested the About page
gboolean
on_about_requested ()
{
  GtkWidget *window, *textview;
  GtkTextBuffer *buffer;
  GdkPixbuf *about;
  GtkTextIter line;

// Create new window to display About information
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
  gtk_window_set_title (GTK_WINDOW (window), "About");
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);
  gtk_widget_set_size_request (window, 666, 666);

// Create a textview object to put image in
  textview = gtk_text_view_new ();
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), FALSE);

// Load image into pixbuf called "about"
  about = gdk_pixbuf_new_from_file ("./about.png", NULL);

// Load pixbuf into textview buffer
  gtk_text_buffer_get_iter_at_line (buffer, &line, 0);
  gtk_text_buffer_insert_pixbuf (buffer, &line, about);

// Add textview to our new window and display
  gtk_container_add (GTK_CONTAINER (window), textview);
  gtk_widget_show_all (window);

  return (TRUE);
}

// Someone has requested the Documentation page
gboolean
on_documentation_activate (SPSData *data)
{
  GtkWidget *window, *scrolled_win, *textview;
  gint i, n;
  gchar *doc, *text;
  GtkTextBuffer *buffer;
  void *tmp;
  FILE *fp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_documentation_activate().\n");
    exit (EXIT_FAILURE);
  }

// Create new window to display documentation
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
  gtk_window_set_title (GTK_WINDOW (window), "SPS Documentation");
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);
  gtk_window_set_default_size (GTK_WINDOW (window), 900, 600);

// Create a textview object to put image in
  textview = gtk_text_view_new ();
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), TRUE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview), GTK_WRAP_WORD);

// Open documentation file
  fp = fopen ("documentation.txt", "r");
  if (fp == NULL) {
    strcpy (text, "Can't open file documentation.txt.");
    data->parent = data->menubar1;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

// Determine size of documentation file
  i = 0;
  while ((n = fgetc (fp)) != EOF) {
    i++;
  }
  rewind (fp);

  tmp = (gchar *) g_malloc0 (i * sizeof (gchar));
  if (tmp != NULL) {
    doc = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'doc' in on_documentation_activate().\n");
    exit (EXIT_FAILURE);
  }

// Read documentation file
   i = 0;
  while ((n = fgetc (fp)) != EOF) {
    doc[i] = n;
    i++;
  }
  fclose (fp);
  
  gtk_text_buffer_set_text (buffer, doc, -1);

  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), textview);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
                                  GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

  gtk_container_add (GTK_CONTAINER (window), scrolled_win);
  gtk_widget_show_all (window);

  g_free (text);
  g_free (doc);
  return (TRUE);
}

// Display TCP packet contents (IPv4)
gboolean
on_button3_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  nbytes = ETH_HDRLEN + IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen;

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button3_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button3_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->tcp4_specify_ether == 1) {
    strcpy (title, "TCP (IPv4) Ethernet Frame Contents");
  } else {
    strcpy (title, "TCP (IPv4) Packet Contents");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
// Ethernet header
  memcpy (packet, data->tcp4ethhdr.dst_mac, 6); c += 6;
  memcpy (packet + c, data->tcp4ethhdr.src_mac, 6); c += 6;
  packet[c] = data->tcp4ethhdr.type_code % 256; c++;
  packet[c] = data->tcp4ethhdr.type_code / 256; c++;
// IPv4 header
  packet[c] = data->ip4hdr_tcp.ip_hl + (data->ip4hdr_tcp.ip_v << 4); c++;
  packet[c] = data->ip4hdr_tcp.ip_tos; c++;
  packet[c] = data->ip4hdr_tcp.ip_len & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp.ip_len >> 8; c++;
  packet[c] = data->ip4hdr_tcp.ip_id & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp.ip_id >> 8; c++;
  packet[c] = data->ip4hdr_tcp.ip_off & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp.ip_off >> 8; c++;
  packet[c] = data->ip4hdr_tcp.ip_ttl; c++;
  packet[c] = data->ip4hdr_tcp.ip_p; c++;
  packet[c] = data->ip4hdr_tcp.ip_sum & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp.ip_sum >> 8; c++;
  packet[c] = data->ip4hdr_tcp.ip_src.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp.ip_src.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp.ip_src.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_tcp.ip_src.s_addr >> 24; c++;
  packet[c] = data->ip4hdr_tcp.ip_dst.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp.ip_dst.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp.ip_dst.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_tcp.ip_dst.s_addr >> 24; c++;

// TCP header
  packet[c] = data->tcp4hdr.th_sport & 0xffu; c++;
  packet[c] = data->tcp4hdr.th_sport >> 8; c++;
  packet[c] = data->tcp4hdr.th_dport & 0xffu; c++;
  packet[c] = data->tcp4hdr.th_dport >> 8; c++;
  packet[c] = data->tcp4hdr.th_seq & 0xfful; c++;
  packet[c] = (data->tcp4hdr.th_seq >> 8) & 0xfful; c++;
  packet[c] = (data->tcp4hdr.th_seq >> 16) & 0xfful; c++;
  packet[c] = data->tcp4hdr.th_seq >> 24; c++;
  packet[c] = data->tcp4hdr.th_ack & 0xfful; c++;
  packet[c] = (data->tcp4hdr.th_ack >> 8) & 0xfful; c++;
  packet[c] = (data->tcp4hdr.th_ack >> 16) & 0xfful; c++;
  packet[c] = data->tcp4hdr.th_ack >> 24; c++;
  packet[c] = (data->tcp4hdr.th_off << 4) + data->tcp4hdr.th_x2; c++;
  packet[c] = data->tcp4hdr.th_flags; c++;
  packet[c] = data->tcp4hdr.th_win & 0xffu; c++;
  packet[c] = data->tcp4hdr.th_win >> 8; c++;
  packet[c] = data->tcp4hdr.th_sum & 0xffu; c++;
  packet[c] = data->tcp4hdr.th_sum >> 8; c++;
  packet[c] = data->tcp4hdr.th_urp & 0xffu; c++;
  packet[c] = data->tcp4hdr.th_urp >> 8; c++;
// TCP data
  for (i=0; i<data->tcp4_payloadlen; i++) {
    packet[c+i] = *(data->tcp4_payload + i);
  }

  show_packet (packet, nbytes, title, data->tcp4_specify_ether, 0);

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display ICMP packet contents (IPv4)
gboolean
on_button2_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  nbytes = ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen;

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button2_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button2_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->icmp4_specify_ether == 1) {
    strcpy (title, "ICMP (IPv4) Ethernet Frame Contents");
  } else {
    strcpy (title, "ICMP (IPv4) Packet Contents");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
// Ethernet header
  memcpy (packet, data->icmp4ethhdr.dst_mac, 6); c += 6;
  memcpy (packet + c, data->icmp4ethhdr.src_mac, 6); c += 6;
  packet[c] = data->icmp4ethhdr.type_code % 256; c++;
  packet[c] = data->icmp4ethhdr.type_code / 256; c++;
// IPv4 header
  packet[c] = data->ip4hdr_icmp.ip_hl + (data->ip4hdr_icmp.ip_v << 4); c++;
  packet[c] = data->ip4hdr_icmp.ip_tos; c++;
  packet[c] = data->ip4hdr_icmp.ip_len & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp.ip_len >> 8; c++;
  packet[c] = data->ip4hdr_icmp.ip_id & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp.ip_id >> 8; c++;
  packet[c] = data->ip4hdr_icmp.ip_off & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp.ip_off >> 8; c++;
  packet[c] = data->ip4hdr_icmp.ip_ttl; c++;
  packet[c] = data->ip4hdr_icmp.ip_p; c++;
  packet[c] = data->ip4hdr_icmp.ip_sum & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp.ip_sum >> 8; c++;
  packet[c] = data->ip4hdr_icmp.ip_src.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp.ip_src.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp.ip_src.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_icmp.ip_src.s_addr >> 24; c++;
  packet[c] = data->ip4hdr_icmp.ip_dst.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp.ip_dst.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp.ip_dst.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_icmp.ip_dst.s_addr >> 24; c++;
// ICMP header
  packet[c] = data->icmp4hdr.icmp_type; c++;
  packet[c] = data->icmp4hdr.icmp_code; c++;
  packet[c] = data->icmp4hdr.icmp_cksum & 0xffu; c++;
  packet[c] = data->icmp4hdr.icmp_cksum >> 8; c++;
  packet[c] = data->icmp4hdr.icmp_id & 0xffu; c++;
  packet[c] = data->icmp4hdr.icmp_id >> 8; c++;
  packet[c] = data->icmp4hdr.icmp_seq & 0xffu; c++;
  packet[c] = data->icmp4hdr.icmp_seq >> 8; c++;
// ICMP data
  for (i=0; i<data->icmp4_payloadlen; i++) {
    packet[c+i] = *(data->icmp4_payload + i);
  }

  show_packet (packet, nbytes, title, data->icmp4_specify_ether, 0);

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display UDP packet contents (IPv4)
gboolean
on_button4_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  nbytes = ETH_HDRLEN + IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen;

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button4_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button4_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->udp4_specify_ether == 1) {
    strcpy (title, "UDP (IPv4) Ethernet Frame Contents");
  } else {
    strcpy (title, "UDP (IPv4) Packet Contents");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
// Ethernet header
  memcpy (packet, data->udp4ethhdr.dst_mac, 6); c += 6;
  memcpy (packet + c, data->udp4ethhdr.src_mac, 6); c += 6;
  packet[c] = data->udp4ethhdr.type_code % 256; c++;
  packet[c] = data->udp4ethhdr.type_code / 256; c++;
// IPv4 header
  packet[c] = data->ip4hdr_udp.ip_hl + (data->ip4hdr_udp.ip_v << 4); c++;
  packet[c] = data->ip4hdr_udp.ip_tos; c++;
  packet[c] = data->ip4hdr_udp.ip_len & 0xffu; c++;
  packet[c] = data->ip4hdr_udp.ip_len >> 8; c++;
  packet[c] = data->ip4hdr_udp.ip_id & 0xffu; c++;
  packet[c] = data->ip4hdr_udp.ip_id >> 8; c++;
  packet[c] = data->ip4hdr_udp.ip_off & 0xffu; c++;
  packet[c] = data->ip4hdr_udp.ip_off >> 8; c++;
  packet[c] = data->ip4hdr_udp.ip_ttl; c++;
  packet[c] = data->ip4hdr_udp.ip_p; c++;
  packet[c] = data->ip4hdr_udp.ip_sum & 0xffu; c++;
  packet[c] = data->ip4hdr_udp.ip_sum >> 8; c++;
  packet[c] = data->ip4hdr_udp.ip_src.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp.ip_src.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp.ip_src.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_udp.ip_src.s_addr >> 24; c++;
  packet[c] = data->ip4hdr_udp.ip_dst.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp.ip_dst.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp.ip_dst.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_udp.ip_dst.s_addr >> 24; c++;
// UDP header
  packet[c] = data->udp4hdr.uh_sport & 0xffu; c++;
  packet[c] = data->udp4hdr.uh_sport >> 8; c++;
  packet[c] = data->udp4hdr.uh_dport & 0xffu; c++;
  packet[c] = data->udp4hdr.uh_dport >> 8; c++;
  packet[c] = data->udp4hdr.uh_ulen & 0xffu; c++;
  packet[c] = data->udp4hdr.uh_ulen >> 8; c++;
  packet[c] = data->udp4hdr.uh_sum & 0xffu; c++;
  packet[c] = data->udp4hdr.uh_sum >> 8; c++;
// UDP data
  for (i=0; i<data->udp4_payloadlen; i++) {
    packet[c+i] = *(data->udp4_payload + i);
  }

  show_packet (packet, nbytes, title, data->udp4_specify_ether, 0);

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display TCP packet contents (IPv6)
gboolean
on_button6_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  if (data->sixto4_flag == 0) {
    nbytes = ETH_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_payloadlen;
  } else {
    nbytes = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_payloadlen;
  }

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button6_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button6_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->sixto4_flag == 0) {
    strcpy (title, "TCP (IPv6) Ethernet Frame Contents");
  } else {
    strcpy (title, "TCP (IPv6 tunneled over IPv4) Ethernet Frame Contents");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
  if (data->sixto4_flag == 0) {
    // IPv6 TCP ethernet header
    memcpy (packet, data->tcp6ethhdr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->tcp6ethhdr.src_mac, 6); c += 6;
    packet[c] = data->tcp6ethhdr.type_code % 256; c++;
    packet[c] = data->tcp6ethhdr.type_code / 256; c++;
  } else {
    // 6to4 ethernet header
    memcpy (packet, data->sixto4ethhdr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->sixto4ethhdr.src_mac, 6); c += 6;
    packet[c] = data->sixto4ethhdr.type_code % 256; c++;
    packet[c] = data->sixto4ethhdr.type_code / 256; c++;
    // 6to4 IPv4 header
    packet[c] = data->ip4hdr_tcp.ip_hl + (data->ip4hdr_tcp.ip_v << 4); c++;
    packet[c] = data->ip4hdr_tcp.ip_tos; c++;
    packet[c] = data->ip4hdr_tcp.ip_len & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp.ip_len >> 8; c++;
    packet[c] = data->ip4hdr_tcp.ip_id & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp.ip_id >> 8; c++;
    packet[c] = data->ip4hdr_tcp.ip_off & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp.ip_off >> 8; c++;
    packet[c] = data->ip4hdr_tcp.ip_ttl; c++;
    packet[c] = data->ip4hdr_tcp.ip_p; c++;
    packet[c] = data->ip4hdr_tcp.ip_sum & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp.ip_sum >> 8; c++;
    packet[c] = data->ip4hdr_tcp.ip_src.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp.ip_src.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp.ip_src.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_tcp.ip_src.s_addr >> 24; c++;
    packet[c] = data->ip4hdr_tcp.ip_dst.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp.ip_dst.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp.ip_dst.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_tcp.ip_dst.s_addr >> 24; c++;
  }
  // IPv6 header
  packet[c] = data->ip6hdr_tcp.ip6_flow & 0xfful; c++;
  packet[c] = (data->ip6hdr_tcp.ip6_flow >> 8) & 0xfful; c++;
  packet[c] = (data->ip6hdr_tcp.ip6_flow >> 16) & 0xfful; c++;
  packet[c] = data->ip6hdr_tcp.ip6_flow >> 24; c++;
  packet[c] = data->ip6hdr_tcp.ip6_plen % 256; c++;
  packet[c] = data->ip6hdr_tcp.ip6_plen / 256; c++;
  packet[c] = data->ip6hdr_tcp.ip6_nxt; c++;
  packet[c] = data->ip6hdr_tcp.ip6_hops; c++;
  memcpy (&packet[c], data->ip6hdr_tcp.ip6_src.s6_addr, 16); c += 16;
  memcpy (&packet[c], data->ip6hdr_tcp.ip6_dst.s6_addr, 16); c += 16;
  // TCP header
  packet[c] = data->tcp6hdr.th_sport & 0xffu; c++;
  packet[c] = data->tcp6hdr.th_sport >> 8; c++;
  packet[c] = data->tcp6hdr.th_dport & 0xffu; c++;
  packet[c] = data->tcp6hdr.th_dport >> 8; c++;
  packet[c] = data->tcp6hdr.th_seq & 0xfful; c++;
  packet[c] = (data->tcp6hdr.th_seq >> 8) & 0xfful; c++;
  packet[c] = (data->tcp6hdr.th_seq >> 16) & 0xfful; c++;
  packet[c] = data->tcp6hdr.th_seq >> 24; c++;
  packet[c] = data->tcp6hdr.th_ack & 0xfful; c++;
  packet[c] = (data->tcp6hdr.th_ack >> 8) & 0xfful; c++;
  packet[c] = (data->tcp6hdr.th_ack >> 16) & 0xfful; c++;
  packet[c] = data->tcp6hdr.th_ack >> 24; c++;
  packet[c] = (data->tcp6hdr.th_off << 4) + data->tcp6hdr.th_x2; c++;
  packet[c] = data->tcp6hdr.th_flags; c++;
  packet[c] = data->tcp6hdr.th_win & 0xffu; c++;
  packet[c] = data->tcp6hdr.th_win >> 8; c++;
  packet[c] = data->tcp6hdr.th_sum & 0xffu; c++;
  packet[c] = data->tcp6hdr.th_sum >> 8; c++;
  packet[c] = data->tcp6hdr.th_urp & 0xffu; c++;
  packet[c] = data->tcp6hdr.th_urp >> 8; c++;
// TCP data
  for (i=0; i<data->tcp6_payloadlen; i++) {
    packet[c+i] = *(data->tcp6_payload + i);
  }

  if (data->sixto4_flag == 0) {
    show_packet (packet, nbytes, title, 1, 0);
  } else {
    show_packet (packet, nbytes, title, data->sixto4_specify_ether, 1);
  }

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display ICMP packet contents (IPv6)
gboolean
on_button17_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  if (data->sixto4_flag == 0) {
    nbytes = ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_payloadlen;
  } else {
    nbytes = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_payloadlen;
  }

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button17_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button17_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->sixto4_flag == 0) {
    strcpy (title, "ICMP (IPv6) Ethernet Frame Contents");
  } else {
    strcpy (title, "ICMP (IPv6 tunneled over IPv4) Ethernet Frame Contents");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
  if (data->sixto4_flag == 0) {
    // IPv6 ICMP ethernet header
    memcpy (packet, data->icmp6ethhdr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->icmp6ethhdr.src_mac, 6); c += 6;
    packet[c] = data->icmp6ethhdr.type_code % 256; c++;
    packet[c] = data->icmp6ethhdr.type_code / 256; c++;
  } else {
    // 6to4 ethernet header
    memcpy (packet, data->sixto4ethhdr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->sixto4ethhdr.src_mac, 6); c += 6;
    packet[c] = data->sixto4ethhdr.type_code % 256; c++;
    packet[c] = data->sixto4ethhdr.type_code / 256; c++;
    // 6to4 IPv4 header
    packet[c] = data->ip4hdr_icmp.ip_hl + (data->ip4hdr_icmp.ip_v << 4); c++;
    packet[c] = data->ip4hdr_icmp.ip_tos; c++;
    packet[c] = data->ip4hdr_icmp.ip_len & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_len >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_id & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_id >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_off & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_off >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_ttl; c++;
    packet[c] = data->ip4hdr_icmp.ip_p; c++;
    packet[c] = data->ip4hdr_icmp.ip_sum & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_sum >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_src.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_src.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_src.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp.ip_src.s_addr >> 24; c++;
    packet[c] = data->ip4hdr_icmp.ip_dst.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_dst.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_dst.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp.ip_dst.s_addr >> 24; c++;
  }
// IPv6 header
  packet[c] = data->ip6hdr_icmp.ip6_flow & 0xfful; c++;
  packet[c] = (data->ip6hdr_icmp.ip6_flow >> 8) & 0xfful; c++;
  packet[c] = (data->ip6hdr_icmp.ip6_flow >> 16) & 0xfful; c++;
  packet[c] = data->ip6hdr_icmp.ip6_flow >> 24; c++;
  packet[c] = data->ip6hdr_icmp.ip6_plen % 256; c++;
  packet[c] = data->ip6hdr_icmp.ip6_plen / 256; c++;
  packet[c] = data->ip6hdr_icmp.ip6_nxt; c++;
  packet[c] = data->ip6hdr_icmp.ip6_hops; c++;
  memcpy (&packet[c], data->ip6hdr_icmp.ip6_src.s6_addr, 16); c += 16;
  memcpy (&packet[c], data->ip6hdr_icmp.ip6_dst.s6_addr, 16); c += 16;
// ICMP header
  packet[c] = data->icmp6hdr.icmp6_type; c++;
  packet[c] = data->icmp6hdr.icmp6_code; c++;
  packet[c] = data->icmp6hdr.icmp6_cksum & 0xffu; c++;
  packet[c] = data->icmp6hdr.icmp6_cksum >> 8; c++;
  packet[c] = data->icmp6hdr.icmp6_id & 0xffu; c++;
  packet[c] = data->icmp6hdr.icmp6_id >> 8; c++;
  packet[c] = data->icmp6hdr.icmp6_seq & 0xffu; c++;
  packet[c] = data->icmp6hdr.icmp6_seq >> 8; c++;
// ICMP data
  for (i=0; i<data->icmp6_payloadlen; i++) {
    packet[c+i] = *(data->icmp6_payload + i);
  }

  if (data->sixto4_flag == 0) {
    show_packet (packet, nbytes, title, 1, 0);
  } else {
    show_packet (packet, nbytes, title, data->sixto4_specify_ether, 1);
  }

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display UDP packet contents (IPv6)
gboolean
on_button13_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  if (data->sixto4_flag == 0) {
    nbytes = ETH_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_payloadlen;
  } else {
    nbytes = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_payloadlen;
  }

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button13_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button13_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->sixto4_flag == 0) {
    strcpy (title, "UDP (IPv6) Ethernet Frame Contents");
  } else {
    strcpy (title, "UDP (IPv6 tunneled over IPv4) Ethernet Frame Contents");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
  if (data->sixto4_flag == 0) {
    // IPv6 UDP ethernet header
    memcpy (packet, data->udp6ethhdr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->udp6ethhdr.src_mac, 6); c += 6;
    packet[c] = data->udp6ethhdr.type_code % 256; c++;
    packet[c] = data->udp6ethhdr.type_code / 256; c++;
  } else {
    // 6to4 ethernet header
    memcpy (packet, data->sixto4ethhdr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->sixto4ethhdr.src_mac, 6); c += 6;
    packet[c] = data->sixto4ethhdr.type_code % 256; c++;
    packet[c] = data->sixto4ethhdr.type_code / 256; c++;
    // 6to4 IPv4 header
    packet[c] = data->ip4hdr_icmp.ip_hl + (data->ip4hdr_icmp.ip_v << 4); c++;
    packet[c] = data->ip4hdr_icmp.ip_tos; c++;
    packet[c] = data->ip4hdr_icmp.ip_len & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_len >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_id & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_id >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_off & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_off >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_ttl; c++;
    packet[c] = data->ip4hdr_icmp.ip_p; c++;
    packet[c] = data->ip4hdr_icmp.ip_sum & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp.ip_sum >> 8; c++;
    packet[c] = data->ip4hdr_icmp.ip_src.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_src.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_src.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp.ip_src.s_addr >> 24; c++;
    packet[c] = data->ip4hdr_icmp.ip_dst.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_dst.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp.ip_dst.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp.ip_dst.s_addr >> 24; c++;
  }
// IPv6 header
  packet[c] = data->ip6hdr_udp.ip6_flow & 0xfful; c++;
  packet[c] = (data->ip6hdr_udp.ip6_flow >> 8) & 0xfful; c++;
  packet[c] = (data->ip6hdr_udp.ip6_flow >> 16) & 0xfful; c++;
  packet[c] = data->ip6hdr_udp.ip6_flow >> 24; c++;
  packet[c] = data->ip6hdr_udp.ip6_plen % 256; c++;
  packet[c] = data->ip6hdr_udp.ip6_plen / 256; c++;
  packet[c] = data->ip6hdr_udp.ip6_nxt; c++;
  packet[c] = data->ip6hdr_udp.ip6_hops; c++;
  memcpy (&packet[c], data->ip6hdr_udp.ip6_src.s6_addr, 16); c += 16;
  memcpy (&packet[c], data->ip6hdr_udp.ip6_dst.s6_addr, 16); c += 16;
// UDP header
  packet[c] = data->udp6hdr.uh_sport & 0xffu; c++;
  packet[c] = data->udp6hdr.uh_sport >> 8; c++;
  packet[c] = data->udp6hdr.uh_dport & 0xffu; c++;
  packet[c] = data->udp6hdr.uh_dport >> 8; c++;
  packet[c] = data->udp6hdr.uh_ulen & 0xffu; c++;
  packet[c] = data->udp6hdr.uh_ulen >> 8; c++;
  packet[c] = data->udp6hdr.uh_sum & 0xffu; c++;
  packet[c] = data->udp6hdr.uh_sum >> 8; c++;
// UDP data
  for (i=0; i<data->udp6_payloadlen; i++) {
    packet[c+i] = *(data->udp6_payload + i);
  }

  if (data->sixto4_flag == 0) {
    show_packet (packet, nbytes, title, 1, 0);
  } else {
    show_packet (packet, nbytes, title, data->sixto4_specify_ether, 1);
  }

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display TCP traceroute packet contents (IPv4)
gboolean
on_button28_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  nbytes = ETH_HDRLEN + IP4_HDRLEN + TCP_HDRLEN + data->tcp4_tr_payloadlen;

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button28_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button28clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->tcp4_tr_specify_ether == 1) {
    strcpy (title, "TCP (IPv4) Ethernet Frame Contents for Traceroute");
  } else {
    strcpy (title, "TCP (IPv4) Packet Contents for Traceroute");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
// Ethernet header
  memcpy (packet, data->tcp4ethhdr_tr.dst_mac, 6); c += 6;
  memcpy (packet + c, data->tcp4ethhdr_tr.src_mac, 6); c += 6;
  packet[c] = data->tcp4ethhdr_tr.type_code % 256; c++;
  packet[c] = data->tcp4ethhdr_tr.type_code / 256; c++;
// IPv4 header
  packet[c] = data->ip4hdr_tcp_tr.ip_hl + (data->ip4hdr_tcp_tr.ip_v << 4); c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_tos; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_len & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_len >> 8; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_id & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_id >> 8; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_off & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_off >> 8; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_ttl; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_p; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_sum & 0xffu; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_sum >> 8; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_src.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp_tr.ip_src.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp_tr.ip_src.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_src.s_addr >> 24; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_dst.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp_tr.ip_dst.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_tcp_tr.ip_dst.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_tcp_tr.ip_dst.s_addr >> 24; c++;

// TCP header
  packet[c] = data->tcp4hdr_tr.th_sport & 0xffu; c++;
  packet[c] = data->tcp4hdr_tr.th_sport >> 8; c++;
  packet[c] = data->tcp4hdr_tr.th_dport & 0xffu; c++;
  packet[c] = data->tcp4hdr_tr.th_dport >> 8; c++;
  packet[c] = data->tcp4hdr_tr.th_seq & 0xfful; c++;
  packet[c] = (data->tcp4hdr_tr.th_seq >> 8) & 0xfful; c++;
  packet[c] = (data->tcp4hdr_tr.th_seq >> 16) & 0xfful; c++;
  packet[c] = data->tcp4hdr_tr.th_seq >> 24; c++;
  packet[c] = data->tcp4hdr_tr.th_ack & 0xfful; c++;
  packet[c] = (data->tcp4hdr_tr.th_ack >> 8) & 0xfful; c++;
  packet[c] = (data->tcp4hdr_tr.th_ack >> 16) & 0xfful; c++;
  packet[c] = data->tcp4hdr_tr.th_ack >> 24; c++;
  packet[c] = (data->tcp4hdr_tr.th_off << 4) + data->tcp4hdr_tr.th_x2; c++;
  packet[c] = data->tcp4hdr_tr.th_flags; c++;
  packet[c] = data->tcp4hdr_tr.th_win & 0xffu; c++;
  packet[c] = data->tcp4hdr_tr.th_win >> 8; c++;
  packet[c] = data->tcp4hdr_tr.th_sum & 0xffu; c++;
  packet[c] = data->tcp4hdr_tr.th_sum >> 8; c++;
  packet[c] = data->tcp4hdr_tr.th_urp & 0xffu; c++;
  packet[c] = data->tcp4hdr_tr.th_urp >> 8; c++;
// TCP data
  for (i=0; i<data->tcp4_tr_payloadlen; i++) {
    packet[c+i] = *(data->tcp4_tr_payload + i);
  }

  show_packet (packet, nbytes, title, data->tcp4_tr_specify_ether, 0);

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display ICMP traceroute packet contents (IPv4)
gboolean
on_button31_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  nbytes = ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_tr_payloadlen;

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button31_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button31_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->icmp4_specify_ether == 1) {
    strcpy (title, "ICMP (IPv4) Ethernet Frame Contents for Traceroute");
  } else {
    strcpy (title, "ICMP (IPv4) Packet Contents for Traceroute");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
// Ethernet header
  memcpy (packet, data->icmp4ethhdr_tr.dst_mac, 6); c += 6;
  memcpy (packet + c, data->icmp4ethhdr_tr.src_mac, 6); c += 6;
  packet[c] = data->icmp4ethhdr_tr.type_code % 256; c++;
  packet[c] = data->icmp4ethhdr_tr.type_code / 256; c++;
// IPv4 header
  packet[c] = data->ip4hdr_icmp_tr.ip_hl + (data->ip4hdr_icmp_tr.ip_v << 4); c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_tos; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_len & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_len >> 8; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_id & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_id >> 8; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_off & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_off >> 8; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_ttl; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_p; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_sum & 0xffu; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_sum >> 8; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_src.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp_tr.ip_src.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp_tr.ip_src.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_src.s_addr >> 24; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_dst.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp_tr.ip_dst.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_icmp_tr.ip_dst.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_icmp_tr.ip_dst.s_addr >> 24; c++;
// ICMP header
  packet[c] = data->icmp4hdr_tr.icmp_type; c++;
  packet[c] = data->icmp4hdr_tr.icmp_code; c++;
  packet[c] = data->icmp4hdr_tr.icmp_cksum & 0xffu; c++;
  packet[c] = data->icmp4hdr_tr.icmp_cksum >> 8; c++;
  packet[c] = data->icmp4hdr_tr.icmp_id & 0xffu; c++;
  packet[c] = data->icmp4hdr_tr.icmp_id >> 8; c++;
  packet[c] = data->icmp4hdr_tr.icmp_seq & 0xffu; c++;
  packet[c] = data->icmp4hdr_tr.icmp_seq >> 8; c++;
// ICMP data
  for (i=0; i<data->icmp4_tr_payloadlen; i++) {
    packet[c+i] = *(data->icmp4_tr_payload + i);
  }

  show_packet (packet, nbytes, title, data->icmp4_tr_specify_ether, 0);

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display UDP traceroute packet contents (IPv4)
gboolean
on_button34_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  nbytes = ETH_HDRLEN + IP4_HDRLEN + UDP_HDRLEN + data->udp4_tr_payloadlen;

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button34_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button34_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->udp4_specify_ether == 1) {
    strcpy (title, "UDP (IPv4) Ethernet Frame Contents for Traceroute");
  } else {
    strcpy (title, "UDP (IPv4) Packet Contents for Traceroute");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
// Ethernet header
  memcpy (packet, data->udp4ethhdr_tr.dst_mac, 6); c += 6;
  memcpy (packet + c, data->udp4ethhdr_tr.src_mac, 6); c += 6;
  packet[c] = data->udp4ethhdr_tr.type_code % 256; c++;
  packet[c] = data->udp4ethhdr_tr.type_code / 256; c++;
// IPv4 header
  packet[c] = data->ip4hdr_udp_tr.ip_hl + (data->ip4hdr_udp_tr.ip_v << 4); c++;
  packet[c] = data->ip4hdr_udp_tr.ip_tos; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_len & 0xffu; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_len >> 8; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_id & 0xffu; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_id >> 8; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_off & 0xffu; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_off >> 8; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_ttl; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_p; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_sum & 0xffu; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_sum >> 8; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_src.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp_tr.ip_src.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp_tr.ip_src.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_src.s_addr >> 24; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_dst.s_addr & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp_tr.ip_dst.s_addr >> 8) & 0xfful; c++;
  packet[c] = (data->ip4hdr_udp_tr.ip_dst.s_addr >> 16) & 0xfful; c++;
  packet[c] = data->ip4hdr_udp_tr.ip_dst.s_addr >> 24; c++;
// UDP header
  packet[c] = data->udp4hdr_tr.uh_sport & 0xffu; c++;
  packet[c] = data->udp4hdr_tr.uh_sport >> 8; c++;
  packet[c] = data->udp4hdr_tr.uh_dport & 0xffu; c++;
  packet[c] = data->udp4hdr_tr.uh_dport >> 8; c++;
  packet[c] = data->udp4hdr_tr.uh_ulen & 0xffu; c++;
  packet[c] = data->udp4hdr_tr.uh_ulen >> 8; c++;
  packet[c] = data->udp4hdr_tr.uh_sum & 0xffu; c++;
  packet[c] = data->udp4hdr_tr.uh_sum >> 8; c++;
// UDP data
  for (i=0; i<data->udp4_tr_payloadlen; i++) {
    packet[c+i] = *(data->udp4_tr_payload + i);
  }

  show_packet (packet, nbytes, title, data->udp4_tr_specify_ether, 0);

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display TCP traceroute packet contents (IPv6)
gboolean
on_button37_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  if (data->sixto4_tr_flag == 0) {
    nbytes = ETH_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_tr_payloadlen;
  } else {
    nbytes = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_tr_payloadlen;
  }

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button37_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button37_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->sixto4_flag == 0) {
    strcpy (title, "TCP (IPv6) Ethernet Frame Contents for Traceroute");
  } else {
    strcpy (title, "TCP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
  if (data->sixto4_tr_flag == 0) {
    // IPv6 TCP ethernet header
    memcpy (packet, data->tcp6ethhdr_tr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->tcp6ethhdr_tr.src_mac, 6); c += 6;
    packet[c] = data->tcp6ethhdr_tr.type_code % 256; c++;
    packet[c] = data->tcp6ethhdr_tr.type_code / 256; c++;
  } else {
    // 6to4 ethernet header
    memcpy (packet, data->sixto4ethhdr_tr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->sixto4ethhdr_tr.src_mac, 6); c += 6;
    packet[c] = data->sixto4ethhdr_tr.type_code % 256; c++;
    packet[c] = data->sixto4ethhdr_tr.type_code / 256; c++;
    // 6to4 IPv4 header
    packet[c] = data->ip4hdr_tcp_tr.ip_hl + (data->ip4hdr_tcp_tr.ip_v << 4); c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_tos; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_len & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_len >> 8; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_id & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_id >> 8; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_off & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_off >> 8; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_ttl; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_p; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_sum & 0xffu; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_sum >> 8; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_src.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp_tr.ip_src.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp_tr.ip_src.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_src.s_addr >> 24; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_dst.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp_tr.ip_dst.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_tcp_tr.ip_dst.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_tcp_tr.ip_dst.s_addr >> 24; c++;
  }
  // IPv6 header
  packet[c] = data->ip6hdr_tcp_tr.ip6_flow & 0xfful; c++;
  packet[c] = (data->ip6hdr_tcp_tr.ip6_flow >> 8) & 0xfful; c++;
  packet[c] = (data->ip6hdr_tcp_tr.ip6_flow >> 16) & 0xfful; c++;
  packet[c] = data->ip6hdr_tcp_tr.ip6_flow >> 24; c++;
  packet[c] = data->ip6hdr_tcp_tr.ip6_plen % 256; c++;
  packet[c] = data->ip6hdr_tcp_tr.ip6_plen / 256; c++;
  packet[c] = data->ip6hdr_tcp_tr.ip6_nxt; c++;
  packet[c] = data->ip6hdr_tcp_tr.ip6_hops; c++;
  memcpy (&packet[c], data->ip6hdr_tcp_tr.ip6_src.s6_addr, 16); c += 16;
  memcpy (&packet[c], data->ip6hdr_tcp_tr.ip6_dst.s6_addr, 16); c += 16;
  // TCP header
  packet[c] = data->tcp6hdr_tr.th_sport & 0xffu; c++;
  packet[c] = data->tcp6hdr_tr.th_sport >> 8; c++;
  packet[c] = data->tcp6hdr_tr.th_dport & 0xffu; c++;
  packet[c] = data->tcp6hdr_tr.th_dport >> 8; c++;
  packet[c] = data->tcp6hdr_tr.th_seq & 0xfful; c++;
  packet[c] = (data->tcp6hdr_tr.th_seq >> 8) & 0xfful; c++;
  packet[c] = (data->tcp6hdr_tr.th_seq >> 16) & 0xfful; c++;
  packet[c] = data->tcp6hdr_tr.th_seq >> 24; c++;
  packet[c] = data->tcp6hdr_tr.th_ack & 0xfful; c++;
  packet[c] = (data->tcp6hdr_tr.th_ack >> 8) & 0xfful; c++;
  packet[c] = (data->tcp6hdr_tr.th_ack >> 16) & 0xfful; c++;
  packet[c] = data->tcp6hdr_tr.th_ack >> 24; c++;
  packet[c] = (data->tcp6hdr_tr.th_off << 4) + data->tcp6hdr_tr.th_x2; c++;
  packet[c] = data->tcp6hdr_tr.th_flags; c++;
  packet[c] = data->tcp6hdr_tr.th_win & 0xffu; c++;
  packet[c] = data->tcp6hdr_tr.th_win >> 8; c++;
  packet[c] = data->tcp6hdr_tr.th_sum & 0xffu; c++;
  packet[c] = data->tcp6hdr_tr.th_sum >> 8; c++;
  packet[c] = data->tcp6hdr_tr.th_urp & 0xffu; c++;
  packet[c] = data->tcp6hdr_tr.th_urp >> 8; c++;
// TCP data
  for (i=0; i<data->tcp6_tr_payloadlen; i++) {
    packet[c+i] = *(data->tcp6_tr_payload + i);
  }

  if (data->sixto4_tr_flag == 0) {
    show_packet (packet, nbytes, title, 1, 0);
  } else {
    show_packet (packet, nbytes, title, data->sixto4_tr_specify_ether, 1);
  }

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display ICMP traceroute packet contents (IPv6)
gboolean
on_button40_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  if (data->sixto4_tr_flag == 0) {
    nbytes = ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_tr_payloadlen;
  } else {
    nbytes = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_tr_payloadlen;
  }

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button40_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button40_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->sixto4_flag == 0) {
    strcpy (title, "ICMP (IPv6) Ethernet Frame Contents for Traceroute");
  } else {
    strcpy (title, "ICMP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
  if (data->sixto4_tr_flag == 0) {
    // IPv6 ICMP ethernet header
    memcpy (packet, data->icmp6ethhdr_tr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->icmp6ethhdr_tr.src_mac, 6); c += 6;
    packet[c] = data->icmp6ethhdr_tr.type_code % 256; c++;
    packet[c] = data->icmp6ethhdr_tr.type_code / 256; c++;
  } else {
    // 6to4 ethernet header
    memcpy (packet, data->sixto4ethhdr_tr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->sixto4ethhdr_tr.src_mac, 6); c += 6;
    packet[c] = data->sixto4ethhdr_tr.type_code % 256; c++;
    packet[c] = data->sixto4ethhdr_tr.type_code / 256; c++;
    // 6to4 IPv4 header
    packet[c] = data->ip4hdr_icmp_tr.ip_hl + (data->ip4hdr_icmp_tr.ip_v << 4); c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_tos; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_len & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_len >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_id & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_id >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_off & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_off >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_ttl; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_p; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_sum & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_sum >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_src.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_src.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_src.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_src.s_addr >> 24; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_dst.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_dst.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_dst.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_dst.s_addr >> 24; c++;
  }
// IPv6 header
  packet[c] = data->ip6hdr_icmp_tr.ip6_flow & 0xfful; c++;
  packet[c] = (data->ip6hdr_icmp_tr.ip6_flow >> 8) & 0xfful; c++;
  packet[c] = (data->ip6hdr_icmp_tr.ip6_flow >> 16) & 0xfful; c++;
  packet[c] = data->ip6hdr_icmp_tr.ip6_flow >> 24; c++;
  packet[c] = data->ip6hdr_icmp_tr.ip6_plen % 256; c++;
  packet[c] = data->ip6hdr_icmp_tr.ip6_plen / 256; c++;
  packet[c] = data->ip6hdr_icmp_tr.ip6_nxt; c++;
  packet[c] = data->ip6hdr_icmp_tr.ip6_hops; c++;
  memcpy (&packet[c], data->ip6hdr_icmp_tr.ip6_src.s6_addr, 16); c += 16;
  memcpy (&packet[c], data->ip6hdr_icmp_tr.ip6_dst.s6_addr, 16); c += 16;
// ICMP header
  packet[c] = data->icmp6hdr_tr.icmp6_type; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_code; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_cksum & 0xffu; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_cksum >> 8; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_id & 0xffu; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_id >> 8; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_seq & 0xffu; c++;
  packet[c] = data->icmp6hdr_tr.icmp6_seq >> 8; c++;
// ICMP data
  for (i=0; i<data->icmp6_tr_payloadlen; i++) {
    packet[c+i] = *(data->icmp6_tr_payload + i);
  }

  if (data->sixto4_tr_flag == 0) {
    show_packet (packet, nbytes, title, 1, 0);
  } else {
    show_packet (packet, nbytes, title, data->sixto4_tr_specify_ether, 1);
  }

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// Display UDP traceroute packet contents (IPv6)
gboolean
on_button43_clicked (GtkButton *button, SPSData *data)
{
  unsigned char *packet;
  gint i, c, nbytes;
  gchar *title;
  void *tmp;

  if (data->sixto4_tr_flag == 0) {
    nbytes = ETH_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_tr_payloadlen;
  } else {
    nbytes = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_tr_payloadlen;
  }

// Allocate memory for various arrays.

// Array of packet contents.
  tmp = (unsigned char *) g_malloc0 (nbytes * sizeof (unsigned char));
  if (tmp != NULL) {
    packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet' in on_button43_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    title = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'title' in on_button43_clicked().\n");
    exit (EXIT_FAILURE);
  }

  if (data->sixto4_flag == 0) {
    strcpy (title, "UDP (IPv6) Ethernet Frame Contents for Traceroute");
  } else {
    strcpy (title, "UDP (IPv6 tunneled over IPv4) Ethernet Frame Contents for Traceroute");
  }

// Pack headers in packet-order into contiguous memory for easy access by loop
  c = 0;
  if (data->sixto4_tr_flag == 0) {
    // IPv6 UDP ethernet header
    memcpy (packet, data->udp6ethhdr_tr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->udp6ethhdr_tr.src_mac, 6); c += 6;
    packet[c] = data->udp6ethhdr_tr.type_code % 256; c++;
    packet[c] = data->udp6ethhdr_tr.type_code / 256; c++;
  } else {
    // 6to4 ethernet header
    memcpy (packet, data->sixto4ethhdr_tr.dst_mac, 6); c += 6;
    memcpy (packet + c, data->sixto4ethhdr_tr.src_mac, 6); c += 6;
    packet[c] = data->sixto4ethhdr_tr.type_code % 256; c++;
    packet[c] = data->sixto4ethhdr_tr.type_code / 256; c++;
    // 6to4 IPv4 header
    packet[c] = data->ip4hdr_icmp_tr.ip_hl + (data->ip4hdr_icmp_tr.ip_v << 4); c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_tos; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_len & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_len >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_id & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_id >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_off & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_off >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_ttl; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_p; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_sum & 0xffu; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_sum >> 8; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_src.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_src.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_src.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_src.s_addr >> 24; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_dst.s_addr & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_dst.s_addr >> 8) & 0xfful; c++;
    packet[c] = (data->ip4hdr_icmp_tr.ip_dst.s_addr >> 16) & 0xfful; c++;
    packet[c] = data->ip4hdr_icmp_tr.ip_dst.s_addr >> 24; c++;
  }
// IPv6 header
  packet[c] = data->ip6hdr_udp_tr.ip6_flow & 0xfful; c++;
  packet[c] = (data->ip6hdr_udp_tr.ip6_flow >> 8) & 0xfful; c++;
  packet[c] = (data->ip6hdr_udp_tr.ip6_flow >> 16) & 0xfful; c++;
  packet[c] = data->ip6hdr_udp_tr.ip6_flow >> 24; c++;
  packet[c] = data->ip6hdr_udp_tr.ip6_plen % 256; c++;
  packet[c] = data->ip6hdr_udp_tr.ip6_plen / 256; c++;
  packet[c] = data->ip6hdr_udp_tr.ip6_nxt; c++;
  packet[c] = data->ip6hdr_udp_tr.ip6_hops; c++;
  memcpy (&packet[c], data->ip6hdr_udp_tr.ip6_src.s6_addr, 16); c += 16;
  memcpy (&packet[c], data->ip6hdr_udp_tr.ip6_dst.s6_addr, 16); c += 16;
// UDP header
  packet[c] = data->udp6hdr_tr.uh_sport & 0xffu; c++;
  packet[c] = data->udp6hdr_tr.uh_sport >> 8; c++;
  packet[c] = data->udp6hdr_tr.uh_dport & 0xffu; c++;
  packet[c] = data->udp6hdr_tr.uh_dport >> 8; c++;
  packet[c] = data->udp6hdr_tr.uh_ulen & 0xffu; c++;
  packet[c] = data->udp6hdr_tr.uh_ulen >> 8; c++;
  packet[c] = data->udp6hdr_tr.uh_sum & 0xffu; c++;
  packet[c] = data->udp6hdr_tr.uh_sum >> 8; c++;
// UDP data
  for (i=0; i<data->udp6_tr_payloadlen; i++) {
    packet[c+i] = *(data->udp6_tr_payload + i);
  }

  if (data->sixto4_tr_flag == 0) {
    show_packet (packet, nbytes, title, 1, 0);
  } else {
    show_packet (packet, nbytes, title, data->sixto4_tr_specify_ether, 1);
  }

  g_free (packet);
  g_free (title);

  return (TRUE);
}

// TCP page (IPv4)

// Specify ethernet header
gboolean
on_checkbutton18_toggled (GtkButton *button, SPSData *data)
{
  if (data->tcp4_specify_ether == 0) {
    data->tcp4_specify_ether = 1;
  } else {
    data->tcp4_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry144_activate (GtkWidget *entry144, SPSData *data)
{
  int i, bad;
  const gchar *entry144_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry144_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry144_text = gtk_entry_get_text (GTK_ENTRY (data->entry144));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry144_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry144_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry144_text[i] > 47) && (entry144_text[i] < 58)) ||    // 0 to 10
        ((entry144_text[i] > 64) && (entry144_text[i] < 71)) ||      // A to Z
        ((entry144_text[i] > 96) && (entry144_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry144_text[i] > 47) && (entry144_text[i] < 58)) ||    // 0 to 10
        ((entry144_text[i] > 64) && (entry144_text[i] < 71)) ||      // A to Z
        ((entry144_text[i] > 96) && (entry144_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp4ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry144), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry144_text, data->tcp4ethhdr.dst_mac);
  }

  g_free (value);
  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry145_activate (GtkWidget *entry145, SPSData *data)
{
  int i, bad;
  const gchar *entry145_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry145_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry145_text = gtk_entry_get_text (GTK_ENTRY (data->entry145));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry145_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry145_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry145_text[i] > 47) && (entry145_text[i] < 58)) ||    // 0 to 10
        ((entry145_text[i] > 64) && (entry145_text[i] < 71)) ||      // A to Z
        ((entry145_text[i] > 96) && (entry145_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry145_text[i] > 47) && (entry145_text[i] < 58)) ||    // 0 to 10
        ((entry145_text[i] > 64) && (entry145_text[i] < 71)) ||      // A to Z
        ((entry145_text[i] > 96) && (entry145_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp4ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry145), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry145_text, data->tcp4ethhdr.src_mac);
  }

  g_free (value);
  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry150_activate (GtkWidget *entry150, SPSData *data)
{
  int i, sd, status;
  const gchar *entry150_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry150_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry150_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry150_text = gtk_entry_get_text (GTK_ENTRY (entry150));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry150_activate().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry150_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry150_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->tcp4_interfacename, entry150_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->tcp4ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->tcp4ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry145), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry146_activate (GtkWidget *entry146, SPSData *data)
{
  const gchar *entry146_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry146_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry146_text = gtk_entry_get_text (GTK_ENTRY (data->entry146));

// Is value too large for a short int?
  if ((atoi (entry146_text) < 0) || (atoi (entry146_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry146), value);
  } else {
    data->tcp4ethhdr.type_code = htons (atoi (entry146_text));
  }

  g_free (value);
  return (TRUE);
}

// IPv4 header length (4 bits)
gboolean
on_entry29_activate (GtkWidget *entry29, SPSData *data)
{
  const gchar *entry29_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry29_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry29_text = gtk_entry_get_text (GTK_ENTRY (data->entry29));

// Is value too large or small?
  if ((atoi (entry29_text) < 0) || (atoi (entry29_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_tcp.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry29), value);
  } else {
    data->ip4hdr_tcp.ip_hl = atoi (entry29_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry30_activate (GtkWidget *entry30, SPSData *data)
{
  const gchar *entry30_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry30_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry30_text = gtk_entry_get_text (GTK_ENTRY (data->entry30));

// Is value too large or small?
  if ((atoi (entry30_text) < 0) || (atoi (entry30_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_tcp.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry30), value);
  } else {
    data->ip4hdr_tcp.ip_v = atoi (entry30_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }
  g_free (value);

  return (TRUE);
}

// Type of service (8 bits)
gboolean
on_entry31_activate (GtkWidget *entry31, SPSData *data)
{
  const gchar *entry31_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry31_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry31_text = gtk_entry_get_text (GTK_ENTRY (data->entry31));

// Is value too large or small?
  if ((atoi (entry31_text) < 0) || (atoi (entry31_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_tcp.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry31), value);
  } else {
    data->ip4hdr_tcp.ip_tos = atoi (entry31_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }
  g_free (value);

  return (TRUE);
}

// Total length of datagram (16 bits)
gboolean
on_entry32_activate (GtkWidget *entry32, SPSData *data)
{
  const gchar *entry32_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry32_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry32_text = gtk_entry_get_text (GTK_ENTRY (data->entry32));

// Is value too large or small?
  if ((atoi (entry32_text) < 0) || (atoi (entry32_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry32), value);
  } else {
    data->ip4hdr_tcp.ip_len = htons (atoi (entry32_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }
  g_free (value);

  return (TRUE);
}

// ID sequence number (16 bits)
gboolean
on_entry33_activate (GtkWidget *entry33, SPSData *data)
{
  const gchar *entry33_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry33_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry33_text = gtk_entry_get_text (GTK_ENTRY (data->entry33));

// Is value too large or small?
  if ((atoi (entry33_text) < 0) || (atoi (entry33_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry33), value);
  } else {
    data->ip4hdr_tcp.ip_id = htons (atoi (entry33_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }
  g_free (value);

  return (TRUE);
}

// Zero (1 bit)
gboolean
on_entry122_activate (GtkWidget *entry122, SPSData *data)
{
  const gchar *entry122_text, *entry37_text, *entry38_text, *entry39_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry122_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry122_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry122_text = gtk_entry_get_text (GTK_ENTRY (data->entry122));

// Is value too large or small?
  if ((atoi (entry122_text) < 0) || (atoi (entry122_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry122), value);
  } else {
    ip_tcp_flags[0] = atoi (entry122_text);                            // Zero (1 bit)
    entry37_text = gtk_entry_get_text (GTK_ENTRY (data->entry37));
    ip_tcp_flags[1] = atoi (entry37_text);                             // Do not fragment flag (1 bit)
    entry38_text = gtk_entry_get_text (GTK_ENTRY (data->entry38));
    ip_tcp_flags[2] = atoi (entry38_text);                             // More fragments following flag (1 bit)
    entry39_text = gtk_entry_get_text (GTK_ENTRY (data->entry39));
    ip_tcp_flags[3] = atoi (entry39_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_tcp.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// Do not fragment flag (1 bit)
gboolean
on_entry37_activate (GtkWidget *entry37, SPSData *data)
{
  const gchar *entry122_text, *entry37_text, *entry38_text, *entry39_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry37_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry37_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry37_text = gtk_entry_get_text (GTK_ENTRY (data->entry37));

// Is value too large or small?
  if ((atoi (entry37_text) < 0) || (atoi (entry37_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_tcp.ip_off) >> 14) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry37), value);
  } else {
    entry122_text = gtk_entry_get_text (GTK_ENTRY (data->entry122));
    ip_tcp_flags[0] = atoi (entry122_text);                            // Zero (1 bit)
    ip_tcp_flags[1] = atoi (entry37_text);                             // Do not fragment flag (1 bit)
    entry38_text = gtk_entry_get_text (GTK_ENTRY (data->entry38));
    ip_tcp_flags[2] = atoi (entry38_text);                             // More fragments following flag (1 bit)
    entry39_text = gtk_entry_get_text (GTK_ENTRY (data->entry39));
    ip_tcp_flags[3] = atoi (entry39_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_tcp.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// More fragments following flag (1 bit)
gboolean
on_entry38_activate (GtkWidget *entry38, SPSData *data)
{
  const gchar *entry122_text, *entry37_text, *entry38_text, *entry39_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry38_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry38_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry38_text = gtk_entry_get_text (GTK_ENTRY (data->entry38));

// Is value too large or small?
  if ((atoi (entry38_text) < 0) || (atoi (entry38_text) > 1u)) {
    sprintf (value, "%hu", ((ntohs (data->ip4hdr_tcp.ip_off) >> 13) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry38), value);
  } else {
    entry122_text = gtk_entry_get_text (GTK_ENTRY (data->entry122));
    ip_tcp_flags[0] = atoi (entry122_text);                            // Zero (1 bit)
    entry37_text = gtk_entry_get_text (GTK_ENTRY (data->entry37));
    ip_tcp_flags[1] = atoi (entry37_text);                             // Do not fragment flag (1 bit)
    ip_tcp_flags[2] = atoi (entry38_text);                             // More fragments following flag (1 bit)
    entry39_text = gtk_entry_get_text (GTK_ENTRY (data->entry39));
    ip_tcp_flags[3] = atoi (entry39_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_tcp.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// Fragmentation offset (13 bits)
gboolean
on_entry39_activate (GtkWidget *entry39, SPSData *data)
{
  const gchar *entry122_text, *entry37_text, *entry38_text, *entry39_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry39_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry39_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry39_text = gtk_entry_get_text (GTK_ENTRY (data->entry39));

// Is value too large or small?
  if ((atoi (entry39_text) < 0) || (atoi (entry39_text) > 0x1fffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_off) & 0x1fffu);
    gtk_entry_set_text (GTK_ENTRY (data->entry39), value);
  } else {
    entry122_text = gtk_entry_get_text (GTK_ENTRY (data->entry122));
    ip_tcp_flags[0] = atoi (entry122_text);                            // Zero (1 bit)
    entry37_text = gtk_entry_get_text (GTK_ENTRY (data->entry37));
    ip_tcp_flags[1] = atoi (entry37_text);                             // Do not fragment flag (1 bit)
    entry38_text = gtk_entry_get_text (GTK_ENTRY (data->entry38));
    ip_tcp_flags[2] = atoi (entry38_text);                             // More fragments following flag (1 bit)
    ip_tcp_flags[3] = atoi (entry39_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_tcp.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// Time-to-Live (8 bits)
gboolean
on_entry40_activate (GtkWidget *entry40, SPSData *data)
{
  const gchar *entry40_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry40_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry40_text = gtk_entry_get_text (GTK_ENTRY (data->entry40));

// Is value too large or small?
  if ((atoi (entry40_text) < 0) || (atoi (entry40_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_tcp.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry40), value);
  } else {
    data->ip4hdr_tcp.ip_ttl = atoi (entry40_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);

  return (TRUE);
}

// Transport layer protocol (8 bits)
gboolean
on_entry74_activate (GtkWidget *entry74, SPSData *data)
{
  const gchar *entry74_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry74_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry74_text = gtk_entry_get_text (GTK_ENTRY (data->entry74));

// Is value too large or small?
  if ((atoi (entry74_text) < 0) || (atoi (entry74_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_tcp.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry74), value);
  } else {
    data->ip4hdr_tcp.ip_p = atoi (entry74_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (32 bits)
gboolean
on_entry35_activate (GtkWidget *entry35, SPSData *data)
{
  int status;
  const gchar *entry35_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry35_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry35_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry35_text = gtk_entry_get_text (GTK_ENTRY (data->entry35));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry35_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_src.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry35), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry35_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry35_text, &data->ip4hdr_tcp.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry35_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);

// TCP header checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), text);

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton5), FALSE);
    data->ran_tcp4_sourceip = 0;
  }

  g_free (text);
  g_free (value);

  return (TRUE);
}

// Randomize source IP address
gboolean
on_checkbutton5_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton5_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' on_checkbutton5_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_tcp4_sourceip == 0) {
    data->ran_tcp4_sourceip = 1;
    inet_pton (AF_INET, ran4_addr (ipaddress), &(data->ip4hdr_tcp.ip_src));
    inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry35), value);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);

// TCP header checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  } else {
    data->ran_tcp4_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Destination IP address (32 bits)
gboolean
on_entry36_activate (GtkWidget *entry36, SPSData *data)
{
  int status;
  const gchar *entry36_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry36_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry36_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry36_text = gtk_entry_get_text (GTK_ENTRY (data->entry36));

// Is it a valid IP address?
  if (is_valid_ip4 (entry36_text) != TRUE) {

// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_dst.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry36), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry36_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry36_text, &data->ip4hdr_tcp.ip_dst)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry36_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp.ip_sum = 0;
    data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), text);
    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// IP checksum (16 bits)
gboolean
on_entry34_activate (GtkWidget *entry34, SPSData *data)
{
  const gchar *entry34_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry34_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry34_text = gtk_entry_get_text (GTK_ENTRY (data->entry34));

// Is value too large or small?
  if ((atoi (entry34_text) < 0) || (atoi (entry34_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), value);
  } else {
    data->ip4hdr_tcp.ip_sum = htons (atoi (entry34_text));  // Update IPv4 header with new value
  }

  g_free (value);

  return (TRUE);
}

// TCP header (IPv4)

// TCP source port (16 bits)
gboolean
on_entry41_activate (GtkWidget *entry41, SPSData *data)
{
  const gchar *entry41_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry41_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry41_text = gtk_entry_get_text (GTK_ENTRY (data->entry41));

// Is value too large or small?
  if ((atoi (entry41_text) < 0) || (atoi (entry41_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry41), value);
  } else {
    data->tcp4hdr.th_sport = htons (atoi (entry41_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton7), FALSE);
    data->ran_tcp4_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize TCP source port address
gboolean
on_checkbutton7_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton7_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_tcp4_sourceport == 0) {
    data->ran_tcp4_sourceport = 1;
    data->tcp4hdr.th_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry41), value);

// TCP header checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  } else {
    data->ran_tcp4_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// TCP destination port (16 bits)
gboolean
on_entry42_activate (GtkWidget *entry42, SPSData *data)
{
  const gchar *entry42_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry42_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry42_text = gtk_entry_get_text (GTK_ENTRY (data->entry42));

// Is value too large or small?
  if ((atoi (entry42_text) < 0) || (atoi (entry42_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry42), value);
  } else {
    data->tcp4hdr.th_dport = htons (atoi (entry42_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP sequence number (32 bits)
gboolean
on_entry43_activate (GtkWidget *entry43, SPSData *data)
{
  const gchar *entry43_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry43_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry43_text = gtk_entry_get_text (GTK_ENTRY (data->entry43));

// Is value too large or small?
  if ((atol (entry43_text) < 0) || (strtoul (entry43_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr.th_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry43), value);
  } else {
    data->tcp4hdr.th_seq = htonl (atol (entry43_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP acknowledgement number (32 bits)
gboolean
on_entry44_activate (GtkWidget *entry44, SPSData *data)
{
  const gchar *entry44_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry44_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry44_text = gtk_entry_get_text (GTK_ENTRY (data->entry44));

// Is value too large or small?
  if ((atol (entry44_text) < 0) || (strtoul (entry44_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr.th_ack));
    gtk_entry_set_text (GTK_ENTRY (data->entry44), value);
  } else {
    data->tcp4hdr.th_ack = htonl (atol (entry44_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP reserved (4 bits)
gboolean
on_entry45_activate (GtkWidget *entry45, SPSData *data)
{
  const gchar *entry45_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry45_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry45_text = gtk_entry_get_text (GTK_ENTRY (data->entry45));

// Is value too large or small?
  if ((atoi (entry45_text) < 0) || (atoi (entry45_text) > 15)) {
    sprintf (value, "%u", data->tcp4hdr.th_x2);
    gtk_entry_set_text (GTK_ENTRY (data->entry45), value);
  } else {
    data->tcp4hdr.th_x2 = atoi (entry45_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP data offset (4 bits)
gboolean
on_entry46_activate (GtkWidget *entry46, SPSData *data)
{
  const gchar *entry46_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry46_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry46_text = gtk_entry_get_text (GTK_ENTRY (data->entry46));

// Is value too large or small?
  if ((atoi (entry46_text) < 0) || (atoi (entry46_text) > 15)) {
    sprintf (value, "%u", data->tcp4hdr.th_off);
    gtk_entry_set_text (GTK_ENTRY (data->entry46), value);
  } else {
    data->tcp4hdr.th_off = atoi (entry46_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP FIN flag (1 bit)
gboolean
on_entry47_activate (GtkWidget *entry47, SPSData *data)
{
  const gchar *entry47_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry47_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry47_text = gtk_entry_get_text (GTK_ENTRY (data->entry47));

// Is value too large or small?
  if ((atoi (entry47_text) < 0) || (atoi (entry47_text) > 1)) {
    sprintf (value, "%u", data->tcp4hdr.th_flags & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry47), value);
  } else {
    if (atoi (entry47_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 254;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 1;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP SYN flag (1 bit)
gboolean
on_entry48_activate (GtkWidget *entry48, SPSData *data)
{
  const gchar *entry48_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry48_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry48_text = gtk_entry_get_text (GTK_ENTRY (data->entry48));

// Is value too large or small?
  if ((atoi (entry48_text) < 0) || (atoi (entry48_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 2) >> 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry48), value);
  } else {
    if (atoi (entry48_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 253;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 2;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP RST flag (1 bit)
gboolean
on_entry49_activate (GtkWidget *entry49, SPSData *data)
{
  const gchar *entry49_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry49_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry49_text = gtk_entry_get_text (GTK_ENTRY (data->entry49));

// Is value too large or small?
  if ((atoi (entry49_text) < 0) || (atoi (entry49_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 4) >> 2);
    gtk_entry_set_text (GTK_ENTRY (data->entry49), value);
  } else {
    if (atoi (entry49_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 251;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 4;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP PSH flag (1 bit)
gboolean
on_entry50_activate (GtkWidget *entry50, SPSData *data)
{
  const gchar *entry50_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry50_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry50_text = gtk_entry_get_text (GTK_ENTRY (data->entry50));

// Is value too large or small?
  if ((atoi (entry50_text) < 0) || (atoi (entry50_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 8) >> 3);
    gtk_entry_set_text (GTK_ENTRY (data->entry50), value);
  } else {
    if (atoi (entry50_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 247;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 8;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ACK flag (1 bit)
gboolean
on_entry51_activate (GtkWidget *entry51, SPSData *data)
{
  const gchar *entry51_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry51_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry51_text = gtk_entry_get_text (GTK_ENTRY (data->entry51));

// Is value too large or small?
  if ((atoi (entry51_text) < 0) || (atoi (entry51_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 16) >> 4);
    gtk_entry_set_text (GTK_ENTRY (data->entry51), value);
  } else {
    if (atoi (entry51_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 239;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 16;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP URG flag (1 bit)
gboolean
on_entry52_activate (GtkWidget *entry52, SPSData *data)
{
  const gchar *entry52_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry52_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry52_text = gtk_entry_get_text (GTK_ENTRY (data->entry52));

// Is value too large or small?
  if ((atoi (entry52_text) < 0) || (atoi (entry52_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 32) >> 5);
    gtk_entry_set_text (GTK_ENTRY (data->entry52), value);
  } else {
    if (atoi (entry52_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 223;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 32;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ECE flag (1 bit)
gboolean
on_entry53_activate (GtkWidget *entry53, SPSData *data)
{
  const gchar *entry53_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry53_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry53_text = gtk_entry_get_text (GTK_ENTRY (data->entry53));

// Is value too large or small?
  if ((atoi (entry53_text) < 0) || (atoi (entry53_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 64) >> 6);
    gtk_entry_set_text (GTK_ENTRY (data->entry53), value);
  } else {
    if (atoi (entry53_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 191;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 64;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP CWR flag (1 bit)
gboolean
on_entry54_activate (GtkWidget *entry54, SPSData *data)
{
  const gchar *entry54_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry54_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry54_text = gtk_entry_get_text (GTK_ENTRY (data->entry54));

// Is value too large or small?
  if ((atoi (entry54_text) < 0) || (atoi (entry54_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr.th_flags & 128) >> 7);
    gtk_entry_set_text (GTK_ENTRY (data->entry54), value);
  } else {
    if (atoi (entry54_text) == 0) {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags & 127;
    } else {
      data->tcp4hdr.th_flags = data->tcp4hdr.th_flags | 128;
    }

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP window size (16 bits)
gboolean
on_entry55_activate (GtkWidget *entry55, SPSData *data)
{
  const gchar *entry55_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry55_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry55_text = gtk_entry_get_text (GTK_ENTRY (data->entry55));

// Is value too large or small?
  if ((atoi (entry55_text) < 0) || (atoi (entry55_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_win));
    gtk_entry_set_text (GTK_ENTRY (data->entry55), value);
  } else {
    data->tcp4hdr.th_win = htons (atoi (entry55_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP checksum (16 bits)
gboolean
on_entry56_activate (GtkWidget *entry56, SPSData *data)
{
  const gchar *entry56_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry56_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry56_text = gtk_entry_get_text (GTK_ENTRY (data->entry56));

// Is value too large or small?
  if ((atoi (entry56_text) < 0) || (atoi (entry56_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  } else {
    data->tcp4hdr.th_sum = htons (atoi (entry56_text));  // Update TCP header with new value
  }

  g_free (value);

  return (TRUE);
}

// TCP urgent pointer (16 bits)
gboolean
on_entry77_activate (GtkWidget *entry77, SPSData *data)
{
  const gchar *entry77_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry77_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry77_text = gtk_entry_get_text (GTK_ENTRY (data->entry77));

// Is value too large or small?
  if ((atoi (entry77_text) < 0) || (atoi (entry77_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_urp));
    gtk_entry_set_text (GTK_ENTRY (data->entry77), value);
  } else {
    data->tcp4hdr.th_urp = htons (atoi (entry77_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton17_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp4 = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton18_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp4 = 1;

  return (TRUE);
}

// TCP data from keyboard
gboolean
on_entry162_activate (GtkWidget *entry162, SPSData *data)
{
  int i, j, c;
  const gchar *entry162_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry162_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry162_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry162_text = gtk_entry_get_text (GTK_ENTRY (data->entry162));

// If user asked for no payload then zero it out.
  if (strlen (entry162_text) == 0) {
    data->tcp4_payloadlen = 0;
    memset (data->tcp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry162_text) != 0) && (data->ascii_hex_tcp4 == 0)) {

    data->tcp4_payloadlen = strlen (entry162_text);
    memset (data->tcp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
    memcpy (data->tcp4_payload, entry162_text, data->tcp4_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry162_text) != 0) && (data->ascii_hex_tcp4 == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry162_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->tcp4_payloadlen = 0;
      memset (data->tcp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry162_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry162_text[i+j] != 32) && (entry162_text[i+j] != 44) && ((i + j) < strlen (entry162_text))) {
          value[j] = entry162_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->tcp4_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry162_activate().\n");
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->tcp4_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry162_text[i] == 32) || (entry162_text[i] == 44)) && (i < strlen (entry162_text))) {
          i++;
        }
      }  // Looping through entry162_text
    }  // else data is in acceptable format
  }  // else data is hex

// Total length of datagram (16 bits): IP header + TCP header + TCP payload
  data->ip4hdr_tcp.ip_len = htons (IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry32), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp.ip_sum = 0;
  data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry34), text);

// TCP header checksum (16 bits)
  data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry56), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// TCP data from file
gboolean
on_button20_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int tcp4_lim = IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button20_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button20_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button20_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > tcp4_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", tcp4_lim);
      data->parent = data->main_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->tcp4_payloadlen = index;

// Clear current data.
  memset (data->tcp4_payload, 0, tcp4_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry162), "");

// Copy payload from buffer.
  memcpy (data->tcp4_payload, buffer, data->tcp4_payloadlen);

// Total length of datagram (16 bits): IP header + TCP header + TCP payload
  data->ip4hdr_tcp.ip_len = htons (IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry32), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp.ip_sum = 0;
  data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry34), text);

// TCP header checksum (16 bits)
  data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry56), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry162), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// TCP page (IPv6)

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry131_activate (GtkWidget *entry131, SPSData *data)
{
  int i, bad;
  const gchar *entry131_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry131_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry131_text = gtk_entry_get_text (GTK_ENTRY (data->entry131));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry131_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry131_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry131_text[i] > 47) && (entry131_text[i] < 58)) ||    // 0 to 10
        ((entry131_text[i] > 64) && (entry131_text[i] < 71)) ||      // A to Z
        ((entry131_text[i] > 96) && (entry131_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry131_text[i] > 47) && (entry131_text[i] < 58)) ||    // 0 to 10
        ((entry131_text[i] > 64) && (entry131_text[i] < 71)) ||      // A to Z
        ((entry131_text[i] > 96) && (entry131_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp6ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry131), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry131_text, data->tcp6ethhdr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry132_activate (GtkWidget *entry132, SPSData *data)
{
  int i, bad;
  const gchar *entry132_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry132_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry132_text = gtk_entry_get_text (GTK_ENTRY (data->entry132));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry132_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry132_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry132_text[i] > 47) && (entry132_text[i] < 58)) ||    // 0 to 10
        ((entry132_text[i] > 64) && (entry132_text[i] < 71)) ||      // A to Z
        ((entry132_text[i] > 96) && (entry132_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry132_text[i] > 47) && (entry132_text[i] < 58)) ||    // 0 to 10
        ((entry132_text[i] > 64) && (entry132_text[i] < 71)) ||      // A to Z
        ((entry132_text[i] > 96) && (entry132_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp6ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry132), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry132_text, data->tcp6ethhdr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry154_activate (GtkWidget *entry154, SPSData *data)
{
  int i, sd, status;
  const gchar *entry154_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry154_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry154_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry154_text = gtk_entry_get_text (GTK_ENTRY (entry154));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for using ioctl() in on_entry154_activate.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry154_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry154_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->tcp6_interfacename, entry154_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->tcp6ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->tcp6ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry132), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry133_activate (GtkWidget *entry133, SPSData *data)
{
  const gchar *entry133_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry133_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry133_text = gtk_entry_get_text (GTK_ENTRY (data->entry133));

// Is value too large for a short int?
  if ((atoi (entry133_text) < 0) || (atoi (entry133_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry133), value);
  } else {
    data->tcp6ethhdr.type_code = htons (atoi (entry133_text));
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry22_activate (GtkWidget *entry22, SPSData *data)
{
  const gchar *entry22_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry22_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry22_text = gtk_entry_get_text (GTK_ENTRY (data->entry22));

// Save value currently in header
  old = ntohl (data->ip6hdr_tcp.ip6_flow);

// Is value too large or small?
  if ((atol (entry22_text) < 0) || (atol (entry22_text) > 15ul)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->ip6hdr_tcp.ip6_flow) >> 28);
    gtk_entry_set_text (GTK_ENTRY (data->entry22), value);
  } else {
    data->ip6hdr_tcp.ip6_flow = htonl (((atol (entry22_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Traffic class (8 bits)
gboolean
on_entry23_activate (GtkWidget *entry23, SPSData *data)
{
  const gchar *entry23_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry23_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry23_text = gtk_entry_get_text (GTK_ENTRY (data->entry23));

// Save value currently in header
  old = ntohl (data->ip6hdr_tcp.ip6_flow);

// Is value too large or small?
  if ((atol (entry23_text) < 0) || (atol (entry23_text) > 0xfful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry23), value);
  } else {
    data->ip6hdr_tcp.ip6_flow = htonl (((atol (entry23_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Flow label (20 bits)
gboolean
on_entry24_activate (GtkWidget *entry24, SPSData *data)
{
  const gchar *entry24_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry24_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry24_text = gtk_entry_get_text (GTK_ENTRY (data->entry24));

// Save value currently in header
  old = ntohl (data->ip6hdr_tcp.ip6_flow);

// Is value too large or small?
  if ((atol (entry24_text) < 0) || (atol (entry24_text) > 0xffffful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry24), value);
  } else {
    data->ip6hdr_tcp.ip6_flow = htonl (atol (entry24_text) + (old & 0xfff00000ul));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Payload length (16 bits)
gboolean
on_entry25_activate (GtkWidget *entry25, SPSData *data)
{
  const gchar *entry25_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry25_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry25_text = gtk_entry_get_text (GTK_ENTRY (data->entry25));

// Is value too large or small?
  if ((atoi (entry25_text) < 0) || (atoi (entry25_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry25), value);
  } else {
    data->ip6hdr_tcp.ip6_plen = htons (atoi (entry25_text));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Next header (8 bits)
gboolean
on_entry26_activate (GtkWidget *entry26, SPSData *data)
{
  const gchar *entry26_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry26_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry26_text = gtk_entry_get_text (GTK_ENTRY (data->entry26));

// Is value too large or small?
  if ((atoi (entry26_text) < 0) || (atoi (entry26_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_tcp.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry26), value);
  } else {
    data->ip6hdr_tcp.ip6_nxt = atoi (entry26_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Hop limit (8 bits)
gboolean
on_entry27_activate (GtkWidget *entry27, SPSData *data)
{
  const gchar *entry27_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry27_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry27_text = gtk_entry_get_text (GTK_ENTRY (data->entry27));

// Is value too large or small?
  if ((atoi (entry27_text) < 0) || (atoi (entry27_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_tcp.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry27), value);
  } else {
    data->ip6hdr_tcp.ip6_hops = atoi (entry27_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (128 bits)
gboolean
on_entry28_activate (GtkWidget *entry28, SPSData *data)
{
  const gchar *entry28_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry28_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry28_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry28_text = gtk_entry_get_text (GTK_ENTRY (data->entry28));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry28_text, &(data->ip6hdr_tcp.ip6_src)) < 1) {
    
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry28), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry28_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), text);

// IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton2), FALSE);
    data->ran_tcp6_sourceip = 0;

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize source IP address
gboolean
on_checkbutton2_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton2_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in on_checkbutton2_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_tcp6_sourceip == 0) {
    data->ran_tcp6_sourceip = 1;
    inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_tcp.ip6_src);
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry28), value);

    // TCP header checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  } else {
    data->ran_tcp6_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Destination IP address (128 bits)
gboolean
on_entry78_activate (GtkWidget *entry78, SPSData *data)
{
  const gchar *entry78_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry78_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry78_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry78_text = gtk_entry_get_text (GTK_ENTRY (data->entry78));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry78_text, &(data->ip6hdr_tcp.ip6_dst)) < 1) {

// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_dst), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry78), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry78_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), text);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// TCP header (IPv6)

// TCP source port (16 bits)
gboolean
on_entry79_activate (GtkWidget *entry79, SPSData *data)
{
  const gchar *entry79_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry79_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry79_text = gtk_entry_get_text (GTK_ENTRY (data->entry79));

// Is value too large or small?
  if ((atoi (entry79_text) < 0) || (atoi (entry79_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry79), value);
  } else {
    data->tcp6hdr.th_sport = htons (atoi (entry79_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton4), FALSE);
    data->ran_tcp6_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize TCP source port address
gboolean
on_checkbutton4_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton4_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_tcp6_sourceport == 0) {
    data->ran_tcp6_sourceport = 1;
    data->tcp6hdr.th_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry79), value);

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  } else {
    data->ran_tcp6_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// TCP destination port (16 bits)
gboolean
on_entry80_activate (GtkWidget *entry80, SPSData *data)
{
  const gchar *entry80_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry80_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry80_text = gtk_entry_get_text (GTK_ENTRY (data->entry80));

// Is value too large or small?
  if ((atoi (entry80_text) < 0) || (atoi (entry80_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry80), value);
  } else {
    data->tcp6hdr.th_dport = htons (atoi (entry80_text)); // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP sequence number (32 bits)
gboolean
on_entry81_activate (GtkWidget *entry81, SPSData *data)
{
  const gchar *entry81_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry81_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry81_text = gtk_entry_get_text (GTK_ENTRY (data->entry81));

// Is value too large or small?
  if ((atol (entry81_text) < 0) || (strtoul (entry81_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr.th_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry81), value);
  } else {
    data->tcp6hdr.th_seq = htonl (atol (entry81_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP acknowledgement number (32 bits)
gboolean
on_entry82_activate (GtkWidget *entry82, SPSData *data)
{
  const gchar *entry82_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry82_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry82_text = gtk_entry_get_text (GTK_ENTRY (data->entry82));

// Is value too large or small?
  if ((atol (entry82_text) < 0) || (strtoul (entry82_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr.th_ack));
    gtk_entry_set_text (GTK_ENTRY (data->entry82), value);
  } else {
    data->tcp6hdr.th_ack = htonl (atol (entry82_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP reserved (4 bits)
gboolean
on_entry83_activate (GtkWidget *entry83, SPSData *data)
{
  const gchar *entry83_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry83_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry83_text = gtk_entry_get_text (GTK_ENTRY (data->entry83));

// Is value too large or small?
  if ((atoi (entry83_text) < 0) || (atoi (entry83_text) > 15)) {
    sprintf (value, "%u", data->tcp6hdr.th_x2);
    gtk_entry_set_text (GTK_ENTRY (data->entry83), value);
  } else {
    data->tcp6hdr.th_x2 = atoi (entry83_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP data offset (4 bits)
gboolean
on_entry84_activate (GtkWidget *entry84, SPSData *data)
{
  const gchar *entry84_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry84_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry84_text = gtk_entry_get_text (GTK_ENTRY (data->entry84));

// Is value too large or small?
  if ((atoi (entry84_text) < 0) || (atoi (entry84_text) > 15)) {
    sprintf (value, "%u", data->tcp6hdr.th_off);
    gtk_entry_set_text (GTK_ENTRY (data->entry84), value);
  } else {
    data->tcp6hdr.th_off = atoi (entry84_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP FIN flag (1 bit)
gboolean
on_entry85_activate (GtkWidget *entry85, SPSData *data)
{
  const gchar *entry85_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry85_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry85_text = gtk_entry_get_text (GTK_ENTRY (data->entry85));

// Is value too large or small?
  if ((atoi (entry85_text) < 0) || (atoi (entry85_text) > 1)) {
    sprintf (value, "%u", data->tcp6hdr.th_flags & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry85), value);
  } else {
    if (atoi (entry85_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 254;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 1;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP SYN flag (1 bit)
gboolean
on_entry86_activate (GtkWidget *entry86, SPSData *data)
{
  const gchar *entry86_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry86_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry86_text = gtk_entry_get_text (GTK_ENTRY (data->entry86));

// Is value too large or small?
  if ((atoi (entry86_text) < 0) || (atoi (entry86_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 2) >> 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry86), value);
  } else {
    if (atoi (entry86_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 253;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 2;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP RST flag (1 bit)
gboolean
on_entry87_activate (GtkWidget *entry87, SPSData *data)
{
  const gchar *entry87_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry87_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry87_text = gtk_entry_get_text (GTK_ENTRY (data->entry87));

// Is value too large or small?
  if ((atoi (entry87_text) < 0) || (atoi (entry87_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 4) >> 2);
    gtk_entry_set_text (GTK_ENTRY (data->entry87), value);
  } else {
    if (atoi (entry87_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 251;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 4;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP PSH flag (1 bit)
gboolean
on_entry88_activate (GtkWidget *entry88, SPSData *data)
{
  const gchar *entry88_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry88_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry88_text = gtk_entry_get_text (GTK_ENTRY (data->entry88));

// Is value too large or small?
  if ((atoi (entry88_text) < 0) || (atoi (entry88_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 8) >> 3);
    gtk_entry_set_text (GTK_ENTRY (data->entry88), value);
  } else {
    if (atoi (entry88_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 247;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 8;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ACK flag (1 bit)
gboolean
on_entry89_activate (GtkWidget *entry89, SPSData *data)
{
  const gchar *entry89_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry89_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry89_text = gtk_entry_get_text (GTK_ENTRY (data->entry89));

// Is value too large or small?
  if ((atoi (entry89_text) < 0) || (atoi (entry89_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 16) >> 4);
    gtk_entry_set_text (GTK_ENTRY (data->entry89), value);
  } else {
    if (atoi (entry89_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 239;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 16;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP URG flag (1 bit)
gboolean
on_entry90_activate (GtkWidget *entry90, SPSData *data)
{
  const gchar *entry90_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry90_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry90_text = gtk_entry_get_text (GTK_ENTRY (data->entry90));

// Is value too large or small?
  if ((atoi (entry90_text) < 0) || (atoi (entry90_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 32) >> 5);
    gtk_entry_set_text (GTK_ENTRY (data->entry90), value);
  } else {
    if (atoi (entry90_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 223;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 32;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ECE flag (1 bit)
gboolean
on_entry91_activate (GtkWidget *entry91, SPSData *data)
{
  const gchar *entry91_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry91_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry91_text = gtk_entry_get_text (GTK_ENTRY (data->entry91));

// Is value too large or small?
  if ((atoi (entry91_text) < 0) || (atoi (entry91_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 64) >> 6);
    gtk_entry_set_text (GTK_ENTRY (data->entry91), value);
  } else {
    if (atoi (entry91_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 191;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 64;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP CWR flag (1 bit)
gboolean
on_entry92_activate (GtkWidget *entry92, SPSData *data)
{
  const gchar *entry92_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry92_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry92_text = gtk_entry_get_text (GTK_ENTRY (data->entry92));

// Is value too large or small?
  if ((atoi (entry92_text) < 0) || (atoi (entry92_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr.th_flags & 128) >> 7);
    gtk_entry_set_text (GTK_ENTRY (data->entry92), value);
  } else {
    if (atoi (entry92_text) == 0) {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags & 127;
    } else {
      data->tcp6hdr.th_flags = data->tcp6hdr.th_flags | 128;
    }

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP window size (16 bits)
gboolean
on_entry93_activate (GtkWidget *entry93, SPSData *data)
{
  const gchar *entry93_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry93_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry93_text = gtk_entry_get_text (GTK_ENTRY (data->entry93));

// Is value too large or small?
  if ((atoi (entry93_text) < 0) || (atoi (entry93_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_win));
    gtk_entry_set_text (GTK_ENTRY (data->entry93), value);
  } else {
    data->tcp6hdr.th_win = htons (atoi (entry93_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP checksum (16 bits)
gboolean
on_entry94_activate (GtkWidget *entry94, SPSData *data)
{
  const gchar *entry94_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry94_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry94_text = gtk_entry_get_text (GTK_ENTRY (data->entry94));

// Is value too large or small?
  if ((atoi (entry94_text) < 0) || (atoi (entry94_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  } else {
    data->tcp6hdr.th_sum = htons (atoi (entry94_text));  // Update TCP header with new value
  }

  g_free (value);

  return (TRUE);
}

// TCP urgent pointer (16 bits)
gboolean
on_entry95_activate (GtkWidget *entry95, SPSData *data)
{
  const gchar *entry95_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry95_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry95_text = gtk_entry_get_text (GTK_ENTRY (data->entry95));

// Is value too large or small?
  if ((atoi (entry95_text) < 0) || (atoi (entry95_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_urp));
    gtk_entry_set_text (GTK_ENTRY (data->entry95), value);
  } else {
    data->tcp6hdr.th_urp = htons (atoi (entry95_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton19_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp6 = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton20_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp6 = 1;

  return (TRUE);
}

// TCP data from keyboard
gboolean
on_entry163_activate (GtkWidget *entry163, SPSData *data)
{
  int i, j, c;
  const gchar *entry163_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry163_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry163_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry163_text = gtk_entry_get_text (GTK_ENTRY (data->entry163));

// If user asked for no payload then zero it out.
  if (strlen (entry163_text) == 0) {
    data->tcp6_payloadlen = 0;
    memset (data->tcp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry163_text) != 0) && (data->ascii_hex_tcp6 == 0)) {

    data->tcp6_payloadlen = strlen (entry163_text);
    memset (data->tcp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
    memcpy (data->tcp6_payload, entry163_text, data->tcp6_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry163_text) != 0) && (data->ascii_hex_tcp6 == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry163_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->tcp6_payloadlen = 0;
      memset (data->tcp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry163_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry163_text[i+j] != 32) && (entry163_text[i+j] != 44) && ((i + j) < strlen (entry163_text))) {
          value[j] = entry163_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->tcp6_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry163_activate().\n");
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->tcp6_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry163_text[i] == 32) || (entry163_text[i] == 44)) && (i < strlen (entry163_text))) {
          i++;
        }
      }  // Looping through entry163_text
    }  // else data is in acceptable format
  }  // else data is hex

// Payload length (16 bits): TCP header + TCP data
  data->ip6hdr_tcp.ip6_plen = htons (TCP_HDRLEN + data->tcp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry25), value);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_tcp.ip_len = htons (TCP_HDRLEN + data->tcp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// TCP header checksum (16 bits)
  data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry94), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// TCP data from file
gboolean
on_button22_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int tcp6_lim = IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button22_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button22_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button22_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > tcp6_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", tcp6_lim);
      data->parent = data->main_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->tcp6_payloadlen = index;

// Clear current data.
  memset (data->tcp6_payload, 0, tcp6_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry163), "");

// Copy payload from buffer.
  memcpy (data->tcp6_payload, buffer, data->tcp6_payloadlen);

// Payload length (16 bits): TCP header + TCP data
  data->ip6hdr_tcp.ip6_plen = htons (TCP_HDRLEN + data->tcp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry25), text);

// TCP header checksum (16 bits)
  data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry94), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry163), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// ICMP page (IPv4)

// Specify ethernet header
gboolean
on_checkbutton17_toggled (GtkButton *button, SPSData *data)
{
  if (data->icmp4_specify_ether == 0) {
    data->icmp4_specify_ether = 1;
  } else {
    data->icmp4_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry141_activate (GtkWidget *entry141, SPSData *data)
{
  int i, bad;
  const gchar *entry141_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry141_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry141_text = gtk_entry_get_text (GTK_ENTRY (data->entry141));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry141_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry141_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry141_text[i] > 47) && (entry141_text[i] < 58)) ||    // 0 to 10
        ((entry141_text[i] > 64) && (entry141_text[i] < 71)) ||      // A to Z
        ((entry141_text[i] > 96) && (entry141_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry141_text[i] > 47) && (entry141_text[i] < 58)) ||    // 0 to 10
        ((entry141_text[i] > 64) && (entry141_text[i] < 71)) ||      // A to Z
        ((entry141_text[i] > 96) && (entry141_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp4ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry141), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry141_text, data->icmp4ethhdr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry142_activate (GtkWidget *entry142, SPSData *data)
{
  int i, bad;
  const gchar *entry142_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry142_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry142_text = gtk_entry_get_text (GTK_ENTRY (data->entry142));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry142_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry142_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry142_text[i] > 47) && (entry142_text[i] < 58)) ||    // 0 to 10
        ((entry142_text[i] > 64) && (entry142_text[i] < 71)) ||      // A to Z
        ((entry142_text[i] > 96) && (entry142_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry142_text[i] > 47) && (entry142_text[i] < 58)) ||    // 0 to 10
        ((entry142_text[i] > 64) && (entry142_text[i] < 71)) ||      // A to Z
        ((entry142_text[i] > 96) && (entry142_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp4ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry142), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry142_text, data->icmp4ethhdr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry151_activate (GtkWidget *entry151, SPSData *data)
{
  int i, sd, status;
  const gchar *entry151_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry151_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry151_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry151_text = gtk_entry_get_text (GTK_ENTRY (entry151));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry151_activate().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry151_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry151_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->icmp4_interfacename, entry151_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->icmp4ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->icmp4ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry142), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry143_activate (GtkWidget *entry143, SPSData *data)
{
  const gchar *entry143_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry143_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry143_text = gtk_entry_get_text (GTK_ENTRY (data->entry143));

// Is value too large for a short int?
  if ((atoi (entry143_text) < 0) || (atoi (entry143_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry143), value);
  } else {
    data->icmp4ethhdr.type_code = htons (atoi (entry143_text));
  }

  g_free (value);

  return (TRUE);
}

// IP header length (4 bits)
gboolean
on_entry57_activate (GtkWidget *entry57, SPSData *data)
{
  const gchar *entry57_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry57_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry57_text = gtk_entry_get_text (GTK_ENTRY (data->entry57));

// Is value too large or small?
  if ((atoi (entry57_text) < 0) || (atoi (entry57_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_icmp.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry57), value);
  } else {
    data->ip4hdr_icmp.ip_hl = atoi (entry57_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry58_activate (GtkWidget *entry58, SPSData *data)
{
  const gchar *entry58_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry58_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry58_text = gtk_entry_get_text (GTK_ENTRY (data->entry58));

// Is value too large or small?
  if ((atoi (entry58_text) < 0) || (atoi (entry58_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_icmp.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry58), value);
  } else {
    data->ip4hdr_icmp.ip_v = atoi (entry58_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// Type of service (8 bits)
gboolean
on_entry59_activate (GtkWidget *entry59, SPSData *data)
{
  const gchar *entry59_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry59_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry59_text = gtk_entry_get_text (GTK_ENTRY (data->entry59));

// Is value too large or small?
  if ((atoi (entry59_text) < 0) || (atoi (entry59_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_icmp.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry59), value);
  } else {
    data->ip4hdr_icmp.ip_tos = atoi (entry59_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// Total length of datagram (16 bits)
gboolean
on_entry60_activate (GtkWidget *entry60, SPSData *data)
{
  const gchar *entry60_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry60_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry60_text = gtk_entry_get_text (GTK_ENTRY (data->entry60));

// Is value too large or small?
  if ((atoi (entry60_text) < 0) || (atoi (entry60_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry60), value);
  } else {
    data->ip4hdr_icmp.ip_len = htons (atoi (entry60_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// ID sequence number (16 bits)
gboolean
on_entry61_activate (GtkWidget *entry61, SPSData *data)
{
  const gchar *entry61_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry61_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry61_text = gtk_entry_get_text (GTK_ENTRY (data->entry61));

// Is value too large or small?
  if ((atoi (entry61_text) < 0) || (atoi (entry61_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry61), value);
  } else {
    data->ip4hdr_icmp.ip_id = htons (atoi (entry61_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// Zero (1 bit)
gboolean
on_entry123_activate (GtkWidget *entry123, SPSData *data)
{
  const gchar *entry123_text, *entry65_text, *entry66_text, *entry67_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry123_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (ICMP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry123_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry123_text = gtk_entry_get_text (GTK_ENTRY (data->entry123));

// Is value too large or small?
  if ((atoi (entry123_text) < 0) || (atoi (entry123_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry123), value);
  } else {
    ip_icmp_flags[0] = atoi (entry123_text);                            // Zero (1 bit)
    entry65_text = gtk_entry_get_text (GTK_ENTRY (data->entry65));
    ip_icmp_flags[1] = atoi (entry65_text);                             // Do not fragment flag (1 bit)
    entry66_text = gtk_entry_get_text (GTK_ENTRY (data->entry66));
    ip_icmp_flags[2] = atoi (entry66_text);                             // More fragments following flag (1 bit)
    entry67_text = gtk_entry_get_text (GTK_ENTRY (data->entry67));
    ip_icmp_flags[3] = atoi (entry67_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_icmp.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// Do not fragment flag (1 bit)
gboolean
on_entry65_activate (GtkWidget *entry65, SPSData *data)
{
  const gchar *entry123_text, *entry65_text, *entry66_text, *entry67_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry65_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (ICMP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry65_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry65_text = gtk_entry_get_text (GTK_ENTRY (data->entry65));

// Is value too large or small?
  if ((atoi (entry65_text) < 0) || (atoi (entry65_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_icmp.ip_off) >> 14) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry65), value);
  } else {
    entry123_text = gtk_entry_get_text (GTK_ENTRY (data->entry123));
    ip_icmp_flags[0] = atoi (entry123_text);                            // Zero (1 bit)
    ip_icmp_flags[1] = atoi (entry65_text);                             // Do not fragment flag (1 bit)
    entry66_text = gtk_entry_get_text (GTK_ENTRY (data->entry66));
    ip_icmp_flags[2] = atoi (entry66_text);                             // More fragments following flag (1 bit)
    entry67_text = gtk_entry_get_text (GTK_ENTRY (data->entry67));
    ip_icmp_flags[3] = atoi (entry67_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_icmp.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// More fragments following flag (1 bit)
gboolean
on_entry66_activate (GtkWidget *entry66, SPSData *data)
{
  const gchar *entry123_text, *entry65_text, *entry66_text, *entry67_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry66_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (ICMP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry66_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry66_text = gtk_entry_get_text (GTK_ENTRY (data->entry66));

// Is value too large or small?
  if ((atoi (entry66_text) < 0) || (atoi (entry66_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_icmp.ip_off) >> 13) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry66), value);
  } else {
    entry123_text = gtk_entry_get_text (GTK_ENTRY (data->entry123));    // Zero (1 bit)
    ip_icmp_flags[0] = atoi (entry123_text);
    entry65_text = gtk_entry_get_text (GTK_ENTRY (data->entry65));
    ip_icmp_flags[1] = atoi (entry65_text);                             // Do not fragment flag (1 bit)
    ip_icmp_flags[2] = atoi (entry66_text);                             // More fragments following flag (1 bit)
    entry67_text = gtk_entry_get_text (GTK_ENTRY (data->entry67));
    ip_icmp_flags[3] = atoi (entry67_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_icmp.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// Fragmentation offset (13 bits)
gboolean
on_entry67_activate (GtkWidget *entry67, SPSData *data)
{
  const gchar *entry123_text, *entry65_text, *entry66_text, *entry67_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry67_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry67_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry67_text = gtk_entry_get_text (GTK_ENTRY (data->entry67));

// Is value too large or small?
  if ((atoi (entry67_text) < 0) || (atoi (entry67_text) > 0x1000u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_off) & 0x1fffu);
    gtk_entry_set_text (GTK_ENTRY (data->entry67), value);
  } else {
    entry123_text = gtk_entry_get_text (GTK_ENTRY (data->entry123));    // Zero (1 bit)
    ip_icmp_flags[0] = atoi (entry123_text);
    entry65_text = gtk_entry_get_text (GTK_ENTRY (data->entry65));
    ip_icmp_flags[1] = atoi (entry65_text);                             // Do not fragment flag (1 bit)
    entry66_text = gtk_entry_get_text (GTK_ENTRY (data->entry66));
    ip_icmp_flags[2] = atoi (entry66_text);                             // More fragments following flag (1 bit)
    ip_icmp_flags[3] = atoi (entry67_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_icmp.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// Time-to-Live (8 bits)
gboolean
on_entry68_activate (GtkWidget *entry68, SPSData *data)
{
  const gchar *entry68_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry68_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry68_text = gtk_entry_get_text (GTK_ENTRY (data->entry68));

// Is value too large or small?
  if ((atoi (entry68_text) < 0) || (atoi (entry68_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_icmp.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry68), value);
  } else {
    data->ip4hdr_icmp.ip_ttl = atoi (entry68_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// Transport layer protocol (8 bits)
gboolean
on_entry75_activate (GtkWidget *entry75, SPSData *data)
{
  const gchar *entry75_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry75_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry75_text = gtk_entry_get_text (GTK_ENTRY (data->entry75));

// Is value too large or small?
  if ((atoi (entry75_text) < 0) || (atoi (entry75_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_icmp.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry75), value);
  } else {
    data->ip4hdr_icmp.ip_p = atoi (entry75_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (32 bits)
gboolean
on_entry63_activate (GtkWidget *entry63, SPSData *data)
{
  int status;
  const gchar *entry63_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry63_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry63_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry63_text = gtk_entry_get_text (GTK_ENTRY (data->entry63));

// Is it a valid IP address?
  if (is_valid_ip4 (entry63_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_src.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry63), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry63_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry63_text, &data->ip4hdr_icmp.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry63_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton9), FALSE);
    data->ran_icmp4_sourceip = 0;

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize source IP address
gboolean
on_checkbutton9_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton9_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in on_checkbutton9_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_icmp4_sourceip == 0) {
    data->ran_icmp4_sourceip = 1;
    inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_icmp.ip_src);
    inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry63), value);

    // IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  } else {
    data->ran_icmp4_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Destination IP address (32 bits)
gboolean
on_entry64_activate (GtkWidget *entry64, SPSData *data)
{
  int status;
  const gchar *entry64_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry64_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry64_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry64_text = gtk_entry_get_text (GTK_ENTRY (data->entry64));

// Is it a valid IP address?
  if (is_valid_ip4 (entry64_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_dst.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry64), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry64_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry64_text, &data->ip4hdr_icmp.ip_dst)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry64_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }
// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp.ip_sum = 0;
    data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// IP checksum (16 bits)
gboolean
on_entry62_activate (GtkWidget *entry62, SPSData *data)
{
  const gchar *entry62_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry62_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry62_text = gtk_entry_get_text (GTK_ENTRY (data->entry62));

// Is value too large or small?
  if ((atoi (entry62_text) < 0) || (atoi (entry62_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), value);
  } else {
    data->ip4hdr_icmp.ip_sum = htons (atoi (entry62_text));  // Update IPv4 header with new value
  }

  g_free (value);

  return (TRUE);
}

// ICMP header (IPv4)

// Message Type (8 bits)
gboolean
on_entry69_activate (GtkWidget *entry69, SPSData *data)
{
  const gchar *entry69_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry69_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry69_text = gtk_entry_get_text (GTK_ENTRY (data->entry69));

// Is value too large or small?
  if ((atoi (entry69_text) < 0) || (atoi (entry69_text) > 255)) {
    sprintf (value, "%u", data->icmp4hdr.icmp_type);
    gtk_entry_set_text (GTK_ENTRY (data->entry69), value);
  } else {
    data->icmp4hdr.icmp_type = atoi (entry69_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry71), value);
  }

  g_free (value);

  return (TRUE);
}

// Message Code (8 bits)
gboolean
on_entry70_activate (GtkWidget *entry70, SPSData *data)
{
  const gchar *entry70_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry70_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry70_text = gtk_entry_get_text (GTK_ENTRY (data->entry70));

// Is value too large or small?
  if ((atoi (entry70_text) < 0) || (atoi (entry70_text) > 255)) {
    sprintf (value, "%u", data->icmp4hdr.icmp_code);
    gtk_entry_set_text (GTK_ENTRY (data->entry70), value);
  } else {
    data->icmp4hdr.icmp_code = atoi (entry70_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry71), value);
  }

  g_free (value);

  return (TRUE);
}

// ICMP checksum (16 bits)
gboolean
on_entry71_activate (GtkWidget *entry71, SPSData *data)
{
  const gchar *entry71_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry71_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry71_text = gtk_entry_get_text (GTK_ENTRY (data->entry71));

// Is value too large or small?
  if ((atoi (entry71_text) < 0) || (atoi (entry71_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry71), value);
  } else {
    data->icmp4hdr.icmp_cksum = htons (atoi (entry71_text));  // Update ICMP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ID (16 bits)
gboolean
on_entry72_activate (GtkWidget *entry72, SPSData *data)
{
  const gchar *entry72_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry72_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry72_text = gtk_entry_get_text (GTK_ENTRY (data->entry72));

// Is value too large or small?
  if ((atoi (entry72_text) < 0) || (atoi (entry72_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry72), value);
  } else {
    data->icmp4hdr.icmp_id = htons (atoi (entry72_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry71), value);
  }

  g_free (value);

  return (TRUE);
}

// Sequence number (16 bits)
gboolean
on_entry73_activate (GtkWidget *entry73, SPSData *data)
{
  const gchar *entry73_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry73_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry73_text = gtk_entry_get_text (GTK_ENTRY (data->entry73));

// Is value too large or small?
  if ((atoi (entry73_text) < 0) || (atoi (entry73_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry73), value);
  } else {
    data->icmp4hdr.icmp_seq = htons (atoi (entry73_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry71), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton9_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp4 = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton10_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp4 = 1;

  return (TRUE);
}

// ICMP data from keyboard
gboolean
on_entry126_activate (GtkWidget *entry126, SPSData *data)
{
  int i, j, c;
  const gchar *entry126_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry126_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry126_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry126_text = gtk_entry_get_text (GTK_ENTRY (data->entry126));

// If user asked for no payload then zero it out.
  if (strlen (entry126_text) == 0) {
    data->icmp4_payloadlen = 0;
    memset (data->icmp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry126_text) != 0) && (data->ascii_hex_icmp4 == 0)) {

    data->icmp4_payloadlen = strlen (entry126_text);
    memset (data->icmp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
    memcpy (data->icmp4_payload, entry126_text, data->icmp4_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry126_text) != 0) && (data->ascii_hex_icmp4 == 1)) {
    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry126_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->icmp4_payloadlen = 0;
      memset (data->icmp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry126_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry126_text[i+j] != 32) && (entry126_text[i+j] != 44) && ((i + j) < strlen (entry126_text))) {
          value[j] = entry126_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->icmp4_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry126_activate().\n");
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->icmp4_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry126_text[i] == 32) || (entry126_text[i] == 44)) && (i < strlen (entry126_text))) {
          i++;
        }
      }  // Looping through entry126_text
    }  // else data is in acceptable format
  }  // else data is hex

// Total length of datagram (16 bits): IP header + ICMP header + ICMP payload
  data->ip4hdr_icmp.ip_len = htons (IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry60), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp.ip_sum = 0;
  data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry62), text);

// ICMP header checksum (16 bits):
  data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry71), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// ICMP data from file
gboolean
on_button7_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int icmp4_lim = IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button7_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button7_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button7_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > icmp4_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", icmp4_lim);
      data->parent = data->main_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->icmp4_payloadlen = index;

// Clear current data.
  memset (data->icmp4_payload, 0, icmp4_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry126), "");

// Copy payload from buffer.
  memcpy (data->icmp4_payload, buffer, data->icmp4_payloadlen);

// Total length of datagram (16 bits): IP header + ICMP header + ICMP payload
  data->ip4hdr_icmp.ip_len = htons (IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry60), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp.ip_sum = 0;
  data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry62), text);

// ICMP header checksum (16 bits):
  data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry71), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry126), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// ICMP page (IPv6)

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry128_activate (GtkWidget *entry128, SPSData *data)
{
  int i, bad;
  const gchar *entry128_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry128_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry128_text = gtk_entry_get_text (GTK_ENTRY (data->entry128));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry128_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry128_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry128_text[i] > 47) && (entry128_text[i] < 58)) ||    // 0 to 10
        ((entry128_text[i] > 64) && (entry128_text[i] < 71)) ||      // A to Z
        ((entry128_text[i] > 96) && (entry128_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry128_text[i] > 47) && (entry128_text[i] < 58)) ||    // 0 to 10
        ((entry128_text[i] > 64) && (entry128_text[i] < 71)) ||      // A to Z
        ((entry128_text[i] > 96) && (entry128_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp6ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry128), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry128_text, data->icmp6ethhdr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry129_activate (GtkWidget *entry129, SPSData *data)
{
  int i, bad;
  const gchar *entry129_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry129_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry129_text = gtk_entry_get_text (GTK_ENTRY (data->entry129));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry129_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry129_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry129_text[i] > 47) && (entry129_text[i] < 58)) ||    // 0 to 10
        ((entry129_text[i] > 64) && (entry129_text[i] < 71)) ||      // A to Z
        ((entry129_text[i] > 96) && (entry129_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry129_text[i] > 47) && (entry129_text[i] < 58)) ||    // 0 to 10
        ((entry129_text[i] > 64) && (entry129_text[i] < 71)) ||      // A to Z
        ((entry129_text[i] > 96) && (entry129_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp6ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry129), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry129_text, data->icmp6ethhdr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry153_activate (GtkWidget *entry153, SPSData *data)
{
  int i, sd, status;
  const gchar *entry153_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry153_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry153_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry153_text = gtk_entry_get_text (GTK_ENTRY (entry153));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "socket() failed to get socket descriptor for using ioctl() in on_entry153_activate().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry153_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry153_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->icmp6_interfacename, entry153_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->icmp6ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->icmp6ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry129), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry130_activate (GtkWidget *entry130, SPSData *data)
{
  const gchar *entry130_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry130_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry130_text = gtk_entry_get_text (GTK_ENTRY (data->entry130));

// Is value too large for a short int?
  if ((atoi (entry130_text) < 0) || (atoi (entry130_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry130), value);
  } else {
    data->icmp6ethhdr.type_code = htons (atoi (entry130_text));
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry110_activate (GtkWidget *entry110, SPSData *data)
{
  const gchar *entry110_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry110_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry110_text = gtk_entry_get_text (GTK_ENTRY (data->entry110));

// Save value currently in header
  old = ntohl (data->ip6hdr_icmp.ip6_flow);

// Is value too large or small?
  if ((atol (entry110_text) < 0) || (atol (entry110_text) > 15ul)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->ip6hdr_icmp.ip6_flow) >> 28);
    gtk_entry_set_text (GTK_ENTRY (data->entry110), value);
  } else {
    data->ip6hdr_icmp.ip6_flow = htonl (((atol (entry110_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Traffic class (8 bits)
gboolean
on_entry111_activate (GtkWidget *entry111, SPSData *data)
{
  const gchar *entry111_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry111_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry111_text = gtk_entry_get_text (GTK_ENTRY (data->entry111));

// Save value currently in header
  old = ntohl (data->ip6hdr_icmp.ip6_flow);

// Is value too large or small?
  if ((atol (entry111_text) < 0) || (atol (entry111_text) > 0xfful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry111), value);
  } else {
    data->ip6hdr_icmp.ip6_flow = htonl (((atol (entry111_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Flow label (20 bits)
gboolean
on_entry112_activate (GtkWidget *entry112, SPSData *data)
{
  const gchar *entry112_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry112_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry112_text = gtk_entry_get_text (GTK_ENTRY (data->entry112));

// Save value currently in header
  old = ntohl (data->ip6hdr_icmp.ip6_flow);

// Is value too large or small?
  if ((atol (entry112_text) < 0) || (atol (entry112_text) > 0xfffffu)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) & 0xfffffu));
    gtk_entry_set_text (GTK_ENTRY (data->entry112), value);
  } else {
    data->ip6hdr_icmp.ip6_flow = htonl (atol (entry112_text) + (old & 0xfff00000ul));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Payload length (16 bits)
gboolean
on_entry113_activate (GtkWidget *entry113, SPSData *data)
{
  const gchar *entry113_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry113_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry113_text = gtk_entry_get_text (GTK_ENTRY (data->entry113));

// Is value too large or small?
  if ((atoi (entry113_text) < 0) || (atoi (entry113_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry113), value);
  } else {
    data->ip6hdr_icmp.ip6_plen = htons (atoi (entry113_text));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Next header (8 bits)
gboolean
on_entry114_activate (GtkWidget *entry114, SPSData *data)
{
  const gchar *entry114_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry114_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry114_text = gtk_entry_get_text (GTK_ENTRY (data->entry114));

// Is value too large or small?
  if ((atoi (entry114_text) < 0) || (atoi (entry114_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_icmp.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry114), value);
  } else {
    data->ip6hdr_icmp.ip6_nxt = atoi (entry114_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Hop limit (8 bits)
gboolean
on_entry115_activate (GtkWidget *entry115, SPSData *data)
{
  const gchar *entry115_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry115_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry115_text = gtk_entry_get_text (GTK_ENTRY (data->entry115));

// Is value too large or small?
  if ((atoi (entry115_text) < 0) || (atoi (entry115_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_icmp.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry115), value);
  } else {
    data->ip6hdr_icmp.ip6_hops = atoi (entry115_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (128 bits)
gboolean
on_entry109_activate (GtkWidget *entry109, SPSData *data)
{
  const gchar *entry109_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry109_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry109_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry109_text = gtk_entry_get_text (GTK_ENTRY (data->entry109));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry109_text, &(data->ip6hdr_icmp.ip6_src)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry109), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry109_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton10), FALSE);
    data->ran_icmp6_sourceip = 0;

// ICMP header checksum (16 bits)
    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize source IP address
gboolean
on_checkbutton10_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton10_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in on_checkbutton10_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_icmp6_sourceip == 0) {
    data->ran_icmp6_sourceip = 1;
    inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_icmp.ip6_src);
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry109), value);

    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);
  } else {
    data->ran_icmp6_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Destination IP address (128 bits)
gboolean
on_entry116_activate (GtkWidget *entry116, SPSData *data)
{
  const gchar *entry116_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry116_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry116_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry116_text = gtk_entry_get_text (GTK_ENTRY (data->entry116));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry116_text, &(data->ip6hdr_icmp.ip6_dst)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_dst), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry116), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry116_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
 } else { 
// ICMP header checksum (16 bits)
    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// ICMP header (IPv6)

// Message Type (8 bits)
gboolean
on_entry117_activate (GtkWidget *entry117, SPSData *data)
{
  const gchar *entry117_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry117_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry117_text = gtk_entry_get_text (GTK_ENTRY (data->entry117));

// Is value too large or small?
  if ((atoi (entry117_text) < 0) || (atoi (entry117_text) > 255)) {
    sprintf (value, "%u", data->icmp6hdr.icmp6_type);
    gtk_entry_set_text (GTK_ENTRY (data->entry117), value);
  } else {
    data->icmp6hdr.icmp6_type = atoi (entry117_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);
  }

  g_free (value);

  return (TRUE);
}

// Message Code (8 bits)
gboolean
on_entry118_activate (GtkWidget *entry118, SPSData *data)
{
  const gchar *entry118_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry118_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry118_text = gtk_entry_get_text (GTK_ENTRY (data->entry118));

// Is value too large or small?
  if ((atoi (entry118_text) < 0) || (atoi (entry118_text) > 255)) {
    sprintf (value, "%u", data->icmp6hdr.icmp6_code);
    gtk_entry_set_text (GTK_ENTRY (data->entry118), value);
  } else {
    data->icmp6hdr.icmp6_code = atoi (entry118_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);
  }

  g_free (value);

  return (TRUE);
}

// ICMP checksum (16 bits)
gboolean
on_entry119_activate (GtkWidget *entry119, SPSData *data)
{
  const gchar *entry119_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry119_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry119_text = gtk_entry_get_text (GTK_ENTRY (data->entry119));

// Is value too large or small?
  if ((atoi (entry119_text) < 0) || (atoi (entry119_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);
  } else {
    data->icmp6hdr.icmp6_cksum = htons (atoi (entry119_text));  // Update ICMP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ID (16 bits)
gboolean
on_entry120_activate (GtkWidget *entry120, SPSData *data)
{
  const gchar *entry120_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry120_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry120_text = gtk_entry_get_text (GTK_ENTRY (data->entry120));

// Is value too large or small?
  if ((atoi (entry120_text) < 0) || (atoi (entry120_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry120), value);
  } else {
    data->icmp6hdr.icmp6_id = htons (atoi (entry120_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);
  }

  g_free (value);

  return (TRUE);
}

// Sequence number (16 bits)
gboolean
on_entry121_activate (GtkWidget *entry121, SPSData *data)
{
  const gchar *entry121_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry121_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry121_text = gtk_entry_get_text (GTK_ENTRY (data->entry121));

// Is value too large or small?
  if ((atoi (entry121_text) < 0) || (atoi (entry121_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry121), value);
  } else {
    data->icmp6hdr.icmp6_seq = htons (atoi (entry121_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton11_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp6 = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton12_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp6 = 1;

  return (TRUE);
}

// ICMP data from keyboard
gboolean
on_entry127_activate (GtkWidget *entry127, SPSData *data)
{
  int i, j, c;
  const gchar *entry127_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry127_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry127_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry127_text = gtk_entry_get_text (GTK_ENTRY (data->entry127));

// If user asked for no payload then zero it out.
  if (strlen (entry127_text) == 0) {
    data->icmp4_payloadlen = 0;
    memset (data->icmp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry127_text) != 0) && (data->ascii_hex_icmp6 == 0)) {

    data->icmp6_payloadlen = strlen (entry127_text);
    memset (data->icmp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
    memcpy (data->icmp6_payload, entry127_text, data->icmp6_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry127_text) != 0) && (data->ascii_hex_icmp6 == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry127_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->icmp6_payloadlen = 0;
      memset (data->icmp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry127_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry127_text[i+j] != 32) && (entry127_text[i+j] != 44) && ((i + j) < strlen (entry127_text))) {
          value[j] = entry127_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->icmp6_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry127_activate().\n");
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->icmp6_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry127_text[i] == 32) || (entry127_text[i] == 44)) && (i < strlen (entry127_text))) {
          i++;
        }
      }  // Looping through entry127_text
    }  // else data is in acceptable format
  }  // else data is hex

// Update IPv6 header payload length = ICMP header + ICMP data
  data->ip6hdr_icmp.ip6_plen = htons (ICMP_HDRLEN + data->icmp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry113), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_icmp.ip_len = htons (ICMP_HDRLEN + data->icmp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// ICMP header checksum (16 bits)
  data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry119), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// ICMP data from file
gboolean
on_button9_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int icmp6_lim = IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button9_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button9_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button9_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > icmp6_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", icmp6_lim);
      data->parent = data->main_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->icmp6_payloadlen = index;

// Clear current data.
  memset (data->icmp6_payload, 0, icmp6_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry127), "");

// Copy payload from buffer.
  memcpy (data->icmp6_payload, buffer, data->icmp6_payloadlen);

// Update IPv6 header payload length = ICMP header + ICMP data
  data->ip6hdr_icmp.ip6_plen = htons (ICMP_HDRLEN + data->icmp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry113), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_icmp.ip_len = htons (ICMP_HDRLEN + data->icmp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// ICMP header checksum (16 bits)
  data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry119), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry127), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// UDP page (IPv4)

// Specify ethernet header
gboolean
on_checkbutton19_toggled (GtkButton *button, SPSData *data)
{
  if (data->udp4_specify_ether == 0) {
    data->udp4_specify_ether = 1;
  } else {
    data->udp4_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry147_activate (GtkWidget *entry147, SPSData *data)
{
  int i, bad;
  const gchar *entry147_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry147_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry147_text = gtk_entry_get_text (GTK_ENTRY (data->entry147));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry147_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry147_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry147_text[i] > 47) && (entry147_text[i] < 58)) ||    // 0 to 10
        ((entry147_text[i] > 64) && (entry147_text[i] < 71)) ||      // A to Z
        ((entry147_text[i] > 96) && (entry147_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry147_text[i] > 47) && (entry147_text[i] < 58)) ||    // 0 to 10
        ((entry147_text[i] > 64) && (entry147_text[i] < 71)) ||      // A to Z
        ((entry147_text[i] > 96) && (entry147_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp4ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry147), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry147_text, data->udp4ethhdr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry148_activate (GtkWidget *entry148, SPSData *data)
{
  int i, bad;
  const gchar *entry148_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry148_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry148_text = gtk_entry_get_text (GTK_ENTRY (data->entry148));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry148_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry148_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry148_text[i] > 47) && (entry148_text[i] < 58)) ||    // 0 to 10
        ((entry148_text[i] > 64) && (entry148_text[i] < 71)) ||      // A to Z
        ((entry148_text[i] > 96) && (entry148_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry148_text[i] > 47) && (entry148_text[i] < 58)) ||    // 0 to 10
        ((entry148_text[i] > 64) && (entry148_text[i] < 71)) ||      // A to Z
        ((entry148_text[i] > 96) && (entry148_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp4ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry148), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry148_text, data->udp4ethhdr.src_mac);
  }

  g_free (value);

  return (TRUE);
}


// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry152_activate (GtkWidget *entry152, SPSData *data)
{
  int i, sd, status;
  const gchar *entry152_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry152_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry152_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry152_text = gtk_entry_get_text (GTK_ENTRY (entry152));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "Iin callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry152_activate().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry152_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry152_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->udp4_interfacename, entry152_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->udp4ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->udp4ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry148), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry149_activate (GtkWidget *entry149, SPSData *data)
{
  const gchar *entry149_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry149_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry149_text = gtk_entry_get_text (GTK_ENTRY (data->entry149));

// Is value too large for a short int?
  if ((atoi (entry149_text) < 0) || (atoi (entry149_text) > 65535)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry149), value);
  } else {
    data->udp4ethhdr.type_code = htons (atoi (entry149_text));
  }

  g_free (value);

  return (TRUE);
}

// IP header length (4 bits)
gboolean
on_entry1_activate (GtkWidget *entry1, SPSData *data)
{
  const gchar *entry1_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry1_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry1_text = gtk_entry_get_text (GTK_ENTRY (data->entry1));

// Is value too large or small?
  if ((atoi (entry1_text) < 0) || (atoi (entry1_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_udp.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry1), value);
  } else {
    data->ip4hdr_udp.ip_hl = atoi (entry1_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// IPv4 version (4 bits)
gboolean
on_entry2_activate (GtkWidget *entry2, SPSData *data)
{
  const gchar *entry2_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry2_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry2_text = gtk_entry_get_text (GTK_ENTRY (data->entry2));

// Is value too large or small?
  if ((atoi (entry2_text) < 0) || (atoi (entry2_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_udp.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry2), value);
  } else {
    data->ip4hdr_udp.ip_v = atoi (entry2_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// Type of service (8 bits)
gboolean
on_entry3_activate (GtkWidget *entry3, SPSData *data)
{
  const gchar *entry3_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry3_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry3_text = gtk_entry_get_text (GTK_ENTRY (data->entry3));

// Is value too large or small?
  if ((atoi (entry3_text) < 0) || (atoi (entry3_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_udp.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry3), value);
  } else {
    data->ip4hdr_udp.ip_tos = atoi (entry3_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// Total length of datagram (16 bits)
gboolean
on_entry4_activate (GtkWidget *entry4, SPSData *data)
{
  const gchar *entry4_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry4_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry4_text = gtk_entry_get_text (GTK_ENTRY (data->entry4));

// Is value too large or small?
  if ((atoi (entry4_text) < 0) || (atoi (entry4_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry4), value);
  } else {
    data->ip4hdr_udp.ip_len = htons (atoi (entry4_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// ID sequence number (16 bits)
gboolean
on_entry5_activate (GtkWidget *entry5, SPSData *data)
{
  const gchar *entry5_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry5_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry5_text = gtk_entry_get_text (GTK_ENTRY (data->entry5));

// Is value too large or small?
  if ((atoi (entry5_text) < 0) || (atoi (entry5_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry5), value);
  } else {
    data->ip4hdr_udp.ip_id = htons (atoi (entry5_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// Zero (1 bit)
gboolean
on_entry124_activate (GtkWidget *entry124, SPSData *data)
{
  const gchar *entry124_text, *entry9_text, *entry10_text, *entry11_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry124_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (UDP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry124_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry124_text = gtk_entry_get_text (GTK_ENTRY (data->entry124));

// Is value too large or small?
  if ((atoi (entry124_text) < 0) || (atoi (entry124_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry124), value);
  } else {
    ip_udp_flags[0] = atoi (entry124_text);                            // Zero (1 bit)
    entry9_text = gtk_entry_get_text (GTK_ENTRY (data->entry9));
    ip_udp_flags[1] = atoi (entry9_text);                              // Do not fragment flag (1 bit)
    entry10_text = gtk_entry_get_text (GTK_ENTRY (data->entry10));
    ip_udp_flags[2] = atoi (entry10_text);                             // More fragments following flag (1 bit)
    entry11_text = gtk_entry_get_text (GTK_ENTRY (data->entry11));
    ip_udp_flags[3] = atoi (entry11_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_udp.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// Do not fragment flag (1 bit)
gboolean
on_entry9_activate (GtkWidget *entry9, SPSData *data)
{
  const gchar *entry124_text, *entry9_text, *entry10_text, *entry11_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry9_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (UDP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry9_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry9_text = gtk_entry_get_text (GTK_ENTRY (data->entry9));

// Is value too large or small?
  if ((atoi (entry9_text) < 0) || (atoi (entry9_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_udp.ip_off) >> 14) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry9), value);
  } else {
    entry124_text = gtk_entry_get_text (GTK_ENTRY (data->entry124));
    ip_udp_flags[0] = atoi (entry124_text);                            // Zero (1 bit)
    ip_udp_flags[1] = atoi (entry9_text);                              // Do not fragment flag (1 bit)
    entry10_text = gtk_entry_get_text (GTK_ENTRY (data->entry10));
    ip_udp_flags[2] = atoi (entry10_text);                             // More fragments following flag (1 bit)
    entry11_text = gtk_entry_get_text (GTK_ENTRY (data->entry11));
    ip_udp_flags[3] = atoi (entry11_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_udp.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// More fragments following flag (1 bit)
gboolean
on_entry10_activate (GtkWidget *entry10, SPSData *data)
{
  const gchar *entry124_text, *entry9_text, *entry10_text, *entry11_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry10_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (UDP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry10_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry10_text = gtk_entry_get_text (GTK_ENTRY (data->entry10));

// Is value too large or small?
  if ((atoi (entry10_text) < 0) || (atoi (entry10_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_udp.ip_off) >> 13) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry10), value);
  } else {
    entry124_text = gtk_entry_get_text (GTK_ENTRY (data->entry124));
    ip_udp_flags[0] = atoi (entry124_text);                            // Zero (1 bit)
    entry9_text = gtk_entry_get_text (GTK_ENTRY (data->entry9));
    ip_udp_flags[1] = atoi (entry9_text);                              // Do not fragment flag (1 bit)
    ip_udp_flags[2] = atoi (entry10_text);                             // More fragments following flag (1 bit)
    entry11_text = gtk_entry_get_text (GTK_ENTRY (data->entry11));
    ip_udp_flags[3] = atoi (entry11_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_udp.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// Fragmentation offset (13 bits)
gboolean
on_entry11_activate (GtkWidget *entry11, SPSData *data)
{
  const gchar *entry124_text, *entry9_text, *entry10_text, *entry11_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry11_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry11_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry11_text = gtk_entry_get_text (GTK_ENTRY (data->entry11));

// Is value too large or small?
  if ((atoi (entry11_text) < 0) || (atoi (entry11_text) > 0x1000u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_off) & 0x1fffu);
    gtk_entry_set_text (GTK_ENTRY (data->entry11), value);
  } else {
    entry124_text = gtk_entry_get_text (GTK_ENTRY (data->entry124));
    ip_udp_flags[0] = atoi (entry124_text);                            // Zero (1 bit)
    entry9_text = gtk_entry_get_text (GTK_ENTRY (data->entry9));
    ip_udp_flags[1] = atoi (entry9_text);                              // Do not fragment flag (1 bit)
    entry10_text = gtk_entry_get_text (GTK_ENTRY (data->entry10));
    ip_udp_flags[2] = atoi (entry10_text);                             // More fragments following flag (1 bit)
    ip_udp_flags[3] = atoi (entry11_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_udp.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// Time-to-Live (8 bits)
gboolean
on_entry12_activate (GtkWidget *entry12, SPSData *data)
{
  const gchar *entry12_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry12_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry12_text = gtk_entry_get_text (GTK_ENTRY (data->entry12));

// Is value too large or small?
  if ((atoi (entry12_text) < 0) || (atoi (entry12_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_udp.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry12), value);
  } else {
    data->ip4hdr_udp.ip_ttl = atoi (entry12_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// Transport layer protocol (8 bits)
gboolean
on_entry76_activate (GtkWidget *entry76, SPSData *data)
{
  const gchar *entry76_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry76_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry76_text = gtk_entry_get_text (GTK_ENTRY (data->entry76));

// Is value too large or small?
  if ((atoi (entry76_text) < 0) || (atoi (entry76_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_udp.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry76), value);
  } else {
    data->ip4hdr_udp.ip_p = atoi (entry76_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (32 bits)
gboolean
on_entry7_activate (GtkWidget *entry7, SPSData *data)
{
  int status;
  const gchar *entry7_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry7_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry7_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry7_text = gtk_entry_get_text (GTK_ENTRY (data->entry7));

// Is it a valid IP address?
  if (is_valid_ip4 (entry7_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_src.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry7), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry7_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry7_text, &data->ip4hdr_udp.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry7_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton1), FALSE);
    data->ran_udp4_sourceip = 0;

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize source IP address
gboolean
on_checkbutton1_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton1_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in on_checkbutton1_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_udp4_sourceip == 0) {
    data->ran_udp4_sourceip = 1;
    inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_udp.ip_src);
    inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry7), value);

    // IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

    // UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  } else {
    data->ran_udp4_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Destination IP address (32 bits)
gboolean
on_entry8_activate (GtkWidget *entry8, SPSData *data)
{
  int status;
  const gchar *entry8_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry8_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry8_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry8_text = gtk_entry_get_text (GTK_ENTRY (data->entry8));

// Is it a valid IP address?
  if (is_valid_ip4 (entry8_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_dst.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry8), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry8_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry8_text, &data->ip4hdr_udp.ip_dst)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry8_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp.ip_sum = 0;
    data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// IP checksum (16 bits)
gboolean
on_entry6_activate (GtkWidget *entry6, SPSData *data)
{
  const gchar *entry6_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry6_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry6_text = gtk_entry_get_text (GTK_ENTRY (data->entry6));

// Is value too large or small?
  if ((atoi (entry6_text) < 0) || (atoi (entry6_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), value);
  } else {
    data->ip4hdr_udp.ip_sum = htons (atoi (entry6_text));  // Update IPv4 header with new value

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// UDP header (IPv4)

// UDP source port (16 bits)
gboolean
on_entry13_activate (GtkWidget *entry13, SPSData *data)
{
  const gchar *entry13_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry13_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry13_text = gtk_entry_get_text (GTK_ENTRY (data->entry13));

// Is value too large or small?
  if ((atoi (entry13_text) < 0) || (atoi (entry13_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry13), value);
  } else {
    data->udp4hdr.uh_sport = htons (atoi (entry13_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton3), FALSE);
    data->ran_udp4_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize UDP source port address
gboolean
on_checkbutton3_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton3_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_udp4_sourceport == 0) {
    data->ran_udp4_sourceport = 1;
    data->udp4hdr.uh_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry13), value);

// UDP checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  } else {
    data->ran_udp4_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// UDP destination port (16 bits)
gboolean
on_entry14_activate (GtkWidget *entry14, SPSData *data)
{
  const gchar *entry14_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry14_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry14_text = gtk_entry_get_text (GTK_ENTRY (data->entry14));

// Is value too large or small?
  if ((atoi (entry14_text) < 0) || (atoi (entry14_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry14), value);
  } else {
    data->udp4hdr.uh_dport = htons (atoi (entry14_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// Length of UDP datagram (16 bits)
gboolean
on_entry15_activate (GtkWidget *entry15, SPSData *data)
{
  const gchar *entry15_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry15_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry15_text = gtk_entry_get_text (GTK_ENTRY (data->entry15));

// Is value too large or small?
  if ((atoi (entry15_text) < 0) || (atoi (entry15_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_ulen));
    gtk_entry_set_text (GTK_ENTRY (data->entry15), value);
  } else {
    data->udp4hdr.uh_ulen = htons (atoi (entry15_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  }

  g_free (value);

  return (TRUE);
}

// UDP checksum (16 bits)
gboolean
on_entry16_activate (GtkWidget *entry16, SPSData *data)
{
  const gchar *entry16_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry16_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry16_text = gtk_entry_get_text (GTK_ENTRY (data->entry16));

// Is value too large or small?
  if ((atoi (entry16_text) < 0) || (atoi (entry16_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), value);
  } else {
    data->udp4hdr.uh_sum = htons (atoi (entry16_text));  // Update UDP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton13_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp4 = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton14_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp4 = 1;

  return (TRUE);
}

// UDP data from keyboard
gboolean
on_entry17_activate (GtkWidget *entry17, SPSData *data)
{
  int i, j, c;
  const gchar *entry17_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry17_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry17_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry17_text = gtk_entry_get_text (GTK_ENTRY (data->entry17));

// If user asked for no payload then zero it out.
  if (strlen (entry17_text) == 0) {
    data->udp4_payloadlen = 0;
    memset (data->udp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry17_text) != 0) && (data->ascii_hex_udp4 == 0)) {

    data->udp4_payloadlen = strlen (entry17_text);
    memset (data->udp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
    memcpy (data->udp4_payload, entry17_text, data->udp4_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry17_text) != 0) && (data->ascii_hex_udp4 == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry17_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->udp4_payloadlen = 0;
      memset (data->udp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry17_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry17_text[i+j] != 32) && (entry17_text[i+j] != 44) && ((i + j) < strlen (entry17_text))) {
          value[j] = entry17_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->udp4_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry17_activate().\n");
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->udp4_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry17_text[i] == 32) || (entry17_text[i] == 44)) && (i < strlen (entry17_text))) {
          i++;
        }
      }  // Looping through entry17_text
    }  // else data is in acceptable format
  }  // else data is hex

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp4hdr.uh_ulen = htons (UDP_HDRLEN + data->udp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry15), text);

// Total length of datagram (16 bits): IP header + UDP header + UDP payload
  data->ip4hdr_udp.ip_len = htons (IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry4), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp.ip_sum = 0;
  data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry6), text);

// UDP header checksum (16 bits)
  data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry16), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// UDP data from file
gboolean
on_button8_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int udp4_lim = IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button8_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button8_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button8_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > udp4_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", udp4_lim);
      data->parent = data->main_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->udp4_payloadlen = index;

// Clear current data.
  memset (data->udp4_payload, 0, udp4_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry17), "");

// Copy payload from buffer.
  memcpy (data->udp4_payload, buffer, data->udp4_payloadlen);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp4hdr.uh_ulen = htons (UDP_HDRLEN + data->udp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry15), text);

// Total length of datagram (16 bits): IP header + UDP header + UDP payload
  data->ip4hdr_udp.ip_len = htons (IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry4), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp.ip_sum = 0;
  data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry6), text);

// UDP header checksum (16 bits)
  data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry16), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry17), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// UDP page (IPv6)

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry134_activate (GtkWidget *entry134, SPSData *data)
{
  int i, bad;
  const gchar *entry134_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry134_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry134_text = gtk_entry_get_text (GTK_ENTRY (data->entry134));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry134_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry134_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry134_text[i] > 47) && (entry134_text[i] < 58)) ||    // 0 to 10
        ((entry134_text[i] > 64) && (entry134_text[i] < 71)) ||      // A to Z
        ((entry134_text[i] > 96) && (entry134_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry134_text[i] > 47) && (entry134_text[i] < 58)) ||    // 0 to 10
        ((entry134_text[i] > 64) && (entry134_text[i] < 71)) ||      // A to Z
        ((entry134_text[i] > 96) && (entry134_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp6ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry134), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry134_text, data->udp6ethhdr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry135_activate (GtkWidget *entry135, SPSData *data)
{
  int i, bad;
  const gchar *entry135_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry135_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry135_text = gtk_entry_get_text (GTK_ENTRY (data->entry135));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry135_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry135_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry135_text[i] > 47) && (entry135_text[i] < 58)) ||    // 0 to 10
        ((entry135_text[i] > 64) && (entry135_text[i] < 71)) ||      // A to Z
        ((entry135_text[i] > 96) && (entry135_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry135_text[i] > 47) && (entry135_text[i] < 58)) ||    // 0 to 10
        ((entry135_text[i] > 64) && (entry135_text[i] < 71)) ||      // A to Z
        ((entry135_text[i] > 96) && (entry135_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp6ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry135), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry135_text, data->udp6ethhdr.src_mac);
  }

  g_free (value);

  return (TRUE);
}


// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry155_activate (GtkWidget *entry155, SPSData *data)
{
  int i, sd, status;
  const gchar *entry155_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry155_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry155_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry155_text = gtk_entry_get_text (GTK_ENTRY (entry155));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text,  "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry155_activate.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry155_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry155_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->udp6_interfacename, entry155_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->udp6ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->udp6ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry135), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry136_activate (GtkWidget *entry136, SPSData *data)
{
  const gchar *entry136_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry136_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry136_text = gtk_entry_get_text (GTK_ENTRY (data->entry136));

// Is value too large for a short int?
  if ((atoi (entry136_text) < 0) || (atoi (entry136_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry136), value);
  } else {
    data->udp6ethhdr.type_code = htons (atoi (entry136_text));
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry97_activate (GtkWidget *entry97, SPSData *data)
{
  const gchar *entry97_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry97_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry97_text = gtk_entry_get_text (GTK_ENTRY (data->entry97));

// Save value currently in header
  old = ntohl (data->ip6hdr_udp.ip6_flow);

// Is value too large or small?
  if ((atol (entry97_text) < 0) || (atol (entry97_text) > 15ul)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->ip6hdr_udp.ip6_flow) >> 28);
    gtk_entry_set_text (GTK_ENTRY (data->entry97), value);
  } else {
    data->ip6hdr_udp.ip6_flow = htonl (((atol (entry97_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Traffic class (8 bits)
gboolean
on_entry98_activate (GtkWidget *entry98, SPSData *data)
{
  const gchar *entry98_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry98_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry98_text = gtk_entry_get_text (GTK_ENTRY (data->entry98));

// Save value currently in header
  old = ntohl (data->ip6hdr_udp.ip6_flow);

// Is value too large or small?
  if ((atol (entry98_text) < 0) || (atol (entry98_text) > 0xfful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry98), value);
  } else {
    data->ip6hdr_udp.ip6_flow = htonl (((atol (entry98_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Flow label (20 bits)
gboolean
on_entry99_activate (GtkWidget *entry99, SPSData *data)
{
  const gchar *entry99_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry99_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry99_text = gtk_entry_get_text (GTK_ENTRY (data->entry99));

// Save value currently in header
  old = ntohl (data->ip6hdr_udp.ip6_flow);

// Is value too large or small?
  if ((atol (entry99_text) < 0) || (atol (entry99_text) > 0xffffful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry99), value);
  } else {
    data->ip6hdr_udp.ip6_flow = htonl (atol (entry99_text) + (old & 0xfff00000ul));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Payload length (16 bits)
gboolean
on_entry100_activate (GtkWidget *entry100, SPSData *data)
{
  const gchar *entry100_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry100_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry100_text = gtk_entry_get_text (GTK_ENTRY (data->entry100));

// Is value too large or small?
  if ((atoi (entry100_text) < 0) || (atoi (entry100_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry100), value);
  } else {
    data->ip6hdr_udp.ip6_plen = htons (atoi (entry100_text));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Next header (8 bits)
gboolean
on_entry101_activate (GtkWidget *entry101, SPSData *data)
{
  const gchar *entry101_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry101_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry101_text = gtk_entry_get_text (GTK_ENTRY (data->entry101));

// Is value too large or small?
  if ((atoi (entry101_text) < 0) || (atoi (entry101_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_udp.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry101), value);
  } else {
    data->ip6hdr_udp.ip6_nxt = atoi (entry101_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Hop limit (8 bits)
gboolean
on_entry102_activate (GtkWidget *entry102, SPSData *data)
{
  const gchar *entry102_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry102_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry102_text = gtk_entry_get_text (GTK_ENTRY (data->entry102));

// Is value too large or small?
  if ((atoi (entry102_text) < 0) || (atoi (entry102_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_udp.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry102), value);
  } else {
    data->ip6hdr_udp.ip6_hops = atoi (entry102_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (128 bits)
gboolean
on_entry96_activate (GtkWidget *entry96, SPSData *data)
{
  const gchar *entry96_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry96_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry96_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry96_text = gtk_entry_get_text (GTK_ENTRY (data->entry96));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry96_text, &(data->ip6hdr_udp.ip6_src)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry96), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry96_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton6), FALSE);
    data->ran_udp6_sourceip = 0;

// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize source IP address
gboolean
on_checkbutton6_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton6_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in on_checkbutton6_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_udp6_sourceip == 0) {
    data->ran_udp6_sourceip = 1;
    inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_udp.ip6_src);
    inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry96), value);

    // UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);
  } else {
    data->ran_udp6_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Destination IP address (128 bits)
gboolean
on_entry103_activate (GtkWidget *entry103, SPSData *data)
{
  const gchar *entry103_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry103_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry103_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry103_text = gtk_entry_get_text (GTK_ENTRY (data->entry103));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry103_text, &(data->ip6hdr_udp.ip6_dst)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_dst), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry103), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry103_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// UDP header (IPv6)

// UDP source port (16 bits)
gboolean
on_entry104_activate (GtkWidget *entry104, SPSData *data)
{
  const gchar *entry104_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry104_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry104_text = gtk_entry_get_text (GTK_ENTRY (data->entry104));

// Is value too large or small?
  if ((atoi (entry104_text) < 0) || (atoi (entry104_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry104), value);
  } else {
    data->udp6hdr.uh_sport = htons (atoi (entry104_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton8), FALSE);
    data->ran_udp6_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize UDP source port address
gboolean
on_checkbutton8_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton8_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_udp6_sourceport == 0) {
    data->ran_udp6_sourceport = 1;
    data->udp6hdr.uh_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry104), value);

// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);
  } else {
    data->ran_udp6_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// UDP destination port (16 bits)
gboolean
on_entry105_activate (GtkWidget *entry105, SPSData *data)
{
  const gchar *entry105_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry105_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry105_text = gtk_entry_get_text (GTK_ENTRY (data->entry105));

// Is value too large or small?
  if ((atoi (entry105_text) < 0) || (atoi (entry105_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry105), value);
  } else {
    data->udp6hdr.uh_dport = htons (atoi (entry105_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);
  }

  g_free (value);

  return (TRUE);
}

// Length of UDP datagram (16 bits)
gboolean
on_entry106_activate (GtkWidget *entry106, SPSData *data)
{
  const gchar *entry106_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry106_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry106_text = gtk_entry_get_text (GTK_ENTRY (data->entry106));

// Is value too large or small?
  if ((atoi (entry106_text) < 0) || (atoi (entry106_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_ulen));
    gtk_entry_set_text (GTK_ENTRY (data->entry106), value);
  } else {
    data->udp6hdr.uh_ulen = htons (atoi (entry106_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);
  }

  g_free (value);

  return (TRUE);
}

// UDP checksum (16 bits)
gboolean
on_entry107_activate (GtkWidget *entry107, SPSData *data)
{
  const gchar *entry107_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry107_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry107_text = gtk_entry_get_text (GTK_ENTRY (data->entry107));

// Is value too large or small?
  if ((atoi (entry107_text) < 0) || (atoi (entry107_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), value);
  } else {
    data->udp6hdr.uh_sum = htons (atoi (entry107_text));  // Update UDP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton15_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp6 = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton16_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp6 = 1;

  return (TRUE);
}

// UDP data from keyboard
gboolean
on_entry108_activate (GtkWidget *entry108, SPSData *data)
{
  int i, j, c;
  const gchar *entry108_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry108_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry108_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry108_text = gtk_entry_get_text (GTK_ENTRY (data->entry108));

// If user asked for no payload then zero it out.
  if (strlen (entry108_text) == 0) {
    data->udp6_payloadlen = 0;
    memset (data->udp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry108_text) != 0) && (data->ascii_hex_udp6 == 0)) {

    data->udp6_payloadlen = strlen (entry108_text);
    memset (data->udp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
    memcpy (data->udp6_payload, entry108_text, data->udp6_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry108_text) != 0) && (data->ascii_hex_udp6 == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry108_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->udp6_payloadlen = 0;
      memset (data->udp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry108_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry108_text[i+j] != 32) && (entry108_text[i+j] != 44) && ((i + j) < strlen (entry108_text))) {
          value[j] = entry108_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->udp6_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry108_activate().\n");
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->udp6_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry108_text[i] == 32) || (entry108_text[i] == 44)) && (i < strlen (entry108_text))) {
          i++;
        }
      }  // Looping through entry108_text
    }  // else data is in acceptable format
  }  // else data is hex

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp6hdr.uh_ulen = htons (UDP_HDRLEN + data->udp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry106), text);

// IPv6 header payload length (16 bits): UDP header + UDP payload
  data->ip6hdr_udp.ip6_plen = htons (UDP_HDRLEN + data->udp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry100), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_udp.ip_len = htons (UDP_HDRLEN + data->udp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// UDP header checksum (16 bits)
  data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry107), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// UDP data from file
gboolean
on_button10_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int udp6_lim = IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button10_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button10_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button10_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > udp6_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", udp6_lim);
      data->parent = data->main_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->udp6_payloadlen = index;

// Clear current data.
  memset (data->udp6_payload, 0, udp6_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry108), "");

// Copy payload from buffer.
  memcpy (data->udp6_payload, buffer, data->udp6_payloadlen);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp6hdr.uh_ulen = htons (UDP_HDRLEN + data->udp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry106), text);

// IPv6 header payload length (16 bits): UDP header + UDP payload
  data->ip6hdr_udp.ip6_plen = htons (UDP_HDRLEN + data->udp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry100), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_udp.ip_len = htons (UDP_HDRLEN + data->udp6_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// UDP header checksum (16 bits)
  data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry107), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry108), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// Send page entries

// TCP (IPv4) packet type
gboolean
on_radiobutton1_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 1;

  return (TRUE);
}

// ICMP (IPv4) packet type
gboolean
on_radiobutton2_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 2;

  return (TRUE);
}

// UDP (IPv4) packet type
gboolean
on_radiobutton3_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 3;

  return (TRUE);
}

// All (IPv4) packet types (i.e., cycle)
gboolean
on_radiobutton4_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 4;

  return (TRUE);
}

// TCP (IPv6) packet type
gboolean
on_radiobutton5_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 5;

  return (TRUE);
}

// ICMP (IPv6) packet type
gboolean
on_radiobutton6_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 6;

  return (TRUE);
}

// UDP (IPv6) packet type
gboolean
on_radiobutton7_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 7;

  return (TRUE);
}

// All (IPv6) packet types (i.e., cycle)
gboolean
on_radiobutton8_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type = 8;

  return (TRUE);
}

// Tunnel IPv6 over IPv4
gboolean
on_checkbutton12_toggled (GtkButton *button, SPSData *data)
{
  if (data->sixto4_flag == 0) {
    data->sixto4_flag = 1;
  } else {
    data->sixto4_flag = 0;
  }

  return (TRUE);
}

// TCP IPv4 Source Address for 6to4 tunnel of IPv6 over IPv4
gboolean
on_entry125_activate (GtkWidget *entry125, SPSData *data)
{
  int status;
  const gchar *entry125_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry125_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry125_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry125_text = gtk_entry_get_text (GTK_ENTRY (data->entry125));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry125_text) != TRUE) {
// Invalid IP - use old value
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_tcp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry125), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry125_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry125_text, &data->sixto4ip4hdr_tcp.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry125_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_tcp.ip_sum = 0;
    data->sixto4ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp, IP4_HDRLEN);

// Note: The TCP checksum does not need updating since it contains the IPv6
// source and destination addresses, not the IPv4 6to4 header addresses.

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton13), FALSE);
    data->ran_tcp6to4_sourceip = 0;

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize TCP source IPv4 address for 6to4 tunneling of IPv6 over IPv4
gboolean
on_checkbutton13_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton13_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' on_checkbutton13_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_tcp6to4_sourceip == 0) {
    data->ran_tcp6to4_sourceip = 1;
    inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_tcp.ip_src);
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_tcp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry125), value);

    // TCP header checksum (16 bits)
    data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);

    // IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_tcp.ip_sum = 0;
    data->sixto4ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
  } else {
    data->ran_tcp6to4_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// ICMP IPv4 Source Address for 6to4 tunnel of IPv6 over IPv4
gboolean
on_entry160_activate (GtkWidget *entry160, SPSData *data)
{
  int status;
  const gchar *entry160_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry160_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry160_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry160_text = gtk_entry_get_text (GTK_ENTRY (data->entry160));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry160_text) != TRUE) {
// Invalid IP - use old value
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_icmp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry160), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry160_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry160_text, &data->sixto4ip4hdr_icmp.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry160_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_icmp.ip_sum = 0;
    data->sixto4ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp, IP4_HDRLEN);

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton26), FALSE);
    data->ran_icmp6to4_sourceip = 0;

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize ICMP source IPv4 address for 6to4 tunneling of IPv6 over IPv4
gboolean
on_checkbutton26_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton26_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' on_checkbutton26_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_icmp6to4_sourceip == 0) {
    data->ran_icmp6to4_sourceip = 1;
    inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_icmp.ip_src);
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_icmp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry160), value);

    // IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_icmp.ip_sum = 0;
    data->sixto4ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp, IP4_HDRLEN);

  } else {
    data->ran_icmp6to4_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// UDP IPv4 Source Address for 6to4 tunnel of IPv6 over IPv4
gboolean
on_entry161_activate (GtkWidget *entry161, SPSData *data)
{
  int status;
  const gchar *entry161_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry161_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry161_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry161_text = gtk_entry_get_text (GTK_ENTRY (data->entry161));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry161_text) != TRUE) {
// Invalid IP - use old value
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_udp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry161), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry161_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry161_text, &data->sixto4ip4hdr_udp.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry161_activate().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_udp.ip_sum = 0;
    data->sixto4ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp, IP4_HDRLEN);

//  IP address has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton27), FALSE);
    data->ran_udp6to4_sourceip = 0;

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Randomize UDP source IPv4 address for 6to4 tunneling of IPv6 over IPv4
gboolean
on_checkbutton27_toggled (GtkButton *button, SPSData *data)
{
  gchar *value, *ipaddress;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton27_toggled().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' on_checkbutton27_toggled().\n");
    exit (EXIT_FAILURE);
  }

  if (data->ran_udp6to4_sourceip == 0) {
    data->ran_udp6to4_sourceip = 1;
    inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_udp.ip_src);
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_udp.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry161), value);

    // UDP header checksum (16 bits)
    data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);

    // IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_udp.ip_sum = 0;
    data->sixto4ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp, IP4_HDRLEN);

  } else {
    data->ran_udp6to4_sourceip = 0;
  }

  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Tunnel IPv6 over IPv4 (6to4) (sps.ui)

// Specify ethernet header
gboolean
on_checkbutton14_toggled (GtkButton *button, SPSData *data)
{
  if (data->sixto4_specify_ether == 0) {
    data->sixto4_specify_ether = 1;
  } else {
    data->sixto4_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry156_activate (GtkWidget *entry156, SPSData *data)
{
  int i, bad;
  const gchar *entry156_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry156_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry156_text = gtk_entry_get_text (GTK_ENTRY (data->entry156));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry156_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry156_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry156_text[i] > 47) && (entry156_text[i] < 58)) ||    // 0 to 10
        ((entry156_text[i] > 64) && (entry156_text[i] < 71)) ||      // A to Z
        ((entry156_text[i] > 96) && (entry156_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry156_text[i] > 47) && (entry156_text[i] < 58)) ||    // 0 to 10
        ((entry156_text[i] > 64) && (entry156_text[i] < 71)) ||      // A to Z
        ((entry156_text[i] > 96) && (entry156_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->sixto4ethhdr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry156), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry156_text, data->sixto4ethhdr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry157_activate (GtkWidget *entry157, SPSData *data)
{
  int i, bad;
  const gchar *entry157_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry157_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry157_text = gtk_entry_get_text (GTK_ENTRY (data->entry157));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry157_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry157_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry157_text[i] > 47) && (entry157_text[i] < 58)) ||    // 0 to 10
        ((entry157_text[i] > 64) && (entry157_text[i] < 71)) ||      // A to Z
        ((entry157_text[i] > 96) && (entry157_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry157_text[i] > 47) && (entry157_text[i] < 58)) ||    // 0 to 10
        ((entry157_text[i] > 64) && (entry157_text[i] < 71)) ||      // A to Z
        ((entry157_text[i] > 96) && (entry157_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->sixto4ethhdr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry157), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry157_text, data->sixto4ethhdr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry158_activate (GtkWidget *entry158, SPSData *data)
{
  int i, sd, status;
  const gchar *entry158_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry158_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry158_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry158_text = gtk_entry_get_text (GTK_ENTRY (entry158));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry158_activate.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry158_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry158_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->sixto4_interfacename, entry158_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->sixto4ethhdr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->sixto4ethhdr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry157), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry159_activate (GtkWidget *entry159, SPSData *data)
{
  const gchar *entry159_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry159_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry159_text = gtk_entry_get_text (GTK_ENTRY (data->entry159));

// Is value too large for a short int?
  if ((atoi (entry159_text) < 0) || (atoi (entry159_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->sixto4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry159), value);
  } else {
    data->sixto4ethhdr.type_code = htons (atoi (entry159_text));
  }

  g_free (value);

  return (TRUE);
}

// Number of packets to send
gboolean
on_entry18_activate (GtkWidget *entry18, SPSData *data)
{
  const gchar *entry18_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry18_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry18_text = gtk_entry_get_text (GTK_ENTRY (data->entry18));

// Is value too large or small?
  if ((atoi (entry18_text) < 0) || (strtoul (entry18_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%u", data->npackets);
    gtk_entry_set_text (GTK_ENTRY (data->entry18), value);
    g_free (value);
    return (TRUE);
  } else {
    data->npackets = atoi (entry18_text);  // Update npackets with new value

    g_free (value);

    return (TRUE);
  }
}

// Set flood mode
gboolean
on_checkbutton11_toggled (GtkButton *button, SPSData *data)
{
  if (flood_mode == 0) {
    flood_mode = 1;
  } else {
    flood_mode= 0;
  }
  return (TRUE);
}

// Send packet(s)
gboolean
on_button1_clicked (GtkButton *button, SPSData *data)
{
  if ((data->packet_type == 1) || (data->packet_type == 2) ||
      (data->packet_type == 3) || (data->packet_type == 4)) {
    ipv4_send (data);

  } else if ((data->packet_type == 5) || (data->packet_type == 6) ||
             (data->packet_type == 7) || (data->packet_type ==8)) {
    if (data->sixto4_flag == 1) {
      sixto4_send (data);
    } else {
      ipv6_send (data);
    }
  }

  return (FALSE);  // Unknown data->packet_type value was encountered
}

// Send IPv4 packet(s)
gboolean
ipv4_send (SPSData *data)
{
  int status, tcp_frame_length, icmp_frame_length, udp_frame_length;
  gint i, j, sd_tcp, sd_icmp, sd_udp;
  gchar *text, *ipaddress;
  unsigned char *tcp_packet, *icmp_packet, *udp_packet;
  unsigned char *tcp_ether_frame, *icmp_ether_frame, *udp_ether_frame;
  const int on = 1;
  struct sockaddr_in sin_tcp, sin_icmp, sin_udp;
  struct sockaddr_ll tcp_device, icmp_device, udp_device;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp_packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp_packet' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    icmp_packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'icmp_packet' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    udp_packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'udp_packet' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp_ether_frame' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    icmp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'icmp_ether_frame' in ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    udp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'udp_ether_frame' ipv4_send().\n");
    exit (EXIT_FAILURE);
  }

// Make sure all interfaces were specified
  if (((data->packet_type == 1) && (data->tcp4_specify_ether == 1) &&
       (strlen (data->tcp4_interfacename) < 2)) ||
      ((data->packet_type == 2) && (data->icmp4_specify_ether == 1) &&
       (strlen (data->icmp4_interfacename) < 2)) ||
      ((data->packet_type == 3) && (data->udp4_specify_ether == 1) &&
       (strlen (data->udp4_interfacename) < 2))) {
    sprintf (text, "In callbacks.c: ipv4_send() failed.\nAppears to be an invalid interface name.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (tcp_packet);
    g_free (icmp_packet);
    g_free (udp_packet);
    g_free (ipaddress);
    g_free (tcp_ether_frame);
    g_free (icmp_ether_frame);
    g_free (udp_ether_frame);
    return (FALSE);
  }

// Resolve interface index.
  if ((data->tcp4_specify_ether == 1) && ((data->packet_type == 1) || (data->packet_type == 4))) {
    if ((tcp_device.sll_ifindex = if_nametoindex (data->tcp4_interfacename)) == 0) {
      status = errno;
      sprintf (text, "In callbacks.c: if_nametoindex() failed in ipv4_send() to obtain interface index for TCP packet.\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }
  if ((data->icmp4_specify_ether == 1) && ((data->packet_type == 2) || (data->packet_type == 4))) {
    if ((icmp_device.sll_ifindex = if_nametoindex (data->icmp4_interfacename)) == 0) {
      status = errno;
      sprintf (text, "In callbacks.c: if_nametoindex() failed in ipv4_send() to obtain interface index for ICMP packet.\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }
  if ((data->udp4_specify_ether == 1) && ((data->packet_type == 3) || (data->packet_type == 4))) {
    if ((udp_device.sll_ifindex = if_nametoindex (data->udp4_interfacename)) == 0) {
      status = errno;
      sprintf (text, "In callbacks.c: if_nametoindex() failed in ipv4_send() to obtain interface index for UDP packet.\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }

// Fill out sockaddr_ll.
  tcp_device.sll_family = AF_PACKET;
  memcpy (tcp_device.sll_addr, data->tcp4ethhdr.src_mac, 6);
  tcp_device.sll_halen = htons (6u);
  icmp_device.sll_family = AF_PACKET;
  memcpy (icmp_device.sll_addr, data->icmp4ethhdr.src_mac, 6);
  icmp_device.sll_halen = htons (6u);
  udp_device.sll_family = AF_PACKET;
  memcpy (udp_device.sll_addr, data->udp4ethhdr.src_mac, 6);
  udp_device.sll_halen = htons (6u);

// Fill out ethernet frame header.

// Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv4 header + TCP/ICMP/UDP header + data)
  tcp_frame_length = ETH_HDRLEN + IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen;
  icmp_frame_length = ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen;
  udp_frame_length = ETH_HDRLEN + IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen;

// Destination MAC address
  memcpy (tcp_ether_frame, data->tcp4ethhdr.dst_mac, 6);
  memcpy (icmp_ether_frame, data->icmp4ethhdr.dst_mac, 6);
  memcpy (udp_ether_frame, data->udp4ethhdr.dst_mac, 6);

// Source MAC address
  memcpy (tcp_ether_frame + 6, data->tcp4ethhdr.src_mac, 6);
  memcpy (icmp_ether_frame + 6, data->icmp4ethhdr.src_mac, 6);
  memcpy (udp_ether_frame + 6, data->udp4ethhdr.src_mac, 6);

// Next is ethernet type code.
// http://www.iana.org/assignments/ethernet-numbers
  memcpy (&tcp_ether_frame[12], &data->tcp4ethhdr.type_code, 2);
  memcpy (&icmp_ether_frame[12], &data->icmp4ethhdr.type_code, 2);
  memcpy (&udp_ether_frame[12], &data->udp4ethhdr.type_code, 2);

// Next is ethernet frame data (IPv4 header + TCP/ICMP/UDP header).

// IPv4 header
  memcpy (tcp_ether_frame + ETH_HDRLEN, &data->ip4hdr_tcp, IP4_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN, &data->ip4hdr_icmp, IP4_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN, &data->ip4hdr_udp, IP4_HDRLEN);

// Upper layer protocol header
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->tcp4hdr, TCP_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->icmp4hdr, ICMP_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->udp4hdr, UDP_HDRLEN);

// Upper layer protocol data
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + TCP_HDRLEN, data->tcp4_payload, data->tcp4_payloadlen);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN, data->icmp4_payload, data->icmp4_payloadlen);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + UDP_HDRLEN, data->udp4_payload, data->udp4_payloadlen);

// Prepare packets (no ethernet header)

// First part is an IPv4 header.
  memcpy (tcp_packet, &data->ip4hdr_tcp, IP4_HDRLEN);
  memcpy (icmp_packet, &data->ip4hdr_icmp, IP4_HDRLEN);
  memcpy (udp_packet, &data->ip4hdr_udp, IP4_HDRLEN);

// Next part of packet is upper layer protocol header.
  memcpy ((tcp_packet + IP4_HDRLEN), &data->tcp4hdr, TCP_HDRLEN);
  memcpy ((icmp_packet + IP4_HDRLEN), &data->icmp4hdr, ICMP_HDRLEN);
  memcpy ((udp_packet + IP4_HDRLEN), &data->udp4hdr, UDP_HDRLEN);

// Append data to end of packet.
  memcpy (tcp_packet + IP4_HDRLEN + TCP_HDRLEN, data->tcp4_payload, data->tcp4_payloadlen);
  memcpy (icmp_packet + IP4_HDRLEN + ICMP_HDRLEN, data->icmp4_payload, data->icmp4_payloadlen);
  memcpy (udp_packet + IP4_HDRLEN + UDP_HDRLEN, data->udp4_payload, data->udp4_payloadlen);

// For packets (no ethernet header):
// The kernel will provide Layer 2 (data link layer) information (MAC addresses).
// To do that, we need to specify a destination for the kernel in order for it
// to decide where to send the raw datagram. We fill in a struct in_addr with
// the desired destination IP address, and pass this structure to the sendto() function.
  memset (&sin_tcp, 0, sizeof (struct sockaddr_in));
  sin_tcp.sin_family = AF_INET;
  sin_tcp.sin_addr.s_addr = data->ip4hdr_tcp.ip_dst.s_addr;
  memset (&sin_icmp, 0, sizeof (struct sockaddr_in));
  sin_icmp.sin_family = AF_INET;
  sin_icmp.sin_addr.s_addr = data->ip4hdr_icmp.ip_dst.s_addr;
  memset (&sin_udp, 0, sizeof (struct sockaddr_in));
  sin_udp.sin_family = AF_INET;
  sin_udp.sin_addr.s_addr = data->ip4hdr_udp.ip_dst.s_addr;

// Display packet activity in Activity Log.
  if (flood_mode == 0) {
    if (data->packet_type == 1) {
      if (data->tcp4_specify_ether == 1) {
        sprintf (text, "Sending IPv4 TCP ethernet frames.\n");
      } else {
        sprintf (text, "Sending IPv4 TCP packets.\n");
      }
    } else if (data->packet_type == 2) {
      if (data->icmp4_specify_ether == 1) {
        sprintf (text, "Sending IPv4 ICMP ethernet frames.\n");
      } else {
        sprintf (text, "Sending IPv4 ICMP packets.\n");
      }
    } else if (data->packet_type == 3) {
      if (data->udp4_specify_ether == 1) {
        sprintf (text, "Sending IPv4 UDP ethernet frames.\n");
      } else {
        sprintf (text, "Sending IPv4 UDP packets.\n");
      }
    } else if (data->packet_type == 4) {
      sprintf (text, "Sending IPv4 TCP+ICMP+UDP packets or ethernet frames.\n");
    }
  } else {
    if (data->packet_type == 1) {
      if (data->tcp4_specify_ether == 1) {
        sprintf (text, "Flooding IPv4 TCP ethernet frames.\n");
      } else {
        sprintf (text, "Flooding IPv4 TCP packets.\n");
      }
    } else if (data->packet_type == 2) {
      if (data->icmp4_specify_ether == 1) {
        sprintf (text, "Flooding IPv4 ICMP ethernet frames.\n");
      } else {
        sprintf (text, "Flooding IPv4 ICMP packets.\n");
      }
    } else if (data->packet_type == 3) {
      if (data->udp4_specify_ether == 1) {
        sprintf (text, "Flooding IPv4 UDP ethernet frames.\n");
      } else {
        sprintf (text, "Flooding IPv4 UDP packets.\n");
      }
    } else if (data->packet_type == 4) {
      sprintf (text, "Flooding IPv4 TCP+ICMP+UDP packets or ethernet frames.\n");
    }
  }
  data->message = text;
  log_activity (data);
  memset (text, 0, 1024);

// Send TCP packet(s).
  if (data->packet_type == 1) {

    // Submit request for a raw socket descriptor.
    // Need separate socket decriptors since packets could
    // be going to difference destination addresses.
    if (data->tcp4_specify_ether == 1) {
      if ((sd_tcp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    } else {
      if ((sd_tcp = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Set flag so socket expects us to provide IPv4 header.
      if (setsockopt (sd_tcp, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      if (data->tcp4_specify_ether == 1) {
        if ((sendto (sd_tcp, tcp_ether_frame, tcp_frame_length, 0, (struct sockaddr *) &tcp_device, sizeof (tcp_device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd_tcp, tcp_packet, IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen, 0, (struct sockaddr *) &sin_tcp, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_tcp4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_tcp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_tcp4_sourceport == 1) {
        data->tcp4hdr.th_sport = htons (ran_val2 ());
      }

      // Re-calculate IPv4 and upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // TCP header checksum (16 bits)
      data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->ip4hdr_tcp.ip_sum = 0;
      data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header and upper layer protocol header.
      memcpy (tcp_ether_frame + ETH_HDRLEN, &data->ip4hdr_tcp, IP4_HDRLEN);
      memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->tcp4hdr, TCP_HDRLEN);

      // Update packet IPv4 header and upper layer protocol header.
      memcpy (tcp_packet, &data->ip4hdr_tcp, IP4_HDRLEN);
      memcpy (tcp_packet + IP4_HDRLEN, &data->tcp4hdr, TCP_HDRLEN);
    }

    close (sd_tcp);

// Send ICMP packet(s)
  } else if (data->packet_type == 2) {

    // Submit request for a raw socket descriptor.
    // Need separate socket decriptors since packets could
    // be going to difference destination addresses.
    if (data->icmp4_specify_ether == 1) {
      if ((sd_icmp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    } else {
      if ((sd_icmp = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Set flag so socket expects us to provide IPv4 header.
      if (setsockopt (sd_icmp, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

   while ((i > 0) && (looping == 1)) {

     if (data->icmp4_specify_ether == 1) {
       if ((sendto (sd_icmp, icmp_ether_frame, icmp_frame_length, 0, (struct sockaddr *) &icmp_device, sizeof (icmp_device))) <= 0) {
         status = errno;
         sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
         data->parent = data->main_window;
         data->message = text;
         report_error (data);
         g_free (text);
         g_free (tcp_packet);
         g_free (icmp_packet);
         g_free (udp_packet);
         g_free (ipaddress);
         g_free (tcp_ether_frame);
         g_free (icmp_ether_frame);
         g_free (udp_ether_frame);
         return (FALSE);
       }
     } else {
       if (sendto (sd_icmp, icmp_packet, IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen, 0, (struct sockaddr *) &sin_icmp, sizeof (struct sockaddr)) < 0)  {
         status = errno;
         sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
         data->parent = data->main_window;
         data->message = text;
         report_error (data);
         g_free (text);
         g_free (tcp_packet);
         g_free (icmp_packet);
         g_free (udp_packet);
         g_free (ipaddress);
         g_free (tcp_ether_frame);
         g_free (icmp_ether_frame);
         g_free (udp_ether_frame);
         return (FALSE);
       }
     }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_icmp4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_icmp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }

        // IPv4 header checksum (16 bits): set to 0 when calculating checksum
        data->ip4hdr_icmp.ip_sum = 0;
        data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);

        // Update ethernet frame IPv4 header.
        // ICMP header and payload data remains unchanged, so don't bother
        // re-copying them.
        memcpy (icmp_ether_frame + ETH_HDRLEN, &data->ip4hdr_icmp, IP4_HDRLEN);

        // Update packet IPv4 header.
        // ICMP header and payload data remains unchanged, so don't bother
        // re-copying them.
        memcpy (icmp_packet, &data->ip4hdr_icmp, IP4_HDRLEN);
      }
    }

    close (sd_icmp);

// Send UDP packet(s)
  } else if (data->packet_type == 3) {

    // Submit request for a raw socket descriptor.
    // Need separate socket decriptors since packets could
    // be going to difference destination addresses.
    if (data->udp4_specify_ether == 1) {
      if ((sd_udp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    } else {
      if ((sd_udp = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Set flag so socket expects us to provide IPv4 header.
      if (setsockopt (sd_udp, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      if (data->udp4_specify_ether == 1) {
        if ((sendto (sd_udp, udp_ether_frame, udp_frame_length, 0, (struct sockaddr *) &udp_device, sizeof (udp_device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd_udp, udp_packet, IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen, 0, (struct sockaddr *) &sin_udp, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_udp4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_udp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_udp4_sourceport == 1) {
        data->udp4hdr.uh_sport = htons (ran_val2 ());
      }

      // Re-calculate IPv4 and upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // UDP header checksum (16 bits)
      data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->ip4hdr_udp.ip_sum = 0;
      data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (udp_ether_frame + ETH_HDRLEN, &data->ip4hdr_udp, IP4_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->udp4hdr, UDP_HDRLEN);

      // Update packet IPv4 header and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (udp_packet, &data->ip4hdr_udp, IP4_HDRLEN);
      memcpy (udp_packet + IP4_HDRLEN, &data->udp4hdr, UDP_HDRLEN);
    }

    close (sd_udp);

// Send all types of packets.
  } else if (data->packet_type == 4) {

    // Submit request for a raw socket descriptor.
    // Need separate socket decriptors since packets could
    // be going to difference destination addresses.
    if (data->tcp4_specify_ether == 1) {
      if ((sd_tcp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket decriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    } else {
      if ((sd_tcp = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
      // Set flag so socket expects us to provide IPv4 header.
      if (setsockopt (sd_tcp, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    }

    if (data->icmp4_specify_ether == 1) {
      if ((sd_icmp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    } else {
      if ((sd_icmp = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
      // Set flag so socket expects us to provide IPv4 header.
      if (setsockopt (sd_icmp, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    }

    if (data->udp4_specify_ether == 1) {
      if ((sd_udp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    } else {
      if ((sd_udp = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
      // Set flag so socket expects us to provide IPv4 header.
      if (setsockopt (sd_udp, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
        status = errno;
        sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in ipv4_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (tcp_packet);
        g_free (icmp_packet);
        g_free (udp_packet);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      // Send TCP
      if (data->tcp4_specify_ether == 1) {
        if ((sendto (sd_tcp, tcp_ether_frame, tcp_frame_length, 0, (struct sockaddr *) &tcp_device, sizeof (tcp_device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd_tcp, tcp_packet, IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen, 0, (struct sockaddr *) &sin_tcp, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Send ICMP
      if (data->icmp4_specify_ether == 1) {
        if ((sendto (sd_icmp, icmp_ether_frame, icmp_frame_length, 0, (struct sockaddr *) &icmp_device, sizeof (icmp_device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd_icmp, icmp_packet, IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen, 0, (struct sockaddr *) &sin_icmp, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Send UDP
      if (data->udp4_specify_ether == 1) {
        if ((sendto (sd_udp, udp_ether_frame, udp_frame_length, 0, (struct sockaddr *) &udp_device, sizeof (udp_device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd_udp, udp_packet, IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen, 0, (struct sockaddr *) &sin_udp, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_tcp4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_tcp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_icmp4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_icmp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_udp4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_udp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_tcp4_sourceport == 1) {
        data->tcp4hdr.th_sport = htons (ran_val2 ());
      }
      if (data->ran_udp4_sourceport == 1) {
        data->udp4hdr.uh_sport = htons (ran_val2 ());
      }

      // Re-calculate IPv4 and upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // Upper layer protocol header checksum (16 bits)
      // ICMP does not protect source and destination addresses in IPv4,
      // so don't re-calculate ICMP checksum.
      data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
      data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->ip4hdr_tcp.ip_sum = 0;
      data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
      data->ip4hdr_icmp.ip_sum = 0;
      data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
      data->ip4hdr_udp.ip_sum = 0;
      data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header and upper layer protocol header.
      // ICMP header and payload data remain unchanged, so don't bother re-copying it.
      memcpy (tcp_ether_frame + ETH_HDRLEN, &data->ip4hdr_tcp, IP4_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN, &data->ip4hdr_icmp, IP4_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN, &data->ip4hdr_udp, IP4_HDRLEN);
      memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->tcp4hdr, TCP_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->udp4hdr, UDP_HDRLEN);

      // Update packet IPv4 header and upper layer protocol header.
      // ICMP and payload data remain unchanged, so don't bother re-copying it.
      memcpy (tcp_packet, &data->ip4hdr_tcp, IP4_HDRLEN);
      memcpy (tcp_packet + IP4_HDRLEN, &data->tcp4hdr, TCP_HDRLEN);
      memcpy (icmp_packet, &data->ip4hdr_icmp, IP4_HDRLEN);
      memcpy (udp_packet, &data->ip4hdr_udp, IP4_HDRLEN);
      memcpy (udp_packet + IP4_HDRLEN, &data->udp4hdr, UDP_HDRLEN);
    }

    close (sd_tcp);
    close (sd_icmp);
    close (sd_udp);
  }

// Display packet activity in Activity Log.
  if (data->packet_type == 1) {
    if (data->tcp4_specify_ether == 1) {
      sprintf (text, "%i IPv4 TCP ethernet frames sent.\n", j);
    } else {
      sprintf (text, "%i IPv4 TCP packets sent.\n", j);
    }
  } else if (data->packet_type == 2) {
    if (data->icmp4_specify_ether == 1) {
      sprintf (text, "%i IPv4 ICMP ethernet frames sent.\n", j);
    } else {
      sprintf (text, "%i IPv4 ICMP packets sent.\n", j);
    }
  } else if (data->packet_type == 3) {
    if (data->udp4_specify_ether == 1) {
      sprintf (text, "%i IPv4 UDP ethernet frames sent.\n", j);
    } else {
      sprintf (text, "%i IPv4 UDP packets sent.\n", j);
    }
  } else if (data->packet_type == 4) {
    sprintf (text, "%i IPv4 TCP+ICMP+UDP packets or ethernet frames sent.\n", j);
  }
  data->message = text;
  log_activity (data);
  memset (text, 0, 1024);

// Free allocated memory.
  g_free (text);
  g_free (tcp_packet);
  g_free (icmp_packet);
  g_free (udp_packet);
  g_free (ipaddress);
  g_free (tcp_ether_frame);
  g_free (icmp_ether_frame);
  g_free (udp_ether_frame);

  return (TRUE);
}

// Send IPv6 packet(s).
gboolean
ipv6_send (SPSData *data)
{
  int status, tcp_frame_length, icmp_frame_length, udp_frame_length;
  gint i, j, sd_tcp, sd_icmp, sd_udp;
  gchar *text, *ipaddress;
  unsigned char *tcp_ether_frame, *icmp_ether_frame, *udp_ether_frame;
  struct sockaddr_ll tcp_device, icmp_device, udp_device;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in ipv6_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (40*sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in ipv6_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp_ether_frame' in ipv6_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    icmp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'icmp_ether_frame' in ipv6_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    udp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'udp_ether_frame' in ipv6_send().\n");
    exit (EXIT_FAILURE);
  }

  // Make sure all interfaces were specified
  if ((data->packet_type == 5) || (data->packet_type == 8)) {
    if (strlen (data->tcp6_interfacename) < 2) {
      sprintf (text, "In callbacks.c: ipv6_send() failed.\nAppears to be an invalid interface name for TCP packet.");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }
  if ((data->packet_type == 6) || (data->packet_type == 8)) {
    if (strlen (data->icmp6_interfacename) < 2) {
      sprintf (text, "In callbacks.c: ipv6_send() failed.\nAppears to be an invalid interface name for ICMP packet.");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }
  if ((data->packet_type == 7) || (data->packet_type == 8)) {
    if (strlen (data->udp6_interfacename) < 2) {
      sprintf (text, "In callbacks.c: ipv6_send() failed.\nAppears to be an invalid interface name for UDP packet.");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }

  // Resolve interface index.
  if ((data->packet_type == 5) || (data->packet_type == 8)) {
    if ((tcp_device.sll_ifindex = if_nametoindex (data->tcp6_interfacename)) == 0) {
      status = errno;
      sprintf (text, "In callbacks.c: if_nametoindex() failed to obtain interface index for TCP packet in ipv6_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }
  if ((data->packet_type == 6) || (data->packet_type == 8)) {
    if ((icmp_device.sll_ifindex = if_nametoindex (data->icmp6_interfacename)) == 0) {
      status = errno;
      sprintf (text, "In callbacks.c: if_nametoindex() failed to obtain interface index for ICMP packet in ipv6_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }
  if ((data->packet_type == 7) || (data->packet_type == 8)) {
    if ((udp_device.sll_ifindex = if_nametoindex (data->udp6_interfacename)) == 0) {
      status = errno;
      sprintf (text,  "In callbacks.c: if_nametoindex() failed to obtain interface index for UDP packet in ipv6_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }

  // Fill out sockaddr_ll.
  tcp_device.sll_family = AF_PACKET;
  memcpy (tcp_device.sll_addr, data->tcp6ethhdr.src_mac, 6);
  tcp_device.sll_halen = htons (6u);
  icmp_device.sll_family = AF_PACKET;
  memcpy (icmp_device.sll_addr, data->icmp6ethhdr.src_mac, 6);
  icmp_device.sll_halen = htons (6u);
  udp_device.sll_family = AF_PACKET;
  memcpy (udp_device.sll_addr, data->udp6ethhdr.src_mac, 6);
  udp_device.sll_halen = htons (6u);

  // Fill out ethernet frame header.

  // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv6 header + TCP/ICMP/UDP header + data)
  tcp_frame_length = ETH_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_payloadlen;
  icmp_frame_length = ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_payloadlen;
  udp_frame_length = ETH_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_payloadlen;

  // Destination MAC address
  memcpy (tcp_ether_frame, data->tcp6ethhdr.dst_mac, 6);
  memcpy (icmp_ether_frame, data->icmp6ethhdr.dst_mac, 6);
  memcpy (udp_ether_frame, data->udp6ethhdr.dst_mac, 6);

  // Source MAC address
  memcpy (tcp_ether_frame + 6, data->tcp6ethhdr.src_mac, 6);
  memcpy (icmp_ether_frame + 6, data->icmp6ethhdr.src_mac, 6);
  memcpy (udp_ether_frame + 6, data->udp6ethhdr.src_mac, 6);

  // Next is ethernet type code.
  // http://www.iana.org/assignments/ethernet-numbers
  memcpy (&tcp_ether_frame[12], &data->tcp6ethhdr.type_code, 2);
  memcpy (&icmp_ether_frame[12], &data->icmp6ethhdr.type_code, 2);
  memcpy (&udp_ether_frame[12], &data->udp6ethhdr.type_code, 2);

  // Next is ethernet frame data (IPv6 header + TCP/ICMP/UDP header).

  // IPv6 header
  memcpy (tcp_ether_frame + ETH_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);

  // Upper layer protocol header
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);

  // Upper layer protocol data
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP6_HDRLEN + TCP_HDRLEN, data->tcp6_payload, data->tcp6_payloadlen);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN, data->icmp6_payload, data->icmp6_payloadlen);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP6_HDRLEN + UDP_HDRLEN, data->udp6_payload, data->udp6_payloadlen);

  // Display packet activity in Activity Log.
  if (flood_mode == 0) {
    if (data->packet_type == 5) {
      strcpy (text, "Sending IPv6 TCP ethernet frames.\n");
    } else if (data->packet_type == 6) {
      strcpy (text, "Sending IPv6 ICMP ethernet frames.\n");
    } else if (data->packet_type == 7) {
      strcpy (text, "Sending IPv6 UDP ethernet frames.\n");
    } else if (data->packet_type == 8) {
      strcpy (text, "Sending IPv6 TCP+ICMP+UDP ethernet frames.\n");
    }
  } else {
    if (data->packet_type == 5) {
      strcpy (text, "Flooding IPv6 TCP ethernet frames.\n");
    } else if (data->packet_type == 6) {
      strcpy (text, "Flooding IPv6 ICMP ethernet frames.\n");
    } else if (data->packet_type == 7) {
      strcpy (text, "Flooding IPv6 UDP ethernet frames.\n");
    } else if (data->packet_type == 8) {
      strcpy (text, "Flooding IPv6 TCP+ICMP+UDP ethernet frames.\n");
    }
  }
  data->message = text;
  log_activity (data);
  memset (text, 0, 1024);

  // Send TCP packet(s).
  if (data->packet_type == 5) {

    // Submit request for a raw socket descriptor.
    // Need separate socket decriptors since packets could
    // be going to difference destination addresses.
    if ((sd_tcp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      status = errno;
      sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv6_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;
    
    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {
      if ((sendto (sd_tcp, tcp_ether_frame, tcp_frame_length, 0, (struct sockaddr *) &tcp_device, sizeof (tcp_device))) <= 0) {
        status = errno;
        sprintf (text, "In callbacks.c: sendto() failed in ipv6_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize IPv6 source address if requested.
      if (data->ran_tcp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_tcp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv6_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_tcp6_sourceport == 1) {
        data->tcp6hdr.th_sport = htons (ran_val2 ());
      }

      // Re-calculate upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // TCP header checksum (16 bits)
      data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);

      // Update ethernet frame IPv6 header and upper layer protocol header.
      memcpy (tcp_ether_frame + ETH_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
      memcpy (tcp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
    }

  close (sd_tcp);

  // Send ICMP packet(s)
  } else if (data->packet_type == 6) {

    if ((sd_icmp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      status = errno;
      sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv6_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {
      if ((sendto (sd_icmp, icmp_ether_frame, icmp_frame_length, 0, (struct sockaddr *) &icmp_device, sizeof (icmp_device))) <= 0) {
        status = errno;
        sprintf (text, "In callbacks.c: sendto() failed in ipv6_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv6 address if requested.
      if (data->ran_icmp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_icmp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv6_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }

        // Re-calculate upper layer protocol checksum.
        // ICMPv6 header checksum (16 bits)
        data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);

        // Update ethernet frame IPv6 header and upper layer protocol header.
        // Note that in IPv6, the source and destination addresses are in ICMPv6 checksum calculation.
        // Payload data remains unchanged, so don't bother re-copying it.
        memcpy (icmp_ether_frame + ETH_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
        memcpy (icmp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
      }
    }

  close (sd_icmp);

  // Send UDP packet(s).
  } else if (data->packet_type == 7) {

    if ((sd_udp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      status = errno;
      sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in ipv6_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {
      if ((sendto (sd_udp, udp_ether_frame, udp_frame_length, 0, (struct sockaddr *) &udp_device, sizeof (udp_device))) <= 0) {
        status = errno;
        sprintf (text, "In callbacks.c: sendto() failed in ipv6_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv6 address if requested.
      if (data->ran_udp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_udp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv6_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_udp6_sourceport == 1) {
        data->udp6hdr.uh_sport = htons (ran_val2 ());
      }

      // Re-calculate upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // UDP header checksum (16 bits)
      data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);

      // Update ethernet frame IPv6 header and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (udp_ether_frame + ETH_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);
    }

  close (sd_udp);

  // Send all packet types.
  } else if (data->packet_type == 8) {

    if ((sd_tcp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      perror ("socket() failed in ipv6_send ");
      exit (EXIT_FAILURE);
    }
    if ((sd_icmp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      perror ("socket() failed in ipv6_send ");
      exit (EXIT_FAILURE);
    }
    if ((sd_udp = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      perror ("socket() failed in ipv6_send ");
      exit (EXIT_FAILURE);
    }

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {
      if ((sendto (sd_tcp, tcp_ether_frame, tcp_frame_length, 0, (struct sockaddr *) &tcp_device, sizeof (tcp_device))) <= 0) {
        status = errno;
        sprintf (text, "In callbacks.c: sendto() failed in ipv6_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
      if ((sendto (sd_icmp, icmp_ether_frame, icmp_frame_length, 0, (struct sockaddr *) &icmp_device, sizeof (icmp_device))) <= 0) {
        status = errno;
        sprintf (text, "In callbacks.c: sendto() failed in ipv6_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }
      if ((sendto (sd_udp, udp_ether_frame, udp_frame_length, 0, (struct sockaddr *) &udp_device, sizeof (udp_device))) <= 0) {
        status = errno;
        sprintf (text, "In callbacks.c: sendto() failed in ipv6_send().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (ipaddress);
        g_free (tcp_ether_frame);
        g_free (icmp_ether_frame);
        g_free (udp_ether_frame);
        return (FALSE);
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv6 address if requested.
      if (data->ran_tcp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_tcp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv6_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_icmp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_icmp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv6_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_udp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_udp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in ipv6_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_tcp6_sourceport == 1) {
        data->tcp6hdr.th_sport = htons (ran_val2 ());
      }
      if (data->ran_udp6_sourceport == 1) {
        data->udp6hdr.uh_sport = htons (ran_val2 ());
      }

      // Re-calculate upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // Upper layer protocol header checksum (16 bits)
      data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
      data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
      data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);

      // Update ethernet frame IPv6 header and upper layer protocol header.
      // Payload data remain unchanged, so don't bother re-copying them.
      memcpy (tcp_ether_frame + ETH_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
      memcpy (tcp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);
    }

  close (sd_tcp);
  close (sd_icmp);
  close (sd_udp);
  }

  // Display packet activity in Activity Log.
  if (data->packet_type == 5) {
    sprintf (text, "%i IPv6 TCP ethernet frames sent.\n", j);
  } else if (data->packet_type == 6) {
    sprintf (text, "%i IPv6 ICMP ethernet frames sent.\n", j);
  } else if (data->packet_type == 7) {
    sprintf (text, "%i IPv6 UDP ethernet frames sent.\n", j);
  } else {
    sprintf (text, "%i IPv6 TCP+ICMP+UDP ethernet frames sent.\n", j);
  }
  data->message = text;
  log_activity (data);
  memset (text, 0, 1024);

  g_free (text);
  g_free (ipaddress);
  g_free (tcp_ether_frame);
  g_free (icmp_ether_frame);
  g_free (udp_ether_frame);

  return (TRUE);
}

// Send IPv6 packet(s) over IPv4 tunnel (6to4).
gboolean
sixto4_send (SPSData *data)
{
  const gchar *entry125_text;
  int status, tcp_frame_length, icmp_frame_length, udp_frame_length;
  gint i, j, sd;
  gchar *text, *ipaddress;
  unsigned char *tcp_packet, *icmp_packet, *udp_packet;
  unsigned char *tcp_ether_frame, *icmp_ether_frame, *udp_ether_frame;
  const int on = 1;
  struct sockaddr_in sin;
  struct sockaddr_ll device;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp_packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp_packet' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    icmp_packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'icmp_packet' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    udp_packet = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'udp_packet' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp_ether_frame' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    icmp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'icmp_ether_frame' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    udp_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'udp_ether_frame' in sixto4_send().\n");
    exit (EXIT_FAILURE);
  }

// Make sure a source IPv4 address was specified.
  entry125_text = gtk_entry_get_text (GTK_ENTRY (data->entry125));
  if (strlen (entry125_text) < 1) {
      sprintf (text, "In callbacks.c: sixto4_send() failed.\nUser needs to enter a source IPv4 address for IPv4 tunnel.");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
  }

// Make sure 6to4 interface was specified if needed.
  if (data->sixto4_specify_ether == 1) {
    if (strlen (data->sixto4_interfacename) < 2) {
      sprintf (text, "In callbacks.c: sixto4_send() failed.\nAppears to be an invalid interface name for IPv4 tunnel.");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }

// Resolve interface index.
  if (data->sixto4_specify_ether == 1) {
    if ((device.sll_ifindex = if_nametoindex (data->sixto4_interfacename)) == 0) {
      status = errno;
      sprintf (text, "In callbacks.c: if_nametoindex() failed to obtain interface index for IPv4 tunnel in sixto4_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }

// Fill out sockaddr_ll.
  device.sll_family = AF_PACKET;
  memcpy (device.sll_addr, data->sixto4ethhdr.src_mac, 6);
  device.sll_halen = htons (6u);

// Make sure 6to4 IPv4 header checksums are up to date
// 6to4 IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->sixto4ip4hdr_tcp.ip_sum = 0;
  data->sixto4ip4hdr_icmp.ip_sum = 0;
  data->sixto4ip4hdr_udp.ip_sum = 0;
  data->sixto4ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
  data->sixto4ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
  data->sixto4ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp, IP4_HDRLEN);

// Fill out ethernet frame header.

// Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv4 header + IPv6 header + TCP/ICMP/UDP header + data)
  tcp_frame_length = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_payloadlen;
  icmp_frame_length = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_payloadlen;
  udp_frame_length = ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_payloadlen;

// Destination MAC address
  memcpy (tcp_ether_frame, data->sixto4ethhdr.dst_mac, 6);
  memcpy (icmp_ether_frame, data->sixto4ethhdr.dst_mac, 6);
  memcpy (udp_ether_frame, data->sixto4ethhdr.dst_mac, 6);

// Source MAC address
  memcpy (tcp_ether_frame + 6, data->sixto4ethhdr.src_mac, 6);
  memcpy (icmp_ether_frame + 6, data->sixto4ethhdr.src_mac, 6);
  memcpy (udp_ether_frame + 6, data->sixto4ethhdr.src_mac, 6);

// Next is ethernet type code.
// http://www.iana.org/assignments/ethernet-numbers
  memcpy (&tcp_ether_frame[12], &data->sixto4ethhdr.type_code, 2);
  memcpy (&icmp_ether_frame[12], &data->sixto4ethhdr.type_code, 2);
  memcpy (&udp_ether_frame[12], &data->sixto4ethhdr.type_code, 2);

// Next is ethernet frame data (IPv4 header + IPv6 header + TCP/ICMP/UDP header + data).

// IPv4 header
  memcpy (tcp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_udp, IP4_HDRLEN);

// IPv6 header
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);

// Upper layer protocol header
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);

// Upper layer protocol data
  memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN, data->tcp6_payload, data->tcp6_payloadlen);
  memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN, data->icmp6_payload, data->icmp6_payloadlen);
  memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN, data->udp6_payload, data->udp6_payloadlen);

// Prepare packets (no ethernet header)

// First part is an IPv4 header.
  memcpy (tcp_packet, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
  memcpy (icmp_packet, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
  memcpy (udp_packet, &data->sixto4ip4hdr_udp, IP4_HDRLEN);

// Next part of packet is an IPv6 header.
  memcpy (tcp_packet + IP4_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
  memcpy (icmp_packet + IP4_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
  memcpy (udp_packet + IP4_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);

// Next part of packet is an upper layer protocol header.
  memcpy (tcp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
  memcpy (icmp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
  memcpy (udp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);

// Append data to end of packet.
  memcpy (tcp_packet + IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN, data->tcp6_payload, data->tcp6_payloadlen);
  memcpy (icmp_packet + IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN, data->icmp6_payload, data->icmp6_payloadlen);
  memcpy (udp_packet + IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN, data->udp6_payload, data->udp6_payloadlen);

// For packets (no ethernet header):
// The kernel will provide Layer 2 (data link layer) information (MAC addresses).
// To do that, we need to specify a destination for the kernel in order for it
// to decide where to send the raw datagram. We fill in a struct in_addr with
// the desired destination IP address, and pass this structure to the sendto() function.
// Here, regardless of IPv6 packet type, we're sending to the IPv4 6to4 anycast address
// which is 192.88.99.1. We arbitrarily grab it from the TCP packet's ip4hdr.
  memset (&sin, 0, sizeof (struct sockaddr_in));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = data->sixto4ip4hdr_tcp.ip_dst.s_addr;

  if (flood_mode == 0) {
    if (data->packet_type == 5) {
      if (data->sixto4_specify_ether == 1) {
        sprintf (text, "Sending IPv6 TCP ethernet frames over IPv4 tunnel.\n");
      } else {
        sprintf (text, "Sending IPv6 TCP packets over IPv4 tunnel.\n");
      }
    } else if (data->packet_type == 6) {
      if (data->sixto4_specify_ether == 1) {
        sprintf (text, "Sending IPv6 ICMP ethernet frames over IPv4 tunnel.\n");
      } else {
        sprintf (text, "Sending IPv6 ICMP packets over IPv4 tunnel.\n");
      }
    } else if (data->packet_type == 7) {
      if (data->sixto4_specify_ether == 1) {
        sprintf (text, "Sending IPv6 UDP ethernet frames over IPv4 tunnel.\n");
      } else {
        sprintf (text, "Sending IPv6 UDP packets over IPv4 tunnel.\n");
      }
    } else if (data->packet_type == 8) {
      sprintf (text, "Sending IPv6 TCP+ICMP+UDP packets or ethernet frames over IPv4 tunnel.\n");
    }
  } else {
    if (data->packet_type == 5) {
      if (data->sixto4_specify_ether == 1) {
        sprintf (text, "Flooding IPv6 TCP ethernet frames over IPv4 tunnel.\n");
      } else {
        sprintf (text, "Flooding IPv6 TCP packets over IPv4 tunnel.\n");
      }
    } else if (data->packet_type == 6) {
      if (data->sixto4_specify_ether == 1) {
        sprintf (text, "Flooding IPv6 ICMP ethernet frames over IPv4 tunnel.\n");
      } else {
        sprintf (text, "Flooding IPv6 ICMP packets over IPv4 tunnel.\n");
      }
    } else if (data->packet_type == 7) {
      if (data->sixto4_specify_ether == 1) {
        sprintf (text, "Flooding IPv6 UDP ethernet frames over IPv4 tunnel.\n");
      } else {
        sprintf (text, "Flooding IPv6 UDP packets over IPv4 tunnel.\n");
      }
    } else if (data->packet_type == 8) {
      sprintf (text, "Flooding IPv6 TCP+ICMP+UDP packets or ethernet frames over IPv4 tunnel.\n");
    }
  }
  data->message = text;
  log_activity (data);
  memset (text, 0, 1024);

// Submit request for a raw IPv4 socket descriptor.
// Can use one socket decriptor since all packets going to
// 6to4 anycast address.
  if (data->sixto4_specify_ether == 1) {
    if ((sd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      status = errno;
      sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in sixto4_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  } else {
    if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
      status = errno;
      sprintf (text, "In callbacks.c: socket() failed to get a socket descriptor in sixto4_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }

  // Set flag so socket expects us to provide IPv4 header.
    if (setsockopt (sd, IPPROTO_IP, IP_HDRINCL, &on, sizeof (on)) < 0) {
      status = errno;
      sprintf (text, "In callbacks.c: setsockopt() failed to set IP_HDRINCL in sixto4_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (ipaddress);
      g_free (tcp_packet);
      g_free (icmp_packet);
      g_free (udp_packet);
      g_free (tcp_ether_frame);
      g_free (icmp_ether_frame);
      g_free (udp_ether_frame);
      return (FALSE);
    }
  }

// Send TCP packet(s).
  if (data->packet_type == 5) {

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;
    
    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      if (data->sixto4_specify_ether == 1) {
        if ((sendto (sd, tcp_ether_frame, tcp_frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd, tcp_packet, IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN, 0, (struct sockaddr *) &sin, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_tcp6to4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_tcp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source IPv6 address if requested.
      if (data->ran_tcp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_tcp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_tcp6_sourceport == 1) {
        data->tcp6hdr.th_sport = htons (ran_val2 ());
      }

      // Re-calculate IPv4 and upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // TCP header checksum (16 bits)
      data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->sixto4ip4hdr_tcp.ip_sum = 0;
      data->sixto4ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header, IPv6 header, and upper layer protocol header.
      memcpy (tcp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
      memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
      memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);

      // Update packet IPv4 header, IPv6 header, and upper layer protocol header.
      memcpy (tcp_packet, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
      memcpy (tcp_packet + IP4_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
      memcpy (tcp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
    }

// Send ICMP packet(s)
  } else if (data->packet_type == 6) {

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      if (data->sixto4_specify_ether == 1) {
        if ((sendto (sd, icmp_ether_frame, icmp_frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd, icmp_packet, IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_payloadlen, 0, (struct sockaddr *) &sin, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_icmp6to4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_icmp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source IPv6 address if requested.
      if (data->ran_icmp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_icmp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Re-calculate IPv4 and upper layer protocol checksum.
      // ICMPv6 header checksum (16 bits)
      data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->sixto4ip4hdr_icmp.ip_sum = 0;
      data->sixto4ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header, IPv6 header, and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (icmp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);

      // Update packet IPv4 header, IPv6 header, and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (icmp_packet, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
      memcpy (icmp_packet + IP4_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
      memcpy (icmp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
    }

// Send UDP packet(s).
  } else if (data->packet_type == 7) {

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      if (data->sixto4_specify_ether == 1) {
        if ((sendto (sd, udp_ether_frame, udp_frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd, udp_packet, IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_payloadlen, 0, (struct sockaddr *) &sin, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_udp6to4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_udp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source IPv6 address if requested.
      if (data->ran_udp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_udp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_udp6_sourceport == 1) {
        data->udp6hdr.uh_sport = htons (ran_val2 ());
      }

      // Re-calculate IPv4 and upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // UDP header checksum (16 bits)
      data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->sixto4ip4hdr_udp.ip_sum = 0;
      data->sixto4ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header, IPv6 header, and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (udp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_udp, IP4_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);

      // Update packet IPv4 header, IPv6 header, and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (udp_packet, &data->sixto4ip4hdr_udp, IP4_HDRLEN);
      memcpy (udp_packet + IP4_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);
      memcpy (udp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);
    }

// Send all packet types.
  } else if (data->packet_type == 8) {

    // This state can be changed by on_button5_clicked()
    // as will flood_mode.
    looping = 1;

    i = data->npackets;  // Number of packets to send
    j = 0;               // Number of packets sent

    while ((i > 0) && (looping == 1)) {

      if (data->sixto4_specify_ether == 1) {
        if ((sendto (sd, tcp_ether_frame, tcp_frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
        if ((sendto (sd, icmp_ether_frame, icmp_frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
        if ((sendto (sd, udp_ether_frame, udp_frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      } else {
        if (sendto (sd, tcp_packet, IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN, 0, (struct sockaddr *) &sin, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
        if (sendto (sd, icmp_packet, IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_payloadlen, 0, (struct sockaddr *) &sin, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
        if (sendto (sd, udp_packet, IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_payloadlen, 0, (struct sockaddr *) &sin, sizeof (struct sockaddr)) < 0)  {
          status = errno;
          sprintf (text, "In callbacks.c: sendto() failed in sixto4_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Check for pending GTK click events.
      while (gtk_events_pending ()) {
        gtk_main_iteration ();
      }

      if (flood_mode == 0) i--;
      j++;

      // Randomize source IPv4 address if requested.
      if (data->ran_tcp6to4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_tcp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_icmp6to4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_icmp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_udp6to4_sourceip == 1) {
        if ((status = inet_pton (AF_INET, ran4_addr (ipaddress), &data->sixto4ip4hdr_udp.ip_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source IPv6 address if requested.
      if (data->ran_tcp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_tcp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_icmp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_icmp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }
      if (data->ran_udp6_sourceip == 1) {
        if ((status = inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_udp.ip6_src)) != 1) {
          sprintf (text, "In callbacks.c: inet_pton() failed in sixto4().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (ipaddress);
          g_free (tcp_packet);
          g_free (icmp_packet);
          g_free (udp_packet);
          g_free (tcp_ether_frame);
          g_free (icmp_ether_frame);
          g_free (udp_ether_frame);
          return (FALSE);
        }
      }

      // Randomize source port if requested.
      if (data->ran_tcp6_sourceport == 1) {
        data->tcp6hdr.th_sport = htons (ran_val2 ());
      }
      if (data->ran_udp6_sourceport == 1) {
        data->udp6hdr.uh_sport = htons (ran_val2 ());
      }

      // Re-calculate IPv4 and upper layer protocol checksum
      // regardless of whether source IP and/or port was changed.

      // Upper layer protocol header checksum (16 bits)
      data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
      data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
      data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);

      // IPv4 header checksum (16 bits): set to 0 when calculating checksum
      data->sixto4ip4hdr_tcp.ip_sum = 0;
      data->sixto4ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
      data->sixto4ip4hdr_icmp.ip_sum = 0;
      data->sixto4ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
      data->sixto4ip4hdr_udp.ip_sum = 0;
      data->sixto4ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp, IP4_HDRLEN);

      // Update ethernet frame IPv4 header, IPv6 header, and upper layer protocol header.
      // Payload data remains unchanged, so don't bother re-copying it.
      memcpy (tcp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN, &data->sixto4ip4hdr_udp, IP4_HDRLEN);

      memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);

      memcpy (tcp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
      memcpy (icmp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
      memcpy (udp_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);

      // Update packet IPv4 header, IPv6 header, and upper layer protocol header.
      // Payload data remain unchanged, so don't bother re-copying it.
      memcpy (tcp_packet, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
      memcpy (icmp_packet, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
      memcpy (udp_packet, &data->sixto4ip4hdr_udp, IP4_HDRLEN);

      memcpy (tcp_packet + IP4_HDRLEN, &data->ip6hdr_tcp, IP6_HDRLEN);
      memcpy (icmp_packet + IP4_HDRLEN, &data->ip6hdr_icmp, IP6_HDRLEN);
      memcpy (udp_packet + IP4_HDRLEN, &data->ip6hdr_udp, IP6_HDRLEN);

      memcpy (tcp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->tcp6hdr, TCP_HDRLEN);
      memcpy (icmp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->icmp6hdr, ICMP_HDRLEN);
      memcpy (udp_packet + IP4_HDRLEN + IP6_HDRLEN, &data->udp6hdr, UDP_HDRLEN);
    }

    close (sd);
  }

// Display packet activity in Activity Log.
  if (data->packet_type == 5) {
    if (data->sixto4_specify_ether == 1) {
      sprintf (text, "%i IPv6 TCP ethernet frames sent over IPv4 tunnel.\n", j);
    } else {
      sprintf (text, "%i IPv6 TCP packets sent over IPv4 tunnel.\n", j);
    }
  } else if (data->packet_type == 6) {
    if (data->sixto4_specify_ether == 1) {
      sprintf (text, "%i IPv6 ICMP ethernet frames sent over IPv4 tunnel.\n", j);
    } else {
      sprintf (text, "%i IPv6 ICMP packets sent over IPv4 tunnel.\n", j);
    }
  } else if (data->packet_type == 7) {
    if (data->sixto4_specify_ether == 1) {
      sprintf (text, "%i IPv6 UDP ethernet frames sent over IPv4 tunnel.\n", j);
    } else {
      sprintf (text, "%i IPv6 UDP packets sent over IPv4 tunnel.\n", j);
    }
  } else if (data->packet_type == 8) {
    sprintf (text, "%i IPv6 TCP+ICMP+UDP packets or ethernet frames sent over IPv4 tunnel.\n", j);
  }
  data->message = text;
  log_activity (data);
  memset (text, 0, 1024);

// Free allocated memory.
  g_free (text);
  g_free (ipaddress);
  g_free (tcp_packet);
  g_free (icmp_packet);
  g_free (udp_packet);
  g_free (tcp_ether_frame);
  g_free (icmp_ether_frame);
  g_free (udp_ether_frame);

  return (TRUE);
}

// Stop sending packets
gboolean
on_button5_clicked (GtkButton *button, SPSData *data)
{
// By toggling the checkbutton, the on_checkbutton11_toggled callback will be called
// and it will set flood_mode to 0 for us.
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton11), FALSE);

// If we're not flooding but sending a finite number of packets but want to stop,
// set looping to 0 to stop.
  looping = 0;

  return (TRUE);
}

// Add a message to the SPS activity log
void
log_activity (SPSData *data)
{
  GtkTextBuffer *buffer;
  GtkTextIter end;
  gchar *text;
  time_t rawtime;
  struct tm *timeinfo;
  char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  void *tmp;

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in log_activity().\n");
    exit (EXIT_FAILURE);
  }

  time (&rawtime);

// Express time as per local timezone, and as broken-down time
  timeinfo = localtime (&rawtime);

  sprintf (text, "%i ", timeinfo->tm_mday);
  sprintf (text, "%s%.3s ", text, mon_name[timeinfo->tm_mon]);
  sprintf (text, "%s%i ", text, timeinfo->tm_year + 1900);
  sprintf (text, "%s%.2d:%.2d:%.2d  ", text, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
  strcat (text, data->message);

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview1));
  gtk_text_buffer_get_end_iter (buffer, &end);
  gtk_text_buffer_insert (buffer, &end, text, -1);

  g_free (text);

  gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (data->textview1), &end, 0.0, FALSE, 0, 0);

}

// Function to convert 8 bit binary to decimal
unsigned char
bin8_to_dec (unsigned char *bin)
{
  int i;
  unsigned char dec;

  dec = 0;
  for (i=0; i<8; i++) {
    dec += (bin[i] << i);
  }

  return (dec);
}

// Function to generate random numbers between 1 and 254.
gushort
ran_val1 ()
{
  return (gushort) (1 + (gushort) (254.0 * rand () / (RAND_MAX + 1.0)));
}


// Function to generate random numbers between 0 and 65535.
gushort
ran_val2 ()
{
  return (gushort) ((gushort) (65535.0 * rand () / (RAND_MAX + 1.0)));
}

// Function to get this hostname and IP
gboolean
this_host (SPSData *data)
{
  int status;
  gchar *value, *hostname, *text;
  GtkTextBuffer *textbuffer4, *textbuffer5;
  struct addrinfo hints, *res, *p;
  char ipstr [INET6_ADDRSTRLEN];
  void *tmp, *ptr;

// Allocate memory for various arrays.
  ptr = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (ptr != NULL) {
    hostname = ptr;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'hostname' in this_host().\n");
    exit (EXIT_FAILURE);
  }

// Array for error messages.
  ptr = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (ptr != NULL) {
    text = ptr;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in this_host().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in this_host().\n");
    exit (EXIT_FAILURE);
  }

  gethostname (hostname, 1023);

  strcat (hostname, ".local");

  textbuffer4 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview4));
  gtk_text_buffer_set_text (textbuffer4, hostname, -1);

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

  if ((status = getaddrinfo (hostname, NULL, &hints, &res)) != 0) {
    sprintf (text, "In callbacks.c: getaddrinfo() failed in this_host().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (value);
    return (FALSE);
  }   
 
  for (p = res; p != NULL; p = p->ai_next) {

    switch (p->ai_family) {
      case AF_INET:
        ptr = &((struct sockaddr_in *) p->ai_addr)->sin_addr;
        break;
      case AF_INET6:
        ptr = &((struct sockaddr_in6 *) p->ai_addr)->sin6_addr;
        break;
    }

    inet_ntop (p->ai_family, ptr, ipstr, sizeof ipstr);
    sprintf (text, "%s%s\n", text, ipstr);
  }

  textbuffer5 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview5));
  gtk_text_buffer_set_text (textbuffer5, text, -1);

  g_free (hostname);
  g_free (text);
  g_free (value);
  freeaddrinfo (res);
  return (TRUE);
}

gboolean
show_packet (unsigned char *packet, gint nbytes, gchar *title, gint ether, gint sixto4)
{
  gint nchunks, i, j, c, remaindr, trunc;
  gchar *packetp, *packet_text;
  GtkWidget *window, *scrolled_win, *textview;
  GtkTextBuffer *buffer;
  PangoFontDescription *textview_font;
  void *tmp;

// Array of packet contents with only ASCII.
  tmp = (gchar *) g_malloc0 (nbytes * sizeof (gchar));
  if (tmp != NULL) {
    packetp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packetp' in show_packet().\n");
    exit (EXIT_FAILURE);
  }

// Same as above * 3 for ASCII and formatting text of packet.
  tmp = (gchar *) g_malloc0 ((nbytes * 30) * sizeof (gchar));
  if (tmp != NULL) {
    packet_text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'packet_text' in show_packet().\n");
    exit (EXIT_FAILURE);
  }

// Create new window to display packet contents
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
  gtk_window_set_title (GTK_WINDOW (window), title);
  gtk_window_set_default_size (GTK_WINDOW (window), 1000, 200);
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);

// Create a textview object
  textview = gtk_text_view_new ();
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), TRUE);
  textview_font = pango_font_description_from_string ("Courier 10 Pitch 12");
  gtk_widget_modify_font (textview, textview_font);

// Only allow alphanumerics and symbols
  for (i=0; i<nbytes; i++) {
    if ((packet[i] < 32) || (packet[i] > 126)) {
      packetp[i] = 32;  // 32 is a space
    } else {
      packetp[i] = packet[i];
    }
  }

// Don't show any ASCII for ethernet header if not specified.
  if (!ether) {
    for (i=0; i<ETH_HDRLEN; i++) {
      packetp[i] = 32;
    }
  }

// Show dashes if ethernet header is not specified.
  c = 0;
  if (!ether) {
    sprintf (packet_text, "%08x :", 0);
    for (i=0; i<8; i++) {
      sprintf (packet_text, "%s --", packet_text);
    }
    strcat (packet_text, " ");
    for (i=0; i<6; i++) {
      sprintf (packet_text, "%s --", packet_text);
    }
    c = ETH_HDRLEN;
  }

// Number of full lines nchunks (16 bytes per line), and
// partial lines remaindr.
  nchunks = (gint) (nbytes / 16);
  remaindr = nbytes % 16;
  // Truncate to only 40 lines if packet is longer.
  // Nobody is going to use SPS to examine their own data.
  // There are better tools for that.
  trunc = 0;
  if (nchunks > 20) {
    nchunks = 40;
    remaindr = 0;
    trunc = 1;
  }

// sprintf packet into packet_text

// Print first line which may or may not include specific ethernet header data.
  c = 0;
  sprintf (packet_text, "%08x :", 0);
  if (ether) {
    // Hexadecimal values
    for (j=0; j<8; j++) sprintf (packet_text, "%s %02x", packet_text, packet[j+c]);
    strcat (packet_text, " ");
    for (j=8; j<16; j++) sprintf (packet_text, "%s %02x", packet_text, packet[j+c]);
    strcat (packet_text, " ");
    // ASCII equivalent
    for (j=0; j<8; j++) sprintf (packet_text, "%s %c", packet_text, packetp[j+c]);
    strcat (packet_text, " ");
    for (j=8; j<16; j++) sprintf (packet_text, "%s %c", packet_text, packetp[j+c]);
    strcat (packet_text, "\n");
    c += 16;
  } else {
    // Put dashes where ethernet header (14 bytes) would be.
    for (j=0; j<8; j++) sprintf (packet_text, "%s --", packet_text);
    strcat (packet_text, " ");
    for (j=8; j<ETH_HDRLEN; j++) sprintf (packet_text, "%s --", packet_text);
    c += ETH_HDRLEN;
    // Print last two hexadecimal bytes.
    sprintf (packet_text, "%s %02x", packet_text, packet[c]);
    c++;
    sprintf (packet_text, "%s %02x", packet_text, packet[c]);
    strcat (packet_text, "\n");
    c++;
  }

// Print rest of packet contents except any partial line.
  for (i=1; i<nchunks; i++) {
    sprintf (packet_text, "%s%08x :", packet_text, (i * 16));
    for (j=0; j<8; j++) sprintf (packet_text, "%s %02x", packet_text, packet[j+c]);
    strcat (packet_text, " ");
    for (j=8; j<16; j++) sprintf (packet_text, "%s %02x", packet_text, packet[j+c]);
    strcat (packet_text, " ");
    for (j=0; j<8; j++) sprintf (packet_text, "%s %c", packet_text, packetp[j+c]);
    strcat (packet_text, " ");
    for (j=8; j<16; j++) sprintf (packet_text, "%s %c", packet_text, packetp[j+c]);
    strcat (packet_text, "\n");
    c += 16;
  }

// Print address for last line
  sprintf (packet_text, "%s%08x : ", packet_text, (nchunks * 16));

// Write the final (partial) line of hexadecimals
  if (remaindr <= 8) {
    for (i=0; i<remaindr; i++) sprintf (packet_text, "%s%02x ", packet_text, packet[(nchunks * 16) + i]);
    for (i=0; i<(8-remaindr); i++) sprintf (packet_text, "%s   ", packet_text);
    strcat (packet_text, " ");
    for (i=0; i<8; i++) sprintf (packet_text, "%s   ", packet_text);
  } else if (remaindr > 8) {
    for (i=0; i<8; i++) sprintf (packet_text, "%s%02x ", packet_text, packet[(nchunks * 16) + i]);
    strcat (packet_text, " ");
    for (i=8; i<remaindr; i++) sprintf (packet_text, "%s%02x ", packet_text, packet[(nchunks * 16) + i]);
    for (i=0; i<(16-remaindr); i++) sprintf (packet_text, "%s   ", packet_text);
  }
  strcat (packet_text, " ");

// Write the final (partial) line of ASCII characters.
  if (remaindr <= 8) {
    for (i=0; i<remaindr; i++) sprintf (packet_text, "%s%c ", packet_text, packetp[(nchunks * 16) + i]);

  } else if (remaindr > 8) {
    for (i=0; i<8; i++) sprintf (packet_text, "%s%c ", packet_text, packetp[(nchunks * 16) + i]);

    strcat (packet_text, " ");

    for (i=8; i<remaindr; i++) sprintf (packet_text, "%s%c ", packet_text, packetp[(nchunks * 16) + i]);
  }

// Add note for explanation of dashes.
  if (!ether) {
   sprintf (packet_text, "%s\n\nNote: -- to be determined by kernel.", packet_text);
  }

// Add note if packet data truncated.
  if (trunc == 1) {
   sprintf (packet_text, "%s\n\nNote: Only first 20 lines of packet shown.", packet_text);
  }

// Add note if packet includes 6to4 ethernet and IPv4 headers.
  if (sixto4 == 1) {
   sprintf (packet_text, "%s\n\nNote: Displayed packet includes 6to4 ethernet and IPv4 headers.", packet_text);
  }

// Put the text into the buffer.
  gtk_text_buffer_set_text (buffer, packet_text, -1);

// Add scrolled window and textview to our new window and display.
  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), textview);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  gtk_container_add (GTK_CONTAINER (window), scrolled_win);
  gtk_widget_show_all (window);

  g_free (packetp);
  g_free (packet_text);

  return (TRUE);
}

// Do some rough checks to see if address is a valid IPv4 address
gboolean
is_valid_ip4 (const gchar *ip)
{
  guint n1, n2, n3, n4;
  gchar *buf;
  void *tmp;

  tmp = (gchar *) g_malloc0 (64 * sizeof (gchar));
  if (tmp != NULL) {
    buf = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buf' in is_valid_ip4().\n");
    exit (EXIT_FAILURE);
  }

// Less than four values given? (truncates extras if more than four)
  if (sscanf (ip, "%u.%u.%u.%u", &n1, &n2, &n3, &n4) != 4) {
    g_free (buf);
    return (FALSE);
  }

// All four are within 0 and 255, and first isn't 0?
  if ((n1 != 0) && (n1 <= 255) && (n2 <= 255) && (n3 <= 255) && (n4 <= 255)) {

//  OK, but any weird characters? (note: negative signs are weird) Compare input ip with reconstructed to check.
    sprintf (buf, "%u.%u.%u.%u", n1, n2, n3, n4);
    if (!strcmp (buf, ip)) {
      g_free (buf);
      return (TRUE);
    } else {
      g_free (buf);
      return (FALSE);  // A weird character was detected.
    }
  } else {
    g_free (buf);
    return (FALSE);  // Out of range value detected.
  }
}

// Create a random IPv4 address.
gchar *
ran4_addr (gchar *ipaddress)
{
  gushort a, b, c, d;

  a = ran_val1 ();
  b = ran_val1 ();
  c = ran_val1 ();
  d = ran_val1 ();

  sprintf (ipaddress,"%u.%u.%u.%u", a, b, c, d);

  return (ipaddress);
}

// Create a random IPv6 address.
gchar *
ran6_addr (gchar *ipaddress)
{
  gushort a, b, c, d, e, f, g, h;

  a = ran_val2 ();
  b = ran_val2 ();
  c = ran_val2 ();
  d = ran_val2 ();
  e = ran_val2 ();
  f = ran_val2 ();
  g = ran_val2 ();
  h = ran_val2 ();

  sprintf (ipaddress,"%x:%x:%x:%x:%x:%x:%x:%x", a, b, c, d, e, f, g, h);

  return (ipaddress);
}

// Report a warning condition to a GtkDialog
gboolean
report_warning (SPSData *data)
{
  GtkWidget *dialog, *content_area, *image, *label, *hbox;

// data->parent is a pointer to the widget whence the warning originated.
// It can point to things like data->main_window or data->menubar1
  dialog = gtk_dialog_new_with_buttons ("Warning", GTK_WINDOW (data->parent),
                                  GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

  gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);  // No separater

  g_signal_connect_swapped (dialog, "response", G_CALLBACK (gtk_widget_destroy), dialog);

  content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));

  label = gtk_label_new (data->message);
  image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_WARNING,
                                    GTK_ICON_SIZE_DIALOG);

  hbox = gtk_hbox_new (FALSE, 5);  // Don't require homogeneous, 5 px spacing
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);  // 10 px
  gtk_box_pack_start_defaults (GTK_BOX (hbox), image);  // Add warning symbol to hbox
  gtk_box_pack_start_defaults (GTK_BOX (hbox), label);  // Add our text to hbox

  // Pack the dialog content into the dialog's GtkVBox.
  gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox);

  gtk_widget_show_all (dialog);

  gtk_dialog_run (GTK_DIALOG (dialog));

  return (TRUE);
}

// Report an error condition to a GtkDialog
gboolean
report_error (SPSData *data)
{
  GtkWidget *dialog, *content_area, *image, *label, *hbox;

// data->parent is a pointer to the widget whence the error originated.
// It can point to things like data->main_window or data->menubar1
  dialog = gtk_dialog_new_with_buttons ("Error", GTK_WINDOW (data->parent),
                                  GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

  gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);  // No separater

  g_signal_connect_swapped (dialog, "response", G_CALLBACK (gtk_widget_destroy), dialog);

  content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));

  label = gtk_label_new (data->message);
  image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_ERROR,
                                    GTK_ICON_SIZE_DIALOG);

  hbox = gtk_hbox_new (FALSE, 5);  // Don't require homogeneous, 5 px spacing
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);  // 10 px
  gtk_box_pack_start_defaults (GTK_BOX (hbox), image);  // Add error symbol to hbox
  gtk_box_pack_start_defaults (GTK_BOX (hbox), label);  // Add our text to hbox

  // Pack the dialog content into the dialog's GtkVBox.
  gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox);

  gtk_widget_show_all (dialog);

  gtk_dialog_run (GTK_DIALOG (dialog));

  return (TRUE);
}

// Send an ARP request to obtain link-layer address.
gboolean
arp (unsigned char *dst_mac, SPSData *data)
{
  int nbytes, status;
  const gchar *entry137_text, *entry140_text;
  gchar *text, *hostname;
  gint i, frame_length, sd;
  gchar *interface, *target, *src_ip;
  arp_hdr *arphdr;
  guchar *src_mac, *ether_frame;
  struct addrinfo hints, *res;
  struct sockaddr_in *ipv4;
  struct sockaddr_ll device;
  struct ifreq ifr;
  struct timeval tv;
  void *tmp;

// Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    hostname = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'hostname' in arp().\n");
    exit (EXIT_FAILURE);
  }

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in arp().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (6 * sizeof (guchar));
  if (tmp != NULL) {
    src_mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'src_mac' in arp().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (IP_MAXPACKET * sizeof (guchar));
  if (tmp != NULL) {
    ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ether_frame' in arp().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    interface = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'interface' in arp().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    target = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'target' in arp().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    src_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'src_ip' in arp().\n");
    exit (EXIT_FAILURE);
  }

  arphdr = (arp_hdr *) (ether_frame + ETH_HDRLEN);

// Get interface name from text entry
  entry140_text = gtk_entry_get_text (GTK_ENTRY (data->entry140));

// Is an interface name specified in text entry 140?
  if (strlen (entry140_text) < 2) {
    sprintf (text, "In callbacks.c: Please specify an interface name for arp().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }

// Interface to send packet through.
  strcpy (interface, entry140_text);

// Clear IPv6 textentry so there's no confusion about which IP belongs to the MAC address.
  gtk_entry_set_text (GTK_ENTRY (data->entry138), "");

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in arp().\nError Message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", interface);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in arp().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }
  close (sd);

// Copy source MAC address.
  memcpy (src_mac, ifr.ifr_hwaddr.sa_data, 6);

// Resolve interface index.
  if ((device.sll_ifindex = if_nametoindex (interface)) == 0) {
    status = errno;
    sprintf (text, "In callbacks.c: if_nametoindex() failed to obtain interface index in arp().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }

// Set destination MAC address: broadcast address
  memset (dst_mac, 0xff, 6);

// Source IPv4 address (this host)
  gethostname (hostname, 1023);
  strcat (hostname, ".local");
  strcpy (src_ip, thishost_ipv4 (hostname));

// Destination IPv4 address
  entry137_text = gtk_entry_get_text (GTK_ENTRY (data->entry137));
  strcpy (target, entry137_text);

// Fill out hints for getaddrinfo().
  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

// Resolve source using getaddrinfo().
  if ((status = getaddrinfo (src_ip, NULL, &hints, &res)) != 0) {
    sprintf (text, "In callbacks.c: getaddrinfo() failed for source host in arp().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }
  ipv4 = (struct sockaddr_in *) res->ai_addr;
  memcpy (&arphdr->sender_ip, &ipv4->sin_addr, 4);
  freeaddrinfo (res);

// Resolve target using getaddrinfo().
  if ((status = getaddrinfo (target, NULL, &hints, &res)) != 0) {
    sprintf (text, "In callbacks.c: getaddrinfo() failed for target host in arp().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }
  ipv4 = (struct sockaddr_in *) res->ai_addr;
  memcpy (&arphdr->target_ip, &ipv4->sin_addr, 4);
  freeaddrinfo (res);

// Fill out sockaddr_ll.
  device.sll_family = AF_PACKET;
  memcpy (device.sll_addr, src_mac, 6);
  device.sll_halen = htons (6u);

// ARP header

// Hardware type (16 bits): 1 for ethernet
  arphdr->htype = htons (1u);

// Protocol type (16 bits): 2048 for IP
  arphdr->ptype = htons (ETH_P_IP);

// Hardware address length (8 bits): 6 bytes for MAC address
  arphdr->hlen = 6;

// Protocol address length (8 bits): 4 bytes for IPv4 address
  arphdr->plen = 4;

// OpCode: 1 for ARP request
  arphdr->opcode = htons (ARPOP_REQUEST);

// Sender hardware address (48 bits): MAC address
  memcpy (&arphdr->sender_mac, src_mac, 6);

// Sender protocol address (32 bits)
// See getaddrinfo() resolution of src_ip.

// Target hardware address (48 bits): zero, since we don't know it yet.
  memset (&arphdr->target_mac, 0, 6);

// Target protocol address (32 bits)
// See getaddrinfo() resolution of target.

// Fill out ethernet frame header.

// Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (ARP header)
  frame_length = ETH_HDRLEN + ARP_HDRLEN;

// Destination and Source MAC addresses
  memcpy (ether_frame, dst_mac, 6);
  memcpy (ether_frame + 6, src_mac, 6);

// Next is ethernet type code (ETH_P_ARP for ARP).
// http://www.iana.org/assignments/ethernet-numbers
  ether_frame[12] = ETH_P_ARP / 256;
  ether_frame[13] = ETH_P_ARP % 256;

// Next is ethernet frame data (ARP header).

// ARP header
  memcpy (ether_frame + ETH_HDRLEN, arphdr, ARP_HDRLEN);

// Submit request for a raw socket descriptor.
  if ((sd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
    perror ("socket() failed ");
    exit (EXIT_FAILURE);
  }

// Send ethernet frame to socket.
  if (sendto (sd, ether_frame, frame_length, 0, (struct sockaddr *) &device, sizeof (device)) <= 0) {
    status = errno;
    sprintf (text, "In callbacks.c: sendto() failed in arp().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (hostname);
    g_free (text);
    g_free (src_mac);
    g_free (ether_frame);
    g_free (interface);
    g_free (target);
    g_free (src_ip);
    return (FALSE);
  }

// Set time for the socket to timeout and give up waiting for an ARP reply.
  tv.tv_sec  = data->timeout;  
  tv.tv_usec = 0;
  setsockopt (sd, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv, sizeof (struct timeval));

// Listen for incoming ethernet frame from socket sd.
// We expect an ARP ethernet frame of the form:
//     MAC (6 bytes) + MAC (6 bytes) + ethernet type (2 bytes)
//     + ethernet data (ARP header) (28 bytes)
// Keep at it for data->timeout seconds, or until we get an ARP reply.
  memset (ether_frame, 0, IP_MAXPACKET);
  arphdr = (arp_hdr *) (ether_frame + ETH_HDRLEN);
  while (((((ether_frame[12]) << 8) + ether_frame[13]) != ETH_P_ARP) || (ntohs (arphdr->opcode) != ARPOP_REPLY)) {
    if ((nbytes = recv (sd, ether_frame, IP_MAXPACKET, 0)) < 0) {
      status = errno;
      if (status == EAGAIN) {
        sprintf (text, "In callbacks.c: recv() failed in arp().\nNo reply from that node within %i seconds.", data->timeout);
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (hostname);
        g_free (text);
        g_free (src_mac);
        g_free (ether_frame);
        g_free (interface);
        g_free (target);
        g_free (src_ip);
        return (FALSE);
      } else if (status == EINTR) {
        memset (ether_frame, 0, IP_MAXPACKET);
        continue;  // Something weird happened, but let's try again.
      } else {
        sprintf (text, "In callbacks.c: recv() failed in arp().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (hostname);
        g_free (text);
        g_free (src_mac);
        g_free (ether_frame);
        g_free (interface);
        g_free (target);
        g_free (src_ip);
        return (FALSE);
      }
    }
  }

// Close socket descriptor.
  close (sd);

// Extract MAC address from received ethernet frame.
  for (i=0; i<6; i++) {
    dst_mac[i] = ether_frame[i+6];
  }

// Free allocated memory.
  g_free (hostname);
  g_free (text);
  g_free (src_mac);
  g_free (ether_frame);
  g_free (interface);
  g_free (target);
  g_free (src_ip);
  return (TRUE);
}

// Send a neighbor solicitation to obtain link-layer address.
gboolean
neighbor_discovery (unsigned char *dst_mac, SPSData *data)
{
  const gchar *entry138_text, *entry140_text;
  gchar *text;
  int NS_HDRLEN = sizeof (struct nd_neighbor_solicit);  // Length of NS message header
  int optlen = 8; // Option Type (1 byte) + Length (1 byte) + Length of MAC address (6 bytes)

  int i, nbytes, sd, on, status, ifindex, cmsglen, hoplimit, psdhdrlen;
  struct addrinfo hints;
  struct addrinfo *res;
  struct sockaddr_in6 *ipv6, src, dst, dstsnmc;
  struct nd_neighbor_solicit *ns;
  struct nd_neighbor_advert *na;
  unsigned char *outpack, *options, *psdhdr, *inpack, *pkt;
  struct msghdr msghdr;
  struct ifreq ifr;
  struct cmsghdr *cmsghdr1, *cmsghdr2;
  struct in6_pktinfo *pktinfo;
  struct iovec iov[2];
  char *target, *source, *interface;
  struct timeval tv;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (20 * sizeof (char));
  if (tmp != NULL) {
    interface = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'interface' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    target = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'target' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    source = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'source' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    outpack = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'outpack' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (optlen * sizeof (unsigned char));
  if (tmp != NULL) {
    options = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'options' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    psdhdr = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'psdhdr' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    inpack = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'inpack' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }

// Get interface name from text entry
  entry140_text = gtk_entry_get_text (GTK_ENTRY (data->entry140));

// Is an interface name specified in text entry 140?
  if (strlen (entry140_text) < 2) {
    sprintf (text, "In callbacks.c: Please specify an interface name for neighbor_discovery().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    return (FALSE);
  }

// Interface to send packet through.
  strcpy (interface, entry140_text);

// Clear IPv4 textentry so there's no confusion about which IP belongs to the MAC address.
  gtk_entry_set_text (GTK_ENTRY (data->entry137), "");

// Submit request for a socket descriptor.
  if ((sd = socket (AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in neighbor_discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", interface);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in neighbor discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    return (FALSE);
  }

// Copy source MAC address into options buffer.
  options[0] = 1;           // Option Type - "source link layer address" (Section 4.6 of RFC 4861)
  options[1] = optlen / 8;  // Option Length - units of 8 octets (RFC 4861)
  for (i=0; i<6; i++) {
    options[i+2] = (unsigned char) ifr.ifr_addr.sa_data[i];
  }

// Retrieve source interface index.
  if ((ifindex = if_nametoindex (interface)) == 0) {
    status = errno;
    sprintf (text, "In callbacks.c: if_nametoindex() failed to obtain interface index in neighbor_discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    return (FALSE);
  }

// Source IPv6 address: the IPv6 unspecified address
  strcpy (source, "::");

// Target URL or IPv6 address (node we're requesting an advertisement from).
  entry138_text = gtk_entry_get_text (GTK_ENTRY (data->entry138));
  strcpy (target, entry138_text);

// Fill out hints for getaddrinfo().
  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_INET6;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

// Resolve source using getaddrinfo().
  if ((status = getaddrinfo (source, NULL, &hints, &res)) != 0) {
    sprintf (text, "In callbacks.c: getaddrinfo() failed to resolve the IPv6 unspecified address (::) in neighbor_discovery().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    return (FALSE);
  }
  memcpy (&src, res->ai_addr, res->ai_addrlen);
  memcpy (psdhdr, src.sin6_addr.s6_addr, 16);  // Copy to checksum pseudo-header
  freeaddrinfo (res);

// Resolve target using getaddrinfo().
  if ((status = getaddrinfo (target, NULL, &hints, &res)) != 0) {
    fprintf (stderr, "getaddrinfo() failed: %s\n", gai_strerror (status));
    sprintf (text, "In callbacks.c: getaddrinfo() failed to resolve target hostname/IPv6 address in neighbor_discovery().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    return (FALSE);
  }
  memcpy (&dst, res->ai_addr, res->ai_addrlen);
  memcpy (&dstsnmc, res->ai_addr, res->ai_addrlen); 

// Report target's unicast address.
  ipv6 = (struct sockaddr_in6 *) res->ai_addr;
  tmp = &(ipv6->sin6_addr);
  memset (target, 0, 40);
  inet_ntop (AF_INET6, tmp, target, 40);
  freeaddrinfo (res);

// Convert target's IPv6 unicast address to solicited-node multicast address.
// Section 2.7.1 of RFC 4291.
  dstsnmc.sin6_addr.s6_addr[0]= 255;
  dstsnmc.sin6_addr.s6_addr[1]=2;
  for (i=2; i<11; i++) {
    dstsnmc.sin6_addr.s6_addr[i] = 0;
  }
  dstsnmc.sin6_addr.s6_addr[11]=1;
  dstsnmc.sin6_addr.s6_addr[12]=255;
  memcpy (psdhdr + 16, dstsnmc.sin6_addr.s6_addr, 16);  // Solicited-node multicast address goes into pseudo-header

// Define first part of buffer outpack to be a neighbor solicit struct.
  ns = (struct nd_neighbor_solicit *) outpack;
  memset (ns, 0, sizeof (*ns));

// Populate icmp6_hdr portion of neighbor solicit struct.
  ns->nd_ns_hdr.icmp6_type = ND_NEIGHBOR_SOLICIT;  // 135 (RFC 4861)
  ns->nd_ns_hdr.icmp6_code = 0;               // zero for neighbor solicitation (RFC 4861)
  ns->nd_ns_hdr.icmp6_cksum = htons(0);      // zero when calculating checksum
  ns->nd_ns_reserved = htonl(0);             // Reserved - must be set to zero (RFC 4861)
  ns->nd_ns_target = dst.sin6_addr;           // Target address (NOT MULTICAST) (as type in6_addr)

// Append options to end of neighbor solicit struct.
  memcpy (outpack + NS_HDRLEN, options, optlen);

// Need a pseudo-header for checksum calculation. Define length. (RFC 2460)
// Length = source IP (16 bytes) + destination IP (16 bytes)
//        + upper layer packet length (4 bytes) + zero (3 bytes)
//        + next header (1 byte)
  psdhdrlen = 16 + 16 + 4 + 3 + 1 + NS_HDRLEN + optlen;

// Prepare msghdr for sendmsg().
  memset (&msghdr, 0, sizeof (msghdr));
  msghdr.msg_name = &dstsnmc;  // Destination IPv6 address (solicited node multicast) (as struct sockaddr_in6)
  msghdr.msg_namelen = sizeof (dstsnmc);
  memset (&iov, 0, sizeof (iov));
  iov[0].iov_base = (unsigned char *) outpack;  // Point msghdr to buffer outpack
  iov[0].iov_len = NS_HDRLEN + optlen;
  msghdr.msg_iov = iov;                 // scatter/gather array
  msghdr.msg_iovlen = 1;                // number of elements in scatter/gather array

// Tell msghdr we're adding cmsghdr data to change hop limit and specify interface.
// Allocate some memory for our cmsghdr data.
  cmsglen = CMSG_SPACE (sizeof (int)) + CMSG_SPACE (sizeof (struct in6_pktinfo));
  tmp = (unsigned char *) malloc (cmsglen * sizeof (unsigned char));
  if (tmp != NULL) {
    msghdr.msg_control = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'msghdr.msg_control' in neighbor_discovery().\n");
    exit (EXIT_FAILURE);
  }
  memset (msghdr.msg_control, 0, cmsglen);
  msghdr.msg_controllen = cmsglen;

// Change hop limit to 255 as required for neighbor solicitation (RFC 4861).
  hoplimit = 255;
  cmsghdr1 = CMSG_FIRSTHDR (&msghdr);
  cmsghdr1->cmsg_level = IPPROTO_IPV6;
  cmsghdr1->cmsg_type = IPV6_HOPLIMIT;  // We want to change hop limit
  cmsghdr1->cmsg_len = CMSG_LEN (sizeof (int));
  *((int *) CMSG_DATA (cmsghdr1)) = hoplimit;

// Specify source interface index for this packet via cmsghdr data.
  cmsghdr2 = CMSG_NXTHDR (&msghdr, cmsghdr1);
  cmsghdr2->cmsg_level = IPPROTO_IPV6;
  cmsghdr2->cmsg_type = IPV6_PKTINFO;  // We want to specify interface here
  cmsghdr2->cmsg_len = CMSG_LEN (sizeof (struct in6_pktinfo));
  pktinfo = (struct in6_pktinfo *) CMSG_DATA (cmsghdr2);
  pktinfo->ipi6_ifindex = ifindex;

// Compute ICMPv6 checksum (RFC 2460).
// psdhdr[0 to 15] = source IPv6 address, set earlier.
// psdhdr[16 to 31] = destination IPv6 address, set earlier.
  psdhdr[32] = 0;  // Length should not be greater than 65535 (i.e., 2 bytes)
  psdhdr[33] = 0;  // Length should not be greater than 65535 (i.e., 2 bytes)
  psdhdr[34] = (NS_HDRLEN + optlen)  / 256;  // Upper layer packet length
  psdhdr[35] = (NS_HDRLEN + optlen)  % 256;  // Upper layer packet length
  psdhdr[36] = 0;  // Must be zero
  psdhdr[37] = 0;  // Must be zero
  psdhdr[38] = 0;  // Must be zero
  psdhdr[39] = IPPROTO_ICMPV6;
  memcpy (psdhdr + 40, outpack, NS_HDRLEN + optlen);
  ns->nd_ns_hdr.icmp6_cksum = checksum ((unsigned short int *) psdhdr, psdhdrlen);

// Send packet.
  if (sendmsg (sd, &msghdr, 0) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: sendmsg() failed in neighbor_discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    g_free (msghdr.msg_control);
    return (FALSE);
  }

// Prepare msghdr for recvmsg().
  memset (&msghdr, 0, sizeof (msghdr));
  msghdr.msg_name = NULL;
  msghdr.msg_namelen = 0;
  memset (&iov, 0, sizeof (iov));
  iov[0].iov_base = (unsigned char *) inpack;
  iov[0].iov_len = IP_MAXPACKET;
  msghdr.msg_iov = iov;
  msghdr.msg_iovlen = 1;

  msghdr.msg_controllen = IP_MAXPACKET * sizeof (unsigned char);

// Set flag so we receive hop limit from recvmsg.
  on = 1;
  if ((status = setsockopt (sd, IPPROTO_IPV6, IPV6_RECVHOPLIMIT, &on, sizeof (on))) < 0) {
    sprintf (text, "In callbacks.c: setsockopt() failed to set IPV6_RECVHOPLIMIT in neighbor_discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    g_free (msghdr.msg_control);
    return (FALSE);
  }

// Set flag so we receive destination address from recvmsg.
  on = 1;
  if ((status = setsockopt (sd, IPPROTO_IPV6, IPV6_RECVPKTINFO, &on, sizeof (on))) < 0) {
    sprintf (text, "In callbacks.c: setsockopt() failed to set IPV6_RECVPKTINFO in neighbor_discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    g_free (msghdr.msg_control);
    return (FALSE);
  }

// Bind socket to interface of this node.
  if ((status = setsockopt (sd, SOL_SOCKET, SO_BINDTODEVICE, (void *) &ifr, sizeof (ifr))) < 0) {
    sprintf (text, "In callbacks.c: setsockopt() failed to bind socket to interface in neighbor_discovery().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (interface);
    g_free (target);
    g_free (source);
    g_free (outpack);
    g_free (options);
    g_free (psdhdr);
    g_free (inpack);
    g_free (msghdr.msg_control);
    return (FALSE);
  }

// Set time for the socket to timeout and give up waiting for a neighbor advertisement.
  tv.tv_sec  = data->timeout;
  tv.tv_usec = 0;
  setsockopt (sd, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv, sizeof (struct timeval));

// Listen for incoming message from socket sd.
// Keep at it until we get a neighbor advertisement.
  na = (struct nd_neighbor_advert *) inpack;
  while (na->nd_na_hdr.icmp6_type != ND_NEIGHBOR_ADVERT) {
    if ((nbytes = recvmsg (sd, &msghdr, 0)) < 0) {
      status = errno;
      if (status == EAGAIN) {
        sprintf (text, "In callbacks.c: recvmsg() failed in neighbor_discovery().\nNo reply from that node within %i seconds.", data->timeout);
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (interface);
        g_free (target);
        g_free (source);
        g_free (outpack);
        g_free (options);
        g_free (psdhdr);
        g_free (inpack);
        g_free (msghdr.msg_control);
        return (FALSE);
      } else if (status == EINTR) {
        continue;  // Something weird happened, but let's try again.
      } else {
        sprintf (text, "In callbacks.c: recvmsg() failed in neighbor_discovery().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (interface);
        g_free (target);
        g_free (source);
        g_free (outpack);
        g_free (options);
        g_free (psdhdr);
        g_free (inpack);
        g_free (msghdr.msg_control);
        return (FALSE);
      }
    }
  }

// Close socket descriptor.
  close (sd);

// Extract MAC address from received ethernet frame.
  pkt = (unsigned char *) inpack;
  for (i=2; i<8; i++) {
    dst_mac[i-2] = pkt[sizeof (struct nd_neighbor_advert) + i];
  }

// Free allocated memory.
  g_free (text);
  g_free (interface);
  g_free (target);
  g_free (source);
  g_free (outpack);
  g_free (options);
  g_free (psdhdr);
  g_free (inpack);
  g_free (msghdr.msg_control);

  return (TRUE);
}

// Convert link-layer (MAC) address from binary to string
gboolean
mac_bin2string (unsigned char *mac, char *string)
{
  int i;

// NOTE: we assume string was set to zero prior to call to this function.

  for (i=0; i<5; i++) {
    sprintf (string, "%s%02x:", string, mac[i]);
  }
  sprintf (string, "%s%02x", string, mac[5]);

  return (TRUE);
}

// Convert link-layer (MAC) address from string to binary
gboolean
mac_string2bin (char *string, unsigned char *mac)
{
  unsigned int a, b, c, d, e, f;
  
  sscanf (string, "%02x:%02x:%02x:%02x:%02x:%02x", &a, &b, &c, &d, &e, &f);

  mac[0] = a;
  mac[1] = b;
  mac[2] = c;
  mac[3] = d;
  mac[4] = e;
  mac[5] = f;

  return (TRUE);
}

// Load file ...
gboolean
on_button21_clicked (GtkButton *button, SPSData *data)
{
  int i, buffer_size, index;
  gint result;
  struct ethhdr *ethhdr;
  struct ip *ip4hdr;
  struct ip6_hdr *ip6hdr;
  struct tcphdr *tcphdr;
  struct icmp *icmp4hdr;
  struct icmp6_hdr *icmp6hdr;
  struct udphdr *udphdr;
  int16_t *ival16;
  int32_t *ival32;
  GtkWidget *dialog;
  gchar *filename, *text;
  unsigned char *buffer, *flags;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Size of buffer
  buffer_size = 400000;

// Array for error messages and various calculated values.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button21_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (buffer_size * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button21_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));
  if (tmp != NULL) {
    flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'flags' in on_button21_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    g_free (flags);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button21_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    g_free (flags);
    return (FALSE);
  }

// Read entire file into buffer.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > buffer_size) {
      sprintf (text, "This does not appear to be an SPS file.");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      fclose (fi);
      g_free (text);
      g_free (buffer);
      g_free (flags);
      return (FALSE);
    }
  }

// Close file descriptor.
  fclose (fi);

// Check for SPS file identification
  index = 0;
  if (strncmp ((char *) buffer, "Simple Packet Sender", 20) != 0) {
    sprintf (text, "This does not appear to be an SPS file.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    g_free (flags);
    return (FALSE);
  }
  index += 80;

// Retrieve the Save Data flags
  for (i=0; i<8; i++) {
    flags[i] = buffer[index+i];
  }
  // Skip through reserved area
  index += (256);

// Send Packet Settings
  if (flags[0] == 1) {

    // Packet type(s) to send
    data->packet_type = buffer[index];
    if (data->packet_type == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton1), TRUE);
    } else if (data->packet_type == 2) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton2), TRUE);
    } else if (data->packet_type == 3) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton3), TRUE);
    } else if (data->packet_type == 4) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton4), TRUE);
    } else if (data->packet_type == 5) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton5), TRUE);
    } else if (data->packet_type == 6) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton6), TRUE);
    } else if (data->packet_type == 7) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton7), TRUE);
    } else if (data->packet_type == 8) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton8), TRUE);
    }
    index++;

    // Number of packets
    ival32 = (int32_t *) (buffer + index);    // Cast ival32 as pointer to 32 bit (4 byte) int within buffer
    data->npackets = (gint) ntohl (*ival32);  // Value was stored in network byte order for consistent endianess across platforms
    index += 4;
    sprintf (text, "%u", data->npackets);
    gtk_entry_set_text (GTK_ENTRY (data->entry18), text);

    // Flag for flood mode
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton11), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton11), TRUE);
    }
    // Variable flood_mode will be changed by on_checkbutton11_toggled(), so set it now.
    flood_mode = buffer[index];
    index++;
  }

// IPv6 over IPv4 (6to4)
  if (flags[1] == 1) {

    // Flag to use 6to4
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton12), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton12), TRUE);
    }
    // Variable data->sixto4_flag will be changed by on_checkbutton12_toggled(), so set it now.
    data->sixto4_flag = buffer[index];
    index++;

    // Flag to specify ethernet header
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton14), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton14), TRUE);
    }
    // Variable data->sixto4_specify_ether will be changed by on_checkbutton14_toggled(), so set it now.
    data->sixto4_specify_ether = buffer[index];
    index++;

    // 6to4 interface name
    for (i=0; i<255; i++) {
      data->sixto4_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry158), data->sixto4_interfacename);

    // 6to4 ethernet header for IPv4 header
    ethhdr = (struct ethhdr *) (buffer + index);       // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->sixto4ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->sixto4ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry156), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->sixto4ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry157), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->sixto4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry159), text);
    index += ETH_HDRLEN;

    // 6to4 IPv4 header for TCP
    ip4hdr = (struct ip *) (buffer + index);               // Cast ip4hdr as pointer to IPv4 header struct within buffer
    memcpy (&data->sixto4ip4hdr_tcp, ip4hdr, IP4_HDRLEN);  // Extract IPv4 header
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_tcp.ip_src), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry125), text);
    index += IP4_HDRLEN;

    // Flag for randomizing IPv4 source address for TCP over 6to4
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton13), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton13), TRUE);
    }
    // Variable data->ran_tcp6to4_sourceip will be changed by on_checkbutton13_toggled(), so set it now.
    data->ran_tcp6to4_sourceip = buffer[index];
    index++;

    // 6to4 IPv4 header for ICMP
    ip4hdr = (struct ip *) (buffer + index);                // Cast ip4hdr as pointer to IPv4 header struct within buffer
    memcpy (&data->sixto4ip4hdr_icmp, ip4hdr, IP4_HDRLEN);  // Extract IPv4 header
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_icmp.ip_src), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry160), text);
    index += IP4_HDRLEN;

    // Flag for randomizing IPv4 source address for ICMP over 6to4
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton26), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton26), TRUE);
    }
    // Variable data->ran_icmp6to4_sourceip will be changed by on_checkbutton26_toggled(), so set it now.
    data->ran_icmp6to4_sourceip = buffer[index];
    index++;

    // 6to4 IPv4 header for UDP
    ip4hdr = (struct ip *) (buffer + index);               // Cast ip4hdr as pointer to IPv4 header struct within buffer
    memcpy (&data->sixto4ip4hdr_udp, ip4hdr, IP4_HDRLEN);  // Extract IPv4 header
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_udp.ip_src), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry161), text);
    index += IP4_HDRLEN;

    // Flag for randomizing IPv4 source address for UDP over 6to4
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton27), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton27), TRUE);
    }
    // Variable data->ran_udp6to4_sourceip will be changed by on_checkbutton27_toggled(), so set it now.
    data->ran_udp6to4_sourceip = buffer[index];
    index++;
  }

// IPv4 TCP
  if (flags[2] == 1) {

    // Flag to specify ethernet header
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton18), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton18), TRUE);
    }
    // Variable data->tcp4_specify_ether will be changed by on_checkbutton18_toggled(), so set it now.
    data->tcp4_specify_ether = buffer[index];
    index++;

    // Interface name for IPv4 TCP
    for (i=0; i<255; i++) {
      data->tcp4_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry150), data->tcp4_interfacename);

    // Ethernet header for IPv4 header for TCP
    ethhdr = (struct ethhdr *) (buffer + index);     // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->tcp4ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->tcp4ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry144), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->tcp4ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry145), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry146), text);
    index += ETH_HDRLEN;

    // IPv4 header for TCP
    ip4hdr = (struct ip *) (buffer + index);         // Cast ip4hdr as pointer to IP4 header struct within buffer
    memcpy (&data->ip4hdr_tcp, ip4hdr, IP4_HDRLEN);  // Extract IPv4 header
    // IP header length
    sprintf (text, "%u", data->ip4hdr_tcp.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry29), text);
    // Protocol version
    sprintf (text, "%u", data->ip4hdr_tcp.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry30), text);
    // Type of service
    sprintf (text, "%u", data->ip4hdr_tcp.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry31), text);
    // Total length of datagram
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry32), text);
    // ID sequence number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry33), text);
    // Zero
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry122), text);
    // Do not fragment
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_tcp.ip_off) >> 14) & 1u);
    gtk_entry_set_text (GTK_ENTRY (data->entry37), text);
    // More fragments following
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_tcp.ip_off) >> 13) & 1u);
    gtk_entry_set_text (GTK_ENTRY (data->entry38), text);
    // Fragmentation offset
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_tcp.ip_off) & 0x1fffu));
    gtk_entry_set_text (GTK_ENTRY (data->entry39), text);
    // Time-to-Live
    sprintf (text, "%u", data->ip4hdr_tcp.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry40), text);
    // Transport layer protocol
    sprintf (text, "%u", data->ip4hdr_tcp.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry74), text);
    // Source IPv4 address
    inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_src), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry35), text);
    // Destination IPv4 address
    inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_dst), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry36), text);
    // IPv4 header checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry34), text);
    index += IP4_HDRLEN;

    // Flag for randomizing IPv4 source address for TCP
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton5), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton5), TRUE);
    }
    // Variable data->ran_tcp4_sourceip will be changed by on_checkbutton5_toggled(), so set it now.
    data->ran_tcp4_sourceip = buffer[index];
    index++;

    // TCP header
    tcphdr = (struct tcphdr *) (buffer + index);  // Cast tcphdr as pointer to TCP header struct within buffer
    memcpy (&data->tcp4hdr, tcphdr, TCP_HDRLEN);  // Extract TCP header
    // Source port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry41), text);
    // Destination port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry42), text);
    // Sequence number
    sprintf (text, "%lu", (unsigned long int) ntohl (data->tcp4hdr.th_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry43), text);
    // Acknowledgement number
    sprintf (text, "%lu", (unsigned long int) ntohl (data->tcp4hdr.th_ack));
    gtk_entry_set_text (GTK_ENTRY (data->entry44), text);
    // Reserved
    sprintf (text, "%u", data->tcp4hdr.th_x2);
    gtk_entry_set_text (GTK_ENTRY (data->entry45), text);
    // Data offset
    sprintf (text, "%u", data->tcp4hdr.th_off);
    gtk_entry_set_text (GTK_ENTRY (data->entry46), text);
    // FIN flag
    sprintf (text, "%u", data->tcp4hdr.th_flags & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry47), text);
    // SYN flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 1) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry48), text);
    // RST flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 2) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry49), text);
    // PSH flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 3) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry50), text);
    // ACK flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 4) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry51), text);
    // URG flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 5) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry52), text);
    // ECE flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 6) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry53), text);
    // CWR flag
    sprintf (text, "%u", (data->tcp4hdr.th_flags >> 7) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry54), text);
    // Window size
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_win));
    gtk_entry_set_text (GTK_ENTRY (data->entry55), text);
    // Urgent pointer
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_urp));
    gtk_entry_set_text (GTK_ENTRY (data->entry77), text);
    // TCP checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry56), text);
    index += TCP_HDRLEN;

    // Flag for randomizing TCP source port
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton7), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton7), TRUE);
    }
    // Variable data->ran_tcp4_sourceport will be changed by on_checkbutton7_toggled(), so set it now.
    data->ran_tcp4_sourceport = buffer[index];
    index++;

    // Flag for ASCII or hexadecimal data entry
    if (buffer[index] == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton17), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton17), TRUE);
    }
    // Variable data->ascii_hex_tcp4 will be changed by on_checkbutton17_toggled(), so set it now.
    data->ascii_hex_tcp4 = buffer[index];
    index++;

    // Length of TCP data
    ival16 = (int16_t *) (buffer + index);           // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    data->tcp4_payloadlen = (gint) ntohs (*ival16);  // Value was stored in network byte order to be consistent across platforms
    index += 2;

    // TCP data
    memcpy (data->tcp4_payload, buffer + index, data->tcp4_payloadlen);
    if (data->tcp4_payloadlen > 0) {
      memset (text, 0, 1024);
      strcpy (text, "Data file loaded.");
      gtk_entry_set_text (GTK_ENTRY (data->entry162), text);
    }

    index += IP_MAXPACKET;
  }

// IPv4 ICMP
  if (flags[3] == 1) {

    // Flag to specify ethernet header
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton17), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton17), TRUE);
    }
    // Variable data->icmp4_specify_ether will be changed by on_checkbutton17_toggled(), so set it now.
    data->icmp4_specify_ether = buffer[index];
    index++;

    // Interface name for IPv4 ICMP
    for (i=0; i<255; i++) {
      data->icmp4_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry151), data->icmp4_interfacename);

    // Ethernet header for IPv4 header for ICMP
    ethhdr = (struct ethhdr *) (buffer + index);      // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->icmp4ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->icmp4ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry141), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->icmp4ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry142), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry143), text);
    index += ETH_HDRLEN;

    // IPv4 header for ICMP
    ip4hdr = (struct ip *) (buffer + index);          // Cast ip4hdr as pointer to IP4 header struct within buffer
    memcpy (&data->ip4hdr_icmp, ip4hdr, IP4_HDRLEN);  // Extract IPv4 header
    // IP header length
    sprintf (text, "%u", data->ip4hdr_icmp.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry57), text);
    // Protocol version
    sprintf (text, "%u", data->ip4hdr_icmp.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry58), text);
    // Type of service
    sprintf (text, "%u", data->ip4hdr_icmp.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry59), text);
    // Total length of datagram
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry60), text);
    // ID sequence number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry61), text);
    // Zero
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry123), text);
    // Do not fragment
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_icmp.ip_off) >> 14) & 1u);
    gtk_entry_set_text (GTK_ENTRY (data->entry65), text);
    // More fragments following
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_icmp.ip_off) >> 13) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry66), text);
    // Fragmentation offset
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_icmp.ip_off) & 8191));
    gtk_entry_set_text (GTK_ENTRY (data->entry67), text);
    // Time-to-Live
    sprintf (text, "%u", data->ip4hdr_icmp.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry68), text);
    // Transport layer protocol
    sprintf (text, "%u", data->ip4hdr_icmp.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry75), text);
    // Source IPv4 address
    inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_src), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry63), text);
    // Destination IPv4 address
    inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_dst), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry64), text);
    // IPv4 header checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry62), text);
    index += IP4_HDRLEN;

    // Flag for randomizing IPv4 source address for ICMP
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton9), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton9), TRUE);
    }
    // Variable data->ran_icmp4_sourceip will be changed by on_checkbutton9_toggled(), so set it now.
    data->ran_icmp4_sourceip = buffer[index];
    index++;

    // ICMP header
    icmp4hdr = (struct icmp *) (buffer + index);      // Cast icmp4hdr as pointer to ICMP header struct within buffer
    memcpy (&data->icmp4hdr, icmp4hdr, ICMP_HDRLEN);  // Extract ICMP header
    // Message Type
    sprintf (text, "%u", data->icmp4hdr.icmp_type);
    gtk_entry_set_text (GTK_ENTRY (data->entry69), text);
    // Message Code
    sprintf (text, "%u", data->icmp4hdr.icmp_code);
    gtk_entry_set_text (GTK_ENTRY (data->entry70), text);
    // Identifier
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry72), text);
    // Sequence Number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry73), text);
    // ICMP header checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry71), text);
    index += ICMP_HDRLEN;

    // Flag for ASCII or hexadecimal data entry
    if (buffer[index] == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton9), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton9), TRUE);
    }
    // Variable data->ascii_hex_icmp4 will be changed by on_checkbutton9_toggled(), so set it now.
    data->ascii_hex_icmp4 = buffer[index];
    index++;

    // Length of ICMP data
    ival16 = (int16_t *) (buffer + index);            // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    data->icmp4_payloadlen = (gint) ntohs (*ival16);  // Value was stored in network byte order to be consistent across platforms
    index += 2;

    // ICMP data
    memcpy (data->icmp4_payload, buffer + index, data->icmp4_payloadlen);
    if (data->icmp4_payloadlen > 0) {
      memset (text, 0, 1024);
      strcpy (text, "Data file loaded.");
      gtk_entry_set_text (GTK_ENTRY (data->entry126), text);
    }

    index += IP_MAXPACKET;
  }

// IPv4 UDP
  if (flags[4] == 1) {

    // Flag to specify ethernet header
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton19), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton19), TRUE);
    }
    // Variable data->udp4_specify_ether will be changed by on_checkbutton19_toggled(), so set it now.
    data->udp4_specify_ether = buffer[index];
    index++;

    // Interface name for IPv4 UDP
    for (i=0; i<255; i++) {
      data->udp4_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry152), data->udp4_interfacename);

    // Ethernet header for IPv4 header for UDP
    ethhdr = (struct ethhdr *) (buffer + index);     // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->udp4ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->udp4ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry147), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->udp4ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry148), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry149), text);
    index += ETH_HDRLEN;

    // IPv4 header for UDP
    ip4hdr = (struct ip *) (buffer + index);         // Cast ip4hdr as pointer to IP4 header struct within buffer
    memcpy (&data->ip4hdr_udp, ip4hdr, IP4_HDRLEN);  // Extract IPv4 header
    // IP header length
    sprintf (text, "%u", data->ip4hdr_udp.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry1), text);
    // Protocol version
    sprintf (text, "%u", data->ip4hdr_udp.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry2), text);
    // Type of service
    sprintf (text, "%u", data->ip4hdr_udp.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry3), text);
    // Total length of datagram
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry4), text);
    // ID sequence number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry5), text);
    // Zero
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry124), text);
    // Do not fragment
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_udp.ip_off) >> 14) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry9), text);
    // More fragments following
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_udp.ip_off) >> 13) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry10), text);
    // Fragmentation offset
    sprintf (text, "%hu", (unsigned short int) (ntohs (data->ip4hdr_udp.ip_off) & 8191));
    gtk_entry_set_text (GTK_ENTRY (data->entry11), text);
    // Time-to-Live
    sprintf (text, "%u", data->ip4hdr_udp.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry12), text);
    // Transport layer protocol
    sprintf (text, "%u", data->ip4hdr_udp.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry76), text);
    // Source IPv4 address
    inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_src), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry7), text);
    // Destination IPv4 address
    inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_dst), text, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry8), text);
    // IPv4 header checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry6), text);
    index += IP4_HDRLEN;

    // Flag for randomizing IPv4 source address for UDP
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton1), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton1), TRUE);
    }
    // Variable data->ran_udp4_sourceip will be changed by on_checkbutton1_toggled(), so set it now.
    data->ran_udp4_sourceip = buffer[index];
    index++;

    // UDP header
    udphdr = (struct udphdr *) (buffer + index);  // Cast udphdr as pointer to UDP header struct within buffer
    memcpy (&data->udp4hdr, udphdr, UDP_HDRLEN);  // Extract UDP header
    // Source port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry13), text);
    // Destination port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry14), text);
    // Length of UDP datagram
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_ulen));
    gtk_entry_set_text (GTK_ENTRY (data->entry15), text);
    // UDP checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry16), text);
    index += UDP_HDRLEN;

    // Flag for randomizing UDP source port
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton3), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton3), TRUE);
    }
    // Variable data->ran_udp4_sourceport will be changed by on_checkbutton3_toggled(), so set it now.
    data->ran_udp4_sourceport = buffer[index];
    index++;

    // Flag for ASCII or hexadecimal data entry
    if (buffer[index] == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton13), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton13), TRUE);
    }
    // Variable data->ascii_hex_udp4 will be changed by on_checkbutton13_toggled(), so set it now.
    data->ascii_hex_udp4 = buffer[index];
    index++;

    // Length of UDP data
    ival16 = (int16_t *) (buffer + index);           // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    data->udp4_payloadlen = (gint) ntohs (*ival16);  // Value was stored in network byte order to be consistent across platforms
    index += 2;

    // UDP data
    memcpy (data->udp4_payload, buffer + index, data->udp4_payloadlen);
    if (data->udp4_payloadlen > 0) {
      memset (text, 0, 1024);
      strcpy (text, "Data file loaded.");
      gtk_entry_set_text (GTK_ENTRY (data->entry17), text);
    }
    index += IP_MAXPACKET;
  }

// IPv6 TCP
  if (flags[5] == 1) {

    // Interface name for IPv6 TCP
    for (i=0; i<255; i++) {
      data->tcp6_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry154), data->tcp6_interfacename);

    // Ethernet header for IPv6 header for TCP
    ethhdr = (struct ethhdr *) (buffer + index);     // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->tcp6ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->tcp6ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry131), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->tcp6ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry132), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry133), text);
    index += ETH_HDRLEN;

    // IPv6 header for TCP
    ip6hdr = (struct ip6_hdr *) (buffer + index);    // Cast ip6hdr as pointer to IP6 header struct within buffer
    memcpy (&data->ip6hdr_tcp, ip6hdr, IP6_HDRLEN);  // Extract IPv4 header
    // IP version
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) >> 28));
    gtk_entry_set_text (GTK_ENTRY (data->entry22), text);
    // Traffic Class
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry23), text);
    // Flow Label
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry24), text);
    // Payload length
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry25), text);
    // Next header
    sprintf (text, "%u", data->ip6hdr_tcp.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry26), text);
    // Hop limit
    sprintf (text, "%u", data->ip6hdr_tcp.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry27), text);
    // Source IPv6 address
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_src), text, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry28), text);
    // Destination IPv6 address
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_dst), text, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry78), text);
    index += IP6_HDRLEN;

    // Flag for randomizing IPv6 source address for TCP
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton2), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton2), TRUE);
    }
    // Variable data->ran_tcp6_sourceip will be changed by on_checkbutton2_toggled(), so set it now.
    data->ran_tcp6_sourceip = buffer[index];
    index++;

    // TCP header (IPv6)
    tcphdr = (struct tcphdr *) (buffer + index);  // Cast tcphdr as pointer to TCP header struct within buffer
    memcpy (&data->tcp6hdr, tcphdr, TCP_HDRLEN);  // Extract TCP header
    // Source port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry79), text);
    // Destination port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry80), text);
    // Sequence number
    sprintf (text, "%lu", (unsigned long int) ntohl (data->tcp6hdr.th_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry81), text);
    // Acknowledgement number
    sprintf (text, "%lu", (unsigned long int) ntohl (data->tcp6hdr.th_ack));
    gtk_entry_set_text (GTK_ENTRY (data->entry82), text);
    // Reserved
    sprintf (text, "%u", data->tcp6hdr.th_x2);
    gtk_entry_set_text (GTK_ENTRY (data->entry83), text);
    // Data offset
    sprintf (text, "%u", data->tcp6hdr.th_off);
    gtk_entry_set_text (GTK_ENTRY (data->entry84), text);
    // FIN flag
    sprintf (text, "%u", data->tcp6hdr.th_flags & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry85), text);
    // SYN flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 1) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry86), text);
    // RST flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 2) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry87), text);
    // PSH flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 3) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry88), text);
    // ACK flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 4) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry89), text);
    // URG flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 5) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry90), text);
    // ECE flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 6) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry91), text);
    // CWR flag
    sprintf (text, "%u", (data->tcp6hdr.th_flags >> 7) & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry92), text);
    // Window size
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_win));
    gtk_entry_set_text (GTK_ENTRY (data->entry93), text);
    // Urgent pointer
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_urp));
    gtk_entry_set_text (GTK_ENTRY (data->entry95), text);
    // TCP checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry94), text);
    index += TCP_HDRLEN;

    // Flag for randomizing TCP source port
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton4), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton4), TRUE);
    }
    // Variable data->ran_tcp6_sourceport will be changed by on_checkbutton4_toggled(), so set it now.
    data->ran_tcp6_sourceport = buffer[index];
    index++;

   // Flag for ASCII or hexadecimal data entry
    if (buffer[index] == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton19), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton19), TRUE);
    }
    // Variable data->ascii_hex_tcp6 will be changed by on_checkbutton19_toggled(), so set it now.
    data->ascii_hex_tcp6 = buffer[index];
    index++;

    // Length of TCP data
    ival16 = (int16_t *) (buffer + index);           // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    data->tcp6_payloadlen = (gint) ntohs (*ival16);  // Value was stored in network byte order to be consistent across platforms
    index += 2;

    // TCP data
    memcpy (data->tcp6_payload, buffer + index, data->tcp6_payloadlen);
    if (data->tcp6_payloadlen > 0) {
      memset (text, 0, 1024);
      strcpy (text, "Data file loaded.");
      gtk_entry_set_text (GTK_ENTRY (data->entry163), text);
    }

    index += IP_MAXPACKET;
  }

// IPv6 ICMP
  if (flags[6] == 1) {

    // Interface name for IPv6 ICMP
    for (i=0; i<255; i++) {
      data->icmp6_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry153), data->icmp6_interfacename);

    // Ethernet header for IPv6 header for ICMP
    ethhdr = (struct ethhdr *) (buffer + index);      // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->icmp6ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->icmp6ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry128), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->icmp6ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry129), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry130), text);
    index += ETH_HDRLEN;

    // IPv6 header for ICMP
    ip6hdr = (struct ip6_hdr *) (buffer + index);     // Cast ip6hdr as pointer to IP6 header struct within buffer
    memcpy (&data->ip6hdr_icmp, ip6hdr, IP6_HDRLEN);  // Extract IPv4 header
    // IP version
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) >> 28));
    gtk_entry_set_text (GTK_ENTRY (data->entry110), text);
    // Traffic Class
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry111), text);
    // Flow Label
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry112), text);
    // Payload length
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry113), text);
    // Next header
    sprintf (text, "%u", data->ip6hdr_icmp.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry114), text);
    // Hop limit
    sprintf (text, "%u", data->ip6hdr_icmp.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry115), text);
    // Source IPv6 address
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_src), text, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry109), text);
    // Destination IPv6 address
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_dst), text, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry116), text);
    index += IP6_HDRLEN;

    // Flag for randomizing IPv6 source address for ICMP
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton10), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton10), TRUE);
    }
    // Variable data->ran_icmp6_sourceip will be changed by on_checkbutton10_toggled(), so set it now.
    data->ran_icmp6_sourceip = buffer[index];
    index++;

    // ICMP header
    icmp6hdr = (struct icmp6_hdr *) (buffer + index);  // Cast icmp6hdr as pointer to ICMP header struct within buffer
    memcpy (&data->icmp6hdr, icmp6hdr, ICMP_HDRLEN);   // Extract ICMP header
    // Message Type
    sprintf (text, "%u", data->icmp6hdr.icmp6_type);
    gtk_entry_set_text (GTK_ENTRY (data->entry117), text);
    // Message Code
    sprintf (text, "%u", data->icmp6hdr.icmp6_code);
    gtk_entry_set_text (GTK_ENTRY (data->entry118), text);
    // Identifier
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry120), text);
    // Sequence Number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry121), text);
    // ICMP header checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry119), text);
    index += ICMP_HDRLEN;

    // Flag for ASCII or hexadecimal data entry
    if (buffer[index] == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton11), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton11), TRUE);
    }
    // Variable data->ascii_hex_icmp6 will be changed by on_checkbutton11_toggled(), so set it now.
    data->ascii_hex_icmp6 = buffer[index];
    index++;

    // Length of ICMP data
    ival16 = (int16_t *) (buffer + index);            // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    data->icmp6_payloadlen = (gint) ntohs (*ival16);  // Value was stored in network byte order to be consistent across platforms
    index += 2;

    // ICMP data
    memcpy (data->icmp6_payload, buffer + index, data->icmp6_payloadlen);
    if (data->icmp6_payloadlen > 0) {
      memset (text, 0, 1024);
      strcpy (text, "Data file loaded.");
      gtk_entry_set_text (GTK_ENTRY (data->entry127), text);
    }
    index += IP_MAXPACKET;
  }

// IPv6 UDP
  if (flags[7] == 1) {

    // Interface name for IPv6 UDP
    for (i=0; i<255; i++) {
      data->udp6_interfacename[i] = buffer[index];
      index++;
    }
    gtk_entry_set_text (GTK_ENTRY (data->entry155), data->udp6_interfacename);

    // Ethernet header for IPv6 header for UDP
    ethhdr = (struct ethhdr *) (buffer + index);     // Cast ethhdr as pointer to ethernet header within buffer
    memcpy (&data->udp6ethhdr, ethhdr, ETH_HDRLEN);  // Extract ethernet header
    // Destination MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->udp6ethhdr.dst_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry134), text);
    // Source MAC address
    memset (text, 0, 1024);
    mac_bin2string (data->udp6ethhdr.src_mac, text);
    gtk_entry_set_text (GTK_ENTRY (data->entry135), text);
    // Ethernet type code
    memset (text, 0, 1024);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6ethhdr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry136), text);
    index += ETH_HDRLEN;

    // IPv6 header for UDP
    ip6hdr = (struct ip6_hdr *) (buffer + index);    // Cast ip6hdr as pointer to IP6 header struct within buffer
    memcpy (&data->ip6hdr_udp, ip6hdr, IP6_HDRLEN);  // Extract IPv4 header
    // IP version
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) >> 28));
    gtk_entry_set_text (GTK_ENTRY (data->entry97), text);
    // Traffic Class
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry98), text);
    // Flow Label
    sprintf (text, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry99), text);
    // Payload length
    sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry100), text);
    // Next header
    sprintf (text, "%u", data->ip6hdr_udp.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry101), text);
    // Hop limit
    sprintf (text, "%u", data->ip6hdr_udp.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry102), text);
    // Source IPv6 address
    inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_src), text, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry96), text);
    // Destination IPv6 address
    inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_dst), text, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry103), text);
    index += IP6_HDRLEN;

    // Flag for randomizing IPv6 source address for UDP
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton6), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton6), TRUE);
    }
    // Variable data->ran_udp6_sourceip will be changed by on_checkbutton6_toggled(), so set it now.
    data->ran_udp6_sourceip = buffer[index];
    index++;

    // UDP header
    udphdr = (struct udphdr *) (buffer + index);  // Cast udphdr as pointer to UDP header struct within buffer
    memcpy (&data->udp6hdr, udphdr, UDP_HDRLEN);  // Extract UDP header
    // Source port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry104), text);
    // Destination port number
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry105), text);
    // Length of UDP datagram
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_ulen));
    gtk_entry_set_text (GTK_ENTRY (data->entry106), text);
    // UDP checksum
    sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry107), text);
    index += UDP_HDRLEN;

    // Flag for randomizing UDP source port
    if (buffer[index] == 0) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton8), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton8), TRUE);
    }
    // Variable data->ran_udp6_sourceport will be changed by on_checkbutton8_toggled(), so set it now.
    data->ran_udp6_sourceport = buffer[index];
    index++;

    // Flag for ASCII or hexadecimal data entry
    if (buffer[index] == 1) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton15), FALSE);
    } else {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton15), TRUE);
    }
    // Variable data->ascii_hex_udp6 will be changed by on_checkbutton15_toggled(), so set it now.
    data->ascii_hex_udp6 = buffer[index];
    index++;

    // Length of UDP data
    ival16 = (int16_t *) (buffer + index);           // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    data->udp6_payloadlen = (gint) ntohs (*ival16);  // Value was stored in network byte order to be consistent across platforms
    index += 2;

    // UDP data
    memcpy (data->udp6_payload, buffer + index, data->udp6_payloadlen);
    if (data->udp6_payloadlen > 0) {
      memset (text, 0, 1024);
      strcpy (text, "Data file loaded.");
      gtk_entry_set_text (GTK_ENTRY (data->entry108), text);
    }
    index += IP_MAXPACKET;
  }

// Free allocated memory.
  g_free (text);
  g_free (buffer);
  g_free (flags);

  return (TRUE);
}

// Save file ...
gboolean
on_button14_clicked (GtkButton *button, SPSData *data)
{
  int i, buffer_size;
  gint result;
  int16_t *ival16;
  int32_t *ival32;
  GtkWidget *dialog;
  gchar *filename, *text;
  unsigned char *buffer;
  void *tmp;
  FILE *fo;

// Allocate memory for various arrays.

// Size of buffer
  buffer_size = 400000;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button14_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (buffer_size * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button14_clicked().\n");
    exit (EXIT_FAILURE);
  }

  // Check to make sure the user selected something to save.
  if ((!data->settings_save) && (!data->sixto4_save) && (!data->ipv4_tcp_save) &&
      (!data->ipv4_icmp_save) && (!data->ipv4_udp_save) && (!data->ipv6_tcp_save) &&
      (!data->ipv6_icmp_save) && (!data->ipv6_udp_save)) {
     sprintf (text, "Nothing was selected for saving.");
     data->parent = data->main_window;
     data->message = text;
     report_error (data);
     g_free (text);
     g_free (buffer);
     return (FALSE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Save File ...", (GtkWindow *) data->main_window, GTK_FILE_CHOOSER_ACTION_SAVE,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open output file.
  if ((fo = fopen (filename, "wb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button14_clicked().");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// SPS file identification
  memset (text, 0, 1024);
  strcpy (text, "Simple Packet Sender");
  for (i=0; i<80; i++) {
    fputc ((unsigned char) text[i], fo);
  }

// Save data flags
  // Send Packet settings
  fputc (data->settings_save, fo);
  // IPv6 over IPv4 (6to4) information
  fputc (data->sixto4_save, fo);
  // TCP (IPv4)
  fputc (data->ipv4_tcp_save, fo);
  // ICMP (IPv4)
  fputc (data->ipv4_icmp_save, fo);
  // UDP (IPv4)
  fputc (data->ipv4_udp_save, fo);
  // TCP (IPv6)
  fputc (data->ipv6_tcp_save, fo);
  // ICMP (IPv6)
  fputc (data->ipv6_icmp_save, fo);
  // UDP (IPv6)
  fputc (data->ipv6_udp_save, fo);
  // Reserved
  for (i=0; i<(256-8);i++) {
    fputc (0, fo);
  }

// Send Packet Settings
  if (data->settings_save == 1) {

    fputc (data->packet_type, fo);                // Packet type(s) to send

    // Number of packets
    ival32 = (int32_t *) buffer;                  // Cast ival32 as pointer to 32 bit (4 byte) int within buffer
    *ival32 = htonl ((int32_t ) data->npackets);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);
    fputc (buffer[2], fo);
    fputc (buffer[3], fo);

    fputc (flood_mode, fo);  // Flag for flood mode
  }

// IPv6 over IPv4 (6to4)
  if (data->sixto4_save == 1) {

    fputc (data->sixto4_flag, fo);  // Flag to use 6to4

    fputc (data->sixto4_specify_ether, fo);  // Flag to specify ethernet header

    // 6to4 interface name
    for (i=0; i<255; i++) {
      fputc (data->sixto4_interfacename[i], fo);
    }

    // 6to4 ethernet header for IPv4 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->sixto4ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // 6to4 IPv4 header for TCP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->sixto4ip4hdr_tcp, IP4_HDRLEN);
    for (i=0; i<IP4_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_tcp6to4_sourceip, fo);  // Flag for randomizing IPv4 source address for TCP over 6to4

    // 6to4 IPv4 header for ICMP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->sixto4ip4hdr_icmp, IP4_HDRLEN);
    for (i=0; i<IP4_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_icmp6to4_sourceip, fo);  // Flag for randomizing IPv4 source address for ICMP over 6to4

    // 6to4 IPv4 header for UDP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->sixto4ip4hdr_udp, IP4_HDRLEN);
    for (i=0; i<IP4_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_udp6to4_sourceip, fo);  // Flag for randomizing IPv4 source address for UDP over 6to4
  }

// IPv4 TCP
  if (data->ipv4_tcp_save == 1) {

    fputc (data->tcp4_specify_ether, fo);  // Flag to specify ethernet header

    // Interface name for IPv4 TCP
    for (i=0; i<255; i++) {
      fputc (data->tcp4_interfacename[i], fo);
    }

    // Ethernet header for IPv4 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->tcp4ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // IPv4 header for TCP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->ip4hdr_tcp, IP4_HDRLEN);
    for (i=0; i<IP4_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_tcp4_sourceip, fo);  // Flag for randomizing IPv4 source address

    // TCP header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->tcp4hdr, TCP_HDRLEN);
    for (i=0; i<TCP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_tcp4_sourceport, fo);  // Flag for randomizing TCP source port

    // Flag for ASCII for hexadecimal data entry
    fputc (data->ascii_hex_tcp4, fo);

    // Length of TCP data
    memset (buffer, 0, buffer_size);
    ival16 = (int16_t *) buffer;                         // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    *ival16 = htons ((int16_t ) data->tcp4_payloadlen);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);

    // TCP data
    memset (buffer, 0, buffer_size);
    memcpy (buffer, data->tcp4_payload, data->tcp4_payloadlen);  // We use buffer because we also want the zeros
    for (i=0; i<IP_MAXPACKET; i++) {
      fputc (buffer[i], fo);
    }
  }

// IPv4 ICMP
  if (data->ipv4_icmp_save == 1) {

    fputc (data->icmp4_specify_ether, fo);  // Flag to specify ethernet header

    // Interface name for IPv4 ICMP
    for (i=0; i<255; i++) {
      fputc (data->icmp4_interfacename[i], fo);
    }

    // Ethernet header for IPv4 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->icmp4ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // IPv4 header for ICMP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->ip4hdr_icmp, IP4_HDRLEN);
    for (i=0; i<IP4_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_icmp4_sourceip, fo);  // Flag for randomizing IPv4 source address

    // ICMP header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->icmp4hdr, ICMP_HDRLEN);
    for (i=0; i<ICMP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Flag for ASCII for hexadecimal data entry
    fputc (data->ascii_hex_icmp4, fo);

    // Length of ICMP data
    memset (buffer, 0, buffer_size);
    ival16 = (int16_t *) buffer;                          // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    *ival16 = htons ((int16_t ) data->icmp4_payloadlen);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);

    // ICMP data
    memset (buffer, 0, buffer_size);
    memcpy (buffer, data->icmp4_payload, data->icmp4_payloadlen);  // We use buffer because we also want the zeros
    for (i=0; i<IP_MAXPACKET; i++) {
      fputc (buffer[i], fo);
    }
  }

// IPv4 UDP
  if (data->ipv4_udp_save == 1) {

    fputc (data->udp4_specify_ether, fo);  // Flag to specify ethernet header

    // Interface name for IPv4 UDP
    for (i=0; i<255; i++) {
      fputc (data->udp4_interfacename[i], fo);
    }

    // Ethernet header for IPv4 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->udp4ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // IPv4 header for UDP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->ip4hdr_udp, IP4_HDRLEN);
    for (i=0; i<IP4_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_udp4_sourceip, fo);  // Flag for randomizing IPv4 source address

    // UDP header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->udp4hdr, UDP_HDRLEN);
    for (i=0; i<UDP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_udp4_sourceport, fo);  // Flag for randomizing UDP source port

    // Flag for ASCII for hexadecimal data entry
    fputc (data->ascii_hex_udp4, fo);

    // Length of UDP data
    memset (buffer, 0, buffer_size);
    ival16 = (int16_t *) buffer;                         // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    *ival16 = htons ((int16_t ) data->udp4_payloadlen);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);

    // UDP data
    memset (buffer, 0, buffer_size);
    memcpy (buffer, data->udp4_payload, data->udp4_payloadlen);  // We use buffer because we also want the zeros
    for (i=0; i<IP_MAXPACKET; i++) {
      fputc (buffer[i], fo);
    }
  }

// IPv6 TCP
  if (data->ipv6_tcp_save == 1) {

    // Interface name for IPv6 TCP
    for (i=0; i<255; i++) {
      fputc (data->tcp6_interfacename[i], fo);
    }

    // Ethernet header for IPv6 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->tcp6ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // IPv6 header for TCP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->ip6hdr_tcp, IP6_HDRLEN);
    for (i=0; i<IP6_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_tcp6_sourceip, fo);  // Flag for randomizing IPv6 source address

    // TCP header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->tcp6hdr, TCP_HDRLEN);
    for (i=0; i<TCP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_tcp6_sourceport, fo);  // Flag for randomizing TCP source port

    // Flag for ASCII for hexadecimal data entry
    fputc (data->ascii_hex_tcp6, fo);

    // Length of TCP data
    memset (buffer, 0, buffer_size);
    ival16 = (int16_t *) buffer;                         // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    *ival16 = htons ((int16_t ) data->tcp6_payloadlen);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);

    // TCP data
    memset (buffer, 0, buffer_size);
    memcpy (buffer, data->tcp6_payload, data->tcp6_payloadlen);  // We use buffer because we also want the zeros
    for (i=0; i<IP_MAXPACKET; i++) {
      fputc (buffer[i], fo);
    }
  }

// IPv6 ICMP
  if (data->ipv6_icmp_save == 1) {

    // Interface name for IPv6 ICMP
    for (i=0; i<255; i++) {
      fputc (data->icmp6_interfacename[i], fo);
    }

    // Ethernet header for IPv6 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->icmp6ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // IPv6 header for ICMP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->ip6hdr_icmp, IP6_HDRLEN);
    for (i=0; i<IP6_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_icmp6_sourceip, fo);  // Flag for randomizing IPv6 source address

    // ICMP header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->icmp6hdr, ICMP_HDRLEN);
    for (i=0; i<ICMP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // Flag for ASCII for hexadecimal data entry
    fputc (data->ascii_hex_icmp6, fo);

    // Length of ICMP data
    memset (buffer, 0, buffer_size);
    ival16 = (int16_t *) buffer;                          // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    *ival16 = htons ((int16_t ) data->icmp6_payloadlen);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);

    // ICMP data
    memset (buffer, 0, buffer_size);
    memcpy (buffer, data->icmp6_payload, data->icmp6_payloadlen);  // We use buffer because we also want the zeros
    for (i=0; i<IP_MAXPACKET; i++) {
      fputc (buffer[i], fo);
    }
  }

// IPv6 UDP
  if (data->ipv6_udp_save == 1) {

    // Interface name for IPv6 UDP
    for (i=0; i<255; i++) {
      fputc (data->udp6_interfacename[i], fo);
    }

    // Ethernet header for IPv6 header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->udp6ethhdr, ETH_HDRLEN);
    for (i=0; i<ETH_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    // IPv6 header for UDP
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->ip6hdr_udp, IP6_HDRLEN);
    for (i=0; i<IP6_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_udp6_sourceip, fo);  // Flag for randomizing IPv6 source address

    // UDP header
    memset (buffer, 0, buffer_size);
    memcpy (buffer, &data->udp6hdr, UDP_HDRLEN);
    for (i=0; i<UDP_HDRLEN; i++) {
      fputc (buffer[i], fo);
    }

    fputc (data->ran_udp6_sourceport, fo);  // Flag for randomizing UDP source port

    // Flag for ASCII for hexadecimal data entry
    fputc (data->ascii_hex_udp6, fo);

    // Length of UDP data
    memset (buffer, 0, buffer_size);
    ival16 = (int16_t *) buffer;                         // Cast ival16 as pointer to 16 bit (2 byte) int within buffer
    *ival16 = htons ((int16_t ) data->udp6_payloadlen);  // Use network byte order for consistent endianess across platforms
    fputc (buffer[0], fo);
    fputc (buffer[1], fo);

    // UDP data
    memset (buffer, 0, buffer_size);
    memcpy (buffer, data->udp6_payload, data->udp6_payloadlen);  // We use buffer because we also want the zeros
    for (i=0; i<IP_MAXPACKET; i++) {
      fputc (buffer[i], fo);
    }
  }

// Close file descriptor.
  fclose (fo);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// Send Packet settings
gboolean
on_checkbutton25_toggled (GtkButton *button, SPSData *data)
{
  if (data->settings_save == 0) {
    data->settings_save = 1;
  } else {
    data->settings_save = 0;
  }

  return (TRUE);
}

// Select to save IPv6 over IPv4 (6to4)
gboolean
on_checkbutton15_toggled (GtkButton *button, SPSData *data)
{
  if (data->sixto4_save == 0) {
    data->sixto4_save = 1;
  } else {
    data->sixto4_save = 0;
  }

  return (TRUE);
}

// Select to save IPv4 TCP packet
gboolean
on_checkbutton16_toggled (GtkButton *button, SPSData *data)
{
  if (data->ipv4_tcp_save == 0) {
    data->ipv4_tcp_save = 1;
  } else {
    data->ipv4_tcp_save = 0;
  }

  return (TRUE);
}

// Select to save IPv4 ICMP packet
gboolean
on_checkbutton20_toggled (GtkButton *button, SPSData *data)
{
  if (data->ipv4_icmp_save == 0) {
    data->ipv4_icmp_save = 1;
  } else {
    data->ipv4_icmp_save = 0;
  }

  return (TRUE);
}

// Select to save IPv4 UDP packet
gboolean
on_checkbutton21_toggled (GtkButton *button, SPSData *data)
{
  if (data->ipv4_udp_save == 0) {
    data->ipv4_udp_save = 1;
  } else {
    data->ipv4_udp_save = 0;
  }

  return (TRUE);
}

// Select to save IPv6 TCP packet
gboolean
on_checkbutton22_toggled (GtkButton *button, SPSData *data)
{
  if (data->ipv6_tcp_save == 0) {
    data->ipv6_tcp_save = 1;
  } else {
    data->ipv6_tcp_save = 0;
  }

  return (TRUE);
}

// Select to save IPv6 ICMP packet
gboolean
on_checkbutton23_toggled (GtkButton *button, SPSData *data)
{
  if (data->ipv6_icmp_save == 0) {
    data->ipv6_icmp_save = 1;
  } else {
    data->ipv6_icmp_save = 0;
  }

  return (TRUE);
}

// Select to save IPv6 UDP packet
gboolean
on_checkbutton24_toggled (GtkButton *button, SPSData *data)
{
  if (data->ipv6_udp_save == 0) {
    data->ipv6_udp_save = 1;
  } else {
    data->ipv6_udp_save = 0;
  }

  return (TRUE);
}

// Check keyboard data entry for acceptable hexadecimal.
gboolean
is_valid_hexdata (const gchar *string, SPSData *data)
{
  int i, j, len, result;
  gchar temp[50], *text;
  gchar **endptr;
  void *tmp;

  len = strlen (string);

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in is_valid_hexdata().\n");
    exit (EXIT_FAILURE);
  }

// Check for !=hex and !=comma and !=space
  for (i=0; i<len; i++) {
    if (!(is_valid_hex (string[i])) && (string[i] != 44) && (string[i] !=32)) {
      g_free (text);
      return (FALSE);
    }
  }

// Check for data composed of only spaces and/or commas (bad).
  j = 0;  // j=1 means we found a hexadecimal digit (good)
  for (i=0; i<len; i++) {
    if (is_valid_hex (string[i])) {
      j = 1;
      break;
    }
  }
  if (j == 0) {
    g_free (text);
    return (FALSE);
  }

// Check if first character is a comma (which is bad).
  i = 0;
  while (i < len) {
    if (!(is_valid_hex (string[i])) && (string[i] != 32)) {
      g_free (text);
      return (FALSE);
    } else if (is_valid_hex (string[i])) {  // This character is hex so we're ok, and done.
      break;
    } else {  // This character is a space so ignore it and keep looping.
      i++;
    }
  }

// Check if last character is a comma (which is bad).
  i = len - 1;
  while (i >= 0) {  // Ignore trailing spaces
    if (!(is_valid_hex (string[i])) && (string[i] != 32)) {
      g_free (text);
      return (FALSE);
    } else if (is_valid_hex (string[i])) {  // This character is hex so we're ok, and done.
      break;
    } else {  // This character is a space so ignore it and keep looping.
      i--;
    }
  }

// We don't want adjacent commas
  for (i=0; i<(len-1); i++) {
    if ((string[i] == 44) && (string[i+1] == 44)) {
      g_free (text);
      return (FALSE);
    }
  }

// We need to limit values to 0xff.
  i = 0;
  while (i < len) {
    // Grab next hexadecimal value.
    memset (temp, 0, 10);
    j = 0;
    while ((string[i+j] != 32) && (string[i+j] != 44) && ((i + j) < len)) {
      temp[j] = string[i+j];
      j++;
    }
    i += j;
    // Clobber leading zeros
    for (j=0; j<strlen (temp); j++) {
      if (temp[j] == 48) {
        temp[j] = 32;  // replace zero with space
      } else if ((temp[j] != 32) && (temp[j] != 48)) {
        break;  // valid digit was found; we're done looking for leading zeros.
      }
    }
    // Convert to base-10 chars and store as payload data.
    endptr = NULL;
    result = strtol (temp, endptr, 16);
    if (endptr != NULL) {
      sprintf (text, "In callbacks.c: strtol() failed in is_valid_hex_data().");
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      return (FALSE);
    }
    if ((result < 0) || (result > 255)) {
      g_free (text);
      return (FALSE);
    }
    // Increment i until we find more hex or hit end of string.
    while (((string[i] == 32) || (string[i] == 44)) && (i < len)) {
      i++;
    }
  }

  g_free (text);

  return (TRUE);
}

// Check for valid hexadecimal digit.
gboolean
is_valid_hex (gchar byte)
{
  // Within ranges 0 to 9, a to f, or A to F is ok.
  if (!(((byte > 64) && (byte < 71)) || ((byte > 96) && (byte < 103)) ||
    ((byte > 47) && (byte < 58)))) {
    return (FALSE);
  }

  return (TRUE);
}

// Timeout for ARP/ND to receive reply
gboolean
on_spinbutton1_value_changed (GtkWidget *spinbutton1, SPSData *data)
{

  data->timeout = (gint) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton1));

  return (TRUE);
}

// Apply IPv4 TCP page default settings
gboolean
tcp4_default (SPSData *data)
{
  unsigned char *ip4_flags, *tcp4_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in tcp4_default().\n");
    exit (EXIT_FAILURE);
  }

// TCP flags
  tmp = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp4_flags' in tcp4_default().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in tcp4_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in tcp4_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Default to have the user specify the ethernet header
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton18), TRUE);
  data->tcp4_specify_ether = 1;  // The toggling done above will trigger the callback, but we set it here to be sure.

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp4ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->tcp4ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry144), value);

// Source Interface Name: defaults to blank
  memset (data->tcp4_interfacename, 0, 255);
  gtk_entry_set_text (GTK_ENTRY (data->entry150), "");

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp4ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->tcp4ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry145), value);

// Ethernet type code (16 bits): default to IPv4
// http://www.iana.org/assignments/ethernet-numbers
  data->tcp4ethhdr.type_code = htons (ETH_P_IP);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry146), value);

// TCP data

// Clear TCP data buffer.
  memset (data->tcp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
  data->tcp4_payloadlen = 0;
  gtk_entry_set_text (GTK_ENTRY (data->entry162), "");

// TCP data entry format: default to hexadecimal input
  data->ascii_hex_tcp4 = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton18), TRUE);

// IPv4 header

// IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr_tcp.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  sprintf (value, "%u", data->ip4hdr_tcp.ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry29), value);

// Internet Protocol version (4 bits): IPv4
  data->ip4hdr_tcp.ip_v = 4u;
  sprintf (value, "%u", data->ip4hdr_tcp.ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry30), value);

// Type of service (8 bits)
  data->ip4hdr_tcp.ip_tos = 0;
  sprintf (value, "%u", data->ip4hdr_tcp.ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry31), value);

// Total length of datagram (16 bits): IP header + TCP header + TCP data
  data->ip4hdr_tcp.ip_len = htons (IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry32), value);

// ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr_tcp.ip_id = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry33), value);

// Zero (1 bit)
  ip4_flags[0] = 0;
  sprintf (value, "%u", ip4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry122), value);

// Do not fragment flag (1 bit)
  ip4_flags[1] = 0;
  sprintf (value, "%u", ip4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry37), value);

// More fragments following flag (1 bit)
  ip4_flags[2] = 0;
  sprintf (value, "%u", ip4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry38), value);

// Fragmentation offset (13 bits)
  ip4_flags[3] = 0;
  sprintf (value, "%u", ip4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry39), value);

// Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr_tcp.ip_off = htons ((ip4_flags[0] << 15) +
   (ip4_flags[1] << 14) + (ip4_flags[2] << 13) + ip4_flags[3]);

// Time-to-Live (8 bits): default to maximum value
  data->ip4hdr_tcp.ip_ttl = 255u;
  sprintf (value, "%u", data->ip4hdr_tcp.ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry40), value);

// Transport layer protocol (8 bits): 6 for TCP
  data->ip4hdr_tcp.ip_p = IPPROTO_TCP;
  sprintf (value, "%u", data->ip4hdr_tcp.ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry74), value);

// Source IPv4 address (32 bits): default to random
  data->ran_tcp4_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton5), TRUE);
  inet_pton (AF_INET, ran4_addr (ipaddress), &(data->ip4hdr_tcp.ip_src));
  inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry35), value);

// Destination IPv4 address (32 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "127.0.0.1");
  inet_pton (AF_INET, ipaddress, &data->ip4hdr_tcp.ip_dst);
  inet_ntop (AF_INET, &(data->ip4hdr_tcp.ip_dst), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry36), value);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp.ip_sum = 0;
  data->ip4hdr_tcp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp, IP4_HDRLEN);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry34), value);

// TCP header (IPv4)

// Source port number (16 bits): default to random
  data->ran_tcp4_sourceport = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton7), TRUE);
  data->tcp4hdr.th_sport = htons (ran_val2 ());
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry41), value);

// Destination port number (16 bits)
  data->tcp4hdr.th_dport = htons (80u);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry42), value);

// Sequence number (32 bits)
  data->tcp4hdr.th_seq = htonl (0);
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr.th_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry43), value);

// Acknowledgement number (32 bits): 0 in first packet
  data->tcp4hdr.th_ack = htonl (0);
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr.th_ack));
  gtk_entry_set_text (GTK_ENTRY (data->entry44), value);

// Reserved (4 bits): should be 0
  data->tcp4hdr.th_x2 = 0;
  sprintf (value, "%u", data->tcp4hdr.th_x2);
  gtk_entry_set_text (GTK_ENTRY (data->entry45), value);

// Data offset (4 bits): size of TCP header in 32-bit words
  data->tcp4hdr.th_off = TCP_HDRLEN / 4;
  sprintf (value, "%u", data->tcp4hdr.th_off);
  gtk_entry_set_text (GTK_ENTRY (data->entry46), value);

// FIN flag (1 bit)
  tcp4_flags[0] = 0;
  sprintf (value, "%u", tcp4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry47), value);

// SYN flag (1 bit): set to 1
  tcp4_flags[1] = 1;
  sprintf (value, "%u", tcp4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry48), value);

// RST flag (1 bit)
  tcp4_flags[2] = 0;
  sprintf (value, "%u", tcp4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry49), value);

// PSH flag (1 bit)
  tcp4_flags[3] = 0;
  sprintf (value, "%u", tcp4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry50), value);

// ACK flag (1 bit)
  tcp4_flags[4] = 0;
  sprintf (value, "%u", tcp4_flags[4]);
  gtk_entry_set_text (GTK_ENTRY (data->entry51), value);

// URG flag (1 bit)
  tcp4_flags[5] = 0;
  sprintf (value, "%u", tcp4_flags[5]);
  gtk_entry_set_text (GTK_ENTRY (data->entry52), value);

// ECE flag (1 bit)
  tcp4_flags[6] = 0;
  sprintf (value, "%u", tcp4_flags[6]);
  gtk_entry_set_text (GTK_ENTRY (data->entry53), value);

// CWR flag (1 bit)
  tcp4_flags[7] = 0;
  sprintf (value, "%u", tcp4_flags[7]);
  gtk_entry_set_text (GTK_ENTRY (data->entry54), value);

// Flags (8 bits): 2 for SYN
  data->tcp4hdr.th_flags = bin8_to_dec (tcp4_flags);

// Window size (16 bits): default to maximum value
  data->tcp4hdr.th_win = htons (0xffffu);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_win));
  gtk_entry_set_text (GTK_ENTRY (data->entry55), value);

// Urgent pointer (16 bits): 0 (only valid if URG flag is set)
  data->tcp4hdr.th_urp = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_urp));
  gtk_entry_set_text (GTK_ENTRY (data->entry77), value);

// TCP checksum (16 bits)
  data->tcp4hdr.th_sum = tcp4_checksum (data->ip4hdr_tcp, data->tcp4hdr, data->tcp4_payload, data->tcp4_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry56), value);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (tcp4_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv6 TCP page default settings
gboolean
tcp6_default (SPSData *data)
{
  unsigned char *tcp6_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// TCP flags
  tmp = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp6_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp6_flags' in tcp6_default().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in tcp6_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in tcp6_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp6ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->tcp6ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry131), value);

// Source Interface Name: defaults to blank
  memset (data->tcp6_interfacename, 0, 255);
  gtk_entry_set_text (GTK_ENTRY (data->entry154), "");

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp6ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->tcp6ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry132), value);

// Ethernet type code (16 bits): default to IPv6
// http://www.iana.org/assignments/ethernet-numbers
  data->tcp6ethhdr.type_code = htons (ETH_P_IPV6);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry133), value);

// TCP data

// Clear TCP data buffer.
  memset (data->tcp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
  data->tcp6_payloadlen = 0;
  gtk_entry_set_text (GTK_ENTRY (data->entry163), "");

// TCP data entry format: default to hexadecimal input
  data->ascii_hex_tcp6 = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton20), TRUE);

// IPv6 header

// IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr_tcp.ip6_flow = htonl ((6 << 28) | (0 << 20) | 0);

// Version (4 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) >> 28));
  gtk_entry_set_text (GTK_ENTRY (data->entry22), value);

// Traffic Class (8 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) >> 20) & 0xfful);
  gtk_entry_set_text (GTK_ENTRY (data->entry23), value);

// Flow Label (20 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp.ip6_flow) & 0xffffful));
  gtk_entry_set_text (GTK_ENTRY (data->entry24), value);

// Payload length (16 bits): TCP header + TCP data
  data->ip6hdr_tcp.ip6_plen = htons (TCP_HDRLEN + data->tcp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry25), value);

// Next header (8 bits): 6 for TCP
  data->ip6hdr_tcp.ip6_nxt = IPPROTO_TCP;
  sprintf (value, "%u", data->ip6hdr_tcp.ip6_nxt);
  gtk_entry_set_text (GTK_ENTRY (data->entry26), value);

// Hop limit (8 bits): default to maximum value
  data->ip6hdr_tcp.ip6_hops = 255u;
  sprintf (value, "%u", data->ip6hdr_tcp.ip6_hops);
  gtk_entry_set_text (GTK_ENTRY (data->entry27), value);

// Source IPv6 address (128 bits): default to random
  data->ran_tcp6_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton2), TRUE);
  inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_tcp.ip6_src);
  inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_src), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry28), value);

// Destination IPv6 address (128 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "::1");
  inet_pton (AF_INET6, ipaddress, &data->ip6hdr_tcp.ip6_dst);
  inet_ntop (AF_INET6, &(data->ip6hdr_tcp.ip6_dst), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry78), value);

// TCP header (IPv6)

// Source port number (16 bits): default to random
  data->ran_tcp6_sourceport = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton4), TRUE);
  data->tcp6hdr.th_sport = htons (ran_val2 ());
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry79), value);

// Destination port number (16 bits)
  data->tcp6hdr.th_dport = htons (80u);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry80), value);

// Sequence number (32 bits)
  data->tcp6hdr.th_seq = htonl (0);
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr.th_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry81), value);

// Acknowledgement number (32 bits): 0 in first packet
  data->tcp6hdr.th_ack = htonl (0);
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr.th_ack));
  gtk_entry_set_text (GTK_ENTRY (data->entry82), value);

// Reserved (4 bits): should be 0
  data->tcp6hdr.th_x2 = 0;
  sprintf (value, "%u", data->tcp6hdr.th_x2);
  gtk_entry_set_text (GTK_ENTRY (data->entry83), value);

// Data offset (4 bits): size of TCP header in 32-bit words
  data->tcp6hdr.th_off = TCP_HDRLEN / 4;
  sprintf (value, "%u", data->tcp6hdr.th_off);
  gtk_entry_set_text (GTK_ENTRY (data->entry84), value);

// FIN flag (1 bit)
  tcp6_flags[0] = 0;
  sprintf (value, "%u", tcp6_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry85), value);

// SYN flag (1 bit): set to 1
  tcp6_flags[1] = 1;
  sprintf (value, "%u", tcp6_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry86), value);

// RST flag (1 bit)
  tcp6_flags[2] = 0;
  sprintf (value, "%u", tcp6_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry87), value);

// PSH flag (1 bit)
  tcp6_flags[3] = 0;
  sprintf (value, "%u", tcp6_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry88), value);

// ACK flag (1 bit)
  tcp6_flags[4] = 0;
  sprintf (value, "%u", tcp6_flags[4]);
  gtk_entry_set_text (GTK_ENTRY (data->entry89), value);

// URG flag (1 bit)
  tcp6_flags[5] = 0;
  sprintf (value, "%u", tcp6_flags[5]);
  gtk_entry_set_text (GTK_ENTRY (data->entry90), value);

// ECE flag (1 bit)
  tcp6_flags[6] = 0;
  sprintf (value, "%u", tcp6_flags[6]);
  gtk_entry_set_text (GTK_ENTRY (data->entry91), value);

// CWR flag (1 bit)
  tcp6_flags[7] = 0;
  sprintf (value, "%u", tcp6_flags[7]);
  gtk_entry_set_text (GTK_ENTRY (data->entry92), value);

// Flags (8 bits): 2 for SYN
  data->tcp6hdr.th_flags = bin8_to_dec (tcp6_flags);

// Window size (16 bits): default to maximum value
  data->tcp6hdr.th_win = htons (0xffffu);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_win));
  gtk_entry_set_text (GTK_ENTRY (data->entry93), value);

// Urgent pointer (16 bits): 0 (only valid if URG flag is set)
  data->tcp6hdr.th_urp = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_urp));
  gtk_entry_set_text (GTK_ENTRY (data->entry95), value);

// TCP checksum (16 bits)
  data->tcp6hdr.th_sum = tcp6_checksum (data->ip6hdr_tcp, data->tcp6hdr, data->tcp6_payload, data->tcp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry94), value);

// Free allocated memory.
  g_free (tcp6_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv4 ICMP page default settings
gboolean
icmp4_default (SPSData *data)
{
  unsigned char *ip4_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in icmp4_default().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in icmp4_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in icmp4_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Default to have the user specify the ethernet header
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton17), TRUE);
  data->icmp4_specify_ether = 1;  // The toggling done above will trigger the callback, but we set it here to be sure.

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp4ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->icmp4ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry141), value);

// Source Interface Name: defaults to blank
  memset (data->icmp4_interfacename, 0, 255);
  gtk_entry_set_text (GTK_ENTRY (data->entry151), "");

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp4ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->icmp4ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry142), value);

// Ethernet type code (16 bits): default to IPv4
// http://www.iana.org/assignments/ethernet-numbers
  data->icmp4ethhdr.type_code = htons (ETH_P_IP);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry143), value);

// ICMP data

// Clear ICMP data buffer.
  memset (data->icmp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
  data->icmp4_payloadlen = 0;
  gtk_entry_set_text (GTK_ENTRY (data->entry126), "");

// ICMP data entry format: default to hexadecimal input
  data->ascii_hex_icmp4 = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton10), TRUE);

// IPv4 header

// IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr_icmp.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  sprintf (value, "%u", data->ip4hdr_icmp.ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry57), value);

// Internet Protocol version (4 bits): IPv4
  data->ip4hdr_icmp.ip_v = 4u;
  sprintf (value, "%u", data->ip4hdr_icmp.ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry58), value);

// Type of service (8 bits): default is 0
  data->ip4hdr_icmp.ip_tos = 0;
  sprintf (value, "%u", data->ip4hdr_icmp.ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry59), value);

// Total length of datagram (16 bits): IP header + ICMP header + ICMP data
  data->ip4hdr_icmp.ip_len = htons (IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry60), value);

// ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr_icmp.ip_id = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry61), value);

// Zero (1 bit)
  ip4_flags[0] = 0;
  sprintf (value, "%u", ip4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry123), value);

// Do not fragment flag (1 bit)
  ip4_flags[1] = 0;
  sprintf (value, "%u", ip4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry65), value);

// More fragments following flag (1 bit)
  ip4_flags[2] = 0;
  sprintf (value, "%u", ip4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry66), value);

// Fragmentation offset (13 bits)
  ip4_flags[3] = 0;
  sprintf (value, "%u", ip4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry67), value);

// Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr_icmp.ip_off = htons ((ip4_flags[0] << 15) +
   (ip4_flags[1] << 14) + (ip4_flags[2] << 13) + ip4_flags[3]);

// Time-to-Live (8 bits): default to maximum value
  data->ip4hdr_icmp.ip_ttl = 255u;
  sprintf (value, "%u", data->ip4hdr_icmp.ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry68), value);

// Transport layer protocol (8 bits): 1 for ICMP
  data->ip4hdr_icmp.ip_p = IPPROTO_ICMP;
  sprintf (value, "%u", data->ip4hdr_icmp.ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry75), value);

// Source IPv4 address (32 bits): default to random
  data->ran_icmp4_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton9), TRUE);
  inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_icmp.ip_src);
  inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry63), value);

// Destination IPv4 address (32 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "127.0.0.1");
  inet_pton (AF_INET, ipaddress, &data->ip4hdr_icmp.ip_dst);
  inet_ntop (AF_INET, &(data->ip4hdr_icmp.ip_dst), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry64), value);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp.ip_sum = 0;
  data->ip4hdr_icmp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp, IP4_HDRLEN);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry62), value);

// ICMP header (IPv4)

// Message Type (8 bits): echo request
  data->icmp4hdr.icmp_type = ICMP_ECHO;
  sprintf (value, "%u", data->icmp4hdr.icmp_type);
  gtk_entry_set_text (GTK_ENTRY (data->entry69), value);

// Message Code (8 bits): echo request
  data->icmp4hdr.icmp_code = 0;
  sprintf (value, "%u", data->icmp4hdr.icmp_code);
  gtk_entry_set_text (GTK_ENTRY (data->entry70), value);

// Identifier (16 bits): usually pid of sending process - we'll arbitrarily use 1000
  data->icmp4hdr.icmp_id = htons (1000u);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry72), value);

// Sequence Number (16 bits): starts at 0
  data->icmp4hdr.icmp_seq = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry73), value);

// ICMP header checksum (16 bits)
  data->icmp4hdr.icmp_cksum = icmp4_checksum (data->icmp4hdr, data->icmp4_payload, data->icmp4_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr.icmp_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry71), value);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv6 ICMP page default settings
gboolean
icmp6_default (SPSData *data)
{
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in icmp6_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in icmp6_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp6ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->icmp6ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry128), value);

// Source Interface Name: defaults to blank
  memset (data->icmp6_interfacename, 0, 255);
  gtk_entry_set_text (GTK_ENTRY (data->entry153), "");

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp6ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->icmp6ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry129), value);
  
// Ethernet type code (16 bits): default to IPv6
// http://www.iana.org/assignments/ethernet-numbers
  data->icmp6ethhdr.type_code = htons (ETH_P_IPV6);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry130), value);

// ICMP data

// Clear ICMP data buffer.
  memset (data->icmp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
  data->icmp6_payloadlen = 0;
  gtk_entry_set_text (GTK_ENTRY (data->entry127), "");

// ICMP data entry format: default to hexadecimal input
  data->ascii_hex_icmp6 = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton12), TRUE);

// IPv6 header

// IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr_icmp.ip6_flow = htonl ((6 << 28) | (0 << 20) | 0);

// Version (4 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) >> 28));
  gtk_entry_set_text (GTK_ENTRY (data->entry110), value);

// Traffic Class (8 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) >> 20) & 0xfful);
  gtk_entry_set_text (GTK_ENTRY (data->entry111), value);

// Flow Label (20 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp.ip6_flow) & 0xffffful));
  gtk_entry_set_text (GTK_ENTRY (data->entry112), value);

// Payload length (16 bits): ICMP header + ICMP data
  data->ip6hdr_icmp.ip6_plen = htons (ICMP_HDRLEN + data->icmp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry113), value);

// Next header (8 bits): 58 for ICMP
  data->ip6hdr_icmp.ip6_nxt = IPPROTO_ICMPV6;
  sprintf (value, "%u", data->ip6hdr_icmp.ip6_nxt);
  gtk_entry_set_text (GTK_ENTRY (data->entry114), value);

// Hop limit (8 bits): default to maximum value
  data->ip6hdr_icmp.ip6_hops = 255u;
  sprintf (value, "%u", data->ip6hdr_icmp.ip6_hops);
  gtk_entry_set_text (GTK_ENTRY (data->entry115), value);

// Source IPv6 address (128 bits): default to random
  data->ran_icmp6_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton10), TRUE);
  inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_icmp.ip6_src);
  inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_src), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry109), value);

// Destination IPv6 address (128 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "::1");
  inet_pton (AF_INET6, ipaddress, &data->ip6hdr_icmp.ip6_dst);
  inet_ntop (AF_INET6, &(data->ip6hdr_icmp.ip6_dst), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry116), value);

// ICMP header (IPv6)

// Message Type (8 bits): 128 is echo request
  data->icmp6hdr.icmp6_type = ICMP6_ECHO_REQUEST;
  sprintf (value, "%u", data->icmp6hdr.icmp6_type);
  gtk_entry_set_text (GTK_ENTRY (data->entry117), value);

// Message Code (8 bits): echo request
  data->icmp6hdr.icmp6_code = 0;
  sprintf (value, "%u", data->icmp6hdr.icmp6_code);
  gtk_entry_set_text (GTK_ENTRY (data->entry118), value);

// Identifier (16 bits): usually pid of sending process - we'll arbitrarily use 1000
  data->icmp6hdr.icmp6_id = htons (1000u);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry120), value);

// Sequence Number (16 bits): starts at 0
  data->icmp6hdr.icmp6_seq = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry121), value);

// ICMP header checksum (16 bits)
  data->icmp6hdr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp, data->icmp6hdr, data->icmp6_payload, data->icmp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr.icmp6_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry119), value);

// Free allocated memory.
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv4 UDP page default settings
gboolean
udp4_default (SPSData *data)
{
  unsigned char *ip4_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in udp4_default().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in udp4_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in udp4_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Default to have the user specify the ethernet header
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton19), TRUE);
  data->udp4_specify_ether = 1;  // The toggling done above will trigger the callback, but we set it here to be sure.

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp4ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->udp4ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry147), value);

// Source Interface Name: defaults to blank
  memset (data->udp4_interfacename, 0, 255);
  gtk_entry_set_text (GTK_ENTRY (data->entry152), "");

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp4ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->udp4ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry148), value);

// Ethernet type code (16 bits): default to IPv4
// http://www.iana.org/assignments/ethernet-numbers
  data->udp4ethhdr.type_code = htons (ETH_P_IP);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry149), value);

// UDP data

// Clear UDP data buffer.
  memset (data->udp4_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
  data->udp4_payloadlen = 0;
  gtk_entry_set_text (GTK_ENTRY (data->entry17), "");

// UDP data entry format: default to hexadecimal input
  data->ascii_hex_udp4 = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton14), TRUE);

// IPv4 header

// IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr_udp.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);
  sprintf (value, "%u", data->ip4hdr_udp.ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry1), value);

// Internet Protocol version (4 bits): IPv4
  data->ip4hdr_udp.ip_v = 4u;
  sprintf (value, "%u", data->ip4hdr_udp.ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry2), value);

// Type of service (8 bits): default is 0
  data->ip4hdr_udp.ip_tos = 0;
  sprintf (value, "%u", data->ip4hdr_udp.ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry3), value);

// Total length of datagram (16 bits): IP header + UDP header + UDP payload
  data->ip4hdr_udp.ip_len = htons (IP4_HDRLEN + UDP_HDRLEN + data->udp4_payloadlen);
  sprintf (value, "%hu", htons (data->ip4hdr_udp.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry4), value);

// ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr_udp.ip_id = htons (0);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry5), value);

// Zero (1 bit)
  ip4_flags[0] = 0;
  sprintf (value, "%u", ip4_flags[0]);
  gtk_entry_set_text (GTK_ENTRY (data->entry124), value);

// Do not fragment flag (1 bit)
  ip4_flags[1] = 0;
  sprintf (value, "%u", ip4_flags[1]);
  gtk_entry_set_text (GTK_ENTRY (data->entry9), value);

// More fragments following flag (1 bit)
  ip4_flags[2] = 0;
  sprintf (value, "%u", ip4_flags[2]);
  gtk_entry_set_text (GTK_ENTRY (data->entry10), value);

// Fragmentation offset (13 bits)
  ip4_flags[3] = 0;
  sprintf (value, "%u", ip4_flags[3]);
  gtk_entry_set_text (GTK_ENTRY (data->entry11), value);

// Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr_udp.ip_off = htons ((ip4_flags[0] << 15) +
   (ip4_flags[1] << 14) + (ip4_flags[2] << 13) + ip4_flags[3]);

// Time-to-Live (8 bits): default to maximum value
  data->ip4hdr_udp.ip_ttl = 255u;
  sprintf (value, "%u", data->ip4hdr_udp.ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry12), value);

// Transport layer protocol (8 bits): 17 for UDP
  data->ip4hdr_udp.ip_p = IPPROTO_UDP;
  sprintf (value, "%u", data->ip4hdr_udp.ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry76), value);

// Source IPv4 address (32 bits): default to random
  data->ran_udp4_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton1), TRUE);
  inet_pton (AF_INET, ran4_addr (ipaddress), &data->ip4hdr_udp.ip_src);
  inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry7), value);

// Destination IPv4 address (32 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "127.0.0.1");
  inet_pton (AF_INET, ipaddress, &data->ip4hdr_udp.ip_dst);
  inet_ntop (AF_INET, &(data->ip4hdr_udp.ip_dst), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry8), value);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp.ip_sum = 0;
  data->ip4hdr_udp.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp, IP4_HDRLEN);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry6), value);

// UDP header (IPv4)

// Source port number (16 bits): default to random
  data->ran_udp4_sourceport = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton3), TRUE);
  data->udp4hdr.uh_sport = htons (ran_val2 ());
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry13), value);

// Destination port number (16 bits): we'll arbitrarily use 4950
  data->udp4hdr.uh_dport = htons (4950u);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry14), value);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp4hdr.uh_ulen = htons (UDP_HDRLEN + data->udp4_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry15), value);

// UDP checksum (16 bits)
  data->udp4hdr.uh_sum = udp4_checksum (data->ip4hdr_udp, data->udp4hdr, data->udp4_payload, data->udp4_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry16), value);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv6 UDP page default settings
gboolean
udp6_default (SPSData *data)
{
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in udp6_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in udp6_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp6ethhdr.dst_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->udp6ethhdr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry134), value);

// Source Interface Name: defaults to blank
  memset (data->udp6_interfacename, 0, 255);
  gtk_entry_set_text (GTK_ENTRY (data->entry155), "");

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp6ethhdr.src_mac, 0, 6);
  memset (value, 0, 40);
  mac_bin2string (data->udp6ethhdr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry135), value);
  
// Ethernet type code (16 bits): default to IPv6
// http://www.iana.org/assignments/ethernet-numbers
  data->udp6ethhdr.type_code = htons (ETH_P_IPV6);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6ethhdr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry136), value);

// UDP data

// Clear UDP data buffer.
  memset (data->udp6_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
  data->udp6_payloadlen = 0;
  gtk_entry_set_text (GTK_ENTRY (data->entry108), "");

// UDP data entry format: default to hexadecimal input
  data->ascii_hex_udp6 = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton16), TRUE);

// IPv6 header

// IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr_udp.ip6_flow = htonl ((6 << 28) | (0 << 20) | 0);

// Version (4 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) >> 28));
  gtk_entry_set_text (GTK_ENTRY (data->entry97), value);

// Traffic Class (8 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) >> 20) & 0xfful);
  gtk_entry_set_text (GTK_ENTRY (data->entry98), value);

// Flow Label (20 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp.ip6_flow) & 0xffffful));
  gtk_entry_set_text (GTK_ENTRY (data->entry99), value);

// Payload length (16 bits): UDP header + UDP payload
  data->ip6hdr_udp.ip6_plen = htons (UDP_HDRLEN + data->udp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry100), value);

// Next header (8 bits): 17 for UDP
  data->ip6hdr_udp.ip6_nxt = IPPROTO_UDP;
  sprintf (value, "%u", data->ip6hdr_udp.ip6_nxt);
  gtk_entry_set_text (GTK_ENTRY (data->entry101), value);

// Hop limit (8 bits): default to maximum value
  data->ip6hdr_udp.ip6_hops = 255u;
  sprintf (value, "%u", data->ip6hdr_udp.ip6_hops);
  gtk_entry_set_text (GTK_ENTRY (data->entry102), value);

// Source IPv6 address (128 bits): default to random
  data->ran_udp6_sourceip = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton6), TRUE);
  inet_pton (AF_INET6, ran6_addr (ipaddress), &data->ip6hdr_udp.ip6_src);
  inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_src), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry96), value);

// Destination IPv6 address (128 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "::1");
  inet_pton (AF_INET6, ipaddress, &data->ip6hdr_udp.ip6_dst);
  inet_ntop (AF_INET6, &(data->ip6hdr_udp.ip6_dst), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry103), value);

// UDP header (IPv6)

// Source port number (16 bits): default to random
  data->ran_udp6_sourceport = 1;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton8), TRUE);
  data->udp6hdr.uh_sport = htons (ran_val2 ());
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry104), value);

// Destination port number (16 bits): we'll arbitrarily use 4950
  data->udp6hdr.uh_dport = htons (4950u);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry105), value);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp6hdr.uh_ulen = htons (UDP_HDRLEN + data->udp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry106), value);

// UDP header checksum (16 bits)
  data->udp6hdr.uh_sum = udp6_checksum (data->ip6hdr_udp, data->udp6hdr, data->udp6_payload, data->udp6_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry107), value);

// Free allocated memory.
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Call tcp4_default()
gboolean
on_button11_clicked (GtkButton *button, SPSData *data)
{
  tcp4_default (data);

  return (TRUE);
}

// Call icmp4_default()
gboolean
on_button12_clicked (GtkButton *button, SPSData *data)
{
  icmp4_default (data);

  return (TRUE);
}

// Call udp4_default()
gboolean
on_button15_clicked (GtkButton *button, SPSData *data)
{
  udp4_default (data);

  return (TRUE);
}

// Call tcp6_default()
gboolean
on_button16_clicked (GtkButton *button, SPSData *data)
{
  tcp6_default (data);

  return (TRUE);
}

// Call icmp6_default()
gboolean
on_button18_clicked (GtkButton *button, SPSData *data)
{
  icmp6_default (data);

  return (TRUE);
}

// Call udp6_default()
gboolean
on_button19_clicked (GtkButton *button, SPSData *data)
{
  udp6_default (data);

  return (TRUE);
}

// Tunnel IPv6 over IPv4 (6to4) (traceroute.ui)

// Specify ethernet header
gboolean
on_checkbutton32_toggled (GtkButton *button, SPSData *data)
{
  if (data->sixto4_tr_specify_ether == 0) {
    data->sixto4_tr_specify_ether = 1;
  } else {
    data->sixto4_tr_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry312_activate (GtkWidget *entry312, SPSData *data)
{
  int i, bad;
  const gchar *entry312_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry312_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry312_text = gtk_entry_get_text (GTK_ENTRY (data->entry312));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry312_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry312_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry312_text[i] > 47) && (entry312_text[i] < 58)) ||    // 0 to 10
        ((entry312_text[i] > 64) && (entry312_text[i] < 71)) ||      // A to Z
        ((entry312_text[i] > 96) && (entry312_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry312_text[i] > 47) && (entry312_text[i] < 58)) ||    // 0 to 10
        ((entry312_text[i] > 64) && (entry312_text[i] < 71)) ||      // A to Z
        ((entry312_text[i] > 96) && (entry312_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->sixto4ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry312), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry312_text, data->sixto4ethhdr_tr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry313_activate (GtkWidget *entry313, SPSData *data)
{
  int i, bad;
  const gchar *entry313_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry313_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry313_text = gtk_entry_get_text (GTK_ENTRY (data->entry313));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry313_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry313_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry313_text[i] > 47) && (entry313_text[i] < 58)) ||    // 0 to 10
        ((entry313_text[i] > 64) && (entry313_text[i] < 71)) ||      // A to Z
        ((entry313_text[i] > 96) && (entry313_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry313_text[i] > 47) && (entry313_text[i] < 58)) ||    // 0 to 10
        ((entry313_text[i] > 64) && (entry313_text[i] < 71)) ||      // A to Z
        ((entry313_text[i] > 96) && (entry313_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->sixto4ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry313), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry313_text, data->sixto4ethhdr_tr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry314_activate (GtkWidget *entry314, SPSData *data)
{
  int i, sd, status;
  const gchar *entry314_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry314_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry314_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry314_text = gtk_entry_get_text (GTK_ENTRY (entry314));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry314_activate.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry314_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry314_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->sixto4_tr_interfacename, entry314_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->sixto4ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->sixto4ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry313), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry315_activate (GtkWidget *entry315, SPSData *data)
{
  const gchar *entry315_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry315_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry315_text = gtk_entry_get_text (GTK_ENTRY (data->entry315));

// Is value too large for a short int?
  if ((atoi (entry315_text) < 0) || (atoi (entry315_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->sixto4ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry315), value);
  } else {
    data->sixto4ethhdr_tr.type_code = htons (atoi (entry315_text));
  }

  g_free (value);

  return (TRUE);
}

// TCP IPv4 Source Address for traceroute 6to4 tunnel of IPv6 over IPv4
gboolean
on_entry316_activate (GtkWidget *entry316, SPSData *data)
{
  int status;
  const gchar *entry316_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry316_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry316_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry316_text = gtk_entry_get_text (GTK_ENTRY (data->entry316));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry316_text) != TRUE) {
// Invalid IP - use old value
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_tcp_tr.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry316), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry316_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry316_text, &data->sixto4ip4hdr_tcp_tr.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry316_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_tcp_tr.ip_sum = 0;
    data->sixto4ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp_tr, IP4_HDRLEN);

// Note: The TCP checksum does not need updating since it contains the IPv6
// source and destination addresses, not the IPv4 6to4 header addresses.

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// ICMP IPv4 Source Address for traceroute 6to4 tunnel of IPv6 over IPv4
gboolean
on_entry317_activate (GtkWidget *entry317, SPSData *data)
{
  int status;
  const gchar *entry317_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry317_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry317_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry317_text = gtk_entry_get_text (GTK_ENTRY (data->entry317));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry317_text) != TRUE) {
// Invalid IP - use old value
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_icmp_tr.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry317), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry317_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry317_text, &data->sixto4ip4hdr_icmp_tr.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry317_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_icmp_tr.ip_sum = 0;
    data->sixto4ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp_tr, IP4_HDRLEN);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// UDP IPv4 Source Address for traceroute 6to4 tunnel of IPv6 over IPv4
gboolean
on_entry318_activate (GtkWidget *entry318, SPSData *data)
{
  int status;
  const gchar *entry318_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry318_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry318_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry318_text = gtk_entry_get_text (GTK_ENTRY (data->entry318));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry318_text) != TRUE) {
// Invalid IP - use old value
    inet_ntop (AF_INET, &(data->sixto4ip4hdr_udp_tr.ip_src), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry318), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry318_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry318_text, &data->sixto4ip4hdr_udp_tr.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry318_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->sixto4ip4hdr_udp_tr.ip_sum = 0;
    data->sixto4ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp_tr, IP4_HDRLEN);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// TCP page (IPv4) for Traceroute

// Specify ethernet header
gboolean
on_checkbutton28_toggled (GtkButton *button, SPSData *data)
{
  if (data->tcp4_tr_specify_ether == 0) {
    data->tcp4_tr_specify_ether = 1;
  } else {
    data->tcp4_tr_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry164_activate (GtkWidget *entry164, SPSData *data)
{
  int i, bad;
  const gchar *entry164_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry164_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry164_text = gtk_entry_get_text (GTK_ENTRY (data->entry164));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry164_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry164_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry164_text[i] > 47) && (entry164_text[i] < 58)) ||    // 0 to 10
        ((entry164_text[i] > 64) && (entry164_text[i] < 71)) ||      // A to Z
        ((entry164_text[i] > 96) && (entry164_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry164_text[i] > 47) && (entry164_text[i] < 58)) ||    // 0 to 10
        ((entry164_text[i] > 64) && (entry164_text[i] < 71)) ||      // A to Z
        ((entry164_text[i] > 96) && (entry164_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp4ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry164), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry164_text, data->tcp4ethhdr_tr.dst_mac);
  }

  g_free (value);
  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry165_activate (GtkWidget *entry165, SPSData *data)
{
  int i, bad;
  const gchar *entry165_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry165_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry165_text = gtk_entry_get_text (GTK_ENTRY (data->entry165));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry165_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry165_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry165_text[i] > 47) && (entry165_text[i] < 58)) ||    // 0 to 10
        ((entry165_text[i] > 64) && (entry165_text[i] < 71)) ||      // A to Z
        ((entry165_text[i] > 96) && (entry165_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry165_text[i] > 47) && (entry165_text[i] < 58)) ||    // 0 to 10
        ((entry165_text[i] > 64) && (entry165_text[i] < 71)) ||      // A to Z
        ((entry165_text[i] > 96) && (entry165_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp4ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry165), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry165_text, data->tcp4ethhdr_tr.src_mac);
  }

  g_free (value);
  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry166_activate (GtkWidget *entry166, SPSData *data)
{
  int i, sd, status;
  const gchar *entry166_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry166_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry166_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry166_text = gtk_entry_get_text (GTK_ENTRY (entry166));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry166_activate().\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry166_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry166_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->tcp4_tr_interfacename, entry166_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->tcp4ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->tcp4ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry165), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry167_activate (GtkWidget *entry167, SPSData *data)
{
  const gchar *entry167_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry167_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry167_text = gtk_entry_get_text (GTK_ENTRY (data->entry167));

// Is value too large for a short int?
  if ((atoi (entry167_text) < 0) || (atoi (entry167_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry167), value);
  } else {
    data->tcp4ethhdr_tr.type_code = htons (atoi (entry167_text));
  }

  g_free (value);
  return (TRUE);
}

// IPv4 header length (4 bits)
gboolean
on_entry168_activate (GtkWidget *entry168, SPSData *data)
{
  const gchar *entry168_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry168_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry168_text = gtk_entry_get_text (GTK_ENTRY (data->entry168));

// Is value too large or small?
  if ((atoi (entry168_text) < 0) || (atoi (entry168_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry168), value);
  } else {
    data->ip4hdr_tcp_tr.ip_hl = atoi (entry168_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry169_activate (GtkWidget *entry169, SPSData *data)
{
  const gchar *entry169_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry169_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry169_text = gtk_entry_get_text (GTK_ENTRY (data->entry169));

// Is value too large or small?
  if ((atoi (entry169_text) < 0) || (atoi (entry169_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry169), value);
  } else {
    data->ip4hdr_tcp_tr.ip_v = atoi (entry169_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }
  g_free (value);

  return (TRUE);
}

// Type of service (8 bits)
gboolean
on_entry170_activate (GtkWidget *entry170, SPSData *data)
{
  const gchar *entry170_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry170_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry170_text = gtk_entry_get_text (GTK_ENTRY (data->entry170));

// Is value too large or small?
  if ((atoi (entry170_text) < 0) || (atoi (entry170_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry170), value);
  } else {
    data->ip4hdr_tcp_tr.ip_tos = atoi (entry170_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }
  g_free (value);

  return (TRUE);
}

// Total length of datagram (16 bits)
gboolean
on_entry171_activate (GtkWidget *entry171, SPSData *data)
{
  const gchar *entry171_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry171_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry171_text = gtk_entry_get_text (GTK_ENTRY (data->entry171));

// Is value too large or small?
  if ((atoi (entry171_text) < 0) || (atoi (entry171_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry171), value);
  } else {
    data->ip4hdr_tcp_tr.ip_len = htons (atoi (entry171_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }
  g_free (value);

  return (TRUE);
}

// ID sequence number (16 bits)
gboolean
on_entry172_activate (GtkWidget *entry172, SPSData *data)
{
  const gchar *entry172_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry172_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry172_text = gtk_entry_get_text (GTK_ENTRY (data->entry172));

// Is value too large or small?
  if ((atoi (entry172_text) < 0) || (atoi (entry172_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry172), value);
  } else {
    data->ip4hdr_tcp_tr.ip_id = htons (atoi (entry172_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }
  g_free (value);

  return (TRUE);
}

// Zero (1 bit)
gboolean
on_entry173_activate (GtkWidget *entry173, SPSData *data)
{
  const gchar *entry173_text, *entry174_text, *entry175_text, *entry176_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry173_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry173_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry173_text = gtk_entry_get_text (GTK_ENTRY (data->entry173));

// Is value too large or small?
  if ((atoi (entry173_text) < 0) || (atoi (entry173_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry173), value);
  } else {
    ip_tcp_flags[0] = atoi (entry173_text);                           // Zero (1 bit)
    entry174_text = gtk_entry_get_text (GTK_ENTRY (data->entry174));  // Do not fragment flag (1 bit)
    ip_tcp_flags[1] = atoi (entry174_text);
    entry175_text = gtk_entry_get_text (GTK_ENTRY (data->entry175));  // More fragments following flag (1 bit)
    ip_tcp_flags[2] = atoi (entry175_text);
    entry176_text = gtk_entry_get_text (GTK_ENTRY (data->entry176));  // Fragmentation offset (13 bits)
    ip_tcp_flags[3] = atoi (entry176_text);
    data->ip4hdr_tcp_tr.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// Do not fragment flag (1 bit)
gboolean
on_entry174_activate (GtkWidget *entry174, SPSData *data)
{
  const gchar *entry174_text, *entry173_text, *entry175_text, *entry176_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry174_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry174_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry174_text = gtk_entry_get_text (GTK_ENTRY (data->entry174));

// Is value too large or small?
  if ((atoi (entry174_text) < 0) || (atoi (entry174_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_tcp_tr.ip_off) >> 14) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry174), value);
  } else {
    entry173_text = gtk_entry_get_text (GTK_ENTRY (data->entry173));  // Zero (1 bit)
    ip_tcp_flags[0] = atoi (entry173_text);
    ip_tcp_flags[1] = atoi (entry174_text);                           // Do not fragment flag (1 bit)
    entry175_text = gtk_entry_get_text (GTK_ENTRY (data->entry175));  // More fragments following flag (1 bit)
    ip_tcp_flags[2] = atoi (entry175_text);
    entry176_text = gtk_entry_get_text (GTK_ENTRY (data->entry176));  // Fragmentation offset (13 bits)
    ip_tcp_flags[3] = atoi (entry176_text);
    data->ip4hdr_tcp_tr.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// More fragments following flag (1 bit)
gboolean
on_entry175_activate (GtkWidget *entry175, SPSData *data)
{
  const gchar *entry175_text, *entry173_text, *entry174_text, *entry176_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry175_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry175_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry175_text = gtk_entry_get_text (GTK_ENTRY (data->entry175));

// Is value too large or small?
  if ((atoi (entry175_text) < 0) || (atoi (entry175_text) > 1u)) {
    sprintf (value, "%hu", ((ntohs (data->ip4hdr_tcp_tr.ip_off) >> 13) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry175), value);
  } else {
    entry173_text = gtk_entry_get_text (GTK_ENTRY (data->entry173));  // Zero (1 bit)
    ip_tcp_flags[0] = atoi (entry173_text);
    entry174_text = gtk_entry_get_text (GTK_ENTRY (data->entry174));  // Do not fragment flag (1 bit)
    ip_tcp_flags[1] = atoi (entry174_text);
    ip_tcp_flags[2] = atoi (entry175_text);                           // More fragments following flag (1 bit)
    entry176_text = gtk_entry_get_text (GTK_ENTRY (data->entry176));  // Fragmentation offset (13 bits)
    ip_tcp_flags[3] = atoi (entry176_text);
    data->ip4hdr_tcp_tr.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// Fragmentation offset (13 bits)
gboolean
on_entry176_activate (GtkWidget *entry176, SPSData *data)
{
  const gchar *entry176_text, *entry173_text, *entry174_text, *entry175_text;
  gushort *ip_tcp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry176_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_tcp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_tcp_flags' in on_entry176_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry176_text = gtk_entry_get_text (GTK_ENTRY (data->entry176));

// Is value too large or small?
  if ((atoi (entry176_text) < 0) || (atoi (entry176_text) > 0x1fffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_off) & 0x1fffu);
    gtk_entry_set_text (GTK_ENTRY (data->entry176), value);
  } else {
    entry173_text = gtk_entry_get_text (GTK_ENTRY (data->entry173));  // Zero (1 bit)
    ip_tcp_flags[0] = atoi (entry173_text);
    entry174_text = gtk_entry_get_text (GTK_ENTRY (data->entry174));  // Do not fragment flag (1 bit)
    ip_tcp_flags[1] = atoi (entry174_text);
    entry175_text = gtk_entry_get_text (GTK_ENTRY (data->entry175));  // More fragments following flag (1 bit)
    ip_tcp_flags[2] = atoi (entry175_text);
    ip_tcp_flags[3] = atoi (entry176_text);                           // Fragmentation offset (13 bits)
    data->ip4hdr_tcp_tr.ip_off = htons ((ip_tcp_flags[0] << 15) +
                       (ip_tcp_flags[1] << 14) +
                       (ip_tcp_flags[2] << 13) +
                        ip_tcp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);
  g_free (ip_tcp_flags);

  return (TRUE);
}

// Time-to-Live (8 bits)
gboolean
on_entry177_activate (GtkWidget *entry177, SPSData *data)
{
  const gchar *entry177_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry177_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry177_text = gtk_entry_get_text (GTK_ENTRY (data->entry177));

// Is value too large or small?
  if ((atoi (entry177_text) < 0) || (atoi (entry177_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry177), value);
  } else {
    data->ip4hdr_tcp_tr.ip_ttl = atoi (entry177_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);

  return (TRUE);
}

// Transport layer protocol (8 bits)
gboolean
on_entry178_activate (GtkWidget *entry178, SPSData *data)
{
  const gchar *entry178_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry178_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry178_text = gtk_entry_get_text (GTK_ENTRY (data->entry178));

// Is value too large or small?
  if ((atoi (entry178_text) < 0) || (atoi (entry178_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry178), value);
  } else {
    data->ip4hdr_tcp_tr.ip_p = atoi (entry178_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (32 bits)
gboolean
on_entry179_activate (GtkWidget *entry179, SPSData *data)
{
  int status;
  const gchar *entry179_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry179_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry179_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry179_text = gtk_entry_get_text (GTK_ENTRY (data->entry179));

// Is it a valid IPv4 address?
  if (is_valid_ip4 (entry179_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_tcp_tr.ip_src.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry179), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry179_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry179_text, &data->ip4hdr_tcp_tr.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry179_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);

// TCP header checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), text);
  }

  g_free (text);
  g_free (value);

  return (TRUE);
}

// Destination IP address (32 bits)
gboolean
on_entry180_activate (GtkWidget *entry180, SPSData *data)
{
  int status;
  const gchar *entry180_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry180_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry180_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry180_text = gtk_entry_get_text (GTK_ENTRY (data->entry180));

// Is it a valid IP address?
  if (is_valid_ip4 (entry180_text) != TRUE) {

// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_tcp_tr.ip_dst.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry180), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry180_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry180_text, &data->ip4hdr_tcp_tr.ip_dst)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry180_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_tcp_tr.ip_sum = 0;
    data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), text);
    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// IP checksum (16 bits)
gboolean
on_entry181_activate (GtkWidget *entry181, SPSData *data)
{
  const gchar *entry181_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry181_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry181_text = gtk_entry_get_text (GTK_ENTRY (data->entry181));

// Is value too large or small?
  if ((atoi (entry181_text) < 0) || (atoi (entry181_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry181), value);
  } else {
    data->ip4hdr_tcp_tr.ip_sum = htons (atoi (entry181_text));  // Update IPv4 header with new value
  }

  g_free (value);

  return (TRUE);
}

// TCP header (IPv4) for traceroute

// TCP source port (16 bits)
gboolean
on_entry182_activate (GtkWidget *entry182, SPSData *data)
{
  const gchar *entry182_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry182_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry182_text = gtk_entry_get_text (GTK_ENTRY (data->entry182));

// Is value too large or small?
  if ((atoi (entry182_text) < 0) || (atoi (entry182_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry182), value);
  } else {
    data->tcp4hdr_tr.th_sport = htons (atoi (entry182_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton30), FALSE);
    data->ran_tcp4_tr_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize TCP source port address
gboolean
on_checkbutton30_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton30_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_tcp4_tr_sourceport == 0) {
    data->ran_tcp4_tr_sourceport = 1;
    data->tcp4hdr_tr.th_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry182), value);

// TCP header checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  } else {
    data->ran_tcp4_tr_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// TCP destination port (16 bits)
gboolean
on_entry183_activate (GtkWidget *entry183, SPSData *data)
{
  const gchar *entry183_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry183_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry183_text = gtk_entry_get_text (GTK_ENTRY (data->entry183));

// Is value too large or small?
  if ((atoi (entry183_text) < 0) || (atoi (entry183_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry183), value);
  } else {
    data->tcp4hdr_tr.th_dport = htons (atoi (entry183_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP sequence number (32 bits)
gboolean
on_entry184_activate (GtkWidget *entry184, SPSData *data)
{
  const gchar *entry184_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry184_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry184_text = gtk_entry_get_text (GTK_ENTRY (data->entry184));

// Is value too large or small?
  if ((atol (entry184_text) < 0) || (strtoul (entry184_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr_tr.th_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry184), value);
  } else {
    data->tcp4hdr_tr.th_seq = htonl (atol (entry184_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP acknowledgement number (32 bits)
gboolean
on_entry185_activate (GtkWidget *entry185, SPSData *data)
{
  const gchar *entry185_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry185_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry185_text = gtk_entry_get_text (GTK_ENTRY (data->entry185));

// Is value too large or small?
  if ((atol (entry185_text) < 0) || (strtoul (entry185_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr_tr.th_ack));
    gtk_entry_set_text (GTK_ENTRY (data->entry185), value);
  } else {
    data->tcp4hdr_tr.th_ack = htonl (atol (entry185_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP reserved (4 bits)
gboolean
on_entry186_activate (GtkWidget *entry186, SPSData *data)
{
  const gchar *entry186_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry186_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry186_text = gtk_entry_get_text (GTK_ENTRY (data->entry186));

// Is value too large or small?
  if ((atoi (entry186_text) < 0) || (atoi (entry186_text) > 15)) {
    sprintf (value, "%u", data->tcp4hdr_tr.th_x2);
    gtk_entry_set_text (GTK_ENTRY (data->entry186), value);
  } else {
    data->tcp4hdr_tr.th_x2 = atoi (entry186_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP data offset (4 bits)
gboolean
on_entry187_activate (GtkWidget *entry187, SPSData *data)
{
  const gchar *entry187_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry187_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry187_text = gtk_entry_get_text (GTK_ENTRY (data->entry187));

// Is value too large or small?
  if ((atoi (entry187_text) < 0) || (atoi (entry187_text) > 15)) {
    sprintf (value, "%u", data->tcp4hdr_tr.th_off);
    gtk_entry_set_text (GTK_ENTRY (data->entry187), value);
  } else {
    data->tcp4hdr_tr.th_off = atoi (entry187_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP FIN flag (1 bit)
gboolean
on_entry188_activate (GtkWidget *entry188, SPSData *data)
{
  const gchar *entry188_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry188_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry188_text = gtk_entry_get_text (GTK_ENTRY (data->entry188));

// Is value too large or small?
  if ((atoi (entry188_text) < 0) || (atoi (entry188_text) > 1)) {
    sprintf (value, "%u", data->tcp4hdr_tr.th_flags & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry188), value);
  } else {
    if (atoi (entry188_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 254;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 1;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP SYN flag (1 bit)
gboolean
on_entry189_activate (GtkWidget *entry189, SPSData *data)
{
  const gchar *entry189_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry189_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry189_text = gtk_entry_get_text (GTK_ENTRY (data->entry189));

// Is value too large or small?
  if ((atoi (entry189_text) < 0) || (atoi (entry189_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 2) >> 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry189), value);
  } else {
    if (atoi (entry189_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 253;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 2;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP RST flag (1 bit)
gboolean
on_entry190_activate (GtkWidget *entry190, SPSData *data)
{
  const gchar *entry190_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry190_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry190_text = gtk_entry_get_text (GTK_ENTRY (data->entry190));

// Is value too large or small?
  if ((atoi (entry190_text) < 0) || (atoi (entry190_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 4) >> 2);
    gtk_entry_set_text (GTK_ENTRY (data->entry190), value);
  } else {
    if (atoi (entry190_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 251;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 4;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP PSH flag (1 bit)
gboolean
on_entry191_activate (GtkWidget *entry191, SPSData *data)
{
  const gchar *entry191_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry191_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry191_text = gtk_entry_get_text (GTK_ENTRY (data->entry191));

// Is value too large or small?
  if ((atoi (entry191_text) < 0) || (atoi (entry191_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 8) >> 3);
    gtk_entry_set_text (GTK_ENTRY (data->entry191), value);
  } else {
    if (atoi (entry191_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 247;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 8;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ACK flag (1 bit)
gboolean
on_entry192_activate (GtkWidget *entry192, SPSData *data)
{
  const gchar *entry192_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry192_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry192_text = gtk_entry_get_text (GTK_ENTRY (data->entry192));

// Is value too large or small?
  if ((atoi (entry192_text) < 0) || (atoi (entry192_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 16) >> 4);
    gtk_entry_set_text (GTK_ENTRY (data->entry192), value);
  } else {
    if (atoi (entry192_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 239;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 16;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP URG flag (1 bit)
gboolean
on_entry193_activate (GtkWidget *entry193, SPSData *data)
{
  const gchar *entry193_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry193_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry193_text = gtk_entry_get_text (GTK_ENTRY (data->entry193));

// Is value too large or small?
  if ((atoi (entry193_text) < 0) || (atoi (entry193_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 32) >> 5);
    gtk_entry_set_text (GTK_ENTRY (data->entry193), value);
  } else {
    if (atoi (entry193_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 223;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 32;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ECE flag (1 bit)
gboolean
on_entry194_activate (GtkWidget *entry194, SPSData *data)
{
  const gchar *entry194_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry194_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry194_text = gtk_entry_get_text (GTK_ENTRY (data->entry194));

// Is value too large or small?
  if ((atoi (entry194_text) < 0) || (atoi (entry194_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 64) >> 6);
    gtk_entry_set_text (GTK_ENTRY (data->entry194), value);
  } else {
    if (atoi (entry194_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 191;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 64;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP CWR flag (1 bit)
gboolean
on_entry195_activate (GtkWidget *entry195, SPSData *data)
{
  const gchar *entry195_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry195_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry195_text = gtk_entry_get_text (GTK_ENTRY (data->entry195));

// Is value too large or small?
  if ((atoi (entry195_text) < 0) || (atoi (entry195_text) > 1)) {
    sprintf (value, "%u", (data->tcp4hdr_tr.th_flags & 128) >> 7);
    gtk_entry_set_text (GTK_ENTRY (data->entry195), value);
  } else {
    if (atoi (entry195_text) == 0) {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags & 127;
    } else {
      data->tcp4hdr_tr.th_flags = data->tcp4hdr_tr.th_flags | 128;
    }

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP window size (16 bits)
gboolean
on_entry196_activate (GtkWidget *entry196, SPSData *data)
{
  const gchar *entry196_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry196_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry196_text = gtk_entry_get_text (GTK_ENTRY (data->entry196));

// Is value too large or small?
  if ((atoi (entry196_text) < 0) || (atoi (entry196_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_win));
    gtk_entry_set_text (GTK_ENTRY (data->entry196), value);
  } else {
    data->tcp4hdr_tr.th_win = htons (atoi (entry196_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP checksum (16 bits)
gboolean
on_entry197_activate (GtkWidget *entry197, SPSData *data)
{
  const gchar *entry197_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry197_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry197_text = gtk_entry_get_text (GTK_ENTRY (data->entry197));

// Is value too large or small?
  if ((atoi (entry197_text) < 0) || (atoi (entry197_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  } else {
    data->tcp4hdr_tr.th_sum = htons (atoi (entry197_text));  // Update TCP header with new value
  }

  g_free (value);

  return (TRUE);
}

// TCP urgent pointer (16 bits)
gboolean
on_entry198_activate (GtkWidget *entry198, SPSData *data)
{
  const gchar *entry198_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry198_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry198_text = gtk_entry_get_text (GTK_ENTRY (data->entry198));

// Is value too large or small?
  if ((atoi (entry198_text) < 0) || (atoi (entry198_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_urp));
    gtk_entry_set_text (GTK_ENTRY (data->entry198), value);
  } else {
    data->tcp4hdr_tr.th_urp = htons (atoi (entry198_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry197), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton29_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp4_tr = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton30_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp4_tr = 1;

  return (TRUE);
}

// TCP data from keyboard
gboolean
on_entry199_activate (GtkWidget *entry199, SPSData *data)
{
  int i, j, c;
  const gchar *entry199_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry199_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry199_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry199_text = gtk_entry_get_text (GTK_ENTRY (data->entry199));

// If user asked for no payload then zero it out.
  if (strlen (entry199_text) == 0) {
    data->tcp4_tr_payloadlen = 0;
    memset (data->tcp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry199_text) != 0) && (data->ascii_hex_tcp4_tr == 0)) {

    data->tcp4_tr_payloadlen = strlen (entry199_text);
    memset (data->tcp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
    memcpy (data->tcp4_tr_payload, entry199_text, data->tcp4_tr_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry199_text) != 0) && (data->ascii_hex_tcp4_tr == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry199_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->tcp4_tr_payloadlen = 0;
      memset (data->tcp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry199_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry199_text[i+j] != 32) && (entry199_text[i+j] != 44) && ((i + j) < strlen (entry199_text))) {
          value[j] = entry199_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->tcp4_tr_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry199_activate().\n");
          data->parent = data->traceroute_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->tcp4_tr_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry199_text[i] == 32) || (entry199_text[i] == 44)) && (i < strlen (entry199_text))) {
          i++;
        }
      }  // Looping through entry199_text
    }  // else data is in acceptable format
  }  // else data is hex

// Total length of datagram (16 bits): IP header + TCP header + TCP payload
  data->ip4hdr_tcp_tr.ip_len = htons (IP4_HDRLEN + TCP_HDRLEN + data->tcp4_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry171), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp_tr.ip_sum = 0;
  data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry181), text);

// TCP header checksum (16 bits)
  data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry197), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// TCP data from file
gboolean
on_button29_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int tcp4_lim = IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button29_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button29_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->traceroute_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button29_clicked().");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > tcp4_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", tcp4_lim);
      data->parent = data->traceroute_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->tcp4_tr_payloadlen = index;

// Clear current data.
  memset (data->tcp4_tr_payload, 0, tcp4_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry199), "");

// Copy payload from buffer.
  memcpy (data->tcp4_tr_payload, buffer, data->tcp4_tr_payloadlen);

// Total length of datagram (16 bits): IP header + TCP header + TCP payload
  data->ip4hdr_tcp_tr.ip_len = htons (IP4_HDRLEN + TCP_HDRLEN + data->tcp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry171), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp_tr.ip_sum = 0;
  data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry181), text);

// TCP header checksum (16 bits)
  data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry197), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry199), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// TCP page (IPv6) for traceroute

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry247_activate (GtkWidget *entry247, SPSData *data)
{
  int i, bad;
  const gchar *entry247_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry247_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry247_text = gtk_entry_get_text (GTK_ENTRY (data->entry247));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry247_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry247_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry247_text[i] > 47) && (entry247_text[i] < 58)) ||    // 0 to 10
        ((entry247_text[i] > 64) && (entry247_text[i] < 71)) ||      // A to Z
        ((entry247_text[i] > 96) && (entry247_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry247_text[i] > 47) && (entry247_text[i] < 58)) ||    // 0 to 10
        ((entry247_text[i] > 64) && (entry247_text[i] < 71)) ||      // A to Z
        ((entry247_text[i] > 96) && (entry247_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp6ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry247), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry247_text, data->tcp6ethhdr_tr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry248_activate (GtkWidget *entry248, SPSData *data)
{
  int i, bad;
  const gchar *entry248_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry248_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry248_text = gtk_entry_get_text (GTK_ENTRY (data->entry248));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry248_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry248_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry248_text[i] > 47) && (entry248_text[i] < 58)) ||    // 0 to 10
        ((entry248_text[i] > 64) && (entry248_text[i] < 71)) ||      // A to Z
        ((entry248_text[i] > 96) && (entry248_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry248_text[i] > 47) && (entry248_text[i] < 58)) ||    // 0 to 10
        ((entry248_text[i] > 64) && (entry248_text[i] < 71)) ||      // A to Z
        ((entry248_text[i] > 96) && (entry248_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->tcp6ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry248), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry248_text, data->tcp6ethhdr_tr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry249_activate (GtkWidget *entry249, SPSData *data)
{
  int i, sd, status;
  const gchar *entry249_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry249_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry249_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry249_text = gtk_entry_get_text (GTK_ENTRY (entry249));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for using ioctl() in on_entry249_activate.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry249_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry249_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->tcp6_tr_interfacename, entry249_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->tcp6ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->tcp6ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry248), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry250_activate (GtkWidget *entry250, SPSData *data)
{
  const gchar *entry250_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry250_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry250_text = gtk_entry_get_text (GTK_ENTRY (data->entry250));

// Is value too large for a short int?
  if ((atoi (entry250_text) < 0) || (atoi (entry250_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry250), value);
  } else {
    data->tcp6ethhdr_tr.type_code = htons (atoi (entry250_text));
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry251_activate (GtkWidget *entry251, SPSData *data)
{
  const gchar *entry251_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry251_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry251_text = gtk_entry_get_text (GTK_ENTRY (data->entry251));

// Save value currently in header
  old = ntohl (data->ip6hdr_tcp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry251_text) < 0) || (atol (entry251_text) > 15ul)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->ip6hdr_tcp_tr.ip6_flow) >> 28);
    gtk_entry_set_text (GTK_ENTRY (data->entry251), value);
  } else {
    data->ip6hdr_tcp_tr.ip6_flow = htonl (((atol (entry251_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Traffic class (8 bits)
gboolean
on_entry252_activate (GtkWidget *entry252, SPSData *data)
{
  const gchar *entry252_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry252_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry252_text = gtk_entry_get_text (GTK_ENTRY (data->entry252));

// Save value currently in header
  old = ntohl (data->ip6hdr_tcp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry252_text) < 0) || (atol (entry252_text) > 0xfful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp_tr.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry252), value);
  } else {
    data->ip6hdr_tcp_tr.ip6_flow = htonl (((atol (entry252_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Flow label (20 bits)
gboolean
on_entry253_activate (GtkWidget *entry253, SPSData *data)
{
  const gchar *entry253_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry253_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry253_text = gtk_entry_get_text (GTK_ENTRY (data->entry253));

// Save value currently in header
  old = ntohl (data->ip6hdr_tcp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry253_text) < 0) || (atol (entry253_text) > 0xffffful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp_tr.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry253), value);
  } else {
    data->ip6hdr_tcp_tr.ip6_flow = htonl (atol (entry253_text) + (old & 0xfff00000ul));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Payload length (16 bits)
gboolean
on_entry254_activate (GtkWidget *entry254, SPSData *data)
{
  const gchar *entry254_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry254_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry254_text = gtk_entry_get_text (GTK_ENTRY (data->entry254));

// Is value too large or small?
  if ((atoi (entry254_text) < 0) || (atoi (entry254_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp_tr.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry254), value);
  } else {
    data->ip6hdr_tcp_tr.ip6_plen = htons (atoi (entry254_text));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Next header (8 bits)
gboolean
on_entry255_activate (GtkWidget *entry255, SPSData *data)
{
  const gchar *entry255_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry255_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry255_text = gtk_entry_get_text (GTK_ENTRY (data->entry255));

// Is value too large or small?
  if ((atoi (entry255_text) < 0) || (atoi (entry255_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_tcp_tr.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry255), value);
  } else {
    data->ip6hdr_tcp_tr.ip6_nxt = atoi (entry255_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Hop limit (8 bits)
gboolean
on_entry256_activate (GtkWidget *entry256, SPSData *data)
{
  const gchar *entry256_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry256_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry256_text = gtk_entry_get_text (GTK_ENTRY (data->entry256));

// Is value too large or small?
  if ((atoi (entry256_text) < 0) || (atoi (entry256_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_tcp_tr.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry256), value);
  } else {
    data->ip6hdr_tcp_tr.ip6_hops = atoi (entry256_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (128 bits)
gboolean
on_entry257_activate (GtkWidget *entry257, SPSData *data)
{
  const gchar *entry257_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry257_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry257_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry257_text = gtk_entry_get_text (GTK_ENTRY (data->entry257));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry257_text, &(data->ip6hdr_tcp_tr.ip6_src)) < 1) {
    
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp_tr.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry257), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry257_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), text);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Destination IP address (128 bits)
gboolean
on_entry258_activate (GtkWidget *entry258, SPSData *data)
{
  const gchar *entry258_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry258_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry258_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry258_text = gtk_entry_get_text (GTK_ENTRY (data->entry258));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry258_text, &(data->ip6hdr_tcp_tr.ip6_dst)) < 1) {

// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp_tr.ip6_dst), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry258), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry258_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), text);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// TCP header (IPv6)

// TCP source port (16 bits)
gboolean
on_entry259_activate (GtkWidget *entry259, SPSData *data)
{
  const gchar *entry259_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry259_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry259_text = gtk_entry_get_text (GTK_ENTRY (data->entry259));

// Is value too large or small?
  if ((atoi (entry259_text) < 0) || (atoi (entry259_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry259), value);
  } else {
    data->tcp6hdr_tr.th_sport = htons (atoi (entry259_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);

// Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton37), FALSE);
    data->ran_tcp6_tr_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize TCP source port address
gboolean
on_checkbutton37_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton37_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_tcp6_tr_sourceport == 0) {
    data->ran_tcp6_tr_sourceport = 1;
    data->tcp6hdr_tr.th_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry259), value);

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  } else {
    data->ran_tcp6_tr_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// TCP destination port (16 bits)
gboolean
on_entry260_activate (GtkWidget *entry260, SPSData *data)
{
  const gchar *entry260_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry260_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry260_text = gtk_entry_get_text (GTK_ENTRY (data->entry260));

// Is value too large or small?
  if ((atoi (entry260_text) < 0) || (atoi (entry260_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry260), value);
  } else {
    data->tcp6hdr_tr.th_dport = htons (atoi (entry260_text)); // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP sequence number (32 bits)
gboolean
on_entry261_activate (GtkWidget *entry261, SPSData *data)
{
  const gchar *entry261_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry261_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry261_text = gtk_entry_get_text (GTK_ENTRY (data->entry261));

// Is value too large or small?
  if ((atol (entry261_text) < 0) || (strtoul (entry261_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr_tr.th_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry261), value);
  } else {
    data->tcp6hdr_tr.th_seq = htonl (atol (entry261_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP acknowledgement number (32 bits)
gboolean
on_entry262_activate (GtkWidget *entry262, SPSData *data)
{
  const gchar *entry262_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry262_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry262_text = gtk_entry_get_text (GTK_ENTRY (data->entry262));

// Is value too large or small?
  if ((atol (entry262_text) < 0) || (strtoul (entry262_text, NULL, 10) == 0xfffffffful)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr_tr.th_ack));
    gtk_entry_set_text (GTK_ENTRY (data->entry262), value);
  } else {
    data->tcp6hdr_tr.th_ack = htonl (atol (entry262_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP reserved (4 bits)
gboolean
on_entry263_activate (GtkWidget *entry263, SPSData *data)
{
  const gchar *entry263_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry263_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry263_text = gtk_entry_get_text (GTK_ENTRY (data->entry263));

// Is value too large or small?
  if ((atoi (entry263_text) < 0) || (atoi (entry263_text) > 15)) {
    sprintf (value, "%u", data->tcp6hdr_tr.th_x2);
    gtk_entry_set_text (GTK_ENTRY (data->entry263), value);
  } else {
    data->tcp6hdr_tr.th_x2 = atoi (entry263_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP data offset (4 bits)
gboolean
on_entry264_activate (GtkWidget *entry264, SPSData *data)
{
  const gchar *entry264_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry264_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry264_text = gtk_entry_get_text (GTK_ENTRY (data->entry264));

// Is value too large or small?
  if ((atoi (entry264_text) < 0) || (atoi (entry264_text) > 15)) {
    sprintf (value, "%u", data->tcp6hdr_tr.th_off);
    gtk_entry_set_text (GTK_ENTRY (data->entry264), value);
  } else {
    data->tcp6hdr_tr.th_off = atoi (entry264_text);  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP FIN flag (1 bit)
gboolean
on_entry265_activate (GtkWidget *entry265, SPSData *data)
{
  const gchar *entry265_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry265_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry265_text = gtk_entry_get_text (GTK_ENTRY (data->entry265));

// Is value too large or small?
  if ((atoi (entry265_text) < 0) || (atoi (entry265_text) > 1)) {
    sprintf (value, "%u", data->tcp6hdr_tr.th_flags & 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry265), value);
  } else {
    if (atoi (entry265_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 254;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 1;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP SYN flag (1 bit)
gboolean
on_entry266_activate (GtkWidget *entry266, SPSData *data)
{
  const gchar *entry266_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry266_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry266_text = gtk_entry_get_text (GTK_ENTRY (data->entry266));

// Is value too large or small?
  if ((atoi (entry266_text) < 0) || (atoi (entry266_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 2) >> 1);
    gtk_entry_set_text (GTK_ENTRY (data->entry266), value);
  } else {
    if (atoi (entry266_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 253;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 2;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP RST flag (1 bit)
gboolean
on_entry267_activate (GtkWidget *entry267, SPSData *data)
{
  const gchar *entry267_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry267_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry267_text = gtk_entry_get_text (GTK_ENTRY (data->entry267));

// Is value too large or small?
  if ((atoi (entry267_text) < 0) || (atoi (entry267_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 4) >> 2);
    gtk_entry_set_text (GTK_ENTRY (data->entry267), value);
  } else {
    if (atoi (entry267_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 251;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 4;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP PSH flag (1 bit)
gboolean
on_entry268_activate (GtkWidget *entry268, SPSData *data)
{
  const gchar *entry268_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry268_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry268_text = gtk_entry_get_text (GTK_ENTRY (data->entry268));

// Is value too large or small?
  if ((atoi (entry268_text) < 0) || (atoi (entry268_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 8) >> 3);
    gtk_entry_set_text (GTK_ENTRY (data->entry268), value);
  } else {
    if (atoi (entry268_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 247;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 8;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ACK flag (1 bit)
gboolean
on_entry269_activate (GtkWidget *entry269, SPSData *data)
{
  const gchar *entry269_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry269_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry269_text = gtk_entry_get_text (GTK_ENTRY (data->entry269));

// Is value too large or small?
  if ((atoi (entry269_text) < 0) || (atoi (entry269_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 16) >> 4);
    gtk_entry_set_text (GTK_ENTRY (data->entry269), value);
  } else {
    if (atoi (entry269_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 239;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 16;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP URG flag (1 bit)
gboolean
on_entry270_activate (GtkWidget *entry270, SPSData *data)
{
  const gchar *entry270_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry270_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry270_text = gtk_entry_get_text (GTK_ENTRY (data->entry270));

// Is value too large or small?
  if ((atoi (entry270_text) < 0) || (atoi (entry270_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 32) >> 5);
    gtk_entry_set_text (GTK_ENTRY (data->entry270), value);
  } else {
    if (atoi (entry270_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 223;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 32;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP ECE flag (1 bit)
gboolean
on_entry271_activate (GtkWidget *entry271, SPSData *data)
{
  const gchar *entry271_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry271_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry271_text = gtk_entry_get_text (GTK_ENTRY (data->entry271));

// Is value too large or small?
  if ((atoi (entry271_text) < 0) || (atoi (entry271_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 64) >> 6);
    gtk_entry_set_text (GTK_ENTRY (data->entry271), value);
  } else {
    if (atoi (entry271_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 191;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 64;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP CWR flag (1 bit)
gboolean
on_entry272_activate (GtkWidget *entry272, SPSData *data)
{
  const gchar *entry272_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry272_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry272_text = gtk_entry_get_text (GTK_ENTRY (data->entry272));

// Is value too large or small?
  if ((atoi (entry272_text) < 0) || (atoi (entry272_text) > 1)) {
    sprintf (value, "%u", (data->tcp6hdr_tr.th_flags & 128) >> 7);
    gtk_entry_set_text (GTK_ENTRY (data->entry272), value);
  } else {
    if (atoi (entry272_text) == 0) {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags & 127;
    } else {
      data->tcp6hdr_tr.th_flags = data->tcp6hdr_tr.th_flags | 128;
    }

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP window size (16 bits)
gboolean
on_entry273_activate (GtkWidget *entry273, SPSData *data)
{
  const gchar *entry273_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry273_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry273_text = gtk_entry_get_text (GTK_ENTRY (data->entry273));

// Is value too large or small?
  if ((atoi (entry273_text) < 0) || (atoi (entry273_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_win));
    gtk_entry_set_text (GTK_ENTRY (data->entry273), value);
  } else {
    data->tcp6hdr_tr.th_win = htons (atoi (entry273_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// TCP checksum (16 bits)
gboolean
on_entry274_activate (GtkWidget *entry274, SPSData *data)
{
  const gchar *entry274_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry274_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry274_text = gtk_entry_get_text (GTK_ENTRY (data->entry274));

// Is value too large or small?
  if ((atoi (entry274_text) < 0) || (atoi (entry274_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  } else {
    data->tcp6hdr_tr.th_sum = htons (atoi (entry274_text));  // Update TCP header with new value
  }

  g_free (value);

  return (TRUE);
}

// TCP urgent pointer (16 bits)
gboolean
on_entry275_activate (GtkWidget *entry275, SPSData *data)
{
  const gchar *entry275_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry275_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry275_text = gtk_entry_get_text (GTK_ENTRY (data->entry275));

// Is value too large or small?
  if ((atoi (entry275_text) < 0) || (atoi (entry275_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_urp));
    gtk_entry_set_text (GTK_ENTRY (data->entry275), value);
  } else {
    data->tcp6hdr_tr.th_urp = htons (atoi (entry275_text));  // Update TCP header with new value

// TCP checksum (16 bits)
    data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry274), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton35_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp6_tr = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton36_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_tcp6_tr = 1;

  return (TRUE);
}

// TCP data from keyboard
gboolean
on_entry276_activate (GtkWidget *entry276, SPSData *data)
{
  int i, j, c;
  const gchar *entry276_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry276_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry276_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry276_text = gtk_entry_get_text (GTK_ENTRY (data->entry276));

// If user asked for no payload then zero it out.
  if (strlen (entry276_text) == 0) {
    data->tcp6_tr_payloadlen = 0;
    memset (data->tcp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry276_text) != 0) && (data->ascii_hex_tcp6_tr == 0)) {

    data->tcp6_tr_payloadlen = strlen (entry276_text);
    memset (data->tcp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
    memcpy (data->tcp6_tr_payload, entry276_text, data->tcp6_tr_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry276_text) != 0) && (data->ascii_hex_tcp6_tr == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry276_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->tcp6_tr_payloadlen = 0;
      memset (data->tcp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry276_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry276_text[i+j] != 32) && (entry276_text[i+j] != 44) && ((i + j) < strlen (entry276_text))) {
          value[j] = entry276_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->tcp6_tr_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry276_activate().\n");
          data->parent = data->traceroute_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->tcp6_tr_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry276_text[i] == 32) || (entry276_text[i] == 44)) && (i < strlen (entry276_text))) {
          i++;
        }
      }  // Looping through entry276_text
    }  // else data is in acceptable format
  }  // else data is hex

// Payload length (16 bits): TCP header + TCP data
  data->ip6hdr_tcp_tr.ip6_plen = htons (TCP_HDRLEN + data->tcp6_tr_payloadlen);
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry254), value);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_tcp_tr.ip_len = htons (TCP_HDRLEN + data->tcp6_tr_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// TCP header checksum (16 bits)
  data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry274), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// TCP data from file
gboolean
on_button38_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int tcp6_lim = IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button38_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button38_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->traceroute_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button38_clicked().");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > tcp6_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", tcp6_lim);
      data->parent = data->traceroute_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->tcp6_tr_payloadlen = index;

// Clear current data.
  memset (data->tcp6_tr_payload, 0, tcp6_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry276), "");

// Copy payload from buffer.
  memcpy (data->tcp6_tr_payload, buffer, data->tcp6_tr_payloadlen);

// Payload length (16 bits): TCP header + TCP data
  data->ip6hdr_tcp_tr.ip6_plen = htons (TCP_HDRLEN + data->tcp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry254), text);

// TCP header checksum (16 bits)
  data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry274), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry276), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// ICMP page (IPv4) for traceroute

// Specify ethernet header
gboolean
on_checkbutton31_toggled (GtkButton *button, SPSData *data)
{
  if (data->icmp4_tr_specify_ether == 0) {
    data->icmp4_tr_specify_ether = 1;
  } else {
    data->icmp4_tr_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry200_activate (GtkWidget *entry200, SPSData *data)
{
  int i, bad;
  const gchar *entry200_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry200_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry200_text = gtk_entry_get_text (GTK_ENTRY (data->entry200));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry200_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry200_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry200_text[i] > 47) && (entry200_text[i] < 58)) ||    // 0 to 10
        ((entry200_text[i] > 64) && (entry200_text[i] < 71)) ||      // A to Z
        ((entry200_text[i] > 96) && (entry200_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry200_text[i] > 47) && (entry200_text[i] < 58)) ||    // 0 to 10
        ((entry200_text[i] > 64) && (entry200_text[i] < 71)) ||      // A to Z
        ((entry200_text[i] > 96) && (entry200_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp4ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry200), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry200_text, data->icmp4ethhdr_tr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry201_activate (GtkWidget *entry201, SPSData *data)
{
  int i, bad;
  const gchar *entry201_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry201_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry201_text = gtk_entry_get_text (GTK_ENTRY (data->entry201));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry201_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry201_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry201_text[i] > 47) && (entry201_text[i] < 58)) ||    // 0 to 10
        ((entry201_text[i] > 64) && (entry201_text[i] < 71)) ||      // A to Z
        ((entry201_text[i] > 96) && (entry201_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry201_text[i] > 47) && (entry201_text[i] < 58)) ||    // 0 to 10
        ((entry201_text[i] > 64) && (entry201_text[i] < 71)) ||      // A to Z
        ((entry201_text[i] > 96) && (entry201_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp4ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry201), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry201_text, data->icmp4ethhdr_tr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry202_activate (GtkWidget *entry202, SPSData *data)
{
  int i, sd, status;
  const gchar *entry202_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry202_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry202_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry202_text = gtk_entry_get_text (GTK_ENTRY (entry202));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry202_activate().\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry202_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry202_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->icmp4_tr_interfacename, entry202_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->icmp4ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->icmp4ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry201), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry203_activate (GtkWidget *entry203, SPSData *data)
{
  const gchar *entry203_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry203_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry203_text = gtk_entry_get_text (GTK_ENTRY (data->entry203));

// Is value too large for a short int?
  if ((atoi (entry203_text) < 0) || (atoi (entry203_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry203), value);
  } else {
    data->icmp4ethhdr_tr.type_code = htons (atoi (entry203_text));
  }

  g_free (value);

  return (TRUE);
}

// IP header length (4 bits)
gboolean
on_entry204_activate (GtkWidget *entry204, SPSData *data)
{
  const gchar *entry204_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry204_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry204_text = gtk_entry_get_text (GTK_ENTRY (data->entry204));

// Is value too large or small?
  if ((atoi (entry204_text) < 0) || (atoi (entry204_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry204), value);
  } else {
    data->ip4hdr_icmp_tr.ip_hl = atoi (entry204_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry205_activate (GtkWidget *entry205, SPSData *data)
{
  const gchar *entry205_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry205_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry205_text = gtk_entry_get_text (GTK_ENTRY (data->entry205));

// Is value too large or small?
  if ((atoi (entry205_text) < 0) || (atoi (entry205_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry205), value);
  } else {
    data->ip4hdr_icmp_tr.ip_v = atoi (entry205_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// Type of service (8 bits)
gboolean
on_entry206_activate (GtkWidget *entry206, SPSData *data)
{
  const gchar *entry206_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry206_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry206_text = gtk_entry_get_text (GTK_ENTRY (data->entry206));

// Is value too large or small?
  if ((atoi (entry206_text) < 0) || (atoi (entry206_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry206), value);
  } else {
    data->ip4hdr_icmp_tr.ip_tos = atoi (entry206_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// Total length of datagram (16 bits)
gboolean
on_entry207_activate (GtkWidget *entry207, SPSData *data)
{
  const gchar *entry207_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry207_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry207_text = gtk_entry_get_text (GTK_ENTRY (data->entry207));

// Is value too large or small?
  if ((atoi (entry207_text) < 0) || (atoi (entry207_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry207), value);
  } else {
    data->ip4hdr_icmp_tr.ip_len = htons (atoi (entry207_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// ID sequence number (16 bits)
gboolean
on_entry208_activate (GtkWidget *entry208, SPSData *data)
{
  const gchar *entry208_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry208_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry208_text = gtk_entry_get_text (GTK_ENTRY (data->entry208));

// Is value too large or small?
  if ((atoi (entry208_text) < 0) || (atoi (entry208_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry208), value);
  } else {
    data->ip4hdr_icmp_tr.ip_id = htons (atoi (entry208_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// Zero (1 bit)
gboolean
on_entry209_activate (GtkWidget *entry209, SPSData *data)
{
  const gchar *entry209_text, *entry210_text, *entry211_text, *entry212_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry209_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (ICMP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry209_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry209_text = gtk_entry_get_text (GTK_ENTRY (data->entry209));

// Is value too large or small?
  if ((atoi (entry209_text) < 0) || (atoi (entry209_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry209), value);
  } else {
    ip_icmp_flags[0] = atoi (entry209_text);                            // Zero (1 bit)
    entry210_text = gtk_entry_get_text (GTK_ENTRY (data->entry210));
    ip_icmp_flags[1] = atoi (entry210_text);                            // Do not fragment flag (1 bit)
    entry211_text = gtk_entry_get_text (GTK_ENTRY (data->entry211));
    ip_icmp_flags[2] = atoi (entry211_text);                            // More fragments following flag (1 bit)
    entry212_text = gtk_entry_get_text (GTK_ENTRY (data->entry212));
    ip_icmp_flags[3] = atoi (entry212_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_icmp_tr.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// Do not fragment flag (1 bit)
gboolean
on_entry210_activate (GtkWidget *entry210, SPSData *data)
{
  const gchar *entry209_text, *entry210_text, *entry211_text, *entry212_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry210_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (ICMP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry210_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry210_text = gtk_entry_get_text (GTK_ENTRY (data->entry210));

// Is value too large or small?
  if ((atoi (entry210_text) < 0) || (atoi (entry210_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_icmp_tr.ip_off) >> 14) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry210), value);
  } else {
    entry209_text = gtk_entry_get_text (GTK_ENTRY (data->entry209));
    ip_icmp_flags[0] = atoi (entry209_text);                            // Zero (1 bit)
    ip_icmp_flags[1] = atoi (entry210_text);                            // Do not fragment flag (1 bit)
    entry211_text = gtk_entry_get_text (GTK_ENTRY (data->entry211));
    ip_icmp_flags[2] = atoi (entry211_text);                            // More fragments following flag (1 bit)
    entry212_text = gtk_entry_get_text (GTK_ENTRY (data->entry212));
    ip_icmp_flags[3] = atoi (entry212_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_icmp_tr.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// More fragments following flag (1 bit)
gboolean
on_entry211_activate (GtkWidget *entry211, SPSData *data)
{
  const gchar *entry209_text, *entry210_text, *entry211_text, *entry212_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry211_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (ICMP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry211_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry211_text = gtk_entry_get_text (GTK_ENTRY (data->entry211));

// Is value too large or small?
  if ((atoi (entry211_text) < 0) || (atoi (entry211_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_icmp_tr.ip_off) >> 13) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry211), value);
  } else {
    entry209_text = gtk_entry_get_text (GTK_ENTRY (data->entry209));    // Zero (1 bit)
    ip_icmp_flags[0] = atoi (entry209_text);
    entry210_text = gtk_entry_get_text (GTK_ENTRY (data->entry210));
    ip_icmp_flags[1] = atoi (entry210_text);                            // Do not fragment flag (1 bit)
    ip_icmp_flags[2] = atoi (entry211_text);                            // More fragments following flag (1 bit)
    entry212_text = gtk_entry_get_text (GTK_ENTRY (data->entry212));
    ip_icmp_flags[3] = atoi (entry212_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_icmp_tr.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// Fragmentation offset (13 bits)
gboolean
on_entry212_activate (GtkWidget *entry212, SPSData *data)
{
  const gchar *entry209_text, *entry210_text, *entry211_text, *entry212_text;
  gushort *ip_icmp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry212_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_icmp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_icmp_flags' in on_entry212_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry212_text = gtk_entry_get_text (GTK_ENTRY (data->entry212));

// Is value too large or small?
  if ((atoi (entry212_text) < 0) || (atoi (entry212_text) > 0x1000u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_off) & 0x1fffu);
    gtk_entry_set_text (GTK_ENTRY (data->entry212), value);
  } else {
    entry209_text = gtk_entry_get_text (GTK_ENTRY (data->entry209));    // Zero (1 bit)
    ip_icmp_flags[0] = atoi (entry209_text);
    entry210_text = gtk_entry_get_text (GTK_ENTRY (data->entry210));
    ip_icmp_flags[1] = atoi (entry210_text);                            // Do not fragment flag (1 bit)
    entry211_text = gtk_entry_get_text (GTK_ENTRY (data->entry211));
    ip_icmp_flags[2] = atoi (entry211_text);                            // More fragments following flag (1 bit)
    ip_icmp_flags[3] = atoi (entry212_text);                             // Fragmentation offset (13 bits)
    data->ip4hdr_icmp_tr.ip_off = htons ((ip_icmp_flags[0] << 15) +
                       (ip_icmp_flags[1] << 14) +
                       (ip_icmp_flags[2] << 13) +
                        ip_icmp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);
  g_free (ip_icmp_flags);

  return (TRUE);
}

// Time-to-Live (8 bits)
gboolean
on_entry213_activate (GtkWidget *entry213, SPSData *data)
{
  const gchar *entry213_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry213_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry213_text = gtk_entry_get_text (GTK_ENTRY (data->entry213));

// Is value too large or small?
  if ((atoi (entry213_text) < 0) || (atoi (entry213_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry213), value);
  } else {
    data->ip4hdr_icmp_tr.ip_ttl = atoi (entry213_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// Transport layer protocol (8 bits)
gboolean
on_entry214_activate (GtkWidget *entry214, SPSData *data)
{
  const gchar *entry214_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry214_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry214_text = gtk_entry_get_text (GTK_ENTRY (data->entry214));

// Is value too large or small?
  if ((atoi (entry214_text) < 0) || (atoi (entry214_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry214), value);
  } else {
    data->ip4hdr_icmp_tr.ip_p = atoi (entry214_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (32 bits)
gboolean
on_entry215_activate (GtkWidget *entry215, SPSData *data)
{
  int status;
  const gchar *entry215_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry215_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry215_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry215_text = gtk_entry_get_text (GTK_ENTRY (data->entry215));

// Is it a valid IP address?
  if (is_valid_ip4 (entry215_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_icmp_tr.ip_src.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry215), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry215_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry215_text, &data->ip4hdr_icmp_tr.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry215_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Destination IP address (32 bits)
gboolean
on_entry216_activate (GtkWidget *entry216, SPSData *data)
{
  int status;
  const gchar *entry216_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry216_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry216_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry216_text = gtk_entry_get_text (GTK_ENTRY (data->entry216));

// Is it a valid IP address?
  if (is_valid_ip4 (entry216_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_icmp_tr.ip_dst.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry216), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry216_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry216_text, &data->ip4hdr_icmp_tr.ip_dst)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry216_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }
// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_icmp_tr.ip_sum = 0;
    data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// IP checksum (16 bits)
gboolean
on_entry217_activate (GtkWidget *entry217, SPSData *data)
{
  const gchar *entry217_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry217_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry217_text = gtk_entry_get_text (GTK_ENTRY (data->entry217));

// Is value too large or small?
  if ((atoi (entry217_text) < 0) || (atoi (entry217_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry217), value);
  } else {
    data->ip4hdr_icmp_tr.ip_sum = htons (atoi (entry217_text));  // Update IPv4 header with new value
  }

  g_free (value);

  return (TRUE);
}

// ICMP header (IPv4)

// Message Type (8 bits)
gboolean
on_entry218_activate (GtkWidget *entry218, SPSData *data)
{
  const gchar *entry218_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry218_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry218_text = gtk_entry_get_text (GTK_ENTRY (data->entry218));

// Is value too large or small?
  if ((atoi (entry218_text) < 0) || (atoi (entry218_text) > 255)) {
    sprintf (value, "%u", data->icmp4hdr_tr.icmp_type);
    gtk_entry_set_text (GTK_ENTRY (data->entry218), value);
  } else {
    data->icmp4hdr_tr.icmp_type = atoi (entry218_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry220), value);
  }

  g_free (value);

  return (TRUE);
}

// Message Code (8 bits)
gboolean
on_entry219_activate (GtkWidget *entry219, SPSData *data)
{
  const gchar *entry219_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry219_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry219_text = gtk_entry_get_text (GTK_ENTRY (data->entry219));

// Is value too large or small?
  if ((atoi (entry219_text) < 0) || (atoi (entry219_text) > 255)) {
    sprintf (value, "%u", data->icmp4hdr_tr.icmp_code);
    gtk_entry_set_text (GTK_ENTRY (data->entry219), value);
  } else {
    data->icmp4hdr_tr.icmp_code = atoi (entry219_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry220), value);
  }

  g_free (value);

  return (TRUE);
}

// ICMP checksum (16 bits)
gboolean
on_entry220_activate (GtkWidget *entry220, SPSData *data)
{
  const gchar *entry220_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry220_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry220_text = gtk_entry_get_text (GTK_ENTRY (data->entry220));

// Is value too large or small?
  if ((atoi (entry220_text) < 0) || (atoi (entry220_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry220), value);
  } else {
    data->icmp4hdr_tr.icmp_cksum = htons (atoi (entry220_text));  // Update ICMP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ID (16 bits)
gboolean
on_entry221_activate (GtkWidget *entry221, SPSData *data)
{
  const gchar *entry221_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry221_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry221_text = gtk_entry_get_text (GTK_ENTRY (data->entry221));

// Is value too large or small?
  if ((atoi (entry221_text) < 0) || (atoi (entry221_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry221), value);
  } else {
    data->icmp4hdr_tr.icmp_id = htons (atoi (entry221_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry220), value);
  }

  g_free (value);

  return (TRUE);
}

// Sequence number (16 bits)
gboolean
on_entry222_activate (GtkWidget *entry222, SPSData *data)
{
  const gchar *entry222_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry222_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry222_text = gtk_entry_get_text (GTK_ENTRY (data->entry222));

// Is value too large or small?
  if ((atoi (entry222_text) < 0) || (atoi (entry222_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry222), value);
  } else {
    data->icmp4hdr_tr.icmp_seq = htons (atoi (entry222_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry220), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton31_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp4_tr = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton32_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp4_tr = 1;

  return (TRUE);
}

// ICMP data from keyboard
gboolean
on_entry223_activate (GtkWidget *entry223, SPSData *data)
{
  int i, j, c;
  const gchar *entry223_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry223_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry223_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry223_text = gtk_entry_get_text (GTK_ENTRY (data->entry223));

// If user asked for no payload then zero it out.
  if (strlen (entry223_text) == 0) {
    data->icmp4_tr_payloadlen = 0;
    memset (data->icmp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry223_text) != 0) && (data->ascii_hex_icmp4_tr == 0)) {

    data->icmp4_tr_payloadlen = strlen (entry223_text);
    memset (data->icmp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
    memcpy (data->icmp4_tr_payload, entry223_text, data->icmp4_tr_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry223_text) != 0) && (data->ascii_hex_icmp4_tr == 1)) {
    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry223_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->icmp4_tr_payloadlen = 0;
      memset (data->icmp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry223_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry223_text[i+j] != 32) && (entry223_text[i+j] != 44) && ((i + j) < strlen (entry223_text))) {
          value[j] = entry223_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->icmp4_tr_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry223_activate().\n");
          data->parent = data->traceroute_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->icmp4_tr_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry223_text[i] == 32) || (entry223_text[i] == 44)) && (i < strlen (entry223_text))) {
          i++;
        }
      }  // Looping through entry223_text
    }  // else data is in acceptable format
  }  // else data is hex

// Total length of datagram (16 bits): IP header + ICMP header + ICMP payload
  data->ip4hdr_icmp_tr.ip_len = htons (IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry207), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp_tr.ip_sum = 0;
  data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry217), text);

// ICMP header checksum (16 bits):
  data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry220), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// ICMP data from file
gboolean
on_button32_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int icmp4_lim = IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button32_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button32_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->traceroute_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button32_clicked().");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > icmp4_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", icmp4_lim);
      data->parent = data->traceroute_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->icmp4_tr_payloadlen = index;

// Clear current data.
  memset (data->icmp4_tr_payload, 0, icmp4_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry223), "");

// Copy payload from buffer.
  memcpy (data->icmp4_tr_payload, buffer, data->icmp4_tr_payloadlen);

// Total length of datagram (16 bits): IP header + ICMP header + ICMP payload
  data->ip4hdr_icmp_tr.ip_len = htons (IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry207), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp_tr.ip_sum = 0;
  data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry217), text);

// ICMP header checksum (16 bits):
  data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry220), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry223), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// ICMP page (IPv6) for traceroute

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry277_activate (GtkWidget *entry277, SPSData *data)
{
  int i, bad;
  const gchar *entry277_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry277_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry277_text = gtk_entry_get_text (GTK_ENTRY (data->entry277));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry277_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry277_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry277_text[i] > 47) && (entry277_text[i] < 58)) ||    // 0 to 10
        ((entry277_text[i] > 64) && (entry277_text[i] < 71)) ||      // A to Z
        ((entry277_text[i] > 96) && (entry277_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry277_text[i] > 47) && (entry277_text[i] < 58)) ||    // 0 to 10
        ((entry277_text[i] > 64) && (entry277_text[i] < 71)) ||      // A to Z
        ((entry277_text[i] > 96) && (entry277_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp6ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry277), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry277_text, data->icmp6ethhdr_tr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry278_activate (GtkWidget *entry278, SPSData *data)
{
  int i, bad;
  const gchar *entry278_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry278_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry278_text = gtk_entry_get_text (GTK_ENTRY (data->entry278));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry278_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry278_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry278_text[i] > 47) && (entry278_text[i] < 58)) ||    // 0 to 10
        ((entry278_text[i] > 64) && (entry278_text[i] < 71)) ||      // A to Z
        ((entry278_text[i] > 96) && (entry278_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry278_text[i] > 47) && (entry278_text[i] < 58)) ||    // 0 to 10
        ((entry278_text[i] > 64) && (entry278_text[i] < 71)) ||      // A to Z
        ((entry278_text[i] > 96) && (entry278_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->icmp6ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry278), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry278_text, data->icmp6ethhdr_tr.src_mac);
  }

  g_free (value);

  return (TRUE);
}

// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry279_activate (GtkWidget *entry279, SPSData *data)
{
  int i, sd, status;
  const gchar *entry279_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry279_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry279_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry279_text = gtk_entry_get_text (GTK_ENTRY (entry279));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "socket() failed to get socket descriptor for using ioctl() in on_entry279_activate().\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry279_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry279_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->icmp6_tr_interfacename, entry279_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->icmp6ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->icmp6ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry278), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry280_activate (GtkWidget *entry280, SPSData *data)
{
  const gchar *entry280_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry280_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry280_text = gtk_entry_get_text (GTK_ENTRY (data->entry280));

// Is value too large for a short int?
  if ((atoi (entry280_text) < 0) || (atoi (entry280_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry280), value);
  } else {
    data->icmp6ethhdr_tr.type_code = htons (atoi (entry280_text));
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry281_activate (GtkWidget *entry281, SPSData *data)
{
  const gchar *entry281_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry281_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry281_text = gtk_entry_get_text (GTK_ENTRY (data->entry281));

// Save value currently in header
  old = ntohl (data->ip6hdr_icmp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry281_text) < 0) || (atol (entry281_text) > 15ul)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->ip6hdr_icmp_tr.ip6_flow) >> 28);
    gtk_entry_set_text (GTK_ENTRY (data->entry281), value);
  } else {
    data->ip6hdr_icmp_tr.ip6_flow = htonl (((atol (entry281_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Traffic class (8 bits)
gboolean
on_entry282_activate (GtkWidget *entry282, SPSData *data)
{
  const gchar *entry282_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry282_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry282_text = gtk_entry_get_text (GTK_ENTRY (data->entry282));

// Save value currently in header
  old = ntohl (data->ip6hdr_icmp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry282_text) < 0) || (atol (entry282_text) > 0xfful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp_tr.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry282), value);
  } else {
    data->ip6hdr_icmp_tr.ip6_flow = htonl (((atol (entry282_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Flow label (20 bits)
gboolean
on_entry283_activate (GtkWidget *entry283, SPSData *data)
{
  const gchar *entry283_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry283_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry283_text = gtk_entry_get_text (GTK_ENTRY (data->entry283));

// Save value currently in header
  old = ntohl (data->ip6hdr_icmp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry283_text) < 0) || (atol (entry283_text) > 0xfffffu)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp_tr.ip6_flow) & 0xfffffu));
    gtk_entry_set_text (GTK_ENTRY (data->entry283), value);
  } else {
    data->ip6hdr_icmp_tr.ip6_flow = htonl (atol (entry283_text) + (old & 0xfff00000ul));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Payload length (16 bits)
gboolean
on_entry284_activate (GtkWidget *entry284, SPSData *data)
{
  const gchar *entry284_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry284_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry284_text = gtk_entry_get_text (GTK_ENTRY (data->entry284));

// Is value too large or small?
  if ((atoi (entry284_text) < 0) || (atoi (entry284_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp_tr.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry284), value);
  } else {
    data->ip6hdr_icmp_tr.ip6_plen = htons (atoi (entry284_text));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Next header (8 bits)
gboolean
on_entry285_activate (GtkWidget *entry285, SPSData *data)
{
  const gchar *entry285_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry285_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry285_text = gtk_entry_get_text (GTK_ENTRY (data->entry285));

// Is value too large or small?
  if ((atoi (entry285_text) < 0) || (atoi (entry285_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_icmp_tr.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry285), value);
  } else {
    data->ip6hdr_icmp_tr.ip6_nxt = atoi (entry285_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Hop limit (8 bits)
gboolean
on_entry286_activate (GtkWidget *entry286, SPSData *data)
{
  const gchar *entry286_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry286_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry286_text = gtk_entry_get_text (GTK_ENTRY (data->entry286));

// Is value too large or small?
  if ((atoi (entry286_text) < 0) || (atoi (entry286_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_icmp_tr.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry286), value);
  } else {
    data->ip6hdr_icmp_tr.ip6_hops = atoi (entry286_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (128 bits)
gboolean
on_entry287_activate (GtkWidget *entry287, SPSData *data)
{
  const gchar *entry287_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry287_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry287_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry287_text = gtk_entry_get_text (GTK_ENTRY (data->entry287));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry287_text, &(data->ip6hdr_icmp_tr.ip6_src)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp_tr.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry287), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry287_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// ICMP header checksum (16 bits)
    data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Destination IP address (128 bits)
gboolean
on_entry288_activate (GtkWidget *entry288, SPSData *data)
{
  const gchar *entry288_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry288_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry288_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry288_text = gtk_entry_get_text (GTK_ENTRY (data->entry288));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry288_text, &(data->ip6hdr_icmp_tr.ip6_dst)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp_tr.ip6_dst), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry288), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry288_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
 } else { 
// ICMP header checksum (16 bits)
    data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// ICMP header (IPv6)

// Message Type (8 bits)
gboolean
on_entry289_activate (GtkWidget *entry289, SPSData *data)
{
  const gchar *entry289_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry289_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry289_text = gtk_entry_get_text (GTK_ENTRY (data->entry289));

// Is value too large or small?
  if ((atoi (entry289_text) < 0) || (atoi (entry289_text) > 255)) {
    sprintf (value, "%u", data->icmp6hdr_tr.icmp6_type);
    gtk_entry_set_text (GTK_ENTRY (data->entry289), value);
  } else {
    data->icmp6hdr_tr.icmp6_type = atoi (entry289_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);
  }

  g_free (value);

  return (TRUE);
}

// Message Code (8 bits)
gboolean
on_entry290_activate (GtkWidget *entry290, SPSData *data)
{
  const gchar *entry290_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry290_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry290_text = gtk_entry_get_text (GTK_ENTRY (data->entry290));

// Is value too large or small?
  if ((atoi (entry290_text) < 0) || (atoi (entry290_text) > 255)) {
    sprintf (value, "%u", data->icmp6hdr_tr.icmp6_code);
    gtk_entry_set_text (GTK_ENTRY (data->entry290), value);
  } else {
    data->icmp6hdr_tr.icmp6_code = atoi (entry290_text);  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);
  }

  g_free (value);

  return (TRUE);
}

// ICMP checksum (16 bits)
gboolean
on_entry291_activate (GtkWidget *entry291, SPSData *data)
{
  const gchar *entry291_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry291_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry291_text = gtk_entry_get_text (GTK_ENTRY (data->entry291));

// Is value too large or small?
  if ((atoi (entry291_text) < 0) || (atoi (entry291_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);
  } else {
    data->icmp6hdr_tr.icmp6_cksum = htons (atoi (entry291_text));  // Update ICMP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ID (16 bits)
gboolean
on_entry292_activate (GtkWidget *entry292, SPSData *data)
{
  const gchar *entry292_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry292_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry292_text = gtk_entry_get_text (GTK_ENTRY (data->entry292));

// Is value too large or small?
  if ((atoi (entry292_text) < 0) || (atoi (entry292_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry292), value);
  } else {
    data->icmp6hdr_tr.icmp6_id = htons (atoi (entry292_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);
  }

  g_free (value);

  return (TRUE);
}

// Sequence number (16 bits)
gboolean
on_entry293_activate (GtkWidget *entry293, SPSData *data)
{
  const gchar *entry293_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry293_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry293_text = gtk_entry_get_text (GTK_ENTRY (data->entry293));

// Is value too large or small?
  if ((atoi (entry293_text) < 0) || (atoi (entry293_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_seq));
    gtk_entry_set_text (GTK_ENTRY (data->entry293), value);
  } else {
    data->icmp6hdr_tr.icmp6_seq = htons (atoi (entry293_text));  // Update ICMP header with new value

// ICMP header checksum (16 bits)
    data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
    gtk_entry_set_text (GTK_ENTRY (data->entry291), value);
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton37_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp6_tr = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton38_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_icmp6_tr = 1;

  return (TRUE);
}

// ICMP data from keyboard
gboolean
on_entry294_activate (GtkWidget *entry294, SPSData *data)
{
  int i, j, c;
  const gchar *entry294_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry294_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry294_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry294_text = gtk_entry_get_text (GTK_ENTRY (data->entry294));

// If user asked for no payload then zero it out.
  if (strlen (entry294_text) == 0) {
    data->icmp4_tr_payloadlen = 0;
    memset (data->icmp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry294_text) != 0) && (data->ascii_hex_icmp6_tr == 0)) {

    data->icmp6_tr_payloadlen = strlen (entry294_text);
    memset (data->icmp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
    memcpy (data->icmp6_tr_payload, entry294_text, data->icmp6_tr_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry294_text) != 0) && (data->ascii_hex_icmp6_tr == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry294_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->icmp6_tr_payloadlen = 0;
      memset (data->icmp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry294_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry294_text[i+j] != 32) && (entry294_text[i+j] != 44) && ((i + j) < strlen (entry294_text))) {
          value[j] = entry294_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->icmp6_tr_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry294_activate().\n");
          data->parent = data->traceroute_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->icmp6_tr_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry294_text[i] == 32) || (entry294_text[i] == 44)) && (i < strlen (entry294_text))) {
          i++;
        }
      }  // Looping through entry294_text
    }  // else data is in acceptable format
  }  // else data is hex

// Update IPv6 header payload length = ICMP header + ICMP data
  data->ip6hdr_icmp_tr.ip6_plen = htons (ICMP_HDRLEN + data->icmp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry284), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_icmp_tr.ip_len = htons (ICMP_HDRLEN + data->icmp6_tr_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// ICMP header checksum (16 bits)
  data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry291), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// ICMP data from file
gboolean
on_button41_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int icmp6_lim = IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button41_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button41_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->traceroute_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button41_clicked().");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > icmp6_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", icmp6_lim);
      data->parent = data->traceroute_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->icmp6_tr_payloadlen = index;

// Clear current data.
  memset (data->icmp6_tr_payload, 0, icmp6_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry294), "");

// Copy payload from buffer.
  memcpy (data->icmp6_tr_payload, buffer, data->icmp6_tr_payloadlen);

// Update IPv6 header payload length = ICMP header + ICMP data
  data->ip6hdr_icmp_tr.ip6_plen = htons (ICMP_HDRLEN + data->icmp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry284), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_icmp_tr.ip_len = htons (ICMP_HDRLEN + data->icmp6_tr_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// ICMP header checksum (16 bits)
  data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry291), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry294), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// UDP page (IPv4) for traceroute

// Specify ethernet header
gboolean
on_checkbutton33_toggled (GtkButton *button, SPSData *data)
{
  if (data->udp4_tr_specify_ether == 0) {
    data->udp4_tr_specify_ether = 1;
  } else {
    data->udp4_tr_specify_ether = 0;
  }

  return (TRUE);
}

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry224_activate (GtkWidget *entry224, SPSData *data)
{
  int i, bad;
  const gchar *entry224_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry224_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry224_text = gtk_entry_get_text (GTK_ENTRY (data->entry224));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry224_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry224_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry224_text[i] > 47) && (entry224_text[i] < 58)) ||    // 0 to 10
        ((entry224_text[i] > 64) && (entry224_text[i] < 71)) ||      // A to Z
        ((entry224_text[i] > 96) && (entry224_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry224_text[i] > 47) && (entry224_text[i] < 58)) ||    // 0 to 10
        ((entry224_text[i] > 64) && (entry224_text[i] < 71)) ||      // A to Z
        ((entry224_text[i] > 96) && (entry224_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp4ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry224), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry224_text, data->udp4ethhdr_tr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry225_activate (GtkWidget *entry225, SPSData *data)
{
  int i, bad;
  const gchar *entry225_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry225_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry225_text = gtk_entry_get_text (GTK_ENTRY (data->entry225));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry225_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry225_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry225_text[i] > 47) && (entry225_text[i] < 58)) ||    // 0 to 10
        ((entry225_text[i] > 64) && (entry225_text[i] < 71)) ||      // A to Z
        ((entry225_text[i] > 96) && (entry225_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry225_text[i] > 47) && (entry225_text[i] < 58)) ||    // 0 to 10
        ((entry225_text[i] > 64) && (entry225_text[i] < 71)) ||      // A to Z
        ((entry225_text[i] > 96) && (entry225_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp4ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry225), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry225_text, data->udp4ethhdr_tr.src_mac);
  }

  g_free (value);

  return (TRUE);
}


// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry226_activate (GtkWidget *entry226, SPSData *data)
{
  int i, sd, status;
  const gchar *entry226_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry226_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry226_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry226_text = gtk_entry_get_text (GTK_ENTRY (entry226));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text, "Iin callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry226_activate().\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry226_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry226_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->udp4_tr_interfacename, entry226_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->udp4ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->udp4ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry225), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry227_activate (GtkWidget *entry227, SPSData *data)
{
  const gchar *entry227_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry227_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry227_text = gtk_entry_get_text (GTK_ENTRY (data->entry227));

// Is value too large for a short int?
  if ((atoi (entry227_text) < 0) || (atoi (entry227_text) > 65535)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry227), value);
  } else {
    data->udp4ethhdr_tr.type_code = htons (atoi (entry227_text));
  }

  g_free (value);

  return (TRUE);
}

// IP header length (4 bits)
gboolean
on_entry228_activate (GtkWidget *entry228, SPSData *data)
{
  const gchar *entry228_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry228_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry228_text = gtk_entry_get_text (GTK_ENTRY (data->entry228));

// Is value too large or small?
  if ((atoi (entry228_text) < 0) || (atoi (entry228_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_udp_tr.ip_hl);
    gtk_entry_set_text (GTK_ENTRY (data->entry228), value);
  } else {
    data->ip4hdr_udp_tr.ip_hl = atoi (entry228_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// IPv4 version (4 bits)
gboolean
on_entry229_activate (GtkWidget *entry229, SPSData *data)
{
  const gchar *entry229_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry229_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry229_text = gtk_entry_get_text (GTK_ENTRY (data->entry229));

// Is value too large or small?
  if ((atoi (entry229_text) < 0) || (atoi (entry229_text) > 15)) {
    sprintf (value, "%u", data->ip4hdr_udp_tr.ip_v);
    gtk_entry_set_text (GTK_ENTRY (data->entry229), value);
  } else {
    data->ip4hdr_udp_tr.ip_v = atoi (entry229_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// Type of service (8 bits)
gboolean
on_entry230_activate (GtkWidget *entry230, SPSData *data)
{
  const gchar *entry230_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry230_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry230_text = gtk_entry_get_text (GTK_ENTRY (data->entry230));

// Is value too large or small?
  if ((atoi (entry230_text) < 0) || (atoi (entry230_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_udp_tr.ip_tos);
    gtk_entry_set_text (GTK_ENTRY (data->entry230), value);
  } else {
    data->ip4hdr_udp_tr.ip_tos = atoi (entry230_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// Total length of datagram (16 bits)
gboolean
on_entry231_activate (GtkWidget *entry231, SPSData *data)
{
  const gchar *entry231_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry231_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry231_text = gtk_entry_get_text (GTK_ENTRY (data->entry231));

// Is value too large or small?
  if ((atoi (entry231_text) < 0) || (atoi (entry231_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_len));
    gtk_entry_set_text (GTK_ENTRY (data->entry231), value);
  } else {
    data->ip4hdr_udp_tr.ip_len = htons (atoi (entry231_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// ID sequence number (16 bits)
gboolean
on_entry232_activate (GtkWidget *entry232, SPSData *data)
{
  const gchar *entry232_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry232_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry232_text = gtk_entry_get_text (GTK_ENTRY (data->entry232));

// Is value too large or small?
  if ((atoi (entry232_text) < 0) || (atoi (entry232_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_id));
    gtk_entry_set_text (GTK_ENTRY (data->entry232), value);
  } else {
    data->ip4hdr_udp_tr.ip_id = htons (atoi (entry232_text));  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// Zero (1 bit)
gboolean
on_entry233_activate (GtkWidget *entry233, SPSData *data)
{
  const gchar *entry233_text, *entry234_text, *entry235_text, *entry236_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry233_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (UDP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry233_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry233_text = gtk_entry_get_text (GTK_ENTRY (data->entry233));

// Is value too large or small?
  if ((atoi (entry233_text) < 0) || (atoi (entry233_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_off) >> 15);
    gtk_entry_set_text (GTK_ENTRY (data->entry233), value);
  } else {
    ip_udp_flags[0] = atoi (entry233_text);                            // Zero (1 bit)
    entry234_text = gtk_entry_get_text (GTK_ENTRY (data->entry234));
    ip_udp_flags[1] = atoi (entry234_text);                            // Do not fragment flag (1 bit)
    entry235_text = gtk_entry_get_text (GTK_ENTRY (data->entry235));
    ip_udp_flags[2] = atoi (entry235_text);                            // More fragments following flag (1 bit)
    entry236_text = gtk_entry_get_text (GTK_ENTRY (data->entry236));
    ip_udp_flags[3] = atoi (entry236_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_udp_tr.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// Do not fragment flag (1 bit)
gboolean
on_entry234_activate (GtkWidget *entry234, SPSData *data)
{
  const gchar *entry233_text, *entry234_text, *entry235_text, *entry236_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry234_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (UDP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry234_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry234_text = gtk_entry_get_text (GTK_ENTRY (data->entry234));

// Is value too large or small?
  if ((atoi (entry234_text) < 0) || (atoi (entry234_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_udp_tr.ip_off) >> 14) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry234), value);
  } else {
    entry233_text = gtk_entry_get_text (GTK_ENTRY (data->entry233));
    ip_udp_flags[0] = atoi (entry233_text);                            // Zero (1 bit)
    ip_udp_flags[1] = atoi (entry234_text);                            // Do not fragment flag (1 bit)
    entry235_text = gtk_entry_get_text (GTK_ENTRY (data->entry235));
    ip_udp_flags[2] = atoi (entry235_text);                            // More fragments following flag (1 bit)
    entry236_text = gtk_entry_get_text (GTK_ENTRY (data->entry236));
    ip_udp_flags[3] = atoi (entry236_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_udp_tr.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// More fragments following flag (1 bit)
gboolean
on_entry235_activate (GtkWidget *entry235, SPSData *data)
{
  const gchar *entry233_text, *entry234_text, *entry235_text, *entry236_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry235_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (UDP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry235_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry235_text = gtk_entry_get_text (GTK_ENTRY (data->entry235));

// Is value too large or small?
  if ((atoi (entry235_text) < 0) || (atoi (entry235_text) > 1u)) {
    sprintf (value, "%hu", (unsigned short int) ((ntohs (data->ip4hdr_udp_tr.ip_off) >> 13) & 1u));
    gtk_entry_set_text (GTK_ENTRY (data->entry235), value);
  } else {
    entry233_text = gtk_entry_get_text (GTK_ENTRY (data->entry233));
    ip_udp_flags[0] = atoi (entry233_text);                            // Zero (1 bit)
    entry234_text = gtk_entry_get_text (GTK_ENTRY (data->entry234));
    ip_udp_flags[1] = atoi (entry234_text);                            // Do not fragment flag (1 bit)
    ip_udp_flags[2] = atoi (entry235_text);                            // More fragments following flag (1 bit)
    entry236_text = gtk_entry_get_text (GTK_ENTRY (data->entry236));
    ip_udp_flags[3] = atoi (entry236_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_udp_tr.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// Fragmentation offset (13 bits)
gboolean
on_entry236_activate (GtkWidget *entry236, SPSData *data)
{
  const gchar *entry233_text, *entry234_text, *entry235_text, *entry236_text;
  gushort *ip_udp_flags;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry236_activate().\n");
    exit (EXIT_FAILURE);
  }

// IP header (TCP) fragmentation flags and offset
  tmp = (gushort *) g_malloc0 (4 * sizeof (gushort));
  if (tmp != NULL) {
    ip_udp_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_udp_flags' in on_entry236_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry236_text = gtk_entry_get_text (GTK_ENTRY (data->entry236));

// Is value too large or small?
  if ((atoi (entry236_text) < 0) || (atoi (entry236_text) > 0x1000u)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_off) & 0x1fffu);
    gtk_entry_set_text (GTK_ENTRY (data->entry236), value);
  } else {
    entry233_text = gtk_entry_get_text (GTK_ENTRY (data->entry233));
    ip_udp_flags[0] = atoi (entry233_text);                            // Zero (1 bit)
    entry234_text = gtk_entry_get_text (GTK_ENTRY (data->entry234));
    ip_udp_flags[1] = atoi (entry234_text);                            // Do not fragment flag (1 bit)
    entry235_text = gtk_entry_get_text (GTK_ENTRY (data->entry235));
    ip_udp_flags[2] = atoi (entry235_text);                            // More fragments following flag (1 bit)
    ip_udp_flags[3] = atoi (entry236_text);                            // Fragmentation offset (13 bits)
    data->ip4hdr_udp_tr.ip_off = htons ((ip_udp_flags[0] << 15) +
                       (ip_udp_flags[1] << 14) +
                       (ip_udp_flags[2] << 13) +
                        ip_udp_flags[3]);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);
  g_free (ip_udp_flags);

  return (TRUE);
}

// Time-to-Live (8 bits)
gboolean
on_entry237_activate (GtkWidget *entry237, SPSData *data)
{
  const gchar *entry237_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry237_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry237_text = gtk_entry_get_text (GTK_ENTRY (data->entry237));

// Is value too large or small?
  if ((atoi (entry237_text) < 0) || (atoi (entry237_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_udp_tr.ip_ttl);
    gtk_entry_set_text (GTK_ENTRY (data->entry237), value);
  } else {
    data->ip4hdr_udp_tr.ip_ttl = atoi (entry237_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// Transport layer protocol (8 bits)
gboolean
on_entry238_activate (GtkWidget *entry238, SPSData *data)
{
  const gchar *entry238_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry238_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry238_text = gtk_entry_get_text (GTK_ENTRY (data->entry238));

// Is value too large or small?
  if ((atoi (entry238_text) < 0) || (atoi (entry238_text) > 255)) {
    sprintf (value, "%u", data->ip4hdr_udp_tr.ip_p);
    gtk_entry_set_text (GTK_ENTRY (data->entry238), value);
  } else {
    data->ip4hdr_udp_tr.ip_p = atoi (entry238_text);  // Update IPv4 header with new value

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (32 bits)
gboolean
on_entry239_activate (GtkWidget *entry239, SPSData *data)
{
  int status;
  const gchar *entry239_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry239_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry239_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry239_text = gtk_entry_get_text (GTK_ENTRY (data->entry239));

// Is it a valid IP address?
  if (is_valid_ip4 (entry239_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_udp_tr.ip_src.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry239), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry239_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry239_text, &data->ip4hdr_udp_tr.ip_src)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry239_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Destination IP address (32 bits)
gboolean
on_entry240_activate (GtkWidget *entry240, SPSData *data)
{
  int status;
  const gchar *entry240_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry240_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry240_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry240_text = gtk_entry_get_text (GTK_ENTRY (data->entry240));

// Is it a valid IP address?
  if (is_valid_ip4 (entry240_text) != TRUE) {
// Invalid IP
    inet_ntop (AF_INET, &(data->ip4hdr_udp_tr.ip_dst.s_addr), value, 16);
    gtk_entry_set_text (GTK_ENTRY (data->entry240), value);
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry240_activate().\nAppears to be an invalid IPv4 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
    if ((status = inet_pton (AF_INET, entry240_text, &data->ip4hdr_udp_tr.ip_dst)) != 1) {
      sprintf (text, "In callbacks.c: inet_pton() failed in on_entry240_activate().\nError message: %s", strerror (status));
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);
    }

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
    data->ip4hdr_udp_tr.ip_sum = 0;
    data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// IP checksum (16 bits)
gboolean
on_entry241_activate (GtkWidget *entry241, SPSData *data)
{
  const gchar *entry241_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry241_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry241_text = gtk_entry_get_text (GTK_ENTRY (data->entry241));

// Is value too large or small?
  if ((atoi (entry241_text) < 0) || (atoi (entry241_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry241), value);
  } else {
    data->ip4hdr_udp_tr.ip_sum = htons (atoi (entry241_text));  // Update IPv4 header with new value

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// UDP header (IPv4) for traceroute

// UDP source port (16 bits)
gboolean
on_entry242_activate (GtkWidget *entry242, SPSData *data)
{
  const gchar *entry242_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry242_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry242_text = gtk_entry_get_text (GTK_ENTRY (data->entry242));

// Is value too large or small?
  if ((atoi (entry242_text) < 0) || (atoi (entry242_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry242), value);
  } else {
    data->udp4hdr_tr.uh_sport = htons (atoi (entry242_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton35), FALSE);
    data->ran_udp4_tr_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize UDP source port address
gboolean
on_checkbutton35_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton35_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_udp4_tr_sourceport == 0) {
    data->ran_udp4_tr_sourceport = 1;
    data->udp4hdr_tr.uh_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry242), value);

// UDP checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  } else {
    data->ran_udp4_tr_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// UDP destination port (16 bits)
gboolean
on_entry243_activate (GtkWidget *entry243, SPSData *data)
{
  const gchar *entry243_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry243_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry243_text = gtk_entry_get_text (GTK_ENTRY (data->entry243));

// Is value too large or small?
  if ((atoi (entry243_text) < 0) || (atoi (entry243_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry243), value);
  } else {
    data->udp4hdr_tr.uh_dport = htons (atoi (entry243_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// Length of UDP datagram (16 bits)
gboolean
on_entry244_activate (GtkWidget *entry244, SPSData *data)
{
  const gchar *entry244_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry244_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry244_text = gtk_entry_get_text (GTK_ENTRY (data->entry244));

// Is value too large or small?
  if ((atoi (entry244_text) < 0) || (atoi (entry244_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_ulen));
    gtk_entry_set_text (GTK_ENTRY (data->entry244), value);
  } else {
    data->udp4hdr_tr.uh_ulen = htons (atoi (entry244_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  }

  g_free (value);

  return (TRUE);
}

// UDP checksum (16 bits)
gboolean
on_entry245_activate (GtkWidget *entry245, SPSData *data)
{
  const gchar *entry245_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry245_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry245_text = gtk_entry_get_text (GTK_ENTRY (data->entry245));

// Is value too large or small?
  if ((atoi (entry245_text) < 0) || (atoi (entry245_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry245), value);
  } else {
    data->udp4hdr_tr.uh_sum = htons (atoi (entry245_text));  // Update UDP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton33_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp4_tr = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton34_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp4_tr = 1;

  return (TRUE);
}

// UDP data from keyboard
gboolean
on_entry246_activate (GtkWidget *entry246, SPSData *data)
{
  int i, j, c;
  const gchar *entry246_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry246_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry246_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry246_text = gtk_entry_get_text (GTK_ENTRY (data->entry246));

// If user asked for no payload then zero it out.
  if (strlen (entry246_text) == 0) {
    data->udp4_tr_payloadlen = 0;
    memset (data->udp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry246_text) != 0) && (data->ascii_hex_udp4_tr == 0)) {

    data->udp4_tr_payloadlen = strlen (entry246_text);
    memset (data->udp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
    memcpy (data->udp4_tr_payload, entry246_text, data->udp4_tr_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry246_text) != 0) && (data->ascii_hex_udp4_tr == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry246_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->udp4_tr_payloadlen = 0;
      memset (data->udp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry246_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry246_text[i+j] != 32) && (entry246_text[i+j] != 44) && ((i + j) < strlen (entry246_text))) {
          value[j] = entry246_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->udp4_tr_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry246_activate().\n");
          data->parent = data->traceroute_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->udp4_tr_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry246_text[i] == 32) || (entry246_text[i] == 44)) && (i < strlen (entry246_text))) {
          i++;
        }
      }  // Looping through entry246_text
    }  // else data is in acceptable format
  }  // else data is hex

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp4hdr_tr.uh_ulen = htons (UDP_HDRLEN + data->udp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry244), text);

// Total length of datagram (16 bits): IP header + UDP header + UDP payload
  data->ip4hdr_udp_tr.ip_len = htons (IP4_HDRLEN + UDP_HDRLEN + data->udp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry231), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp_tr.ip_sum = 0;
  data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry241), text);

// UDP header checksum (16 bits)
  data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry245), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// UDP data from file
gboolean
on_button35_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int udp4_lim = IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button35_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button35_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->traceroute_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button35_clicked().");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > udp4_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", udp4_lim);
      data->parent = data->traceroute_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->udp4_tr_payloadlen = index;

// Clear current data.
  memset (data->udp4_tr_payload, 0, udp4_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry246), "");

// Copy payload from buffer.
  memcpy (data->udp4_tr_payload, buffer, data->udp4_tr_payloadlen);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp4hdr_tr.uh_ulen = htons (UDP_HDRLEN + data->udp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry244), text);

// Total length of datagram (16 bits): IP header + UDP header + UDP payload
  data->ip4hdr_udp_tr.ip_len = htons (IP4_HDRLEN + UDP_HDRLEN + data->udp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry231), text);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp_tr.ip_sum = 0;
  data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry241), text);

// UDP header checksum (16 bits)
  data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry245), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry246), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// UDP page (IPv6) for traceroute

// Destination link-layer (MAC) address (48 bits)
gboolean
on_entry295_activate (GtkWidget *entry295, SPSData *data)
{
  int i, bad;
  const gchar *entry295_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry295_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry295_text = gtk_entry_get_text (GTK_ENTRY (data->entry295));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry295_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry295_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry295_text[i] > 47) && (entry295_text[i] < 58)) ||    // 0 to 10
        ((entry295_text[i] > 64) && (entry295_text[i] < 71)) ||      // A to Z
        ((entry295_text[i] > 96) && (entry295_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry295_text[i] > 47) && (entry295_text[i] < 58)) ||    // 0 to 10
        ((entry295_text[i] > 64) && (entry295_text[i] < 71)) ||      // A to Z
        ((entry295_text[i] > 96) && (entry295_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp6ethhdr_tr.dst_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry295), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry295_text, data->udp6ethhdr_tr.dst_mac);
  }

  g_free (value);

  return (TRUE);
}

// Source link-layer (MAC) address (48 bits)
gboolean
on_entry296_activate (GtkWidget *entry296, SPSData *data)
{
  int i, bad;
  const gchar *entry296_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry296_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry296_text = gtk_entry_get_text (GTK_ENTRY (data->entry296));

// Is new value in appropriate form?

  // Check length
  bad = 0;
  if (strlen (entry296_text) != 17) {
    bad = 1;
  }

  // Check for colons
  i = 2;
  while (i<17) {
    if (entry296_text[i] != ':') {
      bad = 1;
    }
    i += 3;
  }

  // Check for valid hexadecimals
  i = 0;
  while (i<17) {
    if (!(((entry296_text[i] > 47) && (entry296_text[i] < 58)) ||    // 0 to 10
        ((entry296_text[i] > 64) && (entry296_text[i] < 71)) ||      // A to Z
        ((entry296_text[i] > 96) && (entry296_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i++;
    if (!(((entry296_text[i] > 47) && (entry296_text[i] < 58)) ||    // 0 to 10
        ((entry296_text[i] > 64) && (entry296_text[i] < 71)) ||      // A to Z
        ((entry296_text[i] > 96) && (entry296_text[i] < 103)))) {    // a to z
      bad = 1;
    }
    i += 2;
  }
  if (bad == 1) {
    mac_bin2string (data->udp6ethhdr_tr.src_mac, value);
    gtk_entry_set_text (GTK_ENTRY (data->entry296), value);
  } else {
    // Seems ok, so keep it.
    mac_string2bin ((char *) entry296_text, data->udp6ethhdr_tr.src_mac);
  }

  g_free (value);

  return (TRUE);
}


// This host's interface name to be converted to link-layer (MAC) address.
gboolean
on_entry297_activate (GtkWidget *entry297, SPSData *data)
{
  int i, sd, status;
  const gchar *entry297_text;
  gchar *text;
  unsigned char *mac;
  struct ifreq ifr;
  void *tmp;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry297_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (6 * sizeof (unsigned char));
  if (tmp != NULL) {
    mac = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'mac' in on_entry297_activate().\n");
    exit (EXIT_FAILURE);
  }

  entry297_text = gtk_entry_get_text (GTK_ENTRY (entry297));

// Submit request for a socket descriptor to lookup interface.
  if ((sd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    status = errno;
    sprintf (text,  "In callbacks.c: socket() failed to get socket descriptor for ioctl() in on_entry297_activate.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }

// Use ioctl() to lookup interface and get MAC address.
  memset (&ifr, 0, sizeof (ifr));
  snprintf (ifr.ifr_name, sizeof (ifr.ifr_name), "%s", entry297_text);
  if (ioctl (sd, SIOCGIFHWADDR, &ifr) < 0) {
    status = errno;
    sprintf (text, "In callbacks.c: ioctl() failed to get interface hardware address (MAC) in on_entry297_activate().\nAppears to be an invalid interface name.\nError message: %s", strerror (status));
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (mac);
    return (FALSE);
  }
  close (sd);

// Save interface name
  strcpy (data->udp6_tr_interfacename, entry297_text);

// Copy source MAC address.
  memcpy (mac, ifr.ifr_hwaddr.sa_data, 6);

// Put MAC address into a string.
  memset (text, 0, 1024);
  for (i=0; i<5; i++) {
    sprintf (text, "%s%02x:", text, (unsigned char) mac[i]);
  }
  sprintf (text, "%s%02x", text, (unsigned char) mac[5]);

// Update source MAC address
  mac_string2bin ((char *) text, data->udp6ethhdr_tr.src_mac);
  memset (text, 0, 1024);
  mac_bin2string (data->udp6ethhdr_tr.src_mac, text);
  gtk_entry_set_text (GTK_ENTRY (data->entry296), text);

// Free allocated memory.
  g_free (text);
  g_free (mac);

  return (TRUE);
}

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
gboolean
on_entry298_activate (GtkWidget *entry298, SPSData *data)
{
  const gchar *entry298_text;
  gchar *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry298_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry298_text = gtk_entry_get_text (GTK_ENTRY (data->entry298));

// Is value too large for a short int?
  if ((atoi (entry298_text) < 0) || (atoi (entry298_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6ethhdr_tr.type_code));
    gtk_entry_set_text (GTK_ENTRY (data->entry298), value);
  } else {
    data->udp6ethhdr_tr.type_code = htons (atoi (entry298_text));
  }

  g_free (value);

  return (TRUE);
}

// IP version (4 bits)
gboolean
on_entry299_activate (GtkWidget *entry299, SPSData *data)
{
  const gchar *entry299_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry299_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry299_text = gtk_entry_get_text (GTK_ENTRY (data->entry299));

// Save value currently in header
  old = ntohl (data->ip6hdr_udp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry299_text) < 0) || (atol (entry299_text) > 15ul)) {
    sprintf (value, "%lu", (unsigned long int) ntohl (data->ip6hdr_udp_tr.ip6_flow) >> 28);
    gtk_entry_set_text (GTK_ENTRY (data->entry299), value);
  } else {
    data->ip6hdr_udp_tr.ip6_flow = htonl (((atol (entry299_text)) << 28) + (old & 0xffffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Traffic class (8 bits)
gboolean
on_entry300_activate (GtkWidget *entry300, SPSData *data)
{
  const gchar *entry300_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry300_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry300_text = gtk_entry_get_text (GTK_ENTRY (data->entry300));

// Save value currently in header
  old = ntohl (data->ip6hdr_udp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry300_text) < 0) || (atol (entry300_text) > 0xfful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp_tr.ip6_flow) >> 20) & 0xfful);
    gtk_entry_set_text (GTK_ENTRY (data->entry300), value);
  } else {
    data->ip6hdr_udp_tr.ip6_flow = htonl (((atol (entry300_text)) << 20) + (old & 0xf00ffffful));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Flow label (20 bits)
gboolean
on_entry301_activate (GtkWidget *entry301, SPSData *data)
{
  const gchar *entry301_text;
  unsigned long int old;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry301_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry301_text = gtk_entry_get_text (GTK_ENTRY (data->entry301));

// Save value currently in header
  old = ntohl (data->ip6hdr_udp_tr.ip6_flow);

// Is value too large or small?
  if ((atol (entry301_text) < 0) || (atol (entry301_text) > 0xffffful)) {
    sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp_tr.ip6_flow) & 0xffffful));
    gtk_entry_set_text (GTK_ENTRY (data->entry301), value);
  } else {
    data->ip6hdr_udp_tr.ip6_flow = htonl (atol (entry301_text) + (old & 0xfff00000ul));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Payload length (16 bits)
gboolean
on_entry302_activate (GtkWidget *entry302, SPSData *data)
{
  const gchar *entry302_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry302_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry302_text = gtk_entry_get_text (GTK_ENTRY (data->entry302));

// Is value too large or small?
  if ((atoi (entry302_text) < 0) || (atoi (entry302_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp_tr.ip6_plen));
    gtk_entry_set_text (GTK_ENTRY (data->entry302), value);
  } else {
    data->ip6hdr_udp_tr.ip6_plen = htons (atoi (entry302_text));  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Next header (8 bits)
gboolean
on_entry303_activate (GtkWidget *entry303, SPSData *data)
{
  const gchar *entry303_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry303_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry303_text = gtk_entry_get_text (GTK_ENTRY (data->entry303));

// Is value too large or small?
  if ((atoi (entry303_text) < 0) || (atoi (entry303_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_udp_tr.ip6_nxt);
    gtk_entry_set_text (GTK_ENTRY (data->entry303), value);
  } else {
    data->ip6hdr_udp_tr.ip6_nxt = atoi (entry303_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Hop limit (8 bits)
gboolean
on_entry304_activate (GtkWidget *entry304, SPSData *data)
{
  const gchar *entry304_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry304_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry304_text = gtk_entry_get_text (GTK_ENTRY (data->entry304));

// Is value too large or small?
  if ((atoi (entry304_text) < 0) || (atoi (entry304_text) > 255)) {
    sprintf (value, "%u", data->ip6hdr_udp_tr.ip6_hops);
    gtk_entry_set_text (GTK_ENTRY (data->entry304), value);
  } else {
    data->ip6hdr_udp_tr.ip6_hops = atoi (entry304_text);  // Update IPv6 header with new value
  }

  g_free (value);

  return (TRUE);
}

// Source IP address (128 bits)
gboolean
on_entry305_activate (GtkWidget *entry305, SPSData *data)
{
  const gchar *entry305_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry305_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry305_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry305_text = gtk_entry_get_text (GTK_ENTRY (data->entry305));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry305_text, &(data->ip6hdr_udp_tr.ip6_src)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_udp_tr.ip6_src), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry305), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry305_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// UDP header checksum (16 bits)
    data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// Destination IP address (128 bits)
gboolean
on_entry306_activate (GtkWidget *entry306, SPSData *data)
{
  const gchar *entry306_text;
  gchar *value, *text;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry306_activate().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry306_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry306_text = gtk_entry_get_text (GTK_ENTRY (data->entry306));

// Is it a valid IP address? Valid if inet_pton returns 1.
  if (inet_pton (AF_INET6, entry306_text, &(data->ip6hdr_udp_tr.ip6_dst)) < 1) {
// Invalid IP
    inet_ntop (AF_INET6, &(data->ip6hdr_udp_tr.ip6_dst), value, 40);
    gtk_entry_set_text (GTK_ENTRY (data->entry306), value);
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry306_activate().\nAppears to be an invalid IPv6 address.");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (value);
    return (FALSE);

// Valid IP
  } else {
// UDP header checksum (16 bits)
    data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);

    g_free (text);
    g_free (value);

    return (TRUE);
  }
}

// UDP header (IPv6) for traceroute

// UDP source port (16 bits)
gboolean
on_entry307_activate (GtkWidget *entry307, SPSData *data)
{
  const gchar *entry307_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry307_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry307_text = gtk_entry_get_text (GTK_ENTRY (data->entry307));

// Is value too large or small?
  if ((atoi (entry307_text) < 0) || (atoi (entry307_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry307), value);
  } else {
    data->udp6hdr_tr.uh_sport = htons (atoi (entry307_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);

//  Source port has been entered manually. Uncheck "Randomize" checkbutton.
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton40), FALSE);
    data->ran_udp6_tr_sourceport = 0;
  }

  g_free (value);

  return (TRUE);
}

// Randomize UDP source port address
gboolean
on_checkbutton40_toggled (GtkButton *button, SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_checkbutton40_toggled().\n");
    exit (EXIT_FAILURE);
  }
  if (data->ran_udp6_tr_sourceport == 0) {
    data->ran_udp6_tr_sourceport = 1;
    data->udp6hdr_tr.uh_sport = htons (ran_val2 ());
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sport));
    gtk_entry_set_text (GTK_ENTRY (data->entry307), value);

// UDP header checksum (16 bits)
    data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);
  } else {
    data->ran_udp6_tr_sourceport = 0;
  }

  g_free (value);
  return (TRUE);
}

// UDP destination port (16 bits)
gboolean
on_entry308_activate (GtkWidget *entry308, SPSData *data)
{
  const gchar *entry308_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry308_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry308_text = gtk_entry_get_text (GTK_ENTRY (data->entry308));

// Is value too large or small?
  if ((atoi (entry308_text) < 0) || (atoi (entry308_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_dport));
    gtk_entry_set_text (GTK_ENTRY (data->entry308), value);
  } else {
    data->udp6hdr_tr.uh_dport = htons (atoi (entry308_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);
  }

  g_free (value);

  return (TRUE);
}

// Length of UDP datagram (16 bits)
gboolean
on_entry309_activate (GtkWidget *entry309, SPSData *data)
{
  const gchar *entry309_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry309_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry309_text = gtk_entry_get_text (GTK_ENTRY (data->entry309));

// Is value too large or small?
  if ((atoi (entry309_text) < 0) || (atoi (entry309_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_ulen));
    gtk_entry_set_text (GTK_ENTRY (data->entry309), value);
  } else {
    data->udp6hdr_tr.uh_ulen = htons (atoi (entry309_text));  // Update UDP header with new value

// UDP header checksum (16 bits)
    data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);
  }

  g_free (value);

  return (TRUE);
}

// UDP checksum (16 bits)
gboolean
on_entry310_activate (GtkWidget *entry310, SPSData *data)
{
  const gchar *entry310_text;
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (16 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry310_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry310_text = gtk_entry_get_text (GTK_ENTRY (data->entry310));

// Is value too large or small?
  if ((atoi (entry310_text) < 0) || (atoi (entry310_text) > 0xffffu)) {
    sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
    gtk_entry_set_text (GTK_ENTRY (data->entry310), value);
  } else {
    data->udp6hdr_tr.uh_sum = htons (atoi (entry310_text));  // Update UDP header with new value
  }

  g_free (value);

  return (TRUE);
}

// ASCII data entry
gboolean
on_radiobutton39_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp6_tr = 0;

  return (TRUE);
}

// Hexadecimal data entry
gboolean
on_radiobutton40_toggled (GtkButton *button, SPSData *data)
{
  data->ascii_hex_udp6_tr = 1;

  return (TRUE);
}

// UDP data from keyboard
gboolean
on_entry311_activate (GtkWidget *entry311, SPSData *data)
{
  int i, j, c;
  const gchar *entry311_text;
  gchar *text, *value;
  gchar **endptr;
  void *tmp;

// Array to hold error messages and miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry311_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (50 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in on_entry311_activate().\n");
    exit (EXIT_FAILURE);
  }

// Get new value from text entry
  entry311_text = gtk_entry_get_text (GTK_ENTRY (data->entry311));

// If user asked for no payload then zero it out.
  if (strlen (entry311_text) == 0) {
    data->udp6_tr_payloadlen = 0;
    memset (data->udp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
  
// ASCII data entry radiobutton was selected.
  } else if ((strlen (entry311_text) != 0) && (data->ascii_hex_udp6_tr == 0)) {

    data->udp6_tr_payloadlen = strlen (entry311_text);
    memset (data->udp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
    memcpy (data->udp6_tr_payload, entry311_text, data->udp6_tr_payloadlen);

// Hexadecimal data entry radiobutton was selected.
  } else if ((strlen (entry311_text) != 0) && (data->ascii_hex_udp6_tr == 1)) {

    // Check for invalid hexadecimal data.
    if (!is_valid_hexdata (entry311_text, data)) {
      sprintf (text, "There seems to be invalid hexadecimal data.\nHere's an example of acceptable data: 23, 4e,a,cd    ,7");
      data->parent = data->traceroute_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (value);
      return (FALSE);

    // Data is in acceptable format.
    } else {
      data->udp6_tr_payloadlen = 0;
      memset (data->udp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
      // Store values as payload data.
      c = 0;  // index for payload array
      i = 0;  // index for entry array
      while (i < strlen (entry311_text)) {

        // Grab next hexadecimal value.
        memset (value, 0, 10);
        j = 0;
        while ((entry311_text[i+j] != 32) && (entry311_text[i+j] != 44) && ((i + j) < strlen (entry311_text))) {
          value[j] = entry311_text[i+j];
          j++;
        }
        i += j;
        // Convert to base-10 chars and store as payload data.
        endptr = NULL;
        data->udp6_tr_payload[c] = (unsigned char) strtol (value, endptr, 16);
        if (endptr != NULL) {
          sprintf (text, "In callbacks.c, strtol() failed in on_entry311_activate().\n");
          data->parent = data->traceroute_window;
          data->message = text;
          report_error (data);
          g_free (text);
          g_free (value);
          return (FALSE);
        }
        c++;
        data->udp6_tr_payloadlen++;

        // Increment i until we find more hex or hit end of string.
        while (((entry311_text[i] == 32) || (entry311_text[i] == 44)) && (i < strlen (entry311_text))) {
          i++;
        }
      }  // Looping through entry311_text
    }  // else data is in acceptable format
  }  // else data is hex

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp6hdr_tr.uh_ulen = htons (UDP_HDRLEN + data->udp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry309), text);

// IPv6 header payload length (16 bits): UDP header + UDP payload
  data->ip6hdr_udp_tr.ip6_plen = htons (UDP_HDRLEN + data->udp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry302), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_udp_tr.ip_len = htons (UDP_HDRLEN + data->udp6_tr_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// UDP header checksum (16 bits)
  data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry310), text);

// Free allocated memory.
  g_free (text);
  g_free (value);

  return (TRUE);
}

// UDP data from file
gboolean
on_button44_clicked (GtkButton *button, SPSData *data)
{
  int i, index;
  gint result;
  const int udp6_lim = IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN;  // Maximum allowed amount of data
  gchar *text, *filename;
  unsigned char *buffer;
  GtkWidget *dialog;
  void *tmp;
  FILE *fi;

// Allocate memory for various arrays.

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button44_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    buffer = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'buffer' in on_button44_clicked().\n");
    exit (EXIT_FAILURE);
  }

// Create filechooser dialog.
  dialog = gtk_file_chooser_dialog_new ("Load File ...", (GtkWindow *) data->traceroute_window, GTK_FILE_CHOOSER_ACTION_OPEN,
             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

// Run the filechooser dialog.
  result = gtk_dialog_run (GTK_DIALOG (dialog));

// Interpret result returned from filechooser dialog.
  if (result == GTK_RESPONSE_ACCEPT)
  {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    gtk_widget_destroy (dialog);
  } else {
    g_free (text);
    g_free (buffer);
    gtk_widget_destroy (dialog);
    return (TRUE);
  }

// Open file.
  if ((fi = fopen (filename, "rb")) == NULL) {
    sprintf (text, "In callbacks.c: fopen() failed in on_button10_clicked().");
    data->parent = data->traceroute_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (buffer);
    return (FALSE);
  }

// Read file into buffer. Truncate if necessary.
  index = 0;
  while ((i = fgetc (fi)) != EOF) {
    buffer[index] = (unsigned char) i;
    index++;
    if (index > udp6_lim) {
      sprintf (text, "This file has more data than allowed (%u bytes) and will be truncated.", udp6_lim);
      data->parent = data->traceroute_window;
      data->message = text;
      report_warning (data);
      index--;
      break;
    }
  }

// Close file descriptor.
  fclose (fi);

// New payload length based upon file contents.
  data->udp6_tr_payloadlen = index;

// Clear current data.
  memset (data->udp6_tr_payload, 0, udp6_lim);

// Clear text entry since we used a file and not the keyboard.
  gtk_entry_set_text (GTK_ENTRY (data->entry311), "");

// Copy payload from buffer.
  memcpy (data->udp6_tr_payload, buffer, data->udp6_tr_payloadlen);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp6hdr_tr.uh_ulen = htons (UDP_HDRLEN + data->udp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry309), text);

// IPv6 header payload length (16 bits): UDP header + UDP payload
  data->ip6hdr_udp_tr.ip6_plen = htons (UDP_HDRLEN + data->udp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry302), text);

// Update sixto4 IPv4 header total length of datagram value
  data->sixto4ip4hdr_udp_tr.ip_len = htons (UDP_HDRLEN + data->udp6_tr_payloadlen + IP4_HDRLEN + IP6_HDRLEN);

// UDP header checksum (16 bits)
  data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);
  sprintf (text, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry310), text);

// Make user aware a data file is loaded.
  memset (text, 0, 1024);
  strcpy (text, "Data file loaded.");
  gtk_entry_set_text (GTK_ENTRY (data->entry311), text);

// Free allocated memory.
  g_free (text);
  g_free (buffer);

  return (TRUE);
}

// Call tcp4_tr_default()
gboolean
on_button30_clicked (GtkButton *button, SPSData *data)
{
  tcp4_tr_default (data);
  tcp4_tr_show (data);

  return (TRUE);
}

// Call tcp6_tr_default()
gboolean
on_button39_clicked (GtkButton *button, SPSData *data)
{
  tcp6_tr_default (data);
  tcp6_tr_show (data);

  return (TRUE);
}

// Call icmp4_tr_default()
gboolean
on_button33_clicked (GtkButton *button, SPSData *data)
{
  icmp4_tr_default (data);
  icmp4_tr_show (data);

  return (TRUE);
}

// Call icmp6_tr_default()
gboolean
on_button42_clicked (GtkButton *button, SPSData *data)
{
  icmp6_tr_default (data);
  icmp6_tr_show (data);

  return (TRUE);
}

// Call udp4_tr_default()
gboolean
on_button36_clicked (GtkButton *button, SPSData *data)
{
  udp4_tr_default (data);
  udp4_tr_show (data);

  return (TRUE);
}

// Call udp6_tr_default()
gboolean
on_button45_clicked (GtkButton *button, SPSData *data)
{
  udp6_tr_default (data);
  udp6_tr_show (data);

  return (TRUE);
}

// Apply IPv4 TCP traceroute default settings
gboolean
tcp4_tr_default (SPSData *data)
{
  unsigned char *ip4_flags, *tcp4_flags;
  gchar *ipaddress;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in tcp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// TCP flags
  tmp = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp4_flags' in tcp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in tcp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Default to have the user specify the ethernet header
  data->tcp4_tr_specify_ether = 1;

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp4ethhdr_tr.dst_mac, 0, 6);

// Source Interface Name: defaults to blank
  memset (data->tcp4_tr_interfacename, 0, 255);

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp4ethhdr_tr.src_mac, 0, 6);

// Ethernet type code (16 bits): default to IPv4
// http://www.iana.org/assignments/ethernet-numbers
  data->tcp4ethhdr_tr.type_code = htons (ETH_P_IP);

// TCP data

// Clear TCP data buffer.
  memset (data->tcp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - TCP_HDRLEN);
  data->tcp4_tr_payloadlen = 0;

// TCP data entry format: default to hexadecimal input
  data->ascii_hex_tcp4_tr = 1;

// IPv4 header

// IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr_tcp_tr.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);

// Internet Protocol version (4 bits): IPv4
  data->ip4hdr_tcp_tr.ip_v = 4u;

// Type of service (8 bits)
  data->ip4hdr_tcp_tr.ip_tos = 0;

// Total length of datagram (16 bits): IP header + TCP header + TCP data
  data->ip4hdr_tcp_tr.ip_len = htons (IP4_HDRLEN + TCP_HDRLEN + data->tcp4_tr_payloadlen);

// ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr_tcp_tr.ip_id = htons (0);

// Zero (1 bit)
  ip4_flags[0] = 0;

// Do not fragment flag (1 bit)
  ip4_flags[1] = 0;

// More fragments following flag (1 bit)
  ip4_flags[2] = 0;

// Fragmentation offset (13 bits)
  ip4_flags[3] = 0;

// Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr_tcp_tr.ip_off = htons ((ip4_flags[0] << 15) +
   (ip4_flags[1] << 14) + (ip4_flags[2] << 13) + ip4_flags[3]);

// Time-to-Live (8 bits): start at 1
  data->ip4hdr_tcp_tr.ip_ttl = 1u;

// Transport layer protocol (8 bits): 6 for TCP
  data->ip4hdr_tcp_tr.ip_p = IPPROTO_TCP;

// Source IPv4 address (32 bits): default to this host's IPv4 address
  inet_pton (AF_INET, thishost_ipv4 (ipaddress), &(data->ip4hdr_tcp_tr.ip_src));

// Destination IPv4 address (32 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "127.0.0.1");
  inet_pton (AF_INET, ipaddress, &data->ip4hdr_tcp_tr.ip_dst);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp_tr.ip_sum = 0;
  data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);

// TCP header (IPv4)

// Source port number (16 bits): default to not random
  data->ran_tcp4_tr_sourceport = 0;
  data->tcp4hdr_tr.th_sport = htons (80u);

// Destination port number (16 bits)
  data->tcp4hdr_tr.th_dport = htons (80u);

// Sequence number (32 bits)
  data->tcp4hdr_tr.th_seq = htonl (0);

// Acknowledgement number (32 bits): 0 in first packet
  data->tcp4hdr_tr.th_ack = htonl (0);

// Reserved (4 bits): should be 0
  data->tcp4hdr_tr.th_x2 = 0;

// Data offset (4 bits): size of TCP header in 32-bit words
  data->tcp4hdr_tr.th_off = TCP_HDRLEN / 4;

// FIN flag (1 bit)
  tcp4_flags[0] = 0;

// SYN flag (1 bit): set to 1
  tcp4_flags[1] = 1;

// RST flag (1 bit)
  tcp4_flags[2] = 0;

// PSH flag (1 bit)
  tcp4_flags[3] = 0;

// ACK flag (1 bit)
  tcp4_flags[4] = 0;

// URG flag (1 bit)
  tcp4_flags[5] = 0;

// ECE flag (1 bit)
  tcp4_flags[6] = 0;

// CWR flag (1 bit)
  tcp4_flags[7] = 0;

// Flags (8 bits): 2 for SYN
  data->tcp4hdr_tr.th_flags = bin8_to_dec (tcp4_flags);

// Window size (16 bits): default to maximum value
  data->tcp4hdr_tr.th_win = htons (0xffffu);

// Urgent pointer (16 bits): 0 (only valid if URG flag is set)
  data->tcp4hdr_tr.th_urp = htons (0);

// TCP checksum (16 bits)
  data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (tcp4_flags);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv6 TCP traceroute page default settings
gboolean
tcp6_tr_default (SPSData *data)
{
  unsigned char *tcp6_flags;
  gchar *ipaddress;
  void *tmp;

// Allocate memory for various arrays

// TCP flags
  tmp = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp6_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp6_flags' in tcp6_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in tcp6_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp6ethhdr_tr.dst_mac, 0, 6);

// Source Interface Name: defaults to blank
  memset (data->tcp6_tr_interfacename, 0, 255);

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->tcp6ethhdr_tr.src_mac, 0, 6);

// Ethernet type code (16 bits): default to IPv6
// http://www.iana.org/assignments/ethernet-numbers
  data->tcp6ethhdr_tr.type_code = htons (ETH_P_IPV6);

// TCP data

// Clear TCP data buffer.
  memset (data->tcp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - TCP_HDRLEN);
  data->tcp6_tr_payloadlen = 0;

// TCP data entry format: default to hexadecimal input
  data->ascii_hex_tcp6_tr = 1;

// IPv6 header

// IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr_tcp_tr.ip6_flow = htonl ((6 << 28) | (0 << 20) | 0);

// Payload length (16 bits): TCP header + TCP data
  data->ip6hdr_tcp_tr.ip6_plen = htons (TCP_HDRLEN + data->tcp6_tr_payloadlen);

// Next header (8 bits): 6 for TCP
  data->ip6hdr_tcp_tr.ip6_nxt = IPPROTO_TCP;

// Hop limit (8 bits): start with hop limit of 1 (next nodei in route)
  data->ip6hdr_tcp_tr.ip6_hops = 1u;

// Source IPv6 address (128 bits): default to this host's IPv6 address
  inet_pton (AF_INET6, thishost_ipv6 (ipaddress), &data->ip6hdr_tcp_tr.ip6_src);

// Destination IPv6 address (128 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "::1");
  inet_pton (AF_INET6, ipaddress, &data->ip6hdr_tcp_tr.ip6_dst);

// TCP header (IPv6)

// Source port number (16 bits): 80
  data->ran_tcp6_tr_sourceport = 0;
  data->tcp6hdr_tr.th_sport = htons (80u);

// Destination port number (16 bits)
  data->tcp6hdr_tr.th_dport = htons (80u);

// Sequence number (32 bits)
  data->tcp6hdr_tr.th_seq = htonl (0);

// Acknowledgement number (32 bits): 0 in first packet
  data->tcp6hdr_tr.th_ack = htonl (0);

// Reserved (4 bits): should be 0
  data->tcp6hdr_tr.th_x2 = 0;

// Data offset (4 bits): size of TCP header in 32-bit words
  data->tcp6hdr_tr.th_off = TCP_HDRLEN / 4;

// FIN flag (1 bit)
  tcp6_flags[0] = 0;

// SYN flag (1 bit): set to 1
  tcp6_flags[1] = 1;

// RST flag (1 bit)
  tcp6_flags[2] = 0;

// PSH flag (1 bit)
  tcp6_flags[3] = 0;

// ACK flag (1 bit)
  tcp6_flags[4] = 0;

// URG flag (1 bit)
  tcp6_flags[5] = 0;

// ECE flag (1 bit)
  tcp6_flags[6] = 0;

// CWR flag (1 bit)
  tcp6_flags[7] = 0;

// Flags (8 bits): 2 for SYN
  data->tcp6hdr_tr.th_flags = bin8_to_dec (tcp6_flags);

// Window size (16 bits): default to maximum value
  data->tcp6hdr_tr.th_win = htons (0xffffu);

// Urgent pointer (16 bits): 0 (only valid if URG flag is set)
  data->tcp6hdr_tr.th_urp = htons (0);

// TCP checksum (16 bits)
  data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);

// Free allocated memory.
  g_free (tcp6_flags);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv4 ICMP traceroute page default settings
gboolean
icmp4_tr_default (SPSData *data)
{
  unsigned char *ip4_flags;
  gchar *ipaddress;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in icmp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in icmp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Default to have the user specify the ethernet header
  data->icmp4_tr_specify_ether = 1;  // The toggling done above will trigger the callback, but we set it here to be sure.

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp4ethhdr_tr.dst_mac, 0, 6);

// Source Interface Name: defaults to blank
  memset (data->icmp4_tr_interfacename, 0, 255);

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp4ethhdr_tr.src_mac, 0, 6);

// Ethernet type code (16 bits): default to IPv4
// http://www.iana.org/assignments/ethernet-numbers
  data->icmp4ethhdr_tr.type_code = htons (ETH_P_IP);

// ICMP data

// Clear ICMP data buffer.
  memset (data->icmp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - ICMP_HDRLEN);
  data->icmp4_tr_payloadlen = 0;

// ICMP data entry format: default to hexadecimal input
  data->ascii_hex_icmp4_tr = 1;

// IPv4 header

// IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr_icmp_tr.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);

// Internet Protocol version (4 bits): IPv4
  data->ip4hdr_icmp_tr.ip_v = 4u;

// Type of service (8 bits): default is 0
  data->ip4hdr_icmp_tr.ip_tos = 0;

// Total length of datagram (16 bits): IP header + ICMP header + ICMP data
  data->ip4hdr_icmp_tr.ip_len = htons (IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_tr_payloadlen);

// ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr_icmp_tr.ip_id = htons (0);

// Zero (1 bit)
  ip4_flags[0] = 0;

// Do not fragment flag (1 bit)
  ip4_flags[1] = 0;

// More fragments following flag (1 bit)
  ip4_flags[2] = 0;

// Fragmentation offset (13 bits)
  ip4_flags[3] = 0;

// Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr_icmp_tr.ip_off = htons ((ip4_flags[0] << 15) +
   (ip4_flags[1] << 14) + (ip4_flags[2] << 13) + ip4_flags[3]);

// Time-to-Live (8 bits): start at 1
  data->ip4hdr_icmp_tr.ip_ttl = 1u;

// Transport layer protocol (8 bits): 1 for ICMP
  data->ip4hdr_icmp_tr.ip_p = IPPROTO_ICMP;

// Source IPv4 address (32 bits): default to this host's IPv4 address
  inet_pton (AF_INET, thishost_ipv4 (ipaddress), &data->ip4hdr_icmp_tr.ip_src);

// Destination IPv4 address (32 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "127.0.0.1");
  inet_pton (AF_INET, ipaddress, &data->ip4hdr_icmp_tr.ip_dst);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp_tr.ip_sum = 0;
  data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);

// ICMP header (IPv4)

// Message Type (8 bits): echo request
  data->icmp4hdr_tr.icmp_type = ICMP_ECHO;

// Message Code (8 bits): echo request
  data->icmp4hdr_tr.icmp_code = 0;

// Identifier (16 bits): usually pid of sending process - we'll arbitrarily use 1000
  data->icmp4hdr_tr.icmp_id = htons (1000u);

// Sequence Number (16 bits): starts at 0
  data->icmp4hdr_tr.icmp_seq = htons (0);

// ICMP header checksum (16 bits)
  data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv6 ICMP traceroute page default settings
gboolean
icmp6_tr_default (SPSData *data)
{
  gchar *ipaddress;
  void *tmp;

// Allocate memory for various arrays

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in icmp6_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp6ethhdr_tr.dst_mac, 0, 6);

// Source Interface Name: defaults to blank
  memset (data->icmp6_tr_interfacename, 0, 255);

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->icmp6ethhdr_tr.src_mac, 0, 6);
  
// Ethernet type code (16 bits): default to IPv6
// http://www.iana.org/assignments/ethernet-numbers
  data->icmp6ethhdr_tr.type_code = htons (ETH_P_IPV6);

// ICMP data

// Clear ICMP data buffer.
  memset (data->icmp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - ICMP_HDRLEN);
  data->icmp6_tr_payloadlen = 0;

// ICMP data entry format: default to hexadecimal input
  data->ascii_hex_icmp6_tr = 1;

// IPv6 header

// IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr_icmp_tr.ip6_flow = htonl ((6 << 28) | (0 << 20) | 0);

// Payload length (16 bits): ICMP header + ICMP data
  data->ip6hdr_icmp_tr.ip6_plen = htons (ICMP_HDRLEN + data->icmp6_tr_payloadlen);

// Next header (8 bits): 58 for ICMP
  data->ip6hdr_icmp_tr.ip6_nxt = IPPROTO_ICMPV6;

// Hop limit (8 bits): start with 1 (next node in route)
  data->ip6hdr_icmp_tr.ip6_hops = 1u;

// Source IPv6 address (128 bits): default to this host's IPv6 address
  inet_pton (AF_INET6, thishost_ipv6 (ipaddress), &data->ip6hdr_icmp_tr.ip6_src);

// Destination IPv6 address (128 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "::1");
  inet_pton (AF_INET6, ipaddress, &data->ip6hdr_icmp_tr.ip6_dst);

// ICMP header (IPv6)

// Message Type (8 bits): 128 is echo request
  data->icmp6hdr_tr.icmp6_type = ICMP6_ECHO_REQUEST;

// Message Code (8 bits): echo request
  data->icmp6hdr_tr.icmp6_code = 0;

// Identifier (16 bits): usually pid of sending process - we'll arbitrarily use 1000
  data->icmp6hdr_tr.icmp6_id = htons (1000u);

// Sequence Number (16 bits): starts at 0
  data->icmp6hdr_tr.icmp6_seq = htons (0);

// ICMP header checksum (16 bits)
  data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);

// Free allocated memory.
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv4 UDP traceroute page default settings
gboolean
udp4_tr_default (SPSData *data)
{
  unsigned char *ip4_flags;
  gchar *ipaddress;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in udp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in udp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Default to have the user specify the ethernet header
  data->udp4_tr_specify_ether = 1;  // The toggling done above will trigger the callback, but we set it here to be sure.

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp4ethhdr_tr.dst_mac, 0, 6);

// Source Interface Name: defaults to blank
  memset (data->udp4_tr_interfacename, 0, 255);

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp4ethhdr_tr.src_mac, 0, 6);

// Ethernet type code (16 bits): default to IPv4
// http://www.iana.org/assignments/ethernet-numbers
  data->udp4ethhdr_tr.type_code = htons (ETH_P_IP);

// UDP data

// Clear UDP data buffer.
  memset (data->udp4_tr_payload, 0, IP_MAXPACKET - IP4_HDRLEN - UDP_HDRLEN);
  data->udp4_tr_payloadlen = 0;

// UDP data entry format: default to hexadecimal input
  data->ascii_hex_udp4_tr = 1;

// IPv4 header

// IPv4 header length (4 bits): Number of 32-bit words in header = 5
  data->ip4hdr_udp_tr.ip_hl = IP4_HDRLEN / sizeof (unsigned long int);

// Internet Protocol version (4 bits): IPv4
  data->ip4hdr_udp_tr.ip_v = 4u;

// Type of service (8 bits): default is 0
  data->ip4hdr_udp_tr.ip_tos = 0;

// Total length of datagram (16 bits): IP header + UDP header + UDP payload
  data->ip4hdr_udp_tr.ip_len = htons (IP4_HDRLEN + UDP_HDRLEN + data->udp4_tr_payloadlen);

// ID sequence number (16 bits): unused, since single datagram
  data->ip4hdr_udp_tr.ip_id = htons (0);

// Zero (1 bit)
  ip4_flags[0] = 0;

// Do not fragment flag (1 bit)
  ip4_flags[1] = 0;

// More fragments following flag (1 bit)
  ip4_flags[2] = 0;

// Fragmentation offset (13 bits)
  ip4_flags[3] = 0;

// Flags, and Fragmentation offset (3, 13 bits): 0 since single datagram
  data->ip4hdr_udp_tr.ip_off = htons ((ip4_flags[0] << 15) +
   (ip4_flags[1] << 14) + (ip4_flags[2] << 13) + ip4_flags[3]);

// Time-to-Live (8 bits): start at 1
  data->ip4hdr_udp_tr.ip_ttl = 1u;

// Transport layer protocol (8 bits): 17 for UDP
  data->ip4hdr_udp_tr.ip_p = IPPROTO_UDP;

// Source IPv4 address (32 bits): default to this host's IPv4 address
  inet_pton (AF_INET, thishost_ipv4 (ipaddress), &data->ip4hdr_udp_tr.ip_src);

// Destination IPv4 address (32 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "127.0.0.1");
  inet_pton (AF_INET, ipaddress, &data->ip4hdr_udp_tr.ip_dst);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp_tr.ip_sum = 0;
  data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);

// UDP header (IPv4)

// Source port number (16 bits): we'll arbitrarily use 4950
  data->ran_udp4_tr_sourceport = 0;
  data->udp4hdr_tr.uh_sport = htons (4950u);

// Destination port number (16 bits): 33435
  data->udp4hdr_tr.uh_dport = htons (33435u);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp4hdr_tr.uh_ulen = htons (UDP_HDRLEN + data->udp4_tr_payloadlen);

// UDP checksum (16 bits)
  data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (ipaddress);

  return (TRUE);
}

// Apply IPv6 UDP traceroute page default settings
gboolean
udp6_tr_default (SPSData *data)
{
  gchar *ipaddress;
  void *tmp;

// Allocate memory for various arrays

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in udp6_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp6ethhdr_tr.dst_mac, 0, 6);

// Source Interface Name: defaults to blank
  memset (data->udp6_tr_interfacename, 0, 255);

// Source link-layer (MAC) address (48 bits): default to zero; user needs to fill out
  memset (data->udp6ethhdr_tr.src_mac, 0, 6);
  
// Ethernet type code (16 bits): default to IPv6
// http://www.iana.org/assignments/ethernet-numbers
  data->udp6ethhdr_tr.type_code = htons (ETH_P_IPV6);

// UDP data

// Clear UDP data buffer.
  memset (data->udp6_tr_payload, 0, IP_MAXPACKET - IP6_HDRLEN - UDP_HDRLEN);
  data->udp6_tr_payloadlen = 0;

// UDP data entry format: default to hexadecimal input
  data->ascii_hex_udp6_tr = 1;

// IPv6 header

// IPv6 version (4 bits), Traffic class (8 bits), Flow label (20 bits)
  data->ip6hdr_udp_tr.ip6_flow = htonl ((6 << 28) | (0 << 20) | 0);

// Payload length (16 bits): UDP header + UDP payload
  data->ip6hdr_udp_tr.ip6_plen = htons (UDP_HDRLEN + data->udp6_tr_payloadlen);

// Next header (8 bits): 17 for UDP
  data->ip6hdr_udp_tr.ip6_nxt = IPPROTO_UDP;

// Hop limit (8 bits): start with 1 (next node in route)
  data->ip6hdr_udp_tr.ip6_hops = 1u;

// Source IPv6 address (128 bits): default to this host's IPv6 address
  inet_pton (AF_INET6, thishost_ipv6 (ipaddress), &data->ip6hdr_udp_tr.ip6_src);

// Destination IPv6 address (128 bits): default to loopback IP address to be safe
  strcpy (ipaddress, "::1");
  inet_pton (AF_INET6, ipaddress, &data->ip6hdr_udp_tr.ip6_dst);

// UDP header (IPv6)

// Source port number (16 bits): we'll arbitrarily use 4950
  data->ran_udp6_tr_sourceport = 0;
  data->udp6hdr_tr.uh_sport = htons (4950u);

// Destination port number (16 bits): 33435
  data->udp6hdr_tr.uh_dport = htons (33435u);

// Length of UDP datagram (16 bits): UDP header + UDP payload
  data->udp6hdr_tr.uh_ulen = htons (UDP_HDRLEN + data->udp6_tr_payloadlen);

// UDP header checksum (16 bits)
  data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);

// Free allocated memory.
  g_free (ipaddress);

  return (TRUE);
}

// Traceroute page (sps.ui)

// TCP (IPv4) packet type for traceroute (sps.ui)
gboolean
on_radiobutton21_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type_tr = 1;

  return (TRUE);
}

// ICMP (IPv4) packet type for traceroute (sps.ui)
gboolean
on_radiobutton22_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type_tr = 2;

  return (TRUE);
}

// UDP (IPv4) packet type for traceroute (sps.ui)
gboolean
on_radiobutton23_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type_tr = 3;

  return (TRUE);
}

// TCP (IPv6) packet type for traceroute (sps.ui)
gboolean
on_radiobutton26_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type_tr = 4;

  return (TRUE);
}

// ICMP (IPv6) packet type for traceroute (sps.ui)
gboolean
on_radiobutton27_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type_tr = 5;

  return (TRUE);
}

// UDP (IPv6) packet type for traceroute (sps.ui)
gboolean
on_radiobutton28_toggled (GtkButton *button, SPSData *data)
{
  data->packet_type_tr = 6;

  return (TRUE);
}

// Tunnel IPv6 over IPv4 for traceroute (sps.ui).
gboolean
on_checkbutton29_toggled (GtkButton *button, SPSData *data)
{
  if (data->sixto4_tr_flag == 0) {
    data->sixto4_tr_flag = 1;
  } else {
    data->sixto4_tr_flag = 0;
  }

  return (TRUE);
}

// Number of Probes per Hop for Traceroute
gboolean
on_spinbutton2_value_changed (GtkWidget *spinbutton2, SPSData *data)
{
  data->num_probes = (gint) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton2));

  return (TRUE);
}

// Timeout for Reply for Traceroute
gboolean
on_spinbutton4_value_changed (GtkWidget *spinbutton4, SPSData *data)
{

  data->timeout_tr = (gint) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton4));

  return (TRUE);
}

// Maximum Number of Hops for Traceroute
gboolean
on_spinbutton3_value_changed (GtkWidget *spinbutton3, SPSData *data)
{

  data->maxhops = (gint) gtk_spin_button_get_value (GTK_SPIN_BUTTON (spinbutton3));

  return (TRUE);
}

// Resolve hostnames for nodes during Traceroute
gboolean
on_checkbutton34_toggled (GtkButton *button, SPSData *data)
{
  if (data->resolve_tr == 1) {
    data->resolve_tr = 0;
  } else {
    data->resolve_tr = 1;
  }

  return (TRUE);
}

// Launch Traceroute Packet Editing UI
gboolean
on_button23_clicked (GtkButton *button, SPSData *data)
{
  GtkBuilder *builder;
  GError *error = NULL;

  // If a traceroute is already running, return with no action.
  if (tracing == 1) {
    return (FALSE);
  }

  if (data->traceroute_flag == 0) {
    data->traceroute_flag = 1;
    // Create new GtkBuilder object
    builder = gtk_builder_new();
    if (!gtk_builder_add_from_file (builder, "traceroute.ui", &error)) {
      g_warning ("%s", error->message );
      g_free (error);
      return (TRUE);
    }

    // Get objects from UI file
    data->traceroute_window = GTK_WIDGET (gtk_builder_get_object (builder, "traceroute_window"));

    // IPv6 over IPv4 (6to4) (traceroute.ui)

    // Ethernet header
    data->checkbutton32 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton32"));  // Specify ethernet header
    data->entry312 = GTK_WIDGET (gtk_builder_get_object (builder, "entry312"));   // Destination link-layer (MAC) address
    data->entry313 = GTK_WIDGET (gtk_builder_get_object (builder, "entry313"));   // Source link-layer (MAC) address
    data->entry314 = GTK_WIDGET (gtk_builder_get_object (builder, "entry314"));   // Source interface name
    data->entry315 = GTK_WIDGET (gtk_builder_get_object (builder, "entry315"));   // Ethernet type code

    data->entry316   = GTK_WIDGET (gtk_builder_get_object (builder, "entry316"));   // TCP Source IPv4 address
    data->entry317   = GTK_WIDGET (gtk_builder_get_object (builder, "entry317"));   // TCP Source IPv4 address
    data->entry318   = GTK_WIDGET (gtk_builder_get_object (builder, "entry318"));   // TCP Source IPv4 address

    // Traceroute TCP (IPv4) page (traceroute.ui)

    // Ethernet header
    data->checkbutton28 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton28"));  // Specify ethernet header
    data->entry164 = GTK_WIDGET (gtk_builder_get_object (builder, "entry164"));   // Destination link-layer (MAC) address
    data->entry165 = GTK_WIDGET (gtk_builder_get_object (builder, "entry165"));   // Source link-layer (MAC) address
    data->entry166 = GTK_WIDGET (gtk_builder_get_object (builder, "entry166"));   // Source interface name
    data->entry167 = GTK_WIDGET (gtk_builder_get_object (builder, "entry167"));   // Ethernet type code
    // IPv4 header
    data->entry168   = GTK_WIDGET (gtk_builder_get_object (builder, "entry168"));   // IP header length
    data->entry169   = GTK_WIDGET (gtk_builder_get_object (builder, "entry169"));   // Protocol version
    data->entry170   = GTK_WIDGET (gtk_builder_get_object (builder, "entry170"));   // Type of service
    data->entry171   = GTK_WIDGET (gtk_builder_get_object (builder, "entry171"));   // Total length of datagram
    data->entry172   = GTK_WIDGET (gtk_builder_get_object (builder, "entry172"));   // ID sequence number
    data->entry173  = GTK_WIDGET (gtk_builder_get_object (builder, "entry173"));    // Zero
    data->entry174   = GTK_WIDGET (gtk_builder_get_object (builder, "entry174"));   // Do not fragment flag
    data->entry175   = GTK_WIDGET (gtk_builder_get_object (builder, "entry175"));   // More fragments following flag
    data->entry176   = GTK_WIDGET (gtk_builder_get_object (builder, "entry176"));   // Fragmentation offset
    data->entry177   = GTK_WIDGET (gtk_builder_get_object (builder, "entry177"));   // Time-to-live
    data->entry178   = GTK_WIDGET (gtk_builder_get_object (builder, "entry178"));   // Transport layer protocol
    data->entry179   = GTK_WIDGET (gtk_builder_get_object (builder, "entry179"));   // Source IP address
    data->entry180   = GTK_WIDGET (gtk_builder_get_object (builder, "entry180"));   // Destination IP address
    data->entry181   = GTK_WIDGET (gtk_builder_get_object (builder, "entry181"));   // IP checksum
    // TCP header (IPv4)
    data->entry182   = GTK_WIDGET (gtk_builder_get_object (builder, "entry182"));   // Source port number
    data->checkbutton30 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton30"));  // Source port "Randomize" checkbutton
    data->entry183   = GTK_WIDGET (gtk_builder_get_object (builder, "entry183"));   // Destination port number
    data->entry184   = GTK_WIDGET (gtk_builder_get_object (builder, "entry184"));   // Sequence number
    data->entry185   = GTK_WIDGET (gtk_builder_get_object (builder, "entry185"));   // Ack number
    data->entry186   = GTK_WIDGET (gtk_builder_get_object (builder, "entry186"));   // Reserved
    data->entry187   = GTK_WIDGET (gtk_builder_get_object (builder, "entry187"));   // Data offset
    data->entry188   = GTK_WIDGET (gtk_builder_get_object (builder, "entry188"));   // FIN flag
    data->entry189   = GTK_WIDGET (gtk_builder_get_object (builder, "entry189"));   // SYN flag
    data->entry190   = GTK_WIDGET (gtk_builder_get_object (builder, "entry190"));   // RST flag
    data->entry191   = GTK_WIDGET (gtk_builder_get_object (builder, "entry191"));   // PSH flag
    data->entry192   = GTK_WIDGET (gtk_builder_get_object (builder, "entry192"));   // ACK flag
    data->entry193   = GTK_WIDGET (gtk_builder_get_object (builder, "entry193"));   // URG flag
    data->entry194   = GTK_WIDGET (gtk_builder_get_object (builder, "entry194"));   // ECE flag
    data->entry195   = GTK_WIDGET (gtk_builder_get_object (builder, "entry195"));   // CWR flag
    data->entry196   = GTK_WIDGET (gtk_builder_get_object (builder, "entry196"));   // Window size
    data->entry197   = GTK_WIDGET (gtk_builder_get_object (builder, "entry197"));   // TCP Checksum
    data->entry198   = GTK_WIDGET (gtk_builder_get_object (builder, "entry198"));   // Urgent pointer
    data->radiobutton29 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton29"));  // ASCII data entry
    data->radiobutton30 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton30"));  // Hexadecimal data entry
    data->entry199  = GTK_WIDGET (gtk_builder_get_object (builder, "entry199"));  // TCP data from keyboard
    data->button29  = GTK_BUTTON (gtk_builder_get_object (builder, "button29"));  // TCP data from file

    data->button28   = GTK_BUTTON (gtk_builder_get_object (builder, "button28"));   // View packet
    data->button30  = GTK_BUTTON (gtk_builder_get_object (builder, "button30"));  // Restore default values

    // Traceroute TCP (IPv6) page (traceroute.ui)

    // Ethernet header
    data->entry247 = GTK_WIDGET (gtk_builder_get_object (builder, "entry247"));  // Destination link-layer (MAC) address
    data->entry248 = GTK_WIDGET (gtk_builder_get_object (builder, "entry248"));  // Source link-layer (MAC) address
    data->entry249 = GTK_WIDGET (gtk_builder_get_object (builder, "entry249"));  // Source interface name
    data->entry250 = GTK_WIDGET (gtk_builder_get_object (builder, "entry250"));  // Ethernet type code
    // IPv6 header
    data->entry251 = GTK_WIDGET (gtk_builder_get_object (builder, "entry251"));  // IP version
    data->entry252 = GTK_WIDGET (gtk_builder_get_object (builder, "entry252"));  // Traffic class
    data->entry253 = GTK_WIDGET (gtk_builder_get_object (builder, "entry253"));  // Flow label
    data->entry254 = GTK_WIDGET (gtk_builder_get_object (builder, "entry254"));  // Payload length
    data->entry255 = GTK_WIDGET (gtk_builder_get_object (builder, "entry255"));  // Next header
    data->entry256 = GTK_WIDGET (gtk_builder_get_object (builder, "entry256"));  // Hop limit
    data->entry257 = GTK_WIDGET (gtk_builder_get_object (builder, "entry257"));  // Source IP address
    data->entry258 = GTK_WIDGET (gtk_builder_get_object (builder, "entry258"));  // Destination IP address
    // TCP header (IPv6)
    data->entry259 = GTK_WIDGET (gtk_builder_get_object (builder, "entry259"));  // Source port number
    data->checkbutton37 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton37"));  // Source port "Randomize" checkbutton
    data->entry260 = GTK_WIDGET (gtk_builder_get_object (builder, "entry260"));  // Destination port number
    data->entry261 = GTK_WIDGET (gtk_builder_get_object (builder, "entry261"));  // Sequence number
    data->entry262 = GTK_WIDGET (gtk_builder_get_object (builder, "entry262"));  // Ack number
    data->entry263 = GTK_WIDGET (gtk_builder_get_object (builder, "entry263"));  // Reserved
    data->entry264 = GTK_WIDGET (gtk_builder_get_object (builder, "entry264"));  // Data offset
    data->entry265 = GTK_WIDGET (gtk_builder_get_object (builder, "entry265"));  // FIN flag
    data->entry266 = GTK_WIDGET (gtk_builder_get_object (builder, "entry266"));  // SYN flag
    data->entry267 = GTK_WIDGET (gtk_builder_get_object (builder, "entry267"));  // RST flag
    data->entry268 = GTK_WIDGET (gtk_builder_get_object (builder, "entry268"));  // PSH flag
    data->entry269 = GTK_WIDGET (gtk_builder_get_object (builder, "entry269"));  // ACK flag
    data->entry270 = GTK_WIDGET (gtk_builder_get_object (builder, "entry270"));  // URG flag
    data->entry271 = GTK_WIDGET (gtk_builder_get_object (builder, "entry271"));  // ECE flag
    data->entry272 = GTK_WIDGET (gtk_builder_get_object (builder, "entry272"));  // CWR flag
    data->entry273 = GTK_WIDGET (gtk_builder_get_object (builder, "entry273"));  // Window size
    data->entry274 = GTK_WIDGET (gtk_builder_get_object (builder, "entry274"));  // TCP Checksum
    data->entry275 = GTK_WIDGET (gtk_builder_get_object (builder, "entry275"));  // Urgent pointer
    data->radiobutton35 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton35"));  // ASCII data entry
    data->radiobutton36 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton36"));  // Hexadecimal data entry
    data->entry276 = GTK_WIDGET (gtk_builder_get_object (builder, "entry276"));  // TCP data from keyboard
    data->button38 = GTK_BUTTON (gtk_builder_get_object (builder, "button38"));  // TCP data from file

    data->button37 = GTK_BUTTON (gtk_builder_get_object (builder, "button37"));  // View packet
    data->button39 = GTK_BUTTON (gtk_builder_get_object (builder, "button39"));  // Restore default values

    // Traceroute ICMP (IPv4) page (traceroute.ui)

    // Ethernet header
    data->checkbutton31 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton31"));  // Specify ethernet header`
    data->entry200 = GTK_WIDGET (gtk_builder_get_object (builder, "entry200"));  // Destination link-layer (MAC) address
    data->entry201 = GTK_WIDGET (gtk_builder_get_object (builder, "entry201"));  // Source link-layer (MAC) address
    data->entry202 = GTK_WIDGET (gtk_builder_get_object (builder, "entry202"));  // Source interface name
    data->entry203 = GTK_WIDGET (gtk_builder_get_object (builder, "entry203"));  // Ethernet type code
    // IPv4 header
    data->entry204 = GTK_WIDGET (gtk_builder_get_object (builder, "entry204"));  // IP header length
    data->entry205 = GTK_WIDGET (gtk_builder_get_object (builder, "entry205"));  // Protocol version
    data->entry206 = GTK_WIDGET (gtk_builder_get_object (builder, "entry206"));  // Type of service
    data->entry207 = GTK_WIDGET (gtk_builder_get_object (builder, "entry207"));  // Total length of datagram
    data->entry208 = GTK_WIDGET (gtk_builder_get_object (builder, "entry208"));  // ID sequence number
    data->entry209 = GTK_WIDGET (gtk_builder_get_object (builder, "entry209"));  // Zero
    data->entry210 = GTK_WIDGET (gtk_builder_get_object (builder, "entry210"));  // Do not fragment flag
    data->entry211 = GTK_WIDGET (gtk_builder_get_object (builder, "entry211"));  // More fragments following flag
    data->entry212 = GTK_WIDGET (gtk_builder_get_object (builder, "entry212"));  // Fragmentation offset
    data->entry213 = GTK_WIDGET (gtk_builder_get_object (builder, "entry213"));  // Time-to-live
    data->entry214 = GTK_WIDGET (gtk_builder_get_object (builder, "entry214"));  // Transport layer protocol
    data->entry215 = GTK_WIDGET (gtk_builder_get_object (builder, "entry215"));  // Source IP address
    data->entry216 = GTK_WIDGET (gtk_builder_get_object (builder, "entry216"));  // Destination IP address
    data->entry217 = GTK_WIDGET (gtk_builder_get_object (builder, "entry217"));  // IP checksum
    // ICMP header (IPv4)
    data->entry218 = GTK_WIDGET (gtk_builder_get_object (builder, "entry218"));  // Message type
    data->entry219 = GTK_WIDGET (gtk_builder_get_object (builder, "entry219"));  // Message code
    data->entry220 = GTK_WIDGET (gtk_builder_get_object (builder, "entry220"));  // ICMP checksum
    data->entry221 = GTK_WIDGET (gtk_builder_get_object (builder, "entry221"));  // Identifier
    data->entry222 = GTK_WIDGET (gtk_builder_get_object (builder, "entry222"));  // Sequence number
    data->radiobutton31 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton31"));  // ASCII data entry
    data->radiobutton32 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton32"));  // Hexadecimal data entry
    data->entry223 = GTK_WIDGET (gtk_builder_get_object (builder, "entry223"));  // ICMP data from keyboard
    data->button32 = GTK_BUTTON (gtk_builder_get_object (builder, "button32"));  // ICMP data from file

    data->button31 = GTK_BUTTON (gtk_builder_get_object (builder, "button31"));  // View packet
    data->button33 = GTK_BUTTON (gtk_builder_get_object (builder, "button33"));  // Restore default values

    // Traceroute ICMP (IPv6) page (traceroute.ui)

    // Ethernet header
    data->entry277 = GTK_WIDGET (gtk_builder_get_object (builder, "entry277"));  // Destination link-layer (MAC) address
    data->entry278 = GTK_WIDGET (gtk_builder_get_object (builder, "entry278"));  // Source link-layer (MAC) address
    data->entry279 = GTK_WIDGET (gtk_builder_get_object (builder, "entry279"));  // Source interface name
    data->entry280 = GTK_WIDGET (gtk_builder_get_object (builder, "entry280"));  // Ethernet type code
    // IPv6 header
    data->entry281 = GTK_WIDGET (gtk_builder_get_object (builder, "entry281"));  // IP version
    data->entry282 = GTK_WIDGET (gtk_builder_get_object (builder, "entry282"));  // Traffic class
    data->entry283 = GTK_WIDGET (gtk_builder_get_object (builder, "entry283"));  // Flow label
    data->entry284 = GTK_WIDGET (gtk_builder_get_object (builder, "entry284"));  // Payload length
    data->entry285 = GTK_WIDGET (gtk_builder_get_object (builder, "entry285"));  // Next header
    data->entry286 = GTK_WIDGET (gtk_builder_get_object (builder, "entry286"));  // Hop limit
    data->entry287 = GTK_WIDGET (gtk_builder_get_object (builder, "entry287"));  // Source IP address
    data->entry288 = GTK_WIDGET (gtk_builder_get_object (builder, "entry288"));  // Destination IP address
    // ICMP header (IPv6)
    data->entry289 = GTK_WIDGET (gtk_builder_get_object (builder, "entry289"));  // Message type
    data->entry290 = GTK_WIDGET (gtk_builder_get_object (builder, "entry290"));  // Message code
    data->entry291 = GTK_WIDGET (gtk_builder_get_object (builder, "entry291"));  // ICMP checksum
    data->entry292 = GTK_WIDGET (gtk_builder_get_object (builder, "entry292"));  // Identifier
    data->entry293 = GTK_WIDGET (gtk_builder_get_object (builder, "entry293"));  // Sequence number
    data->radiobutton37 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton37"));  // ASCII data entry
    data->radiobutton38 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton38"));  // Hexadecimal data entry
    data->entry294 = GTK_WIDGET (gtk_builder_get_object (builder, "entry294"));  // ICMP data from keyboard
    data->button41 = GTK_BUTTON (gtk_builder_get_object (builder, "button41"));  // ICMP data from file

    data->button40 = GTK_BUTTON (gtk_builder_get_object (builder, "button40"));  // View packet
    data->button42 = GTK_BUTTON (gtk_builder_get_object (builder, "button42"));  // Restore default values

    // Traceroute UDP (IPv4) page (traceroute.ui)

    // Ethernet header
    data->checkbutton33 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton33"));  // Specify ethernet header
    data->entry224 = GTK_WIDGET (gtk_builder_get_object (builder, "entry224"));  // Destination link-layer (MAC) address
    data->entry225 = GTK_WIDGET (gtk_builder_get_object (builder, "entry225"));  // Source link-layer (MAC) address
    data->entry226 = GTK_WIDGET (gtk_builder_get_object (builder, "entry226"));  // Source interface name
    data->entry227 = GTK_WIDGET (gtk_builder_get_object (builder, "entry227"));  // Ethernet type code
    // IPv4 header
    data->entry228 = GTK_WIDGET (gtk_builder_get_object (builder, "entry228"));  // IP header length
    data->entry229 = GTK_WIDGET (gtk_builder_get_object (builder, "entry229"));  // Protocol version
    data->entry230 = GTK_WIDGET (gtk_builder_get_object (builder, "entry230"));  // Type of service
    data->entry231 = GTK_WIDGET (gtk_builder_get_object (builder, "entry231"));  // Total length of datagram
    data->entry232 = GTK_WIDGET (gtk_builder_get_object (builder, "entry232"));  // ID sequence number
    data->entry233 = GTK_WIDGET (gtk_builder_get_object (builder, "entry233"));  // Zero
    data->entry234 = GTK_WIDGET (gtk_builder_get_object (builder, "entry234"));  // Do not fragment flag
    data->entry235 = GTK_WIDGET (gtk_builder_get_object (builder, "entry235"));  // More fragments following flag
    data->entry236 = GTK_WIDGET (gtk_builder_get_object (builder, "entry236"));  // Fragmentation offset
    data->entry237 = GTK_WIDGET (gtk_builder_get_object (builder, "entry237"));  // Time-to-live
    data->entry238 = GTK_WIDGET (gtk_builder_get_object (builder, "entry238"));  // Transport layer protocol
    data->entry239 = GTK_WIDGET (gtk_builder_get_object (builder, "entry239"));  // Source IP address
    data->entry240 = GTK_WIDGET (gtk_builder_get_object (builder, "entry240"));  // Destination IP address
    data->entry241 = GTK_WIDGET (gtk_builder_get_object (builder, "entry241"));  // IP checksum
    // UDP header (IPv4)
    data->entry242 = GTK_WIDGET (gtk_builder_get_object (builder, "entry242"));  // Source port
    data->checkbutton35 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton35"));  // Source IP "Randomize" checkbutton
    data->entry243 = GTK_WIDGET (gtk_builder_get_object (builder, "entry243"));  // Destination port
    data->entry244 = GTK_WIDGET (gtk_builder_get_object (builder, "entry244"));  // Length of UDP datagram
    data->entry245 = GTK_WIDGET (gtk_builder_get_object (builder, "entry245"));  // UDP checksum
    data->radiobutton33 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton33"));  // ASCII data entry
    data->radiobutton34 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton34"));  // Hexadecimal data entry
    data->entry246 = GTK_WIDGET (gtk_builder_get_object (builder, "entry246"));  // UDP data from keyboard
    data->button35 = GTK_BUTTON (gtk_builder_get_object (builder, "button35"));  // UDP data from file

    data->button34 = GTK_BUTTON (gtk_builder_get_object (builder, "button34"));  // View packet
    data->button36 = GTK_BUTTON (gtk_builder_get_object (builder, "button36"));  // Restore default values

    // Traceroute UDP (IPv6) page (traceroute.ui)

    // Ethernet header
    data->entry295 = GTK_WIDGET (gtk_builder_get_object (builder, "entry295"));  // Destination link-layer (MAC) address
    data->entry296 = GTK_WIDGET (gtk_builder_get_object (builder, "entry296"));  // Source link-layer (MAC) address
    data->entry297 = GTK_WIDGET (gtk_builder_get_object (builder, "entry297"));  // Source interface name
    data->entry298 = GTK_WIDGET (gtk_builder_get_object (builder, "entry298"));  // Ethernet type code
    // IPv6 header
    data->entry299 = GTK_WIDGET (gtk_builder_get_object (builder, "entry299"));  // IP version
    data->entry300 = GTK_WIDGET (gtk_builder_get_object (builder, "entry300"));  // Traffic class
    data->entry301 = GTK_WIDGET (gtk_builder_get_object (builder, "entry301"));  // Flow label
    data->entry302 = GTK_WIDGET (gtk_builder_get_object (builder, "entry302"));  // Payload length
    data->entry303 = GTK_WIDGET (gtk_builder_get_object (builder, "entry303"));  // Next header
    data->entry304 = GTK_WIDGET (gtk_builder_get_object (builder, "entry304"));  // Hop limit
    data->entry305 = GTK_WIDGET (gtk_builder_get_object (builder, "entry305"));  // Source IP address
    data->entry306 = GTK_WIDGET (gtk_builder_get_object (builder, "entry306"));  // Destination IP address
    // UDP header (IPv6)
    data->entry307 = GTK_WIDGET (gtk_builder_get_object (builder, "entry307"));  // Source port
    data->checkbutton40 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "checkbutton40"));  // Source IP "Randomize" checkbutton
    data->entry308 = GTK_WIDGET (gtk_builder_get_object (builder, "entry308"));  // Destination port
    data->entry309 = GTK_WIDGET (gtk_builder_get_object (builder, "entry309"));  // Length of UDP datagram
    data->entry310 = GTK_WIDGET (gtk_builder_get_object (builder, "entry310"));  // UDP checksum
    data->radiobutton39 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton39"));  // ASCII data entry
    data->radiobutton40 = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "radiobutton40"));  // Hexadecimal data entry
    data->entry311 = GTK_WIDGET (gtk_builder_get_object (builder, "entry311"));  // UDP data from keyboard
    data->button44 = GTK_BUTTON (gtk_builder_get_object (builder, "button44"));  // UDP data from file

    data->button43 = GTK_BUTTON (gtk_builder_get_object (builder, "button43"));  // View packet
    data->button45 = GTK_BUTTON (gtk_builder_get_object (builder, "button45"));  // Restore default values

    // Connect signals
    gtk_builder_connect_signals (builder, data);

    // Destroy builder, since we don't need it anymore
    g_object_unref (G_OBJECT (builder));

    // Show window. All other widgets are automatically shown by GtkBuilder
    gtk_widget_show (data->traceroute_window);

    // Populate entries in traceroute.ui window.
    sixto4_tr_show (data);
    tcp4_tr_show (data);
    tcp6_tr_show (data);
    icmp4_tr_show (data);
    icmp6_tr_show (data);
    udp4_tr_show (data);
    udp6_tr_show (data);
  }

  return (TRUE);
}

// Destroy Traceroute window if closed by user.
gboolean
on_traceroute_window_destroy (GtkWidget *widget, SPSData *data)
{
  data->traceroute_flag = 0;

  return (TRUE);
}

// Populate IPv6 over IPv4 (6to4) entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
sixto4_tr_show (SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in sixto4_tr_show().\n");
    exit (EXIT_FAILURE);
  }

  // Specify the ethernet header or not.
  if (data->sixto4_tr_specify_ether == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton32), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton32), FALSE);
  }

  // Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 16);
  mac_bin2string (data->sixto4ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry312), value);

  // Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry314), data->sixto4_tr_interfacename);

  // Source link-layer (MAC) address (48 bits)
  memset (value, 0, 16);
  mac_bin2string (data->sixto4ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry313), value);

  // Ethernet type code (16 bits)
  // http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->sixto4ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry315), value);

  // Source IPv4 addresses (32 bits)
  inet_ntop (AF_INET, &(data->sixto4ip4hdr_tcp_tr.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry316), value);
  inet_ntop (AF_INET, &(data->sixto4ip4hdr_icmp_tr.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry317), value);
  inet_ntop (AF_INET, &(data->sixto4ip4hdr_udp_tr.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry318), value);

  return (TRUE);

}

// Populate IPv4 TCP entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
tcp4_tr_show (SPSData *data)
{
  gchar *value;
  void *tmp;

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in tcp4_tr_show().\n");
    exit (EXIT_FAILURE);
  }

  // Specify the ethernet header or not.
  if (data->tcp4_tr_specify_ether == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton28), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton28), FALSE);
  }

  // Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 16);
  mac_bin2string (data->tcp4ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry164), value);

  // Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry166), data->tcp4_tr_interfacename);

  // Source link-layer (MAC) address (48 bits)
  memset (value, 0, 16);
  mac_bin2string (data->tcp4ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry165), value);

  // Ethernet type code (16 bits)
  // http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry167), value);

  // TCP data

  // TCP data buffer.
  gtk_entry_set_text (GTK_ENTRY (data->entry199), "");

  // TCP data entry format
  if (data->ascii_hex_tcp4_tr == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton30), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton30), FALSE);
  }

  // IPv4 header

  // IPv4 header length (4 bits)
  sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry168), value);

  // Internet Protocol version (4 bits)
  sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry169), value);

  // Type of service (8 bits)
  sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry170), value);

  // Total length of datagram (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry171), value);

  // ID sequence number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry172), value);

  // Zero (1 bit)
  sprintf (value, "%u", (data->ip4hdr_tcp_tr.ip_off >> 15));
  gtk_entry_set_text (GTK_ENTRY (data->entry173), value);

  // Do not fragment flag (1 bit)
  sprintf (value, "%u", (data->ip4hdr_tcp_tr.ip_off >> 14) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry174), value);

  // More fragments following flag (1 bit)
  sprintf (value, "%u", (data->ip4hdr_tcp_tr.ip_off >> 13) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry175), value);

  // Fragmentation offset (13 bits)
  sprintf (value, "%u", (data->ip4hdr_tcp_tr.ip_off & (7 << 13)));
  gtk_entry_set_text (GTK_ENTRY (data->entry176), value);

  // Time-to-Live (8 bits)
  sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry177), value);

  // Transport layer protocol (8 bits)
  sprintf (value, "%u", data->ip4hdr_tcp_tr.ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry178), value);

  // Source IPv4 address (32 bits)
  inet_ntop (AF_INET, &(data->ip4hdr_tcp_tr.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry179), value);

  // Destination IPv4 address (32 bits)
  inet_ntop (AF_INET, &(data->ip4hdr_tcp_tr.ip_dst), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry180), value);

  // IPv4 header checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_tcp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry181), value);

  // TCP header (IPv4)

  // Source port number (16 bits)
  if (data->ran_tcp4_tr_sourceport == 0) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton30), FALSE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton30), TRUE);
  }
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry182), value);

  // Destination port number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry183), value);

  // Sequence number (32 bits)
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr_tr.th_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry184), value);

  // Acknowledgement number (32 bits)
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp4hdr_tr.th_ack));
  gtk_entry_set_text (GTK_ENTRY (data->entry185), value);

  // Reserved (4 bits)
  sprintf (value, "%u", data->tcp4hdr_tr.th_x2);
  gtk_entry_set_text (GTK_ENTRY (data->entry186), value);

  // Data offset (4 bits)
  sprintf (value, "%u", data->tcp4hdr_tr.th_off);
  gtk_entry_set_text (GTK_ENTRY (data->entry187), value);

  // FIN flag (1 bit)
  sprintf (value, "%u",  data->tcp4hdr_tr.th_flags & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry188), value);

  // SYN flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 1) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry189), value);

  // RST flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 2) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry190), value);

  // PSH flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 3) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry191), value);

  // ACK flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 4) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry192), value);

  // URG flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 5) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry193), value);

  // ECE flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 6) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry194), value);

  // CWR flag (1 bit)
  sprintf (value, "%u", (data->tcp4hdr_tr.th_flags >> 7) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry195), value);

  // Window size (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_win));
  gtk_entry_set_text (GTK_ENTRY (data->entry196), value);

  // Urgent pointer (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_urp));
  gtk_entry_set_text (GTK_ENTRY (data->entry198), value);

  // TCP checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp4hdr_tr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry197), value);

  // Free allocated memory.
  g_free (value);

  return (TRUE);
}

// Populate IPv6 TCP entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
tcp6_tr_show (SPSData *data)
{
  unsigned char *tcp6_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// TCP flags
  tmp = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));
  if (tmp != NULL) {
    tcp6_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'tcp6_flags' in tcp6_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in tcp6_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in tcp6_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->tcp6ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry247), value);

// Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry249), data->tcp6_tr_interfacename);

// Source link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->tcp6ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry248), value);

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry250), value);

// TCP data

// TCP data buffer
  gtk_entry_set_text (GTK_ENTRY (data->entry276), "");

// TCP data entry format
  if (data->ascii_hex_tcp6_tr == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton36), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton36), FALSE);
  }

// IPv6 header

// Version (4 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp_tr.ip6_flow) >> 28));
  gtk_entry_set_text (GTK_ENTRY (data->entry251), value);

// Traffic Class (8 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp_tr.ip6_flow) >> 20) & 0xfful);
  gtk_entry_set_text (GTK_ENTRY (data->entry252), value);

// Flow Label (20 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_tcp_tr.ip6_flow) & 0xffffful));
  gtk_entry_set_text (GTK_ENTRY (data->entry253), value);

// Payload length (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_tcp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry254), value);

// Next header (8 bits)
  sprintf (value, "%u", data->ip6hdr_tcp_tr.ip6_nxt);
  gtk_entry_set_text (GTK_ENTRY (data->entry255), value);

// Hop limit (8 bits)
  sprintf (value, "%u", data->ip6hdr_tcp_tr.ip6_hops);
  gtk_entry_set_text (GTK_ENTRY (data->entry256), value);

// Source IPv6 address (128 bits)
  inet_ntop (AF_INET6, &(data->ip6hdr_tcp_tr.ip6_src), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry257), value);

// Destination IPv6 address (128 bits)
  inet_ntop (AF_INET6, &(data->ip6hdr_tcp_tr.ip6_dst), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry258), value);

// TCP header (IPv6)

// Source port number (16 bits)
  if (data->ran_tcp6_tr_sourceport == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton37), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton37), FALSE);
  }
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry259), value);

// Destination port number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry260), value);

// Sequence number (32 bits)
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr_tr.th_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry261), value);

// Acknowledgement number (32 bits)
  sprintf (value, "%lu", (unsigned long int) ntohl (data->tcp6hdr_tr.th_ack));
  gtk_entry_set_text (GTK_ENTRY (data->entry262), value);

// Reserved (4 bits)
  sprintf (value, "%u", data->tcp6hdr_tr.th_x2);
  gtk_entry_set_text (GTK_ENTRY (data->entry263), value);

// Data offset (4 bits)
  sprintf (value, "%u", data->tcp6hdr_tr.th_off);
  gtk_entry_set_text (GTK_ENTRY (data->entry264), value);

  // FIN flag (1 bit)
  sprintf (value, "%u",  data->tcp6hdr_tr.th_flags & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry265), value);

  // SYN flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 1) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry266), value);

  // RST flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 2) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry267), value);

  // PSH flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 3) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry268), value);

  // ACK flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 4) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry269), value);

  // URG flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 5) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry270), value);

  // ECE flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 6) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry271), value);

  // CWR flag (1 bit)
  sprintf (value, "%u", (data->tcp6hdr_tr.th_flags >> 7) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry272), value);

// Window size (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_win));
  gtk_entry_set_text (GTK_ENTRY (data->entry273), value);

// Urgent pointer (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_urp));
  gtk_entry_set_text (GTK_ENTRY (data->entry275), value);

// TCP checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->tcp6hdr_tr.th_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry274), value);

// Free allocated memory.
  g_free (tcp6_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Populate IPv4 ICMP entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
icmp4_tr_show (SPSData *data)
{
  unsigned char *ip4_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in icmp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in icmp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in icmp4_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Specify the ethernet header or not.
  if (data->icmp4_tr_specify_ether ==1 ) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton31), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton31), FALSE);
  }

// Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->icmp4ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry200), value);

// Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry202), data->icmp4_tr_interfacename);

// Source link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->icmp4ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry201), value);

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry203), value);

// ICMP data

// ICMP data buffer
  gtk_entry_set_text (GTK_ENTRY (data->entry223), "");

// ICMP data entry format: default to hexadecimal input
  if (data->ascii_hex_icmp4_tr == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton32), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton32), FALSE);
  }

// IPv4 header

// IPv4 header length (4 bits)
  sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry204), value);

// Internet Protocol version (4 bits)
  sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry205), value);

// Type of service (8 bits)
  sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry206), value);

// Total length of datagram (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry207), value);

// ID sequence number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry208), value);

  // Zero (1 bit)
  sprintf (value, "%u", (data->ip4hdr_icmp_tr.ip_off >> 15));
  gtk_entry_set_text (GTK_ENTRY (data->entry209), value);

  // Do not fragment flag (1 bit)
  sprintf (value, "%u", (data->ip4hdr_icmp_tr.ip_off >> 14) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry210), value);

  // More fragments following flag (1 bit)
  sprintf (value, "%u", (data->ip4hdr_icmp_tr.ip_off >> 13) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry211), value);

  // Fragmentation offset (13 bits)
  sprintf (value, "%u", (data->ip4hdr_icmp_tr.ip_off & (7 << 13)));
  gtk_entry_set_text (GTK_ENTRY (data->entry212), value);

// Time-to-Live (8 bits)
  sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry213), value);

// Transport layer protocol (8 bits)
  sprintf (value, "%u", data->ip4hdr_icmp_tr.ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry214), value);

// Source IPv4 address (32 bits)
  inet_ntop (AF_INET, &(data->ip4hdr_icmp_tr.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry215), value);

// Destination IPv4 address (32 bits)
  inet_ntop (AF_INET, &(data->ip4hdr_icmp_tr.ip_dst), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry216), value);

// IPv4 header checksum (16 bits): set to 0 when calculating checksum
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_icmp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry217), value);

// ICMP header (IPv4)

// Message Type (8 bits)
  sprintf (value, "%u", data->icmp4hdr_tr.icmp_type);
  gtk_entry_set_text (GTK_ENTRY (data->entry218), value);

// Message Code (8 bits)
  sprintf (value, "%u", data->icmp4hdr_tr.icmp_code);
  gtk_entry_set_text (GTK_ENTRY (data->entry219), value);

// Identifier (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry221), value);

// Sequence Number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry222), value);

// ICMP header checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp4hdr_tr.icmp_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry220), value);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Populate IPv6 ICMP entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
icmp6_tr_show (SPSData *data)
{
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in icmp6_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in icmp6_tr_default().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->icmp6ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry277), value);

// Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry279), data->icmp6_tr_interfacename);

// Source link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->icmp6ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry278), value);
  
// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry280), value);

// ICMP data

// ICMP data buffer
  gtk_entry_set_text (GTK_ENTRY (data->entry294), "");

// ICMP data entry format
  if (data->ascii_hex_icmp6_tr == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton38), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton38), FALSE);
  }

// IPv6 header

// Version (4 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp_tr.ip6_flow) >> 28));
  gtk_entry_set_text (GTK_ENTRY (data->entry281), value);

// Traffic Class (8 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp_tr.ip6_flow) >> 20) & 0xfful);
  gtk_entry_set_text (GTK_ENTRY (data->entry282), value);

// Flow Label (20 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_icmp_tr.ip6_flow) & 0xffffful));
  gtk_entry_set_text (GTK_ENTRY (data->entry283), value);

// Payload length (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_icmp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry284), value);

// Next header (8 bits)
  sprintf (value, "%u", data->ip6hdr_icmp_tr.ip6_nxt);
  gtk_entry_set_text (GTK_ENTRY (data->entry285), value);

// Hop limit (8 bits)
  sprintf (value, "%u", data->ip6hdr_icmp_tr.ip6_hops);
  gtk_entry_set_text (GTK_ENTRY (data->entry286), value);

// Source IPv6 address (128 bits)
  inet_ntop (AF_INET6, &(data->ip6hdr_icmp_tr.ip6_src), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry287), value);

// Destination IPv6 address (128 bits)
  inet_ntop (AF_INET6, &(data->ip6hdr_icmp_tr.ip6_dst), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry288), value);

// ICMP header (IPv6)

// Message Type (8 bits)
  sprintf (value, "%u", data->icmp6hdr_tr.icmp6_type);
  gtk_entry_set_text (GTK_ENTRY (data->entry289), value);

// Message Code (8 bits)
  sprintf (value, "%u", data->icmp6hdr_tr.icmp6_code);
  gtk_entry_set_text (GTK_ENTRY (data->entry290), value);

// Identifier (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry292), value);

// Sequence Number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_seq));
  gtk_entry_set_text (GTK_ENTRY (data->entry293), value);

// ICMP header checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->icmp6hdr_tr.icmp6_cksum));
  gtk_entry_set_text (GTK_ENTRY (data->entry291), value);

// Free allocated memory.
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Populate IPv4 UDP entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
udp4_tr_show (SPSData *data)
{
  unsigned char *ip4_flags;
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// IPv4 header fragmentation flags and offset
  tmp = (unsigned char *) g_malloc0 (4 * sizeof (unsigned char));
  if (tmp != NULL) {
    ip4_flags = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip4_flags' in udp4_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in udp4_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in udp4_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Specify the ethernet header or not.
  if (data->udp4_tr_specify_ether == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton33), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton33), FALSE);
  }

// Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->udp4ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry224), value);

// Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry226), data->udp4_tr_interfacename);

// Source link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->udp4ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry225), value);

// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry227), value);

// UDP data

// UDP data buffer
  gtk_entry_set_text (GTK_ENTRY (data->entry246), "");

// UDP data entry format
  if (data->ascii_hex_udp4_tr == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton34), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton34), FALSE);
  }

// IPv4 header

// IPv4 header length (4 bits)
  sprintf (value, "%u", data->ip4hdr_udp_tr.ip_hl);
  gtk_entry_set_text (GTK_ENTRY (data->entry228), value);

// Internet Protocol version (4 bits)
  sprintf (value, "%u", data->ip4hdr_udp_tr.ip_v);
  gtk_entry_set_text (GTK_ENTRY (data->entry229), value);

// Type of service (8 bits)
  sprintf (value, "%u", data->ip4hdr_udp_tr.ip_tos);
  gtk_entry_set_text (GTK_ENTRY (data->entry230), value);

// Total length of datagram (16 bits)
  sprintf (value, "%hu", htons (data->ip4hdr_udp_tr.ip_len));
  gtk_entry_set_text (GTK_ENTRY (data->entry231), value);

// ID sequence number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_id));
  gtk_entry_set_text (GTK_ENTRY (data->entry232), value);

  // Zero (1 bit)
  sprintf (value, "%u", (data->ip4hdr_udp_tr.ip_off >> 15));
  gtk_entry_set_text (GTK_ENTRY (data->entry233), value);

  // Do not fragment flag (1 bit)
  sprintf (value, "%u", (data->ip4hdr_udp_tr.ip_off >> 14) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry234), value);

  // More fragments following flag (1 bit)
  sprintf (value, "%u", (data->ip4hdr_udp_tr.ip_off >> 13) & 1);
  gtk_entry_set_text (GTK_ENTRY (data->entry235), value);

  // Fragmentation offset (13 bits)
  sprintf (value, "%u", (data->ip4hdr_udp_tr.ip_off & (7 << 13)));
  gtk_entry_set_text (GTK_ENTRY (data->entry236), value);

// Time-to-Live (8 bits)
  sprintf (value, "%u", data->ip4hdr_udp_tr.ip_ttl);
  gtk_entry_set_text (GTK_ENTRY (data->entry237), value);

// Transport layer protocol (8 bits)
  sprintf (value, "%u", data->ip4hdr_udp_tr.ip_p);
  gtk_entry_set_text (GTK_ENTRY (data->entry238), value);

// Source IPv4 address (32 bits)
  inet_ntop (AF_INET, &(data->ip4hdr_udp_tr.ip_src), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry239), value);

// Destination IPv4 address (32 bits)
  inet_ntop (AF_INET, &(data->ip4hdr_udp_tr.ip_dst), value, 16);
  gtk_entry_set_text (GTK_ENTRY (data->entry240), value);

// IPv4 header checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip4hdr_udp_tr.ip_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry241), value);

// UDP header (IPv4)

// Source port number (16 bits)
  if (data->ran_udp4_tr_sourceport == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton35), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton35), FALSE);
  }
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry242), value);

// Destination port number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry243), value);

// Length of UDP datagram (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry244), value);

// UDP checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp4hdr_tr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry245), value);

// Free allocated memory.
  g_free (ip4_flags);
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Populate IPv6 UDP entries on Traceroute page
// **ONLY TO BE CALLED WHEN TRACEROUTE WINDOW IS ACTIVATED**
gboolean
udp6_tr_show (SPSData *data)
{
  gchar *ipaddress, *value;
  void *tmp;

// Allocate memory for various arrays

// String to hold miscellaneous calculated values
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in udp6_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// Array to hold an IPv4 or IPv6 address + '\0'
  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ipaddress = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ipaddress' in udp6_tr_show().\n");
    exit (EXIT_FAILURE);
  }

// Ethernet header

// Destination link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->udp6ethhdr_tr.dst_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry295), value);

// Source Interface Name
  gtk_entry_set_text (GTK_ENTRY (data->entry297), data->udp6_tr_interfacename);

// Source link-layer (MAC) address (48 bits)
  memset (value, 0, 40);
  mac_bin2string (data->udp6ethhdr_tr.src_mac, value);
  gtk_entry_set_text (GTK_ENTRY (data->entry296), value);
  
// Ethernet type code (16 bits)
// http://www.iana.org/assignments/ethernet-numbers
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6ethhdr_tr.type_code));
  gtk_entry_set_text (GTK_ENTRY (data->entry298), value);

// UDP data

// UDP data buffer
  gtk_entry_set_text (GTK_ENTRY (data->entry311), "");

// UDP data entry format
  if (data->ascii_hex_udp6 == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton40), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->radiobutton40), FALSE);
  }

// IPv6 header

// Version (4 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp_tr.ip6_flow) >> 28));
  gtk_entry_set_text (GTK_ENTRY (data->entry299), value);

// Traffic Class (8 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp_tr.ip6_flow) >> 20) & 0xfful);
  gtk_entry_set_text (GTK_ENTRY (data->entry300), value);

// Flow Label (20 bits)
  sprintf (value, "%lu", (unsigned long int) (ntohl (data->ip6hdr_udp_tr.ip6_flow) & 0xffffful));
  gtk_entry_set_text (GTK_ENTRY (data->entry301), value);

// Payload length (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->ip6hdr_udp_tr.ip6_plen));
  gtk_entry_set_text (GTK_ENTRY (data->entry302), value);

// Next header (8 bits)
  sprintf (value, "%u", data->ip6hdr_udp_tr.ip6_nxt);
  gtk_entry_set_text (GTK_ENTRY (data->entry303), value);

// Hop limit (8 bits)
  sprintf (value, "%u", data->ip6hdr_udp_tr.ip6_hops);
  gtk_entry_set_text (GTK_ENTRY (data->entry304), value);

// Source IPv6 address (128 bits)
  inet_ntop (AF_INET6, &(data->ip6hdr_udp_tr.ip6_src), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry305), value);

// Destination IPv6 address (128 bits)
  inet_ntop (AF_INET6, &(data->ip6hdr_udp_tr.ip6_dst), value, 40);
  gtk_entry_set_text (GTK_ENTRY (data->entry306), value);

// UDP header (IPv6)

// Source port number (16 bits)
  if (data->ran_udp6_tr_sourceport == 1) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton40), TRUE);
  } else {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (data->checkbutton40), FALSE);
  }
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sport));
  gtk_entry_set_text (GTK_ENTRY (data->entry307), value);

// Destination port number (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_dport));
  gtk_entry_set_text (GTK_ENTRY (data->entry308), value);

// Length of UDP datagram (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_ulen));
  gtk_entry_set_text (GTK_ENTRY (data->entry309), value);

// UDP header checksum (16 bits)
  sprintf (value, "%hu", (unsigned short int) ntohs (data->udp6hdr_tr.uh_sum));
  gtk_entry_set_text (GTK_ENTRY (data->entry310), value);

// Free allocated memory.
  g_free (value);
  g_free (ipaddress);

  return (TRUE);
}

// Start traceroute.
gboolean
on_button24_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (tracing != 1) {
    tracing = 1;

    if ((data->packet_type_tr == 1) || (data->packet_type_tr == 2) ||
        (data->packet_type_tr == 3)) {

      // Spawn thread to perform traceroute.
      thread = g_thread_create ((GThreadFunc) ipv4_tr_send, data, FALSE, &error);
      if (! thread) {
        g_print ("Error: Unable to create new thread for IPv4 traceroute.%s\n", error->message);
        exit (EXIT_FAILURE);
      }

    } else if ((data->packet_type_tr == 4) || (data->packet_type_tr == 5) ||
               (data->packet_type_tr == 6)) {
      if (data->sixto4_tr_flag == 1) {

        // Spawn thread to perform traceroute.
        thread = g_thread_create ((GThreadFunc) sixto4_tr_send, data, FALSE, &error);
        if (! thread) {
          g_print ("Error: Unable to create new thread for IPv6 over IPv4 (6to4) traceroute.%s\n", error->message);
          exit (EXIT_FAILURE);
        }
      } else {

        // Spawn thread to perform traceroute.
        thread = g_thread_create ((GThreadFunc) ipv6_tr_send, data, FALSE, &error);
        if (! thread) {
          g_print ("Error: Unable to create new thread for IPv6 over IPv4 (6to4) traceroute.%s\n", error->message);
          exit (EXIT_FAILURE);
        }
      }
    }
  }  // End If tracing != 1

  return (TRUE);
}

// Start an instance of tr_message() idle function.
gboolean
post_tr_message (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    data->tr_message = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->tr_message' in post_tr_message().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->tr_message, message);  // Copy message.

  g_idle_add ((GSourceFunc) tr_message, data);
  sleep (1);

  return (TRUE);
}

// Idle function to update traceroute textview with a message.
gboolean
tr_message (SPSData *data)
{
  GtkTextBuffer *textbuffer7;
  GtkTextIter end;

  textbuffer7 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview7));
  gtk_text_buffer_get_end_iter (textbuffer7, &end);
  gtk_text_buffer_insert (textbuffer7, &end, data->tr_message, -1);
  gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (data->textview7), &end, 0.0, FALSE, 0, 0);

  // Free memory allocated for message string.
  g_free (data->tr_message);

  return (FALSE);
}

// Start an IPv4 traceroute.
// This function will be executed as a separate thread.
gboolean
ipv4_tr_send (SPSData *data)
{
  int status, frame_length, offset, sendsd, recsd, probes;
  int bytes, timeout, node, trylim, trycount, done;
  char *src_ip, *dst_ip, *rec_ip;
  char hostname[NI_MAXHOST];
  struct ip *iphdr;
  struct tcphdr *tcphdr;
  struct icmp *icmphdr;
  unsigned char *snd_ether_frame, *rec_ether_frame;
  struct sockaddr_in sa, sin;
  struct sockaddr from;
  struct sockaddr_ll device;
  socklen_t fromlen;
  struct timeval wait, t1, t2;
  struct timezone tz;
  double dt;
  time_t rawtime;
  struct tm *timeinfo;
  char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  gchar *text;
  void *tmp;

  tmp = (char *) g_malloc0 (16 * sizeof (char));
  if (tmp != NULL) {
    rec_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rec_ip' in ipv4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    snd_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'snd_ether_frame' in ipv4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    rec_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rec_ether_frame' in ipv4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (16 * sizeof (char));
  if (tmp != NULL) {
    src_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'src_ip' in ipv4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (16 * sizeof (char));
  if (tmp != NULL) {
    dst_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'dst_ip' in ipv4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in ipv4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  // Make sure all interface names were specified.
  if (((data->packet_type_tr == 1) && (data->tcp4_tr_specify_ether == 1) &&
       (strlen (data->tcp4_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 2) && (data->icmp4_tr_specify_ether == 1) &&
       (strlen (data->icmp4_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 3) && (data->udp4_tr_specify_ether == 1) &&
       (strlen (data->udp4_tr_interfacename) < 2))) {
    sprintf (data->error_text, "In callbacks.c: ipv4_tr_send() failed.\nAppears to be an invalid interface name.");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  // Fill out sockaddr_ll struct.
  device.sll_family = AF_PACKET;  // Address family
  device.sll_halen = htons (6u);  // Length of address

  // Set source MAC address and resolve interface index from interface name.
  if ((data->packet_type_tr == 1) && (data->tcp4_tr_specify_ether == 1)) {
    memcpy (device.sll_addr, data->tcp4ethhdr_tr.src_mac, 6);
    if ((device.sll_ifindex = if_nametoindex (data->tcp4_tr_interfacename)) == 0) {
      status = errno;
    }
  } else if ((data->packet_type_tr == 2) && (data->icmp4_tr_specify_ether == 1)) {
    memcpy (device.sll_addr, data->icmp4ethhdr_tr.src_mac, 6);
    if ((device.sll_ifindex = if_nametoindex (data->icmp4_tr_interfacename)) == 0) {
      status = errno;
    }
  } else if ((data->packet_type_tr == 3) && (data->udp4_tr_specify_ether == 1)) {
    memcpy (device.sll_addr, data->udp4ethhdr_tr.src_mac, 6);
    if ((device.sll_ifindex = if_nametoindex (data->udp4_tr_interfacename)) == 0) {
      status = errno;
    }
  }
  if (device.sll_ifindex == 0) {
    sprintf (data->error_text, "In callbacks.c: if_nametoindex() failed in ipv4_tr_send() to obtain interface index.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  // For packets (no ethernet header):
  // The kernel will provide Layer 2 (data link layer) information (MAC addresses).
  // To do that, we need to specify a destination for the kernel in order for it
  // to decide where to send the raw datagram. We fill in a struct in_addr with
  // the desired destination IP address, and pass this structure to the sendto() function.
  sin.sin_family = AF_INET;
  memset (&sin, 0, sizeof (struct sockaddr_in));
  if (data->packet_type_tr == 1) {
    sin.sin_addr.s_addr = data->ip4hdr_tcp_tr.ip_dst.s_addr;
  } else if (data->packet_type_tr == 2) {
    sin.sin_addr.s_addr = data->ip4hdr_icmp_tr.ip_dst.s_addr;
  } else if (data->packet_type_tr == 3) {
    sin.sin_addr.s_addr = data->ip4hdr_udp_tr.ip_dst.s_addr;
  }

  // Grab system time.
  time (&rawtime);

  // Express time as per local timezone, and as broken-down time
  timeinfo = localtime (&rawtime);

  sprintf (text, "%i ", timeinfo->tm_mday);
  sprintf (text, "%s%.3s ", text, mon_name[timeinfo->tm_mon]);
  sprintf (text, "%s%i ", text, timeinfo->tm_year + 1900);
  sprintf (text, "%s%.2d:%.2d:%.2d\n", text, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);

  // Show target of traceroute.
  sa.sin_family = AF_INET;
  if (data->packet_type_tr == 1) {
    inet_ntop (AF_INET, &(data->ip4hdr_tcp_tr.ip_dst.s_addr), dst_ip, 16);
  } else if (data->packet_type_tr == 2) {
    inet_ntop (AF_INET, &(data->ip4hdr_icmp_tr.ip_dst.s_addr), dst_ip, 16);
  } else if (data->packet_type_tr == 3) {
    inet_ntop (AF_INET, &(data->ip4hdr_udp_tr.ip_dst.s_addr), dst_ip, 16);
  }
  inet_pton (AF_INET, dst_ip, &sa.sin_addr);

  if ((status = getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
    sprintf (data->error_text, "In callbacks.c: getnameinfo() failed in ipv4_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  memset (text, 0, 1024 * sizeof (gchar));
  if (data->packet_type_tr == 1) {
    sprintf (text, "%sTCP (IPv4) traceroute to %s (%s)\n", text, dst_ip, hostname);
  } else if (data->packet_type_tr == 2) {
    sprintf (text, "%sICMP (IPv4) traceroute to %s (%s)\n", text, dst_ip, hostname);
  } else if (data->packet_type_tr == 3) {
    sprintf (text, "%sUDP (IPv4) traceroute to %s (%s)\n", text, dst_ip ,hostname);
  }
  post_tr_message (text, data);

  // Submit request for raw socket descriptors - one to send, one to receive.
  if (((data->packet_type_tr == 1) && (data->tcp4_tr_specify_ether == 1)) ||
      ((data->packet_type_tr == 2) && (data->icmp4_tr_specify_ether == 1)) ||
      ((data->packet_type_tr == 3) && (data->udp4_tr_specify_ether == 1))) {
    if ((sendsd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      status = errno;
      sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
      return (FALSE);
    }
  } else {
    if ((sendsd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
      status = errno;
      sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
      return (FALSE);
    }
  }

  if ((recsd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
    status = errno;
    sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in ipv4_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  // Set maximum number of tries for a host before incrementing TTL and moving on.
  trylim = 3;

// LOOP: incrementing TTL each time, exiting when we get our target IP address.
  iphdr = (struct ip *) (rec_ether_frame + ETH_HDRLEN);
  icmphdr = (struct icmp *) (rec_ether_frame + ETH_HDRLEN + IP4_HDRLEN);
  tcphdr = (struct tcphdr *) (rec_ether_frame + ETH_HDRLEN + IP4_HDRLEN);
  done = 0;
  trycount = 0;
  node = 1;
  probes = 0;

// SEND LOOP
  for (;;) {

    if (tracing == 0) {  // Break out of Send loop if Stop button was pressed.
      done = 2;
      break;
    }

    // Create probe ethernet frames.
    memset (snd_ether_frame, 0, IP_MAXPACKET);
    if (data->packet_type_tr == 1) {
      create_tcp4_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IP header + TCP header)
      frame_length = 6 + 6 + 2 + IP4_HDRLEN + TCP_HDRLEN + data->tcp4_tr_payloadlen;
    } else if (data->packet_type_tr == 2) {
      create_icmp4_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IP header + ICMP header)
      frame_length = 6 + 6 + 2 + IP4_HDRLEN + ICMP_HDRLEN + data->icmp4_tr_payloadlen;
    } else if (data->packet_type_tr == 3) {
      create_udp4_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IP header + UDP header)
      frame_length = 6 + 6 + 2 + IP4_HDRLEN + UDP_HDRLEN + data->udp4_tr_payloadlen;
    }

    // If we're sending packets (i.e., letting the kernel fill out ethernet header), move pointer to IPv4 header.
    offset = 0;
    if (((data->packet_type_tr == 1) && (data->tcp4_tr_specify_ether == 0)) ||
       ((data->packet_type_tr == 2) && (data->icmp4_tr_specify_ether == 0)) ||
       ((data->packet_type_tr == 3) && (data->udp4_tr_specify_ether == 0))) {
      offset = 14;
    }

    // Send ethernet frame.
    if (((data->packet_type_tr == 1) && (data->tcp4_tr_specify_ether == 1)) ||
        ((data->packet_type_tr == 2) && (data->icmp4_tr_specify_ether == 1)) ||
        ((data->packet_type_tr == 3) && (data->udp4_tr_specify_ether == 1))) {
      if ((bytes = sendto (sendsd, snd_ether_frame, frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
        status = errno;
      }

    // Send a packet (no ethernet header) to socket.
    } else {
      if ((bytes = sendto (sendsd, snd_ether_frame + offset, frame_length - offset, 0, (struct sockaddr *) &sin, sizeof (sin))) <= 0) {
        status = errno;
      }
    }
    if (bytes <= 0) {
      sprintf (data->error_text, "In callbacks.c: sendto() failed in ipv4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
      return (FALSE);
    }

    probes++;

    // Start timer.
    (void) gettimeofday (&t1, &tz);

    // Set time for the socket to timeout and give up waiting for an ARP reply.
    timeout = data->timeout_tr;
    wait.tv_sec  = timeout;  
    wait.tv_usec = 0;
    if (setsockopt (recsd, SOL_SOCKET, SO_RCVTIMEO, (char *) &wait, sizeof (struct timeval)) < 0) {
      status = errno;
      sprintf (data->error_text, "In callbacks.c: setsockopt() failed to set timeout in ipv4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
      return (FALSE);
    }

    // Listen for incoming ethernet frame from socket sd.
    // We expect an ICMP ethernet frame of the form:
    //     MAC (6 bytes) + MAC (6 bytes) + ethernet type (2 bytes)
    //     + ethernet data (IP header + ICMP header + IP header + TCP/ICMP/UDP header)
    // Keep at it for data->timeout seconds, or until we get an ICMP reply.

    // RECEIVE LOOP
    for (;;) {

      memset (rec_ether_frame, 0, IP_MAXPACKET);
      memset (&from, 0, sizeof (from));
      fromlen = sizeof (from);
      if ((bytes = recvfrom (recsd, rec_ether_frame, IP_MAXPACKET, 0, (struct sockaddr *) &from, &fromlen)) < 0) {

        status = errno;

        // Deal with error conditions first.
        if (status == EAGAIN) {  // EAGAIN = 11
          memset (text, 0, 1024 * sizeof (gchar));
          sprintf (text, "  %i No reply within %i seconds.\n", node, timeout);
          post_tr_message (text, data);
          trycount++;
          break;  // Break out of Receive loop.
        } else if (status == EINTR) {  // EINTR = 4
          continue;  // Something weird happened, but let's keep listening.
        } else {
          sprintf (data->error_text, "In callbacks.c: recvfrom() failed in ipv4_tr_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          G_LOCK (message_available);
          message_available = 1;
          G_UNLOCK (message_available);
          G_LOCK (tracing);
          tracing = 0;
          G_UNLOCK (tracing);
          free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
          return (FALSE);
        }
      }  // End of error handling conditionals.

      // Check for an IP ethernet frame. If not, ignore and keep listening.
      if (((rec_ether_frame[12] << 8) + rec_ether_frame[13]) == ETH_P_IP) {

        // Did we get an ICMP_TIME_EXCEEDED?
        if ((iphdr->ip_p == IPPROTO_ICMP) && (icmphdr->icmp_type == ICMP_TIME_EXCEEDED)) {

          trycount = 0;
          // Stop timer and calculate how long it took to get a reply.
          (void) gettimeofday (&t2, &tz);
          dt = (double) (t2.tv_sec - t1.tv_sec) * 1000.0 + (double) (t2.tv_usec - t1.tv_usec) / 1000.0;

          // Extract source IP address from received ethernet frame.
          inet_ntop (AF_INET, &(iphdr->ip_src.s_addr), rec_ip, 16);

          // Report source IP address and time for reply.
          if (data->resolve_tr == 0) {
            memset (text, 0, 1024 * sizeof (gchar));
            sprintf (text, "%2i  %s  %g ms (%i bytes received)", node, rec_ip, dt, bytes);
          } else if (data->resolve_tr == 1) {
            sa.sin_family = AF_INET;
            if ((status = inet_pton (AF_INET, rec_ip, &sa.sin_addr)) != 1) {
              fprintf (stderr, "inet_pton() failed in ipv4_tr_send().\nError message: %s", strerror (status));
              exit (EXIT_FAILURE);
            }
            if ((status = getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
              fprintf (stderr, "getnameinfo() failed in ipv4_tr_send().\nError message: %s", strerror (status));
              exit (EXIT_FAILURE);
            }
            memset (text, 0, 1024 * sizeof (gchar));
            sprintf (text, "%2i  %s (%s)  %g ms (%i bytes received)", node, rec_ip, hostname, dt, bytes);
          }
          if (probes < data->num_probes) {
            sprintf (text, "%s : ", text);
            post_tr_message (text, data);
            break;  // Break out of Receive loop and probe this node again.
          } else {
            sprintf (text, "%s\n", text);
            post_tr_message (text, data);
            node++;
            probes = 0;
            break;  // Break out of Receive loop and probe next node in route.
          }
        }  // End of ICMP_TIME_EXCEEDED conditional.

        // Did we reach our destination?
        // TCP SYN-ACK means TCP SYN packet reached destination node.
        // ICMP echo reply means ICMP echo request packet reached destination node.
        // ICMP port unreachable means UDP packet reached destination node.
        if (((iphdr->ip_p == IPPROTO_TCP) && (tcphdr->th_flags == 18)) ||  // (18 = SYN, ACK)
            ((iphdr->ip_p == IPPROTO_ICMP) && (icmphdr->icmp_type == ICMP_ECHOREPLY) && (icmphdr->icmp_code == 0)) ||  // ECHO REPLY
            ((iphdr->ip_p == IPPROTO_ICMP) && (icmphdr->icmp_type == ICMP_PORT_UNREACH) && (icmphdr->icmp_code == 3))) {  // PORT UNREACHABLE

          // Stop timer and calculate how long it took to get a reply.
          (void) gettimeofday (&t2, &tz);
          dt = (double) (t2.tv_sec - t1.tv_sec) * 1000.0 + (double) (t2.tv_usec - t1.tv_usec) / 1000.0;

          // Extract source IP address from received ethernet frame.
          inet_ntop (AF_INET, &(iphdr->ip_src.s_addr), rec_ip, 16);

          // Report source IP address and time for reply.
          memset (text, 0, 1024 * sizeof (gchar));
          sprintf (text, "%2i  %s  %g ms", node, rec_ip, dt);
          if (probes < data->num_probes) {
            sprintf (text, "%s : ", text);
            post_tr_message (text, data);
            break;  // Break out of Receive loop and probe this node again.
          } else {
            sprintf (text, "%s\n", text);
            post_tr_message (text, data);
            done = 1;
            break;  // Break out of Receive loop and finish.
          }
        }  // End of Reached Destination conditional.
      }  // End of Was IP Frame conditional.
    }  // End of Receive loop.


    // Reached destination node or Stop button pressed.
    if ((done == 1) || (done == 2)) {
      break;  // Break out of Send loop.

    // Reached maxhops.
    } else if (node > data->maxhops) {
      memset (text, 0, 1024 * sizeof (gchar));
      sprintf (text, "Reached maximum number of hops. Maximum is set to %i hops.\n", data->maxhops);
      post_tr_message (text, data);
      done = 3;
      break;  // Break out of Send loop.
    }

    // We ran out of tries, let's move on to next node unless we reached maxhops limit.
    if (trycount == trylim) {
      memset (text, 0, 1024 * sizeof (gchar));
      sprintf (text, "%2i  Node won't respond after %i probes.\n", node, trylim);
      post_tr_message (text, data);
      node++;
      probes = 0;
      trycount = 0;
      continue;
    }

  }  // End of Send loop.

  if (done == 1) {
    memset (text, 0, 1024 * sizeof (gchar));
    sprintf (text, "End of traceroute.\n\n");
    post_tr_message (text, data);

  } else if ((done == 2) || (done == 3)) {
    memset (text, 0, 1024 * sizeof (gchar));
    sprintf (text, "Traceroute aborted.\n\n");
    post_tr_message (text, data);
  }

  // Close socket descriptors.
  close (sendsd);
  close (recsd);

  // Free allocated memory.
  free_ipv4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);

  // Clear tracing flag.
  G_LOCK (tracing);
  tracing = 0;
  G_UNLOCK (tracing);

  return (TRUE);
}

// Free allocated memory used in ipv4_tr_send().
gboolean
free_ipv4_tr_mem (char *rec_ip, unsigned char *snd_ether_frame, unsigned char *rec_ether_frame, char *src_ip, char *dst_ip, gchar *text)
{
  g_free (snd_ether_frame);
  g_free (rec_ether_frame);
  g_free (src_ip);
  g_free (dst_ip);
  g_free (rec_ip);
  g_free (text);

  return (TRUE);
}

// Start an IPv6 traceroute.
// This function will be executed as a separate thread.
gboolean
ipv6_tr_send (SPSData *data)
{
  int status, frame_length, sendsd, recsd, probes;
  int bytes, timeout, node, trylim, trycount, done;
  char *src_ip, *dst_ip, *rec_ip;
  char hostname[NI_MAXHOST];
  struct ip6_hdr *iphdr;
  struct tcphdr *tcphdr;
  struct icmp6_hdr *icmphdr;
  unsigned char *snd_ether_frame, *rec_ether_frame;
  struct sockaddr_in6 sa;
  struct sockaddr from;
  struct sockaddr_ll device;
  socklen_t fromlen;
  struct timeval wait, t1, t2;
  struct timezone tz;
  double dt;
  time_t rawtime;
  struct tm *timeinfo;
  char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  gchar *text;
  void *tmp;

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    rec_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rec_ip' in ipv6_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    snd_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'snd_ether_frame' in ipv6_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    rec_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rec_ether_frame' in ipv6_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    src_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'src_ip' in ipv6_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    dst_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'dst_ip' in ipv6_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in ipv6_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  // Make sure all interface names were specified.
  if (((data->packet_type_tr == 4) && (strlen (data->tcp6_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 5) && (strlen (data->icmp6_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 6) && (strlen (data->udp6_tr_interfacename) < 2))) {
    sprintf (data->error_text, "In callbacks.c: ipv6_tr_send() failed.\nAppears to be an invalid interface name.");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  // Resolve interface index.
  if (data->packet_type_tr == 4) {
    if ((device.sll_ifindex = if_nametoindex (data->tcp6_tr_interfacename)) == 0) {
      status = errno;
    }
  } else if (data->packet_type_tr == 5) {
    if ((device.sll_ifindex = if_nametoindex (data->icmp6_tr_interfacename)) == 0) {
      status = errno;
    }
  } else if (data->packet_type_tr == 6) {
    if ((device.sll_ifindex = if_nametoindex (data->udp6_tr_interfacename)) == 0) {
      status = errno;
    }
  }
  if (device.sll_ifindex == 0) {
    sprintf (data->error_text, "In callbacks.c: if_nametoindex() failed to obtain interface index in ipv6_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  // Fill out sockaddr_ll struct.
  device.sll_family = AF_PACKET;
  if (data->packet_type_tr == 4) {
    memcpy (device.sll_addr, data->tcp6ethhdr_tr.src_mac, 6);
  } else if (data->packet_type_tr == 5) {
    memcpy (device.sll_addr, data->icmp6ethhdr_tr.src_mac, 6);
  } else if (data->packet_type_tr == 6) {
    memcpy (device.sll_addr, data->udp6ethhdr_tr.src_mac, 6);
  }
  device.sll_halen = htons (6u);

  // Grab system time.
  time (&rawtime);

  // Express time as per local timezone, and as broken-down time
  timeinfo = localtime (&rawtime);

  sprintf (text, "%i ", timeinfo->tm_mday);
  sprintf (text, "%s%.3s ", text, mon_name[timeinfo->tm_mon]);
  sprintf (text, "%s%i ", text, timeinfo->tm_year + 1900);
  sprintf (text, "%s%.2d:%.2d:%.2d\n", text, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);

  // Show target of traceroute.
  sa.sin6_family = AF_INET6;
  if (data->packet_type_tr == 4) {
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp_tr.ip6_dst), dst_ip, 40);
  } else if (data->packet_type_tr == 5) {
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp_tr.ip6_dst), dst_ip, 40);
  } else if (data->packet_type_tr == 6) {
    inet_ntop (AF_INET6, &(data->ip6hdr_udp_tr.ip6_dst), dst_ip, 40);
  }
  inet_pton (AF_INET6, dst_ip, &sa.sin6_addr);

  if ((status = getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
    sprintf (data->error_text, "In callbacks.c: getnameinfo() failed in ipv6_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  if (data->packet_type_tr == 4) {
    sprintf (text, "%sTCP (IPv6) traceroute to %s (%s)\n", text, dst_ip, hostname);
  } else if (data->packet_type_tr == 5) {
    sprintf (text, "%sICMP (IPv6) traceroute to %s (%s)\n", text, dst_ip, hostname);
  } else if (data->packet_type_tr == 6) {
    sprintf (text, "%sUDP (IPv6) traceroute to %s (%s)\n", text, dst_ip ,hostname);
  }
  post_tr_message (text, data);

  // Submit request for raw socket descriptors - one to send, one to receive.
  if ((sendsd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
    status = errno;
    sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in ipv6_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  if ((recsd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
    status = errno;
    sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in ipv6_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
    return (FALSE);
  }

  // Set maximum number of tries for a host before incrementing TTL and moving on.
  trylim = 3;

  // LOOP: incrementing TTL each time, exiting when we get our target IP address.
  iphdr = (struct ip6_hdr *) (rec_ether_frame + ETH_HDRLEN);
  icmphdr = (struct icmp6_hdr *) (rec_ether_frame + ETH_HDRLEN + IP6_HDRLEN);
  tcphdr = (struct tcphdr *) (rec_ether_frame + ETH_HDRLEN + IP6_HDRLEN);
  done = 0;
  trycount = 0;
  node = 1;
  probes = 0;

  // SEND LOOP
  for (;;) {

    if (tracing == 0) {  // Break out of Send loop if Stop button was pressed.
      done = 2;
      break;
    }

    // Create probe ethernet frames.
    memset (snd_ether_frame, 0, IP_MAXPACKET);
    if (data->packet_type_tr == 4) {
      create_tcp6_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv6 header + TCP header)
      frame_length = 6 + 6 + 2 + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_tr_payloadlen;
    } else if (data->packet_type_tr == 5) {
      create_icmp6_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv6 header + ICMP header)
      frame_length = 6 + 6 + 2 + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_tr_payloadlen;
    } else if (data->packet_type_tr == 6) {
      create_udp6_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv6 header + UDP header)
      frame_length = 6 + 6 + 2 + IP6_HDRLEN + UDP_HDRLEN + data->udp6_tr_payloadlen;
    }

    // Send ethernet frame.
    if ((bytes = sendto (sendsd, snd_ether_frame, frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
      status = errno;
      sprintf (data->error_text, "In callbacks.c: sendto() failed in ipv6_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
      return (FALSE);
    }

    probes++;

    // Start timer.
    (void) gettimeofday (&t1, &tz);

    // Set time for the socket to timeout and give up waiting for an ARP reply.
    timeout = data->timeout_tr;
    wait.tv_sec  = timeout;  
    wait.tv_usec = 0;
    setsockopt (recsd, SOL_SOCKET, SO_RCVTIMEO, (char *) &wait, sizeof (struct timeval));

    // Listen for incoming ethernet frame from socket sd.
    // We expect an ICMP ethernet frame of the form:
    //     MAC (6 bytes) + MAC (6 bytes) + ethernet type (2 bytes)
    //     + ethernet data (IPv6 header + ICMP header + IPv6 header + TCP/ICMP/UDP header)
    // Keep at it for data->timeout seconds, or until we get an ICMP reply.

    // RECEIVE LOOP
    for (;;) {

      memset (rec_ether_frame, 0, IP_MAXPACKET);
      memset (&from, 0, sizeof (from));
      fromlen = sizeof (from);
      if ((bytes = recvfrom (recsd, rec_ether_frame, IP_MAXPACKET, 0, (struct sockaddr *) &from, &fromlen)) < 0) {

        status = errno;

        // Deal with error conditions first.
        if (status == EAGAIN) {  // EAGAIN = 11
          memset (text, 0, 1024 * sizeof (gchar));
          sprintf (text, "  %i No reply within %i seconds.\n", node, timeout);
          post_tr_message (text, data);
          trycount++;
          break;  // Break out of Receive loop.
        } else if (status == EINTR) {  // EINTR = 4
          continue;  // Something weird happened, but let's keep listening.
        } else {
          sprintf (data->error_text, "In callbacks.c: recvfrom() failed in ipv6_tr_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          G_LOCK (message_available);
          message_available = 1;
          G_UNLOCK (message_available);
          G_LOCK (tracing);
          tracing = 0;
          G_UNLOCK (tracing);
          free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);
          return (FALSE);
        }
      }  // End of error handling conditionals.

      // Check for an IP ethernet frame. If not, ignore and keep listening.
      if (((rec_ether_frame[12] << 8) + rec_ether_frame[13]) == ETH_P_IPV6) {

        // Did we get an ICMP6_TIME_EXCEEDED?
        if ((iphdr->ip6_nxt == IPPROTO_ICMPV6) && (icmphdr->icmp6_type == ICMP6_TIME_EXCEEDED)) {

          trycount = 0;
          // Stop timer and calculate how long it took to get a reply.
          (void) gettimeofday (&t2, &tz);
          dt = (double) (t2.tv_sec - t1.tv_sec) * 1000.0 + (double) (t2.tv_usec - t1.tv_usec) / 1000.0;

          // Extract source IP address from received ethernet frame.
          inet_ntop (AF_INET6, &(iphdr->ip6_src), rec_ip, 40);

          // Report source IP address and time for reply.
          if (data->resolve_tr == 0) {
            memset (text, 0, 1024 * sizeof (gchar));
            sprintf (text, "%2i  %s  %g ms (%i bytes received)", node, rec_ip, dt, bytes);
          } else if (data->resolve_tr == 1) {
            sa.sin6_family = AF_INET6;
            if ((status = inet_pton (AF_INET6, rec_ip, &sa.sin6_addr)) < 1) {
              fprintf (stderr, "inet_pton() failed.\nError message: %s", strerror (status));
              exit (EXIT_FAILURE);
            }
            if ((status = getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
              fprintf (stderr, "getnameinfo() failed.\nError message: %s", strerror (status));
              exit (EXIT_FAILURE);
            }
            memset (text, 0, 1024 * sizeof (gchar));
            sprintf (text, "%2i  %s (%s)  %g ms (%i bytes received)", node, rec_ip, hostname, dt, bytes);
          }
          if (probes < data->num_probes) {
            sprintf (text, "%s : ", text);
            post_tr_message (text, data);
            break;  // Break out of Receive loop and probe this node again.
          } else {
            sprintf (text, "%s\n", text);
            post_tr_message (text, data);
            node++;
            probes = 0;
            break;  // Break out of Receive loop and probe next node in route.
          }
        }  // End of ICMP6_TIME_EXCEEDED conditional.

        // Did we reach our destination?
        // TCP SYN-ACK means TCP SYN packet reached destination node.
        // ICMP echo reply means ICMP echo request packet reached destination node.
        // ICMP port unreachable means UDP packet reached destination node.
        if (((iphdr->ip6_nxt == IPPROTO_TCP) && (tcphdr->th_flags == 18)) ||  // (18 = SYN, ACK)
            ((iphdr->ip6_nxt == IPPROTO_ICMPV6) && (icmphdr->icmp6_type == ICMP6_ECHO_REPLY) && (icmphdr->icmp6_code == 0)) ||  // ECHO REPLY
            ((iphdr->ip6_nxt == IPPROTO_ICMPV6) && (icmphdr->icmp6_type == ICMP6_DST_UNREACH) && (icmphdr->icmp6_code == ICMP6_DST_UNREACH_NOPORT))) {  // PORT UNREACHABLE

          // Stop timer and calculate how long it took to get a reply.
          (void) gettimeofday (&t2, &tz);
          dt = (double) (t2.tv_sec - t1.tv_sec) * 1000.0 + (double) (t2.tv_usec - t1.tv_usec) / 1000.0;

          // Extract source IP address from received ethernet frame.
          inet_ntop (AF_INET6, &(iphdr->ip6_src), rec_ip, 40);

          // Report source IP address and time for reply.
          memset (text, 0, 1024 * sizeof (gchar));
          sprintf (text, "%2i  %s  %g ms", node, rec_ip, dt);
          if (probes < data->num_probes) {
            sprintf (text, "%s : ", text);
            post_tr_message (text, data);
            break;  // Break out of Receive loop and probe this node again.
          } else {
            sprintf (text, "%s\n", text);
            post_tr_message (text, data);
            done = 1;
            break;  // Break out of Receive loop and finish.
          }
        }  // End of Reached Destination conditional.
      }  // End of Was IP Frame conditional.
    }  // End of Receive loop.


    // Reached destination node or Stop button pressed.
    if ((done == 1) || (done == 2)) {
      break;  // Break out of Send loop.

    // Reached maxhops.
    } else if (node > data->maxhops) {
      memset (text, 0, 1024 * sizeof (gchar));
      sprintf (text, "Reached maximum number of hops. Maximum is set to %i hops.\n", data->maxhops);
      post_tr_message (text, data);
      done = 3;
      break;  // Break out of Send loop.
    }

    // We ran out of tries, let's move on to next node unless we reached maxhops limit.
    if (trycount == trylim) {
      memset (text, 0, 1024 * sizeof (gchar));
      sprintf (text, "%2i  Node won't respond after %i probes.\n", node, trylim);
      post_tr_message (text, data);
      node++;
      probes = 0;
      trycount = 0;
      continue;
    }

  }  // End of Send loop.

  if (done == 1) {
    memset (text, 0, 1024 * sizeof (gchar));
    sprintf (text, "End of traceroute.\n\n");
    post_tr_message (text, data);

  } else if ((done == 2) || (done == 3)) {
    memset (text, 0, 1024 * sizeof (gchar));
    sprintf (text, "Traceroute aborted.\n\n");
    post_tr_message (text, data);
  }

  // Close socket descriptors.
  close (sendsd);
  close (recsd);

  // Free allocated memory.
  free_ipv6_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src_ip, dst_ip, text);

  // Clear tracing flag.
  G_LOCK (tracing);
  tracing = 0;
  G_UNLOCK (tracing);

  return (TRUE);
}

// Free allocated memory used in ipv6_tr_send().
gboolean
free_ipv6_tr_mem (char *rec_ip, unsigned char *snd_ether_frame, unsigned char *rec_ether_frame, char *src_ip, char *dst_ip, gchar *text)
{
  g_free (snd_ether_frame);
  g_free (rec_ether_frame);
  g_free (src_ip);
  g_free (dst_ip);
  g_free (rec_ip);
  g_free (text);

  return (TRUE);
}

// Start an IPv6 over IPv4 (6to4) traceroute.
// This function will be executed as a separate thread.
gboolean
sixto4_tr_send (SPSData *data)
{
  int status, frame_length, offset, sendsd, recsd, probes;
  int bytes, timeout, node, trylim, trycount, done;
  char *src4_ip, *src6_ip, *dst_ip, *rec_ip;
  char hostname[NI_MAXHOST];
  struct ip6_hdr *ip6hdr;
  struct tcphdr *tcphdr;
  struct icmp6_hdr *icmphdr;
  unsigned char *snd_ether_frame, *rec_ether_frame;
  struct sockaddr_in sin;
  struct sockaddr_in6 sa;
  struct sockaddr from;
  struct sockaddr_ll device;
  socklen_t fromlen;
  struct timeval wait, t1, t2;
  struct timezone tz;
  double dt;
  time_t rawtime;
  struct tm *timeinfo;
  char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  gchar *text;
  void *tmp;

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    rec_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rec_ip' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    snd_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'snd_ether_frame' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (unsigned char *) g_malloc0 (IP_MAXPACKET * sizeof (unsigned char));
  if (tmp != NULL) {
    rec_ether_frame = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'rec_ether_frame' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (16 * sizeof (char));
  if (tmp != NULL) {
    src4_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'src4_ip' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    src6_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'src6_ip' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (char *) g_malloc0 (40 * sizeof (char));
  if (tmp != NULL) {
    dst_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'dst_ip' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in sixto4_tr_send().\n");
    exit (EXIT_FAILURE);
  }

  // Make sure all interface names were specified.
  if (((data->sixto4_tr_specify_ether == 1) && (strlen (data->sixto4_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 4) && (strlen (data->tcp6_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 5) && (strlen (data->icmp6_tr_interfacename) < 2)) ||
      ((data->packet_type_tr == 6) && (strlen (data->udp6_tr_interfacename) < 2))) {
    sprintf (data->error_text, "In callbacks.c: sixto4_tr_send() failed.\nAppears to be an invalid interface name.");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
    return (FALSE);
  }

  // Fill out sockaddr_ll struct.
  device.sll_family = AF_PACKET;  // Address family
  device.sll_halen = htons (6u);  // Length of address

  // Set source MAC address and resolve interface index from interface name.
  if (data->sixto4_tr_specify_ether == 1) {
    memcpy (device.sll_addr, data->sixto4ethhdr_tr.src_mac, 6);
    if ((device.sll_ifindex = if_nametoindex (data->sixto4_tr_interfacename)) == 0) {
      status = errno;
    }
  }
  if (device.sll_ifindex == 0) {
    sprintf (data->error_text, "In callbacks.c: if_nametoindex() failed in sixto4_tr_send() to obtain interface index.\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
    return (FALSE);
  }

  // For packets (no ethernet header):
  // The kernel will provide Layer 2 (data link layer) information (MAC addresses).
  // To do that, we need to specify a destination for the kernel in order for it
  // to decide where to send the raw datagram. We fill in a struct in_addr with
  // the desired destination IP address, and pass this structure to the sendto() function.
  // Here, regardless of IPv6 packet type, we're sending to the IPv4 6to4 anycast address
  // which is 192.88.99.1. We arbitrarily grab it from the TCP packet's ip4hdr.
  memset (&sin, 0, sizeof (struct sockaddr_in));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = data->sixto4ip4hdr_tcp.ip_dst.s_addr;

  // Grab system time.
  time (&rawtime);

  // Express time as per local timezone, and as broken-down time
  timeinfo = localtime (&rawtime);

  sprintf (text, "%i ", timeinfo->tm_mday);
  sprintf (text, "%s%.3s ", text, mon_name[timeinfo->tm_mon]);
  sprintf (text, "%s%i ", text, timeinfo->tm_year + 1900);
  sprintf (text, "%s%.2d:%.2d:%.2d\n", text, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);

  // Show target of traceroute.
  sa.sin6_family = AF_INET6;
  if (data->packet_type_tr == 4) {
    inet_ntop (AF_INET6, &(data->ip6hdr_tcp_tr.ip6_dst), dst_ip, 40);
  } else if (data->packet_type_tr == 5) {
    inet_ntop (AF_INET6, &(data->ip6hdr_icmp_tr.ip6_dst), dst_ip, 40);
  } else if (data->packet_type_tr == 6) {
    inet_ntop (AF_INET6, &(data->ip6hdr_udp_tr.ip6_dst), dst_ip, 40);
  }
  inet_pton (AF_INET6, dst_ip, &sa.sin6_addr);

  if ((status = getnameinfo ((struct sockaddr *)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
    sprintf (data->error_text, "In callbacks.c: getnameinfo() failed in sixto4_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
    return (FALSE);
  }

  if (data->packet_type_tr == 4) {
    sprintf (text, "%sTCP (IPv6 over IPv4) traceroute to %s (%s)\n", text, dst_ip, hostname);
  } else if (data->packet_type_tr == 5) {
    sprintf (text, "%sICMP (IPv6 over IPv4) traceroute to %s (%s)\n", text, dst_ip, hostname);
  } else if (data->packet_type_tr == 6) {
    sprintf (text, "%sUDP (IPv6 over IPv4) traceroute to %s (%s)\n", text, dst_ip ,hostname);
  }
  post_tr_message (text, data);

  // Submit request for raw socket descriptors - one to send, one to receive.
  if (data->sixto4_tr_specify_ether == 1) {
    if ((sendsd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
      status = errno;
      sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in sixto4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
      return (FALSE);
    }
  } else {
    if ((sendsd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
      status = errno;
      sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in sixto4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
      return (FALSE);
    }
  }

  if ((recsd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0) {
    status = errno;
    sprintf (data->error_text, "In callbacks.c: socket() failed to get a socket descriptor in sixto4_tr_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    G_LOCK (tracing);
    tracing = 0;
    G_UNLOCK (tracing);
    free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
    return (FALSE);
  }

  // Set maximum number of tries for a host before incrementing TTL and moving on.
  trylim = 3;

  // LOOP: incrementing TTL each time, exiting when we get our target IP address.
  ip6hdr = (struct ip6_hdr *) (rec_ether_frame + ETH_HDRLEN + IP4_HDRLEN);
  icmphdr = (struct icmp6_hdr *) (rec_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN);
  tcphdr = (struct tcphdr *) (rec_ether_frame + ETH_HDRLEN + IP4_HDRLEN + IP6_HDRLEN);
  done = 0;
  trycount = 0;
  node = 1;
  probes = 0;

  // SEND LOOP
  for (;;) {

    if (tracing == 0) {  // Break out of Send loop if Stop button was pressed.
      done = 2;
      break;
    }

    // Create probe ethernet frames.
    memset (snd_ether_frame, 0, IP_MAXPACKET);
    if (data->packet_type_tr == 4) {
      create_tcp6_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv4 header + IPv6 header + TCP header)
      frame_length = 6 + 6 + 2 + IP4_HDRLEN + IP6_HDRLEN + TCP_HDRLEN + data->tcp6_tr_payloadlen;
    } else if (data->packet_type_tr == 5) {
      create_icmp6_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv4 header + IPv6 header + ICMP header)
      frame_length = 6 + 6 + 2 + IP4_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + data->icmp6_tr_payloadlen;
    } else if (data->packet_type_tr == 6) {
      create_udp6_frame (snd_ether_frame, node, data);
      // Ethernet frame length = ethernet header (MAC + MAC + ethernet type) + ethernet data (IPv4 header + IPv6 header + UDP header)
      frame_length = 6 + 6 + 2 + IP4_HDRLEN + IP6_HDRLEN + UDP_HDRLEN + data->udp6_tr_payloadlen;
    }

    // If we're sending packets (i.e., letting the kernel fill out ethernet header), move pointer to IPv4 header.
    offset = 0;
    if (data->sixto4_tr_specify_ether == 0) {
      offset = 14;
    }

    // Send ethernet frame.
    if (data->sixto4_tr_specify_ether == 1) {
      if ((bytes = sendto (sendsd, snd_ether_frame, frame_length, 0, (struct sockaddr *) &device, sizeof (device))) <= 0) {
        status = errno;
      }

    // Send a packet (no ethernet header) to socket.
    } else {
      if ((bytes = sendto (sendsd, snd_ether_frame + offset, frame_length - offset, 0, (struct sockaddr *) &sin, sizeof (sin))) <= 0) {
        status = errno;
      }
    }
    if (bytes <= 0) {
      sprintf (data->error_text, "In callbacks.c: sendto() failed in sixto4_tr_send().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      G_LOCK (tracing);
      tracing = 0;
      G_UNLOCK (tracing);
      free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
      return (FALSE);
    }

    probes++;

    // Start timer.
    (void) gettimeofday (&t1, &tz);

    // Set time for the socket to timeout and give up waiting for an ARP reply.
    timeout = data->timeout_tr;
    wait.tv_sec  = timeout;  
    wait.tv_usec = 0;
    setsockopt (recsd, SOL_SOCKET, SO_RCVTIMEO, (char *) &wait, sizeof (struct timeval));

    // Listen for incoming ethernet frame from socket sd.
    // We expect an ICMP ethernet frame of the form:
    //     MAC (6 bytes) + MAC (6 bytes) + ethernet type (2 bytes)
    //     + ethernet data (IPv4 header + IPv6 header + ICMP header + IPv6 header + TCP/ICMP/UDP header)
    // Keep at it for data->timeout seconds, or until we get an ICMP reply.

    // RECEIVE LOOP
    for (;;) {

      memset (rec_ether_frame, 0, IP_MAXPACKET);
      memset (&from, 0, sizeof (from));
      fromlen = sizeof (from);
      if ((bytes = recvfrom (recsd, rec_ether_frame, IP_MAXPACKET, 0, (struct sockaddr *) &from, &fromlen)) < 0) {

        status = errno;

        // Deal with error conditions first.
        if (status == EAGAIN) {  // EAGAIN = 11
          memset (text, 0, 1024 * sizeof (gchar));
          sprintf (text, "  %i No reply within %i seconds.\n", node, timeout);
          post_tr_message (text, data);
          trycount++;
          break;  // Break out of Receive loop.
        } else if (status == EINTR) {  // EINTR = 4
          continue;  // Something weird happened, but let's keep listening.
        } else {
          sprintf (data->error_text, "In callbacks.c: recvfrom() failed in sixto4_tr_send().\nError message: %s", strerror (status));
          data->parent = data->main_window;
          G_LOCK (message_available);
          message_available = 1;
          G_UNLOCK (message_available);
          G_LOCK (tracing);
          tracing = 0;
          G_UNLOCK (tracing);
          free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);
          return (FALSE);
        }
      }  // End of error handling conditionals.

      // Check for an IP ethernet frame. If not, ignore and keep listening.
      if (((rec_ether_frame[12] << 8) + rec_ether_frame[13]) == ETH_P_IP) {

        // Did we get an ICMP6_TIME_EXCEEDED?
        if ((ip6hdr->ip6_nxt == IPPROTO_ICMPV6) && (icmphdr->icmp6_type == ICMP6_TIME_EXCEEDED)) {

          trycount = 0;
          // Stop timer and calculate how long it took to get a reply.
          (void) gettimeofday (&t2, &tz);
          dt = (double) (t2.tv_sec - t1.tv_sec) * 1000.0 + (double) (t2.tv_usec - t1.tv_usec) / 1000.0;

          // Extract source IP address from received ethernet frame.
          inet_ntop (AF_INET6, &(ip6hdr->ip6_src), rec_ip, 40);

          // Report source IP address and time for reply.
          if (data->resolve_tr == 0) {
            memset (text, 0, 1024 * sizeof (gchar));
            sprintf (text, "%2i  %s  %g ms (%i bytes received)", node, rec_ip, dt, bytes);
          } else if (data->resolve_tr == 1) {
            sa.sin6_family = AF_INET6;
            if ((status = inet_pton (AF_INET6, rec_ip, &sa.sin6_addr)) != 1) {
              fprintf (stderr, "inet_pton() failed.\nError message: %s", strerror (status));
              exit (EXIT_FAILURE);
            }
            if ((status = getnameinfo ((struct sockaddr*)&sa, sizeof (sa), hostname, sizeof (hostname), NULL, 0, 0)) != 0) {
              fprintf (stderr, "getnameinfo() failed.\nError message: %s", strerror (status));
              exit (EXIT_FAILURE);
            }
            memset (text, 0, 1024 * sizeof (gchar));
            sprintf (text, "%2i  %s (%s)  %g ms (%i bytes received)", node, rec_ip, hostname, dt, bytes);
          }
          if (probes < data->num_probes) {
            sprintf (text, "%s : ", text);
            post_tr_message (text, data);
            break;  // Break out of Receive loop and probe this node again.
          } else {
            sprintf (text, "%s\n", text);
            post_tr_message (text, data);
            node++;
            probes = 0;
            break;  // Break out of Receive loop and probe next node in route.
          }
        }  // End of ICMP6_TIME_EXCEEDED conditional.

        // Did we reach our destination?
        // TCP SYN-ACK means TCP SYN packet reached destination node.
        // ICMP echo reply means ICMP echo request packet reached destination node.
        // ICMP port unreachable means UDP packet reached destination node.
        if (((ip6hdr->ip6_nxt == IPPROTO_TCP) && (tcphdr->th_flags == 18)) ||  // (18 = SYN, ACK)
            ((ip6hdr->ip6_nxt == IPPROTO_ICMPV6) && (icmphdr->icmp6_type == ICMP6_ECHO_REPLY) && (icmphdr->icmp6_code == 0)) ||  // ECHO REPLY
            ((ip6hdr->ip6_nxt == IPPROTO_ICMPV6) && (icmphdr->icmp6_type == ICMP6_DST_UNREACH) && (icmphdr->icmp6_code == ICMP6_DST_UNREACH_NOPORT))) {  // PORT UNREACHABLE

          // Stop timer and calculate how long it took to get a reply.
          (void) gettimeofday (&t2, &tz);
          dt = (double) (t2.tv_sec - t1.tv_sec) * 1000.0 + (double) (t2.tv_usec - t1.tv_usec) / 1000.0;

          // Extract source IP address from received ethernet frame.
          inet_ntop (AF_INET6, &(ip6hdr->ip6_src), rec_ip, 40);

          // Report source IP address and time for reply.
          memset (text, 0, 1024 * sizeof (gchar));
          sprintf (text, "%2i  %s  %g ms", node, rec_ip, dt);
          if (probes < data->num_probes) {
            sprintf (text, "%s : ", text);
            post_tr_message (text, data);
            break;  // Break out of Receive loop and probe this node again.
          } else {
            sprintf (text, "%s\n", text);
            post_tr_message (text, data);
            done = 1;
            break;  // Break out of Receive loop and finish.
          }
        }  // End of Reached Destination conditional.
      }  // End of Was IP Frame conditional.
    }  // End of Receive loop.


    // Reached destination node or Stop button pressed.
    if ((done == 1) || (done == 2)) {
      break;  // Break out of Send loop.

    // Reached maxhops.
    } else if (node > data->maxhops) {
      memset (text, 0, 1024 * sizeof (gchar));
      sprintf (text, "Reached maximum number of hops. Maximum is set to %i hops.\n", data->maxhops);
      post_tr_message (text, data);
      done = 3;
      break;  // Break out of Send loop.
    }

    // We ran out of tries, let's move on to next node unless we reached maxhops limit.
    if (trycount == trylim) {
      memset (text, 0, 1024 * sizeof (gchar));
      sprintf (text, "%2i  Node won't respond after %i probes.\n", node, trylim);
      post_tr_message (text, data);
      node++;
      probes = 0;
      trycount = 0;
      continue;
    }
  }  // End of Send loop.

  if (done == 1) {
    memset (text, 0, 1024 * sizeof (gchar));
    sprintf (text, "End of traceroute.\n\n");
    post_tr_message (text, data);

  } else if ((done == 2) || (done == 3)) {
    memset (text, 0, 1024 * sizeof (gchar));
    sprintf (text, "Traceroute aborted.\n\n");
    post_tr_message (text, data);
  }

  // Close socket descriptors.
  close (sendsd);
  close (recsd);

  // Free allocated memory.
  free_sixto4_tr_mem (rec_ip, snd_ether_frame, rec_ether_frame, src4_ip, src6_ip, dst_ip, text);

  // Clear tracing flag.
  G_LOCK (tracing);
  tracing = 0;
  G_UNLOCK (tracing);

  return (TRUE);
}

// Free allocated memory used in sixto4_tr_send().
gboolean
free_sixto4_tr_mem (char *rec_ip, unsigned char *snd_ether_frame, unsigned char *rec_ether_frame, char *src4_ip, char *src6_ip, char *dst_ip, gchar *text)
{
  g_free (snd_ether_frame);
  g_free (rec_ether_frame);
  g_free (src4_ip);
  g_free (src6_ip);
  g_free (dst_ip);
  g_free (rec_ip);
  g_free (text);

  return (TRUE);
}

// Set traceroute mode (0 = terminate traceroute)
gboolean
on_button25_clicked (GtkButton *button, SPSData *data)
{
  tracing = 0;

  return (TRUE);
}

// Create IPv4 TCP ethernet frame for traceroute.
gboolean
create_tcp4_frame (unsigned char *snd_ether_frame, int ttl, SPSData *data)
{
  // Time-to-Live (8 bits)
  data->ip4hdr_tcp_tr.ip_ttl = ttl;

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_tcp_tr.ip_sum = 0;
  data->ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_tcp_tr, IP4_HDRLEN);

  // TCP checksum (16 bits)
  data->tcp4hdr_tr.th_sum = tcp4_checksum (data->ip4hdr_tcp_tr, data->tcp4hdr_tr, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);

  // Fill out ethernet frame header.

  // Destination MAC address
  memcpy (snd_ether_frame, data->tcp4ethhdr_tr.dst_mac, 6);

  // Source MAC address
  memcpy (snd_ether_frame + 6, data->tcp4ethhdr_tr.src_mac, 6);

  // Next is ethernet type code.
  // http://www.iana.org/assignments/ethernet-numbers
  memcpy (&snd_ether_frame[12], &data->tcp4ethhdr_tr.type_code, 2);

  // Next is ethernet frame data (IPv4 header + TCP header + TCP data).

  // IPv4 header
  memcpy (snd_ether_frame + 14, &data->ip4hdr_tcp_tr, IP4_HDRLEN);

  // TCP header
  memcpy (snd_ether_frame + 14 + IP4_HDRLEN, &data->tcp4hdr_tr, TCP_HDRLEN);

  // TCP data
  memcpy (snd_ether_frame + 14 + IP4_HDRLEN + TCP_HDRLEN, data->tcp4_tr_payload, data->tcp4_tr_payloadlen);

  return (TRUE);
}

// Create IPv4 ICMP ethernet frame for traceroute.
gboolean
create_icmp4_frame (unsigned char *snd_ether_frame, int ttl, SPSData *data)
{
  // Time-to-Live (8 bits)
  data->ip4hdr_icmp_tr.ip_ttl = ttl;

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_icmp_tr.ip_sum = 0;
  data->ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_icmp_tr, IP4_HDRLEN);

  // ICMP header checksum (16 bits)
  data->icmp4hdr_tr.icmp_cksum = icmp4_checksum (data->icmp4hdr_tr, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);

  // Fill out ethernet frame header.

  // Destination MAC address
  memcpy (snd_ether_frame, data->icmp4ethhdr_tr.dst_mac, 6);

  // Source MAC address
  memcpy (snd_ether_frame + 6, data->icmp4ethhdr_tr.src_mac, 6);

  // Next is ethernet type code.
  // http://www.iana.org/assignments/ethernet-numbers
  memcpy (&snd_ether_frame[12], &data->icmp4ethhdr_tr.type_code, 2);

  // Next is ethernet frame data (IPv4 header + ICMP header + ICMP data).

  // IPv4 header
  memcpy (snd_ether_frame + 14, &data->ip4hdr_icmp_tr, IP4_HDRLEN);

  // ICMP header
  memcpy (snd_ether_frame + 14 + IP4_HDRLEN, &data->icmp4hdr_tr, ICMP_HDRLEN);

  // ICMP data
  memcpy (snd_ether_frame + 14 + IP4_HDRLEN + ICMP_HDRLEN, data->icmp4_tr_payload, data->icmp4_tr_payloadlen);

  return (TRUE);
}

// Create IPv4 UDP ethernet frame for traceroute.
gboolean
create_udp4_frame (unsigned char *snd_ether_frame, int ttl, SPSData *data)
{
  // Time-to-Live (8 bits)
  data->ip4hdr_udp_tr.ip_ttl = ttl;

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->ip4hdr_udp_tr.ip_sum = 0;
  data->ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->ip4hdr_udp_tr, IP4_HDRLEN);

  // UDP checksum (16 bits)
  data->udp4hdr_tr.uh_sum = udp4_checksum (data->ip4hdr_udp_tr, data->udp4hdr_tr, data->udp4_tr_payload, data->udp4_tr_payloadlen);

  // Fill out ethernet frame header.

  // Destination MAC address
  memcpy (snd_ether_frame, data->udp4ethhdr_tr.dst_mac, 6);

  // Source MAC address
  memcpy (snd_ether_frame + 6, data->udp4ethhdr_tr.src_mac, 6);

  // Next is ethernet type code.
  // http://www.iana.org/assignments/ethernet-numbers
  memcpy (&snd_ether_frame[12], &data->udp4ethhdr_tr.type_code, 2);

  // Next is ethernet frame data (IPv4 header + UDP header + UDP data).
  // IPv4 header
  memcpy (snd_ether_frame + 14, &data->ip4hdr_udp_tr, IP4_HDRLEN);

  // UDP header
  memcpy (snd_ether_frame + 14 + IP4_HDRLEN, &data->udp4hdr_tr, UDP_HDRLEN);

  // UDP data
  memcpy (snd_ether_frame + 14 + IP4_HDRLEN + UDP_HDRLEN, data->udp4_tr_payload, data->udp4_tr_payloadlen);

  return (TRUE);
}

// Create IPv6 TCP ethernet frame for traceroute.
gboolean
create_tcp6_frame (unsigned char *snd_ether_frame, int ttl, SPSData *data)
{
  int offset;

  // Hop limit (8 bits)
  data->ip6hdr_tcp_tr.ip6_hops = ttl;

  // TCP checksum (16 bits)
  data->tcp6hdr_tr.th_sum = tcp6_checksum (data->ip6hdr_tcp_tr, data->tcp6hdr_tr, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->sixto4ip4hdr_tcp_tr.ip_sum = 0;
  data->sixto4ip4hdr_tcp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_tcp_tr, IP4_HDRLEN);

  // Fill out ethernet frame header.
  if (data->sixto4_tr_flag == 0) {
    memcpy (snd_ether_frame, data->tcp6ethhdr_tr.dst_mac, 6);  // Destination link-layer (MAC) address (48 bits)
    memcpy (snd_ether_frame + 6, data->tcp6ethhdr_tr.src_mac, 6);  // Source link-layer (MAC) address (48 bits)
    memcpy (&snd_ether_frame[12], &data->tcp6ethhdr_tr.type_code, 2);  // Ethernet type code (16 bits) (http://www.iana.org/assignments/ethernet-numbers
  } else if (data->sixto4_tr_flag == 1) {
    memcpy (snd_ether_frame, data->sixto4ethhdr_tr.dst_mac, 6);  // Destination link-layer (MAC) address (48 bits)
    memcpy (snd_ether_frame + 6, data->sixto4ethhdr_tr.src_mac, 6);  // Source link-layer (MAC) address (48 bits)
    memcpy (&snd_ether_frame[12], &data->sixto4ethhdr_tr.type_code, 2);  // Ethernet type code (16 bits) (http://www.iana.org/assignments/ethernet-numbers
  }

  // Next is ethernet frame data ([IPv4 header] + IPv6 header + TCP header + TCP data)

  // IPv4 header
  offset = 0;
  if (data->sixto4_tr_flag == 1) {
    memcpy (snd_ether_frame + 14, &data->sixto4ip4hdr_tcp_tr, IP4_HDRLEN);
    offset = IP4_HDRLEN;
  }

  // IPv6 header
  memcpy (snd_ether_frame + 14 + offset, &data->ip6hdr_tcp_tr, IP6_HDRLEN);

  // TCP header
  memcpy (snd_ether_frame + 14 + offset + IP6_HDRLEN, &data->tcp6hdr_tr, TCP_HDRLEN);

  // TCP data
  memcpy (snd_ether_frame + 14 + offset + IP6_HDRLEN + TCP_HDRLEN, data->tcp6_tr_payload, data->tcp6_tr_payloadlen);

  return (TRUE);
}

// Create IPv6 ICMP ethernet frame for traceroute.
gboolean
create_icmp6_frame (unsigned char *snd_ether_frame, int ttl, SPSData *data)
{
  int offset;

  // Hop limit (8 bits)
  data->ip6hdr_icmp_tr.ip6_hops = ttl;

  // ICMP checksum (16 bits)
  data->icmp6hdr_tr.icmp6_cksum = icmp6_checksum (data->ip6hdr_icmp_tr, data->icmp6hdr_tr, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->sixto4ip4hdr_icmp_tr.ip_sum = 0;
  data->sixto4ip4hdr_icmp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_icmp_tr, IP4_HDRLEN);

  // Fill out ethernet frame header.
  if (data->sixto4_tr_flag == 0) {
    memcpy (snd_ether_frame, data->icmp6ethhdr_tr.dst_mac, 6);  // Destination link-layer (MAC) address (48 bits)
    memcpy (snd_ether_frame + 6, data->icmp6ethhdr_tr.src_mac, 6);  // Source link-layer (MAC) address (48 bits)
    memcpy (&snd_ether_frame[12], &data->icmp6ethhdr_tr.type_code, 2);  // Ethernet type code (16 bits) (http://www.iana.org/assignments/ethernet-numbers
  } else if (data->sixto4_tr_flag == 1) {
    memcpy (snd_ether_frame, data->sixto4ethhdr_tr.dst_mac, 6);  // Destination link-layer (MAC) address (48 bits)
    memcpy (snd_ether_frame + 6, data->sixto4ethhdr_tr.src_mac, 6);  // Source link-layer (MAC) address (48 bits)
    memcpy (&snd_ether_frame[12], &data->sixto4ethhdr_tr.type_code, 2);  // Ethernet type code (16 bits) (http://www.iana.org/assignments/ethernet-numbers
  }

  // Next is ethernet frame data ([IPv4 header] + IPv6 header + ICMP header + ICMP data)

  // IPv4 header
  offset = 0;
  if (data->sixto4_tr_flag == 1) {
    memcpy (snd_ether_frame + 14, &data->sixto4ip4hdr_icmp_tr, IP4_HDRLEN);
    offset = IP4_HDRLEN;
  }

  // IPv6 header
  memcpy (snd_ether_frame + 14 + offset, &data->ip6hdr_icmp_tr, IP6_HDRLEN);

  // ICMP header
  memcpy (snd_ether_frame + 14 + offset + IP6_HDRLEN, &data->icmp6hdr_tr, ICMP_HDRLEN);

  // ICMP data
  memcpy (snd_ether_frame + 14 + offset + IP6_HDRLEN + ICMP_HDRLEN, data->icmp6_tr_payload, data->icmp6_tr_payloadlen);

  return (TRUE);
}

// Create IPv6 UDP ethernet frame for traceroute.
gboolean
create_udp6_frame (unsigned char *snd_ether_frame, int ttl, SPSData *data)
{
  int offset;

  // Hop limit (8 bits)
  data->ip6hdr_udp_tr.ip6_hops = ttl;

  // UDP checksum (16 bits)
  data->udp6hdr_tr.uh_sum = udp6_checksum (data->ip6hdr_udp_tr, data->udp6hdr_tr, data->udp6_tr_payload, data->udp6_tr_payloadlen);

  // IPv4 header checksum (16 bits): set to 0 when calculating checksum
  data->sixto4ip4hdr_udp_tr.ip_sum = 0;
  data->sixto4ip4hdr_udp_tr.ip_sum = checksum ((unsigned short int *) &data->sixto4ip4hdr_udp_tr, IP4_HDRLEN);

  // Fill out ethernet frame header.
  if (data->sixto4_tr_flag == 0) {
    memcpy (snd_ether_frame, data->udp6ethhdr_tr.dst_mac, 6);  // Destination link-layer (MAC) address (48 bits)
    memcpy (snd_ether_frame + 6, data->udp6ethhdr_tr.src_mac, 6);  // Source link-layer (MAC) address (48 bits)
    memcpy (&snd_ether_frame[12], &data->udp6ethhdr_tr.type_code, 2);  // Ethernet type code (16 bits) (http://www.iana.org/assignments/ethernet-numbers
  } else if (data->sixto4_tr_flag == 1) {
    memcpy (snd_ether_frame, data->sixto4ethhdr_tr.dst_mac, 6);  // Destination link-layer (MAC) address (48 bits)
    memcpy (snd_ether_frame + 6, data->sixto4ethhdr_tr.src_mac, 6);  // Source link-layer (MAC) address (48 bits)
    memcpy (&snd_ether_frame[12], &data->sixto4ethhdr_tr.type_code, 2);  // Ethernet type code (16 bits) (http://www.iana.org/assignments/ethernet-numbers
  }

  // Next is ethernet frame data ([IPv4 header] + IPv6 header + UDP header + UDP data)

  // IPv4 header
  offset = 0;
  if (data->sixto4_tr_flag == 1) {
    memcpy (snd_ether_frame + 14, &data->sixto4ip4hdr_udp_tr, IP4_HDRLEN);
    offset = IP4_HDRLEN;
  }

  // IPv6 header
  memcpy (snd_ether_frame + 14 + offset, &data->ip6hdr_udp_tr, IP6_HDRLEN);

  // UDP header
  memcpy (snd_ether_frame + 14 + offset + IP6_HDRLEN, &data->udp6hdr_tr, UDP_HDRLEN);

  // UDP data
  memcpy (snd_ether_frame + 14 + offset + IP6_HDRLEN + UDP_HDRLEN, data->udp6_tr_payload, data->udp6_tr_payloadlen);

  return (TRUE);
}

gchar *
thishost_ipv4 (char *host)
{
  int status;
  gchar *text;
  struct addrinfo hints, *res, *p;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in thishost_ipv4().\n");
    exit (EXIT_FAILURE);
  }

  gethostname (host, 1023);

  strcat (host, ".local");

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

  if ((status = getaddrinfo (host, NULL, &hints, &res)) != 0) {
    host[0] = 0;
    g_free (text);
    return (host);
  }

  for (p = res; p != NULL; p = p->ai_next) {

    switch (p->ai_family) {
      case AF_INET:
        tmp = &((struct sockaddr_in *) p->ai_addr)->sin_addr;
        break;
      case AF_INET6:
        continue;
    }
    inet_ntop (AF_INET, tmp, host, 16);
  }

  g_free (text);

  return (host);
}

gchar *
thishost_ipv6 (char *host)
{
  int status;
  gchar *text;
  struct addrinfo hints, *res, *p;
  void *tmp;

// Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in thishost_ipv6().\n");
    exit (EXIT_FAILURE);
  }

  gethostname (host, 1023);

  strcat (host, ".local");

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = hints.ai_flags | AI_CANONNAME;

  if ((status = getaddrinfo (host, NULL, &hints, &res)) != 0) {
    host[0] = 0;
    g_free (text);
    return (host);
  }

  for (p = res; p != NULL; p = p->ai_next) {

    switch (p->ai_family) {
      case AF_INET:
        continue;
      case AF_INET6:
        tmp = &((struct sockaddr_in6 *) p->ai_addr)->sin6_addr;
        break;
    }
    inet_ntop (AF_INET6, tmp, host, 40);
  }

  g_free (text);

  return (host);
}

// Request a socket descriptor for FTP control socket.
int
cntl_sock (gchar *host, gchar *cntl_ip, SPSData *data)
{
  int sd, status;
  struct addrinfo hints, *servinfo, *p;
  struct timeval ts;
  gchar s[INET6_ADDRSTRLEN];

  // Populate struct hints for getaddrinfo().
  memset (&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  // Resolve RIR URL to IP address.
  if ((status = getaddrinfo (host, FTP_PORT, &hints, &servinfo)) != 0) {
    sprintf (data->error_text, "In callbacks.c: getaddrinfo() failed in cntl_sock().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }

  // Loop through all the results and connect to the first we can.
  // Request socket descriptor.
  for (p = servinfo; p != NULL; p = p->ai_next) {
    if ((sd = socket (p->ai_family, p->ai_socktype, p->ai_protocol)) < 0) {
      status = errno;
      continue;
    }

    // Set socket timeout option for control connection.
    memset (&ts, 0, sizeof (struct timeval));
    ts.tv_sec = FTP_TIMEOUT;
    if ((setsockopt (sd, SOL_SOCKET, SO_RCVTIMEO, &ts, sizeof (ts))) != 0) {
      status = errno;
      close (sd);
      continue;
    }

    // Establish connection to remote host.
    if (connect (sd, p->ai_addr, p->ai_addrlen) < 0) {
      status = errno;
      close (sd);
      continue;
    }
    break;
  }

  if (p == NULL) {
    sprintf (data->error_text, "In callbacks.c: Failed to establish a connection to remote host in cntl_sock().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }

  // Save family for calling function to use.
  data->ftp_family = p->ai_family;

  // Get IP address so calling function can report it.
  inet_ntop (p->ai_family, get_in_addr ((struct sockaddr *)p->ai_addr), s, INET6_ADDRSTRLEN);
  strcpy (cntl_ip, s);
  freeaddrinfo (servinfo);

  return (sd);
}

// Request a socket descriptor for FTP data socket.
int
data_sock (gchar *data_ip, gint data_port, SPSData *data)
{
  int sd, status;
  struct addrinfo hints, *servinfo, *p;
  struct timeval ts;
  gchar buf[6];
  const gchar *port;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    port = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'port' in data_sock().\n");
    exit (EXIT_FAILURE);
  }

  sprintf (buf, "%d", data_port);
  port = (const gchar *) &buf;

  // Populate struct hints for getaddrinfo().
  memset (&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  // Populate struct servinfo.
  if ((status = getaddrinfo (data_ip, port, &hints, &servinfo)) != 0) {
    sprintf (data->error_text, "In callbacks.c: getaddrinfo() failed in data_sock().\nError message: %s", gai_strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }

  // Loop through all the results and connect to the first we can.
  // Request socket descriptor.
  for (p = servinfo; p != NULL; p = p->ai_next) {

    if ((sd = socket (p->ai_family, p->ai_socktype, p->ai_protocol)) < 0) {
      status = errno;
      continue;
    }

    // Set socket timeout option for control connection.
    memset (&ts, 0, sizeof (struct timeval));
    ts.tv_sec = FTP_TIMEOUT;
    if ((setsockopt (sd, SOL_SOCKET, SO_RCVTIMEO, &ts, sizeof (ts))) != 0) {
      status = errno;
      close (sd);
      continue;
    }

    // Establish connection to remote host.
    if (connect (sd, p->ai_addr, p->ai_addrlen) < 0) {
      status = errno;
      close (sd);
      continue;
    }
    break;
  }

  if (p == NULL) {
    sprintf (data->error_text, "In callbacks.c: Failed to establish a connection to remote host in data_sock().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }

  freeaddrinfo (servinfo);

  return (sd);
}

// Send a command to FTP control socket.
int
cntl_send (int sd_cntl, gchar *command, SPSData *data)
{
  int status;

  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("Sending command: %s\n", command);
  }
  G_UNLOCK (show_dl_messages);

  if (write (sd_cntl, command, strlen (command)) < 1) {
    status = errno;
    sprintf (data->error_text, "In callbacks.c: write() failed to write bytes to socket in cntl_send().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  sleep (1);  // Allow 1 second for target node to respond.

  return (EXIT_SUCCESS);
}

// Receive FTP command responses from remote host on the control socket.
int
cntl_recv (int sd_cntl, SPSData *data)
{
  int nbytes, status;

  // Note: we expect to get entire response in cntl-data in one recv() call, so we don't loop it.
  memset (data->ftp_data, 0, IP_MAXPACKET);
  if ((nbytes = recv (sd_cntl, data->ftp_data, IP_MAXPACKET, 0)) < 0) {
    status = errno;
    if (status == EAGAIN) {  // No response within TIMEOUT seconds.
      sprintf (data->error_text,"In callbacks.c: recv() failed to receive a response within timeout in cntl_recv().\nSuggest you try again.\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      return (EXIT_FAILURE);
    } else {
      sprintf (data->error_text, "In callbacks.c: recv() failed in cntl_recv().\nError message: %s", strerror (status));
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      return (EXIT_FAILURE);
    }
  }

  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("Received message: %s\n", data->ftp_data);
  }
  G_UNLOCK (show_dl_messages);

  // If a command isn't accepted/fails, then the response code starts with a 4 or 5 like HTTP.
  // If it starts with 2 it's ok.
  if ((data->ftp_data[0] == '4') || (data->ftp_data[0] == '5')) {
    sprintf (data->error_text, "In callbacks.c: FTP command wasn't accepted/failed in cntl_recv().\nSuggest you try again.");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

// Extract IPv4 address and port number for creating FTP data socket in passive mode.
int
parse_ip4_port (gchar *data_ip, int *data_port, SPSData *data)
{
  int i, j, k, p1, p2;
  gchar *value;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in parse_ip4_port().\n");
    exit (EXIT_FAILURE);
  }

  // Find first bracket enclosing IP and port information.
  i = 1;
  while ((data->ftp_data[i] != '(') && (i < 1024)) {
    i++;
  }
  if (i == 1024) {
    sprintf (data->error_text, "In callbacks.c: Can't find data connection IP address in server's reply in parse_ip4_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  i++;

  // Extract data connection IP address.
  memset (data_ip, 0, 1024);
  j = 0;
  k = 0;
  while (j < 4) {
    data_ip[k] = data->ftp_data[i];
    if (data_ip[k] == ',') {
      data_ip[k] = '.';
      j++;
    }
    i++;
    k++;
  }
  data_ip[strlen (data_ip)-1] = 0;
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("Data connection IP address: %s\n", data_ip);
  }
  G_UNLOCK (show_dl_messages);

  // Extract data connection port number.
  memset (value, 0, 80);
  k = 0;
  while ((data->ftp_data[i] != ',') && (i < 1024)) {
    value[k] = data->ftp_data[i];
    i++;
    k++;
  }
  if (i == 1024) {
    sprintf (data->error_text, "In callbacks.c: Can't find data connection port number in server's reply in parse_ip4_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  p1 = atoi (value);
  i++;
  memset (value, 0, 80);
  k = 0;
  while ((data->ftp_data[i] != ')') && (i < 1024)) {
    value[k] = data->ftp_data[i];
    i++;
    k++;
  }
  if (i == 1024) {
    sprintf (data->error_text, "In callbacks.c: Can't find data connection port number in server's reply in parse_ip4_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  p2 = atoi (value);
  *data_port = (p1 * 256) + p2;
  if (*data_port > 65535) {
    sprintf (data->error_text, "In callbacks.c: Data connection port number in server's reply appears to be too large in parse_ip4_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("Data connection port: %i\n", *data_port);
  }
  G_UNLOCK (show_dl_messages);

  g_free (value);

  return (EXIT_SUCCESS);
}

// Extract port number for creating an IPv6 FTP data socket in passive mode.
// For IPv6 FTP, data IP address is the same as the control IP address.
int
parse_ip6_port (int *data_port, SPSData *data)
{
  int i, k;
  gchar *value;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (80 * sizeof (gchar));
  if (tmp != NULL) {
    value = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'value' in parse_ip6_port().\n");
    exit (EXIT_FAILURE);
  }

  // Find first bracket enclosing port information.
  i = 1;
  while ((data->ftp_data[i] != '(') && (i < 1024)) {
    i++;
  }
  if (i == 1024) {
    sprintf (data->error_text, "In callbacks.c: Can't find data connection IP address in server's reply in parse_ip6_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  i++;

  // Find a digit. i.e., where port number starts
  while ((data->ftp_data[i] < 48) && (data->ftp_data[i] > 57) && (i < 1024)) {
    i++;
  }

  // Extract data connection port number.
  memset (value, 0, 80);
  k = 0;
  while ((data->ftp_data[i] > 47) && (data->ftp_data[i] < 58) && (i < 1024)) {
    value[k] = data->ftp_data[i];
    i++;
    k++;
  }
  if (i == 1024) {
    sprintf (data->error_text, "In callbacks.c: Can't find data connection port number in server's reply in parse_ip6_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  *data_port = atoi (value);
  if (*data_port > 65535) {
    sprintf (data->error_text, "In callbacks.c: Data connection port number in server's reply appears to be too large in parse_ip6_port().\nSuggest you try again.\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (value);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("Data connection port: %i\n", *data_port);
  }
  G_UNLOCK (show_dl_messages);

  g_free (value);

  return (EXIT_SUCCESS);
}

// Receive and save data on the FTP data socket.
int
data_recv (SPSData *data)
{
  int i, nbytes, status;
  FILE *fo;

  // Open file to cache the IP delegation data.
  fo = fopen (rir_file[data->rir_index], "wb");
  if (fo == NULL) {
    sprintf (data->error_text, "In callbacks.c: Can't open a file for saving the IP delegation data in data_recv().");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    return (EXIT_FAILURE);
  }

  nbytes = 1;
  while (nbytes != 0) {
    memset (data->ftp_data, 0, IP_MAXPACKET);
    if ((nbytes = recv (data->sd_data, data->ftp_data, IP_MAXPACKET, 0)) < 0) {
      status = errno;
      if (status == EAGAIN) {  // No response within TIMEOUT seconds.
        sprintf (data->error_text, "In callbacks.c: recv() failed to receive a response within timeout in data_recv().\nSuggest you try again.\nError message: %s", strerror (status));
        data->parent = data->main_window;
        G_LOCK (message_available);
        message_available = 1;
        G_UNLOCK (message_available);
        g_idle_add ((GSourceFunc) rir_not_available, data);
        return (EXIT_FAILURE);
      } else if (status == EINTR) {  // EINTR = 4
          continue;  // Something weird happened, but let's keep listening.
      } else {
        sprintf (data->error_text, "In callbacks.c: recvfrom() failed in data_recv().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        G_LOCK (message_available);
        message_available = 1;
        G_UNLOCK (message_available);
        g_idle_add ((GSourceFunc) rir_not_available, data);
        return (EXIT_FAILURE);
      }
    }
    // Write this batch of data to file.
    for (i=0; i<nbytes; i++) {
      fputc (data->ftp_data[i], fo);
    }
  }

  fclose (fo);

  return (EXIT_SUCCESS);
}

// Toggle whether or not to show sent and received
// messages on stdout during file downloads.
gboolean
on_checkbutton38_toggled (GtkButton *button, SPSData *data)
{
  if (show_dl_messages == 1) {
    show_dl_messages = 0;
  } else {
    show_dl_messages = 1;
  }

  return (TRUE);
}

// Set RIR delegation file status to "Loading". This function is
// executed as an idle function, thus it returns FALSE in order to stop.
gboolean
rir_loading (SPSData *data)
{
  GtkTextBuffer *textbuffer;

 // Update file status on ui.
  if (data->rir_index == 0) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview8));
  } else if (data->rir_index == 1) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview9));
  } else if (data->rir_index == 2) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview12));
  } else if (data->rir_index == 3) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview13));
  } else if (data->rir_index == 4) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview14));
  }

  gtk_text_buffer_set_text (textbuffer, "Loading", -1);

  return (FALSE);
}

// Set RIR delegation file status to "Not Available" because
// an error occurred. This function is executed as an idle function,
// thus it returns FALSE in order to stop.
gboolean
rir_not_available (SPSData *data)
{
  GtkTextBuffer *textbuffer;

 // Update file status on ui.
  if (data->rir_index == 0) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview8));
  } else if (data->rir_index == 1) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview9));
  } else if (data->rir_index == 2) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview12));
  } else if (data->rir_index == 3) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview13));
  } else if (data->rir_index == 4) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview14));
  }
  gtk_text_buffer_set_text (textbuffer, "Not Available", -1);

  return (FALSE);
}

// Update RIR delegation file status. This function is executed
// as an idle function, thus it returns FALSE in order to stop.
gboolean
rir_available (SPSData *data)
{
  GtkTextBuffer *textbuffer;

 // Update file status on ui.
  if (data->rir_index == 0) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview8));
  } else if (data->rir_index == 1) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview9));
  } else if (data->rir_index == 2) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview12));
  } else if (data->rir_index == 3) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview13));
  } else if (data->rir_index == 4) {
    textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview14));
  }

  if (file_exists (rir_file[data->rir_index])) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->rir_files_have[data->rir_index] = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->rir_files_have[data->rir_index] = 0;
  }

  return (FALSE);
}

// Timeout function to monitor for error/finished messages from child threads.
gboolean
thread_listen (SPSData *data)
{
  if (message_available == 1) {
    GtkWidget *dialog, *content_area, *image, *label, *hbox;

    // data->parent is a pointer to the widget whence the error originated.
    // It can point to things like data->main_window, data->menubar1, or data->traceroute_window
    dialog = gtk_dialog_new_with_buttons ("Error", GTK_WINDOW (data->parent),
                                  GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);  // No separater

    g_signal_connect_swapped (dialog, "response", G_CALLBACK (gtk_widget_destroy), dialog);

    content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));

    label = gtk_label_new (data->error_text);
    image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_ERROR,
                                    GTK_ICON_SIZE_DIALOG);

    hbox = gtk_hbox_new (FALSE, 5);  // Don't require homogeneous, 5 px spacing
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);  // 10 px
    gtk_box_pack_start_defaults (GTK_BOX (hbox), image);  // Add error symbol to hbox
    gtk_box_pack_start_defaults (GTK_BOX (hbox), label);  // Add our text to hbox

    // Pack the dialog content into the dialog's GtkVBox.
    gtk_box_pack_start_defaults (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox);

    gtk_widget_show_all (dialog);

    gtk_dialog_run (GTK_DIALOG (dialog));

    // Clear the Message Available flag since we've displayed it.
    // Timeout functions are outside of the main GTK+ lock, so we must
    // lock this flag manually.
    G_LOCK (message_available);
    message_available = 0;
    G_UNLOCK (message_available);

    return (TRUE);
  }

  // This timeout will continue if no errors occur.
  // So stop it if it isn't needed for monitoring a thread anymore.
  if (stop_thread_listen == 1) {
    stop_thread_listen = 0;
    return (FALSE);
  }

  return (TRUE);
}

// Retrieve the Regional Internet Registry (RIR) IP delegation files
// using FTP protocol.
// This function will be executed as a separate thread.
gboolean
ftp_rir (SPSData *data)
{
  int sd_cntl, data_port;
  gchar *cntl_ip, *data_ip, *text;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in ftp_rir().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    cntl_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'cntl_ip' in ftp_rir().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    data_ip = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data_ip' in ftp_rir().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (IP_MAXPACKET * sizeof (gchar));
  if (tmp != NULL) {
    data->ftp_data = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ftp_data' in ftp_rir().\n");
    exit (EXIT_FAILURE);
  }

  // Set RIR delegation file status to "Loading".
  g_idle_add ((GSourceFunc) rir_loading, data);

  // Set up control connection.
  if ((sd_cntl = cntl_sock (rir_url[data->rir_index], cntl_ip, data)) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("\nConnected to Regional Internet Registry (RIR): %s (%s)\n", rir_name[data->rir_index], cntl_ip);
  }
  G_UNLOCK (show_dl_messages);

  // Authenticate with target node.
  if (cntl_send (sd_cntl, "USER anonymous\r\n", data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  if (cntl_send (sd_cntl, "PASS password\r\n", data) == EXIT_FAILURE) {  // Anything is acceptable for anonymous user.
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  // Request type of operating system
  if (cntl_send (sd_cntl, "SYST\r\n", data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  // Set to passive mode.
  if (data->ftp_family == AF_INET) {
    if (cntl_send (sd_cntl, "PASV\r\n", data) == EXIT_FAILURE) { // Tell target node we need it to provide IP and port to us.
      free_rir_mem (text, cntl_ip, data_ip, data);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      G_LOCK (downloading);
      downloading = 0;
      G_UNLOCK (downloading);
      return (FALSE);
    }
    if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
      free_rir_mem (text, cntl_ip, data_ip, data);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      G_LOCK (downloading);
      downloading = 0;
      G_UNLOCK (downloading);
      return (FALSE);
    }

    if (data->ftp_data[0] == '2') {
      parse_ip4_port (data_ip, &data_port, data);
    }

    // Set up data connection.
    if ((data->sd_data = data_sock (data_ip, data_port, data)) == EXIT_FAILURE) {
      free_rir_mem (text, cntl_ip, data_ip, data);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      G_LOCK (downloading);
      downloading = 0;
      G_UNLOCK (downloading);
      return (FALSE);
    }
  } else if (data->ftp_family == AF_INET6) {
    if (cntl_send (sd_cntl, "EPSV\r\n", data) == EXIT_FAILURE) { // Tell target node we need it to provide IP and port to us.
      free_rir_mem (text, cntl_ip, data_ip, data);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      G_LOCK (downloading);
      downloading = 0;
      G_UNLOCK (downloading);
      return (FALSE);
    }
    if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
      free_rir_mem (text, cntl_ip, data_ip, data);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      G_LOCK (downloading);
      downloading = 0;
      G_UNLOCK (downloading);
      return (FALSE);
    }

    if (data->ftp_data[0] == '2') {
      parse_ip6_port (&data_port, data);
    }
    strcpy (data_ip, cntl_ip);  // For IPv6, we use the same IP address as the control connection.

    // Set up data connection.
    if ((data->sd_data = data_sock (data_ip, data_port, data)) == EXIT_FAILURE) {
      free_rir_mem (text, cntl_ip, data_ip, data);
      g_idle_add ((GSourceFunc) rir_not_available, data);
      G_LOCK (downloading);
      downloading = 0;
      G_UNLOCK (downloading);
      return (FALSE);
    }
  }
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("\nData connection established.\n");
  }
  G_UNLOCK (show_dl_messages);

  // Set to binary transfer mode.
  if (cntl_send (sd_cntl, "TYPE I\r\n", data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  // Change directory.
  memset (text, 0 ,1024);
  sprintf (text, "CWD %s\r\n", rir_dir[data->rir_index]);
  if (cntl_send (sd_cntl, text, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  // Retrieve delegation file.
  memset (text, 0, 1024);
  sprintf (text, "RETR %s\r\n", rir_file[data->rir_index]);
  if (cntl_send (sd_cntl, text, data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (data_recv (data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  // Close control and data connections.
  if (cntl_send (sd_cntl, "QUIT\r\n", data) == EXIT_FAILURE) {
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }
  if (cntl_recv (sd_cntl, data) == EXIT_FAILURE) {  // Expect "221 Goodbye."
    free_rir_mem (text, cntl_ip, data_ip, data);
    g_idle_add ((GSourceFunc) rir_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (FALSE);
  }

  // Close control and data sockets.
  close (sd_cntl);
  close (data->sd_data);

  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("DONE!\n");
  }
  G_UNLOCK (show_dl_messages);

  // Free allocated memory.
  free_rir_mem (text, cntl_ip, data_ip, data);

  G_LOCK (downloading);
  downloading = 0;
  G_UNLOCK (downloading);

  // Update RIR delegation file status.
  g_idle_add ((GSourceFunc) rir_available, data);

  // Load RIR files and country codes into memory.
  if (files_available (data) != 0) {
    read_delegations_and_countries (data);
  }

  return (TRUE);
}

// Free allocated memory used in ftp_rir().
gboolean
free_rir_mem (gchar *text, gchar *cntl_ip, gchar *data_ip, SPSData *data)
{
  g_free (text);
  g_free (cntl_ip);
  g_free (data_ip);
  g_free (data->ftp_data);

  return (TRUE);
}

gint
file_exists (gchar *filename)
{
  FILE *fp;

  // Does file exist?
  fp = fopen (filename, "r");
  if (fp == NULL) {
    return (0);
  } else {
    fclose (fp);
  }

  return (1);
}

// Download ARIN delegation file.
gboolean
on_button50_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Select ARIN.
    data->rir_index = 0;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) ftp_rir, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// Download RIPE delegation file.
gboolean
on_button51_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Select RIPE.
    data->rir_index = 1;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) ftp_rir, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// Download AFRINIC delegation file.
gboolean
on_button52_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Select AFRINIC.
    data->rir_index = 2;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) ftp_rir, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// Download APNIC delegation file.
gboolean
on_button53_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Select APNIC.
    data->rir_index = 3;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) ftp_rir, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// Download LACNIC delegation file.
gboolean
on_button54_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Select LACNIC.
    data->rir_index = 4;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) ftp_rir, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// Download all RIR ASN / IP delegation files and www.iso.org's country code file.
gboolean
on_button57_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) download_all, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// Download all RIR ASN / IP delegation files and www.iso.org's country code file.
gboolean
download_all (SPSData *data)
{
  gint i;

  // Download all RIR ASN / IP delegation files.
  for (i=0; i<5; i++) {
    data->rir_index = i;
    ftp_rir (data);
  }

  // Download www.iso.org's country code file.
  get_countries (data);

  return (TRUE);
}

// Set country code file status to "Loading". This function is
// executed as an idle function, thus it returns FALSE in order to stop.
gboolean
cc_loading (SPSData *data)
{
  GtkTextBuffer *textbuffer;

  // Update file status on ui.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview15));
  gtk_text_buffer_set_text (textbuffer, "Loading", -1);

  return (FALSE);
}

// Set country code file status to "Not Available" because an
// error occurred. This function is executed as an idle function,
// thus it returns FALSE in order to stop.
gboolean
cc_not_available (SPSData *data)
{
  GtkTextBuffer *textbuffer;

  // Update file status on ui.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview15));
  gtk_text_buffer_set_text (textbuffer, "Not Available", -1);

  return (FALSE);
}

// Set country code file status to "Available". This function is
// executed as an idle function, thus it returns FALSE in order to stop.
gboolean
cc_available (SPSData *data)
{
  GtkTextBuffer *textbuffer;

  // Update file status on ui.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview15));
  if (file_exists (COUNTRY_FILE)) {
    gtk_text_buffer_set_text (textbuffer, "Available", -1);
    data->countries_have = 1;
  } else {
    gtk_text_buffer_set_text (textbuffer, "Not Available", -1);
    data->countries_have = 0;
  }

  return (FALSE);
}

// Retrieve the www.iso.org's ISO 3166-1 country code list.
gboolean
get_countries (SPSData *data)
{
  int sd;
  gchar *url, *ip_addr, *get;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    url = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'url' in get_countries().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (40 * sizeof (gchar));
  if (tmp != NULL) {
    ip_addr = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ip_addr' in get_countries.\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    get = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'get' in get_countries().\n");
    exit (EXIT_FAILURE);
  }

  // Set country code file status to "Loading".
  g_idle_add ((GSourceFunc) cc_loading, data);

  // URL for International Organization for Standardization (ISO).
  strcpy (url, "www.iso.org");

  // HTTP GET statement to retrieve the ISO 3166-1 country code list.
  sprintf (get, "GET %s%s HTTP/1.0\r\nUser-Agent: http.c (linux)\r\nAccept: */*\r\nHost: www.iso.org\r\nConnection: Keep-Alive\r\n\r\n", COUNTRY_DIR, COUNTRY_FILE);

  // Make TCP connection to remote host.
  if ((sd = http_sock (url, ip_addr, HTTP_PORT, data)) < 0) {
    g_free (url);
    g_free (ip_addr);
    g_free (get);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (EXIT_FAILURE);
  }
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("\nConnected to %s (%s)\n", url, ip_addr);
  }
  G_UNLOCK (show_dl_messages);

  // Request the country code file.
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("\nRequesting country code file.\n");
  }
  G_UNLOCK (show_dl_messages);
  if (send_get (sd, get, data) == EXIT_FAILURE) {
    g_free (url);
    g_free (ip_addr);
    g_free (get);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (EXIT_FAILURE);
  }

  // Receive and save the country code file.
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("\nReceiving country code file.\n");
  }
  G_UNLOCK (show_dl_messages);
  if (http_recv (sd, COUNTRY_FILE, data) == EXIT_FAILURE) {
    g_free (url);
    g_free (ip_addr);
    g_free (get);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    G_LOCK (downloading);
    downloading = 0;
    G_UNLOCK (downloading);
    return (EXIT_FAILURE);
  }
  G_LOCK (show_dl_messages);
  if (show_dl_messages == 1) {
    printf ("\nCountry code file received and saved.\n");
    printf ("\nDONE!\n");
  }
  G_UNLOCK (show_dl_messages);

  // Close socket.
  close (sd);

  G_LOCK (downloading);
  downloading = 0;
  G_UNLOCK (downloading);

  // Update country code file status.
  g_idle_add ((GSourceFunc) cc_available, data);

  // Load RIR files and country codes into memory.
  if (files_available (data) != 0) {
    read_delegations_and_countries (data);
  }

  // Free allocated memory.
  g_free (url);
  g_free (ip_addr);
  g_free (get);

  return (TRUE);
}

// Request socket descriptor for HTTP connection.
int
http_sock (gchar *host, gchar *ip_addr, const gchar *port, SPSData *data)
{
  int sd, status;
  struct addrinfo hints, *servinfo, *p;
  struct timeval ts;

  // Populate struct hints for getaddrinfo().
  memset (&hints, 0, sizeof (hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  // Resolve URL to IP address.
  if ((status = getaddrinfo (host, port, &hints, &servinfo)) != 0) {
    sprintf (data->error_text, "In callbacks.c: getaddrinfo() failed in http_sock().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }

  // Loop through all the results and connect to the first we can.
  // Request socket descriptor.
  for (p = servinfo; p != NULL; p = p->ai_next) {
    if ((sd = socket (p->ai_family, p->ai_socktype, p->ai_protocol)) < 0) {
      status = errno;
      continue;
    }

    // Set socket timeout option for connection.
    memset (&ts, 0, sizeof (struct timeval));
    ts.tv_sec = HTTP_TIMEOUT;
    if ((setsockopt (sd, SOL_SOCKET, SO_RCVTIMEO, &ts, sizeof (ts))) != 0) {
      status = errno;
      close (sd);
      continue;
    }

    // Establish connection to remote host.
    if (connect (sd, p->ai_addr, p->ai_addrlen) < 0) {
      status = errno;
      close (sd);
      continue;
    }
    break;
  }

  if (p == NULL) {
    sprintf (data->error_text, "In callbacks.c: Failed to establish a connection to remote host in http_sock().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }

  // Get IP address so calling function can report it.
  inet_ntop (p->ai_family, get_in_addr ((struct sockaddr *)p->ai_addr), ip_addr, INET6_ADDRSTRLEN);

  freeaddrinfo (servinfo);

  return (sd);
}

// Send an HTTP GET.
int
send_get (int sd, gchar *command, SPSData *data)
{
  int status;

  if (write (sd, command, strlen (command)) < 1) {
    status = errno;
    sprintf (data->error_text, "In callbacks.c: write() failed to write bytes to socket in send_get().\nError message: %s", strerror (status));
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }
  sleep (1);  // Allow 1 second for target node to respond.

  return (EXIT_SUCCESS);
}

// Receive and save data resulting from an HTTP GET.
int
http_recv (int sd, gchar *filename, SPSData *data)
{
  int i, j, k, nbytes, status, total_bytes, data_len, code;
  gchar *text, *temp_array, *line, *http_data, *err_msg;
  void *tmp;
  FILE *fo;

  // Allocate memopry for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in http_recv().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    err_msg = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'err_msg' in http_recv().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (100000 * sizeof (gchar));  // Reserve more memory than we'll really need, just to be safe.
  if (tmp != NULL) {
    temp_array = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp_array' in http_recv().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    line = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'line' in http_recv().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (IP_MAXPACKET * sizeof (gchar));
  if (tmp != NULL) {
    http_data = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'http_data' in http_recv().\n");
    exit (EXIT_FAILURE);
  }

  i = 0;
  nbytes = 1;
  total_bytes = 0;
  while (nbytes != 0) {
    memset (http_data, 0, IP_MAXPACKET);
    if ((nbytes = recv (sd, http_data, IP_MAXPACKET, 0)) < 0) {
      status = errno;
      if (status == EAGAIN) {  // No response within TIMEOUT seconds.
        sprintf (data->error_text, "In callbacks.c: recv() timed-out in http_recv().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        G_LOCK (message_available);
        message_available = 1;
        G_UNLOCK (message_available);
        g_free (text);
        g_free (err_msg);
        g_free (temp_array);
        g_free (line);
        g_free (http_data);
        g_idle_add ((GSourceFunc) cc_not_available, data);
        return (EXIT_FAILURE);
      } else if (status == EINTR) {  // EINTR = 4
          continue;  // Something weird happened, but let's keep listening.
      } else {
        sprintf (data->error_text, "In callbacks.c: recv() failed in http_recv().\nError message: %s", strerror (status));
        data->parent = data->main_window;
        G_LOCK (message_available);
        message_available = 1;
        G_UNLOCK (message_available);
        g_free (text);
        g_free (err_msg);
        g_free (temp_array);
        g_free (line);
        g_free (http_data);
        g_idle_add ((GSourceFunc) cc_not_available, data);
        return (EXIT_FAILURE);
      }
    }
    for (j=0; j<nbytes; j++) {
      temp_array[i] = http_data[j];
      i++;
    }
    total_bytes += nbytes;
  }

  // Check status code for error message in response.
  // See Sections 6.1 and 6.1.1 of RFC 2616.
  memset (text, 0, 1024 * sizeof (gchar));
  i = 0;
  while ((temp_array[i] != ' ') && (i < 1024)) {
    i++;
  }
  if (i == 1024) {
    sprintf (data->error_text, "In callbacks.c: Couldn't find status code in HTTP response in http_recv().");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (text);
    g_free (err_msg);
    g_free (temp_array);
    g_free (line);
    g_free (http_data);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }
  i++;
  text[0] = temp_array[i]; i++;
  text[1] = temp_array[i]; i++;
  text[2] = temp_array[i]; i++;
  code = atoi (text);
  if (code != 200) {
    i++;  // Skip space
    memset (err_msg, 0, 1024 * sizeof (gchar));
    j = 0;
    while ((temp_array[i] != '\r') && (temp_array[i] != '\n') && (i < 1024)) {
      err_msg[j] = temp_array[i];
      i++;
      j++;
    }
    if (i == 1024) {
      sprintf (data->error_text, "In callbacks.c: Couldn't find reason phrase in HTTP response in http_recv().");
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      g_free (text);
      g_free (err_msg);
      g_free (temp_array);
      g_free (line);
      g_free (http_data);
      g_idle_add ((GSourceFunc) cc_not_available, data);
      return (EXIT_FAILURE);
    }
    sprintf (data->error_text, "In callbacks.c: HTTP response indicates a problem in http_recv().\n%i %s", code, err_msg);
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (text);
    g_free (err_msg);
    g_free (temp_array);
    g_free (line);
    g_free (http_data);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }

  // Find content length.
  i = 0;  // Index to temp_array.
  k = 0;  // Line count.
  memset (line, 0, 1024);
  while ((strcmp (line, "Content-Length:") != 0) && (k < 80)) {  // Limit k to 80 to avoid endless loops.
    j = 0;
    memset (line, 0, 1024);
    while ((temp_array[i] != '\r') && (j < 80)) {  // Limit j to 80 to avoid endless loops.
      line[j] = temp_array[i];
      i++;
      j++;
    }
    if (i == 80) {
      sprintf (data->error_text, "In callbacks.c: Can't read line from received data in http_recv().\n");
      data->parent = data->main_window;
      G_LOCK (message_available);
      message_available = 1;
      G_UNLOCK (message_available);
      g_free (text);
      g_free (err_msg);
      g_free (temp_array);
      g_free (line);
      g_free (http_data);
      g_idle_add ((GSourceFunc) cc_not_available, data);
      return (EXIT_FAILURE);
    }
    i += 2;  // Skip the \r\n to get to next line.
    line[15] = 0;  // Truncate line length to allow for only "Content-Length:" + '\0'.
    k++;  // Increment line count.
  }
  if (k == 80) {
    sprintf (data->error_text, "In callbacks.c: Can't find 'CONTENT-LENGTH:' in HTTP data received data in http_recv().\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (text);
    g_free (err_msg);
    g_free (temp_array);
    g_free (line);
    g_free (http_data);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }
  data_len = atoi (line+16);

  // Open file to save received data.
  fo = fopen (filename, "wb");
  if (fo == NULL) {
    sprintf (data->error_text, "In callbacks.c: Can't open file to save data in http_recv().\n");
    data->parent = data->main_window;
    G_LOCK (message_available);
    message_available = 1;
    G_UNLOCK (message_available);
    g_free (text);
    g_free (err_msg);
    g_free (temp_array);
    g_free (line);
    g_free (http_data);
    g_idle_add ((GSourceFunc) cc_not_available, data);
    return (EXIT_FAILURE);
  }

  // Write data to file, excluding received HTTP header info.
  for (i=(total_bytes - data_len); i<total_bytes; i++) {
    fputc (temp_array[i], fo);
  }

  // Close file.
  fclose (fo);

  // Free allocated memory.
  g_free (text);
  g_free (err_msg);
  g_free (temp_array);
  g_free (line);
  g_free (http_data);

  return (EXIT_SUCCESS);
}

// Download www.iso.org's country code file.
// Note that www.iso.org only supports IPv4.
gboolean
on_button55_clicked (GtkButton *button, SPSData *data)
{
  GThread *thread;
  GError *error = NULL;

  if (downloading != 1) {
    downloading = 1;

    // Spawn thread to download file.
    thread = g_thread_create ((GThreadFunc) get_countries, data, FALSE, &error);
    if (! thread) {
      g_print ("Error: Unable to create new thread for file download.%s\n", error->message);
      return (FALSE);
    }
  }

  return (TRUE);
}

// View www.iso.org's ISO 3166-1 country code list file.
gboolean
on_button56_clicked (GtkButton *button, SPSData *data)
{
  int i, n;
  gchar *text, *filetext;
  GtkWidget *window, *scrolled_win, *textview;
  GtkTextBuffer *textbuffer;
  void *tmp;
  FILE *fp;

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_button56_clicked().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (100000 * sizeof (gchar));
  if (tmp != NULL) {
    filetext = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'filetext' in on_button56_clicked().\n");
    exit (EXIT_FAILURE);
  }

  // Report error if we don't have the file.
  if (data->countries_have != 1) {
    sprintf (text, "The country code file is not available to display. Please download it first.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (filetext);
    return (FALSE);
  }

  // Create new window to display file contents
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (window), TRUE);
  gtk_window_set_title (GTK_WINDOW (window), "ISO 3166-1 Country Code List (www.iso.org)");
  gtk_window_set_default_size (GTK_WINDOW (window), 800, 600);
  gtk_container_set_border_width (GTK_CONTAINER (window), 1);

  // Create a textview object
  textview = gtk_text_view_new ();
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview), TRUE);

  // Open www.iso.org's country code file.
  fp = fopen (COUNTRY_FILE, "r");
  if (fp == NULL) {
    sprintf (text, "Can't open file %s in on_button56_clicked().", COUNTRY_FILE);
    data->parent = data->menubar1;
    data->message = text;
    report_error (data);
    g_free (text);
    g_free (filetext);
    return (FALSE);
  }

  // Transfer file contents to a string.
  i = 0;
  while ((n = fgetc (fp)) != EOF) {
    filetext[i] = n;
    i++;
  }
  fclose (fp);

  // Put the text into the buffer.
  gtk_text_buffer_set_text (textbuffer, filetext, -1);

  // Add scrolled window and textview to our new window and display.
  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (scrolled_win), textview);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  gtk_container_add (GTK_CONTAINER (window), scrolled_win);
  gtk_widget_show_all (window);

  g_free (text);
  g_free (filetext);

  return (TRUE);
}

// Read Regional Internet Registry ASN / IP delegation files and place into arrays in memory.
gboolean
read_delegations_and_countries (SPSData *data)
{
  int i, j, k, m, n, p;
  unsigned char *num;
  gchar *text;
  void *tmp;
  FILE *fi;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in read_delegations_and_countries().\n");
    exit (EXIT_FAILURE);
  }
  
  tmp = (unsigned char *) g_malloc0 (1024 * sizeof (unsigned char));
  if (tmp != NULL) {
    num = tmp;}
  else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'num' in read_delegations_and_countries().\n");
    exit (EXIT_FAILURE);
  }

  if (files_available (data) == 2) {
    //  0 = no files available
    //  1 = only country code file available
    //  2 = all RIR delegation files and country code file available

    for (i=0; i<5; i++) {

      // Open Regional Internet Registry (RIR) IP delegation file.
      fi = fopen (rir_file[i], "rb");
      if (fi == NULL) {
        sprintf (text, "In read_delegations_and_countries() in callbacks.c: Cannot open RIR delegation file %s.", rir_file[i]);
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (num);
        return (FALSE);
      }

      // Check for empty file.
      if (fgetc (fi) == EOF) {
        sprintf (text, "In read_delegations_and_countries() in callbacks.c: RIR delegation file %s appears to be empty.", rir_file[i]);
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (num);
        return (FALSE);
      } else {

        // Count all lines in file (excluding comments and blank lines).
        // We will use this for allocating memory.
        rewind (fi);
        data->nrecords[i] = 0;

        for (;;) {  // Loop through all records (lines) in file.

          // Grab next line from file.
          n = grab_line (fi, num);

          // End-of-file was reached with zero line length.
          // Exit record loop for this file.
          if ((n == 1) && (num[0] == 0)) {
            break;
          }

          // If line has zero length, then we hit a blank line.
          // Don't count as having extracted a record.
          if (num[0] == 0) {
            continue;
          }

          // Ignore title line, comments, or lines starting with a space.
          // Don't count as having extracted a record.
          if ((num[0] == '#') || (num[0] == '2') || (num[0] == ' ')) {
            continue;
          }

          // Ignore summary lines.
          // Don't count as having extracted a record.
          if ((num[5] == '*') || (num[6] == '*') || (num[7] == '*') || (num[8] == '*') || (num[9] == '*') ||
             (num[10] == '*')) {
            continue;
          }

          data->nrecords[i]++;
        }
      }
      fclose (fi);
    }  // End loop of RIR files.

    // Allocate memory for array[rir][record][field][unsigned char].
    data->array = (unsigned char ****) g_malloc0 (5 * sizeof (unsigned char ***));
    for (i=0; i<5; i++) {
      data->array[i] = (unsigned char ***) g_malloc0 (data->nrecords[i] * sizeof (unsigned char **));
      for (j=0; j<data->nrecords[i]; j++) {
        data->array[i][j] = (unsigned char **) g_malloc0 (7 * sizeof (unsigned char **));
        data->array[i][j][0] = (unsigned char *) g_malloc0 (8 * sizeof (unsigned char));  // RIR name
        data->array[i][j][1] = (unsigned char *) g_malloc0 (3 * sizeof (unsigned char));  // Country code
        data->array[i][j][2] = (unsigned char *) g_malloc0 (5 * sizeof (unsigned char));  // asn/ipv4/ipv6
        data->array[i][j][3] = (unsigned char *) g_malloc0 (40 * sizeof (unsigned char));  // Start
        data->array[i][j][4] = (unsigned char *) g_malloc0 (10 * sizeof (unsigned char));  // Value
        data->array[i][j][5] = (unsigned char *) g_malloc0 (10 * sizeof (unsigned char));  // Date
        data->array[i][j][6] = (unsigned char *) g_malloc0 (10 * sizeof (unsigned char)) ;  // Status
      }
    }

    // Populate array[rir][record][field][char].
    for (i=0; i<5; i++) {  // Loop though RIR files.

      // Open Regional Internet Registry (RIR) IP delegation file.
      fi = fopen (rir_file[i], "rb");
      if (fi == NULL) {
        sprintf (text, "In read_delegations_and_countries() in callbacks.c: Cannot open RIR delegation file %s.", rir_file[i]);
        data->parent = data->main_window;
        data->message = text;
        report_error (data);
        g_free (text);
        g_free (num);
        return (FALSE);
      } else {
        for (j=0; j<data->nrecords[i]; j++) {  // Loop through records.

          // Grab next line from file.
          n = grab_line (fi, num);

          // End-of-file was reached with zero line length.
          // Exit record loop for this file.
          if ((n == 1) && (num[0] == 0)) {
            break;
          }

          // If line has zero length, then we hit a blank line.
          if (num[0] == 0) {
            j--;  // Don't count as having extracted a record.
            continue;
          }

          // Ignore title line, comments, or lines starting with a space.
          if ((num[0] == '#') || (num[0] == '2') || (num[0] == ' ')) {
            j--;  // Don't count as having extracted a record.
            continue;
          }

          // Ignore summary lines.
          if ((num[5] == '*') || (num[6] == '*') || (num[7] == '*') || (num[8] == '*') || (num[9] == '*') ||
             (num[10] == '*')) {
            j--;  // Don't count as having extracted a record.
            continue;
          }

          // Extract field data.
          p = 0;  // P is character index for current line.
          for (k = 0; k<7; k++) {  // Loop through fields.
            m = 0;
            while ((num[p] != '|') && (num[p] != 0) && (num[p] != ' ')) {
              data->array[i][j][k][m] = num[p];
              m++;
              p++;
            }
            p++;
          }  // End Loop fields
        }  // End Loop records
        fclose (fi);
      }  // End If opened file.
    }  // End Loop of RIR files.

  }  // End if files_available == 2

  if ((files_available (data) == 1) || (files_available (data) == 2)) {

    // Open ISO 3166-1 country code list file.
    fi = fopen (COUNTRY_FILE, "rb");
    if (fi == NULL) {
      sprintf (text, "In read_delegations_and_countries() in callbacks.c: Cannot open country code file %s.", COUNTRY_FILE);
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (num);
      return (FALSE);
    }

    // Check for empty file.
    if (fgetc (fi) == EOF) {
      sprintf (text, "In read_delegations_and_countries() in callbacks.c: Country code file %s appears to be empty.", COUNTRY_FILE);
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (num);
      return (FALSE);
    } else {
      grab_line (fi, num);  // Skip the first line; it's just comments.
      data->ncountry = 0;

      for (;;) {  // Loop through all records (lines) in file.
        // Grab next line from file.
        n = grab_line (fi, num);

        // End-of-file was reached with zero line length.
        // Exit record loop for this file.
        // Don't count as having extracted a record.
        if ((n == 1) && (num[0] == 0)) {
          break;
        }

        // If line has zero length, then we hit a blank line.
        // Don't count as having extracted a record.
        if (num[0] == 0) {
          continue;
        }

        // Ignore comments or lines starting with a space.
        // Don't count as having extracted a record.
        if ((num[0] == '#') || (num[0] == ' ')) {
          continue;
        }

      data->ncountry++;
      }
    }
    fclose (fi);

    // Allocate memory for country code array country[record][field][char].
    data->cc = (unsigned char ***) g_malloc0 (data->ncountry * sizeof (unsigned char **));
    for (i=0; i<data->ncountry; i++) {
      data->cc[i] = (unsigned char **) g_malloc0 (2 * sizeof (unsigned char *));
      for (j=0; j<2; j++) {
        data->cc[i][j] = (unsigned char *) g_malloc0 (256 * sizeof (unsigned char));
      }
    }

    // Populate country code array country[record][field][char].
    // Open ISO 3166-1 country code list file.
    fi = fopen (COUNTRY_FILE, "rb");
    if (fi == NULL) {
      sprintf (text, "In read_delegations_and_countries() in callbacks.c: Cannot open country code file %s.", COUNTRY_FILE);
      data->parent = data->main_window;
      data->message = text;
      report_error (data);
      g_free (text);
      g_free (num);
      return (FALSE);
    } else {
      grab_line (fi, num);  // Skip the first line; it's just column headings.
      for (i=0; i<data->ncountry; i++) {  // Loop through records.

        // Grab next line from file.
        n = grab_line (fi, num);

        // End-of-file was reached with zero line length.
        // Exit record loop for this file.
        if ((n == 1) && (num[0] == 0)) {
          break;
        }

        // If line has zero length, then we hit a blank line.
        if (num[0] == 0) {
          i--;  // Don't count as having extracted a record.
          continue;
        }

        // Ignore comments or lines starting with a space.
        if ((num[0] == '#') || (num[0] == ' ')) {
          i--;  // Don't count as having extracted a record.
          continue;
        }

        // Extract field data.
        p = 0;  // P is character index for current line.
        for (j=0; j<2; j++) {  // Loop through fields.
          k = 0;
          while ((num[p] != ';') && (num[p] != 0)) {
            data->cc[i][j][k] = num[p];
            k++;
            p++;
          }
          p++;
        }  // End Loop fields
      }  // End Loop records
      fclose (fi);
    }  // End If opened file.

  }  // End if files available

  // Free allocated memory.
  g_free (text);
  g_free (num);

  return (TRUE);
}

int
grab_line (FILE *fi, unsigned char *line)
{
  int i, val;
  unsigned int n;

  val = 0;  // First assume no end-of-file will be reached.

  memset (line, 0, 1024 * sizeof (unsigned char));
  i = 0;  // i is pointer to byte in line.

  for (;;) {

    // Grab next byte from file.
    n = fgetc (fi);

    // End of file reached.
    if (n == EOF) {
      line[i] = 0;
      val = 1;
      break;
    }

    // Found a carriage return.
    if (n == '\r') {
      line[i] = 0;  // Replace with null.
      // Check if next byte is a newline or end of file.
      n = fgetc (fi);
      if (n == EOF) {
        val = 1;
        break;
      }
      if (n == '\n') {
        line[i] = 0;  // Replace with null.
      } else {
        fseek (fi, -1, SEEK_CUR);  // Backstep one byte; seems to be good data.
      }
      break;
    }

    // Found a new line.
    if (n == '\n') {
      line[i] = 0;  // Replace with null.
      // Check if next byte is a carriage return or end of file.
      n = fgetc (fi);
      if (n == EOF) {
        val = 1;
        break;
      }
      if (n == '\r') {
        line[i] = 0;  // Replace with null.
      } else {
        fseek (fi, -1, SEEK_CUR);  // Backstep one byte; seems to be good data.
      }
      break;
    }

    // Seems to be a valid character. Keep it.
    line[i] = n;
    i++;
  }

  return (val);
}

// Check to see if RIR delegation and country code files are available.
// Return values:
//  0 = no files available
//  1 = only country code file available
//  2 = all RIR delegation files and country code file available
gint
files_available (SPSData *data)
{
  int i, j;

  for (i=0; i<5; i++) {
    if (file_exists (rir_file[i]) == 1) {
      data->rir_files_have[i] = 1;
    } else {
      data->rir_files_have[i] = 0;
    }
  }

  if (file_exists (COUNTRY_FILE) == 1) {
    data->countries_have = 1;
  } else {
    data->countries_have = 0;
  }

  // RIR delegation files.
  i = 2;  // Assume we have all RIR files and country code file available.
  for (j=0; j<5; j++) {
    if (data->rir_files_have[j] == 0) {
      i = 1;  // We're missing at least one RIR file. Assume we have country code file though.
      break;
    }
  }

  // Country code file.
  if (data->countries_have == 0) {
    i = 0;  // We don't have a country code file.
  }

  return (i);
}

// Search www.iso.org's country name/code list for a country name.
// Return country code.
gboolean
find_country_code (gchar *cname, gchar *ccode, SPSData *data)
{
  int i, fail;
  gchar *text;
  void *tmp;

  // Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in find_country_code().\n");
    exit (EXIT_FAILURE);
  }

  memset (ccode, 0, 3 * sizeof (gchar));
  fail = 1;  // Initially assume we won't find the country name in the list.
  for (i=0; i<data->ncountry; i++) {
    if (compare (data->cc[i][0], (unsigned char *) cname) == 0) {
      memcpy (ccode, (gchar *) data->cc[i][1], 2 * sizeof (gchar));
      fail = 0;
      break;
    }
  }

  if (fail == 1) {
    sprintf (text, "Failed to find country name in list. Suggest viewing country code file for spelling.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    gtk_entry_set_text (GTK_ENTRY (data->entry501), "");
    return (FALSE);
  }

  g_free (text);

  return (TRUE);
}

// Search www.iso.org's country name/code list for a country code.
// Return country name.
gboolean
find_country_name (gchar *cname, gchar *ccode, SPSData *data)
{
  int i, fail;
  gchar *text;
  void *tmp;

  // Array for error messages.
  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in find_country_name().\n");
    exit (EXIT_FAILURE);
  }

  memset (cname, 0, 1024 * sizeof (gchar));
  fail = 1;  // Initially assume we won't find the country code in the list.
  for (i=0; i<data->ncountry; i++) {
    if (compare (data->cc[i][1], (unsigned char *) ccode) == 0) {
      memcpy (cname, (gchar *) data->cc[i][0], 256 * sizeof (gchar));
      fail = 0;
      break;
    }
  }

  if (fail == 1) {
    sprintf (text, "Failed to find country code in list. Suggest viewing country code file for spelling.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    gtk_entry_set_text (GTK_ENTRY (data->entry500), "");
    return (FALSE);
  }

  g_free (text);

  return (TRUE);
}

// Search for ASN delegations by country code.
// This function will be executed as a separate thread.
gboolean
search_asn_by_ccode (SPSData *data)
{
  int i, j, start, value, year, month, day, fail;
  gchar asn[4], temp[5], *text;
  gchar mon_name[12][10] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
  };
  void *tmp;

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in search_asn_by_ccode().\n");
    exit (EXIT_FAILURE);
  }

  strcpy (asn, "asn");

  memset (text, 0, 5000000 * sizeof (gchar));
  fail = 1;
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      if ((compare (data->array[i][j][1], (unsigned char *) data->ccode) == 0) &&
          (compare (data->array[i][j][2], (unsigned char *) asn) == 0)) {
        start = atoi ((gchar *) data->array[i][j][3]);
        value = atoi ((gchar *) data->array[i][j][4]);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) data->array[i][j][5], 4 * sizeof (gchar));
        year = atoi (temp);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 4), 2 * sizeof (gchar));
        month = atoi (temp) - 1;
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 6), 2 * sizeof (gchar));
        day = atoi (temp);
        if (year != 0) {  // Date is known if not equal to zero.
          if (value == 1) {
           sprintf (text, "%sASN %i %s to %s (%s) by %s on %s %i, %i.\n", text, start, (gchar *) data->array[i][j][6], data->cname, data->ccode, rir_www[i], mon_name[month], day, year);
          } else {
             sprintf (text, "%sASN %i to ASN %i %s to %s (%s) by %s on %s %i, %i.\n", text, start, start + value - 1, (gchar *) data->array[i][j][6], data->cname, data->ccode, rir_www[i], mon_name[month], day, year);
          }
        } else {  // Date is 00000000, which means it is unavailable.
          if (value == 1) {
           sprintf (text, "%sASN %i %s to %s (%s) by %s (date unavailable).\n", text, start, (gchar *) data->array[i][j][6], data->cname, data->ccode, rir_www[i]);
          } else {
             sprintf (text, "%sASN %i to ASN %i %s to %s (%s) by %s (date unavailable).\n", text, start, start + value - 1, (gchar *) data->array[i][j][6], data->cname, data->ccode, rir_www[i]);
          }
        }
        fail = 0;
      }
    }
  }

  if (fail == 1) {
    memset (text, 0, 5000000 * sizeof (gchar));
    sprintf (text, "Failed to find any ASN delegations for %s (%s).\n", data->cname, data->ccode);
  }

  post_asnsearch_results (text, data);

  // Clear ip_asn_searching flag.
  G_LOCK (ip_asn_searching);
  ip_asn_searching = 0;
  G_UNLOCK (ip_asn_searching);

  g_free (text);

  return (TRUE);
}

// Start idle function which posts ASN delegation search results.
gboolean
post_asnsearch_results (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    data->asn_results = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->asn_results' in post_asnsearch_results().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->asn_results, message);  // Copy message.

  g_idle_add ((GSourceFunc) asn_results, data);
  sleep (1);

  return (TRUE);
}

// Idle function to show ASN delegation search results in textview.
gboolean
asn_results (SPSData *data)
{
  GtkTextBuffer *textbuffer16;

  textbuffer16 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview16));
  gtk_text_buffer_set_text (textbuffer16, data->asn_results, -1);

  // Free memory allocated for message string.
  g_free (data->asn_results);

  return (FALSE);
}

// Search delegation files for asn. Return country code/name.
// This function will be executed as a separate thread.
gboolean
rev_search_asn (SPSData *data)
{
  int i, j, start, value, year, month, day, fail;
  gchar cname[1024], asns[4], temp[5], *text;
  gchar mon_name[12][10] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
  };
  void *tmp;

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in rev_search_asn().\n");
    exit (EXIT_FAILURE);
  }

  strcpy (asns, "asn");
  memset (cname, 0, 1024 * sizeof (gchar));

  memset (text, 0, 5000000 * sizeof (gchar));
  fail = 1;
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      if (compare (data->array[i][j][2], (unsigned char *) asns) == 0) {
        start = atoi ((gchar *) data->array[i][j][3]);
        value = atoi ((gchar *) data->array[i][j][4]);
        if ((start <= data->asn) && (data->asn <= (start + value - 1))) {
          find_country_name (cname, (gchar *) data->array[i][j][1], data);
          memset (temp, 0, 5 * sizeof (gchar));
          memcpy (temp, (gchar *) data->array[i][j][5], 4 * sizeof (gchar));
          year = atoi (temp);
          memset (temp, 0, 5 * sizeof (gchar));
          memcpy (temp, (gchar *) (data->array[i][j][5] + 4), 2 * sizeof (gchar));
          month = atoi (temp) - 1;
          memset (temp, 0, 5 * sizeof (gchar));
          memcpy (temp, (gchar *) (data->array[i][j][5] + 6), 2 * sizeof (gchar));
          day = atoi (temp);
          if (year != 0) {  // Date is known if not equal to zero.
            if (value == 1) {
              sprintf (text, "%sASN %i %s to %s (%s) by %s on %s %i, %i.\n", text, start, (char *) data->array[i][j][6], cname, (char *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
            } else {
              sprintf (text, "%sASN %i to ASN %i %s to %s (%s) by %s on %s %i, %i.\n", text, start, start + value - 1, (char *) data->array[i][j][6], cname, (char *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
            }
          } else {
            if (value == 1) {
              sprintf (text, "%sASN %i %s to %s (%s) by %s (date unavailable).\n", text, start, (char *) data->array[i][j][6], cname, (char *) data->array[i][j][1], rir_www[i]);
            } else {
              sprintf (text, "%sASN %i to ASN %i %s to %s (%s) by %s (date unavailable).\n", text, start, start + value - 1, (char *) data->array[i][j][6], cname, (char *) data->array[i][j][1], rir_www[i]);
            }
          }
          fail = 0;
          break;
        }
      }
    }
    if (fail == 0) break;
  }

  if (fail == 1) {
    memset (text, 0, 5000000 * sizeof (gchar));
    sprintf (text, "Failed to find a delegation for ASN %i.\n", data->asn);
  }

  post_asnrevsearch_results (text, data);

  // Clear ip_asn_searching flag.
  G_LOCK (ip_asn_searching);
  ip_asn_searching = 0;
  G_UNLOCK (ip_asn_searching);

  g_free (text);

  return (TRUE);
}

// Start idle function which posts ASN delegation search results.
gboolean
post_asnrevsearch_results (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    data->asn_results = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->asn_results' in post_asnrevsearch_results().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->asn_results, message);  // Copy message.

  g_idle_add ((GSourceFunc) revasn_results, data);
  sleep (1);

  return (TRUE);
}

// Idle function to show ASN delegation search results in textview.
gboolean
revasn_results (SPSData *data)
{
  GtkTextBuffer *textbuffer17;

  textbuffer17 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview17));
  gtk_text_buffer_set_text (textbuffer17, data->asn_results, -1);

  // Free memory allocated for message string.
  g_free (data->asn_results);

  return (FALSE);
}

// Search for IPv4 address delegations by country code.
// This function will be executed as a separate thread.
gboolean
search_ipv4_by_ccode (SPSData *data)
{
  unsigned int value;
  int i, j, year, month, day, fail;
  gchar ipv4a[5], temp[5], *text;
  gchar mon_name[12][10] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
  };
  void *tmp;

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in search_ipv4_by_ccode().\n");
    exit (EXIT_FAILURE);
  }

  strcpy (ipv4a, "ipv4");

  memset (text, 0, 5000000 * sizeof (gchar));
  fail = 1;
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      if ((compare (data->array[i][j][1], (unsigned char *) data->ccode) == 0) &&
          (compare (data->array[i][j][2], (unsigned char *) ipv4a) == 0)) {
        value = atoi ((gchar *) data->array[i][j][4]);
        parse_ipv4 ((gchar *) data->array[i][j][3], value, text);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) data->array[i][j][5], 4 * sizeof (gchar));
        year = atoi (temp);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 4), 2 * sizeof (gchar));
        month = atoi (temp) - 1;
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 6), 2 * sizeof (gchar));
        day = atoi (temp);
        if (year != 0) {  // Date is known if not equal to zero.
          if (value == 1) {
            sprintf (text, "%s %s to %s (%s) by %s on %s %i, %i.\n", text, (gchar *) data->array[i][j][6], data->cname, (gchar *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
          } else {
            sprintf (text, "%s %s to %s (%s) by %s on %s %i, %i.\n", text, (gchar *) data->array[i][j][6], data->cname, (gchar *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
          }
        } else {
          if (value == 1) {
            sprintf (text, "%s %s to %s (%s) by %s (date unavailable).\n", text, (gchar *) data->array[i][j][6], data->cname, (gchar *) data->array[i][j][1], rir_www[i]);
          } else {
            sprintf (text, "%s %s to %s (%s) by %s (date unavailable).\n", text, (gchar *) data->array[i][j][6], data->cname, (gchar *) data->array[i][j][1], rir_www[i]);
          }
        }
        fail = 0;
      }
    }
  }

  if (fail == 1) {
    memset (text, 0, 5000000 * sizeof (gchar));
    sprintf (text, "Failed to find any IPv4 address delegations for %s (%s).\n", data->cname, data->ccode);
  }

  post_ip4search_results (text, data);

  // Clear ip_asn_searching flag.
  G_LOCK (ip_asn_searching);
  ip_asn_searching = 0;
  G_UNLOCK (ip_asn_searching);

  g_free (text);

  return (TRUE);
}

// Start idle function which posts IPv4 address delegation search results.
gboolean
post_ip4search_results (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    data->ip4_results = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ip4_results' in post_ip4search_results().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->ip4_results, message);  // Copy message.

  g_idle_add ((GSourceFunc) ip4_results, data);
  sleep (1);

  return (TRUE);
}

// Idle function to show IPv4 address delegation search results in textview.
gboolean
ip4_results (SPSData *data)
{
  GtkTextBuffer *textbuffer18;

  textbuffer18 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview18));
  gtk_text_buffer_set_text (textbuffer18, data->ip4_results, -1);

  // Free memory allocated for message string.
  g_free (data->ip4_results);

  return (FALSE);
}

// Search delegation files for IPv4 address. Return country code/name.
// This function will be executed as a separate thread.
gboolean
rev_search_ipv4 (SPSData *data)
{
  unsigned int value;
  int i, j, year, month, day, fail;
  gchar cname[1024], ipv4a[5], temp[5], *text;
  gchar mon_name[12][10] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
  };
  void *tmp;

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in rev_search_ipv4().\n");
    exit (EXIT_FAILURE);
  }

  strcpy (ipv4a, "ipv4");
  memset (cname, 0, 1024 * sizeof (gchar));
  memset (text, 0, 5000000 * sizeof (gchar));
  fail = 1;
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      value = atoi ((gchar *) data->array[i][j][4]);
      if ((compare (data->array[i][j][2], (unsigned char *) ipv4a) == 0) &&
          (interval_test_ipv4 (data->ip4string, (gchar *) data->array[i][j][3], value) == 1)) {
        parse_ipv4 ((gchar *) data->array[i][j][3], value, text);
        find_country_name (cname, (gchar *) data->array[i][j][1], data);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) data->array[i][j][5], 4 * sizeof (gchar));
        year = atoi (temp);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 4), 2 * sizeof (gchar));
        month = atoi (temp) - 1;
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 6), 2 * sizeof (gchar));
        day = atoi (temp);
        if (year != 0) {  // Date is known if not equal to zero.
          if (value == 1) {
            sprintf (text, "%s %s to %s (%s) by %s on %s %i, %i.\n", text, (gchar *) data->array[i][j][6], cname, (gchar *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
          } else {
            sprintf (text, "%s %s to %s (%s) by %s on %s %i, %i.\n", text, (gchar *) data->array[i][j][6], cname, (gchar *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
          }
        } else {
          if (value == 1) {
            sprintf (text, "%s %s to %s (%s) by %s (date unavailable).\n", text, (gchar *) data->array[i][j][6], cname, (gchar *) data->array[i][j][1], rir_www[i]);
          } else {
            sprintf (text, "%s %s to %s (%s) by %s (date unavailable).\n", text, (gchar *) data->array[i][j][6], cname, (gchar *) data->array[i][j][1], rir_www[i]);
          }
        }
        fail = 0;
        break;
      }
    }
    if (fail == 0) break;
  }

  if (fail == 1) {
    memset (text, 0, 5000000 * sizeof (gchar));
    sprintf (text, "Failed to find a delegation for IPv4 address %s.\n", data->ip4string);
  }

  post_ip4revsearch_results (text, data);

  // Clear ip_asn_searching flag.
  G_LOCK (ip_asn_searching);
  ip_asn_searching = 0;
  G_UNLOCK (ip_asn_searching);

  g_free (text);

  return (TRUE);
}

// Start idle function which posts IPv4 address delegation search results.
gboolean
post_ip4revsearch_results (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    data->ip4_results = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ip4_results' in post_ip4revsearch_results().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->ip4_results, message);  // Copy message.

  g_idle_add ((GSourceFunc) revip4_results, data);
  sleep (1);

  return (TRUE);
}

// Idle function to show IPv4 address delegation search results in textview.
gboolean
revip4_results (SPSData *data)
{
  GtkTextBuffer *textbuffer20;

  textbuffer20 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview20));
  gtk_text_buffer_set_text (textbuffer20, data->ip4_results, -1);

  // Free memory allocated for message string.
  g_free (data->ip4_results);

  return (FALSE);
}

// Search for IPv6 address delegations by country code.
// This function will be executed as a separate thread.
gboolean
search_ipv6_by_ccode (SPSData *data)
{
  unsigned int value;
  int i, j, year, month, day, fail;
  gchar ipv6a[5], temp[5], *text;
  gchar mon_name[12][10] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
  };
  void *tmp;

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in search_ipv6_by_ccode().\n");
    exit (EXIT_FAILURE);
  }

  strcpy (ipv6a, "ipv6");

  memset (text, 0, 5000000 * sizeof (gchar));
  fail = 1;
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      if ((compare (data->array[i][j][1], (unsigned char *) data->ccode) == 0) &&
          (compare (data->array[i][j][2], (unsigned char *) ipv6a) == 0)) {
        value = atoi ((gchar *) data->array[i][j][4]);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) data->array[i][j][5], 4 * sizeof (gchar));
        year = atoi (temp);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 4), 2 * sizeof (gchar));
        month = atoi (temp) - 1;
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 6), 2 * sizeof (gchar));
        day = atoi (temp);
        if (year != 0) {  // Date is known if not equal to zero.
          sprintf (text, "%s%s/%i %s to %s (%s) by %s on %s %i, %i.\n", text, (gchar *) data->array[i][j][3], value, (gchar *) data->array[i][j][6], data->cname, (gchar *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
        } else {
          sprintf (text, "%s%s/%i %s to %s (%s) by %s (date unavailable).\n", text, (gchar *) data->array[i][j][3], value, (gchar *) data->array[i][j][6], data->cname, (gchar *) data->array[i][j][1], rir_www[i]);
        }
        fail = 0;
      }
    }
  }

  if (fail == 1) {
    memset (text, 0, 5000000 * sizeof (gchar));
    sprintf (text, "Failed to find any IPv6 address delegations for %s (%s).\n", data->cname, data->ccode);
  }

  post_ip6search_results (text, data);

  // Clear ip_asn_searching flag.
  G_LOCK (ip_asn_searching);
  ip_asn_searching = 0;
  G_UNLOCK (ip_asn_searching);

  g_free (text);

  return (TRUE);
}

// Start idle function which posts IPv6 address delegation search results.
gboolean
post_ip6search_results (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) { 
    data->ip6_results = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ip6_results' in post_ip6search_results().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->ip6_results, message);  // Copy message.
  
  g_idle_add ((GSourceFunc) ip6_results, data);
  sleep (1);
  
  return (TRUE);
}

// Idle function to show IPv6 address delegation search results in textview.
gboolean
ip6_results (SPSData *data)
{
  GtkTextBuffer *textbuffer19;

  textbuffer19 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview19));
  gtk_text_buffer_set_text (textbuffer19, data->ip6_results, -1);

  // Free memory allocated for message string.
  g_free (data->ip6_results);

  return (FALSE);
}

// Search delegation files for IPv6 address. Return country code/name.
// This function will be executed as a separate thread.
gboolean
rev_search_ipv6 (SPSData *data)
{
  unsigned int value;
  int i, j, year, month, day, fail;
  gchar cname[1024], ipv6a[5], temp[5], *text;
  gchar mon_name[12][10] = {
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
  };
  void *tmp;

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in rev_search_ipv6().\n");
    exit (EXIT_FAILURE);
  }

  strcpy (ipv6a, "ipv6");
  memset (cname, 0, 1024 * sizeof (gchar));
  memset (text, 0, 5000000 * sizeof (gchar));
  fail = 1;
  for (i=0; i<5; i++) {
    for (j=0; j<data->nrecords[i]; j++) {
      value = atoi ((gchar *) data->array[i][j][4]);
      if ((compare (data->array[i][j][2], (unsigned char *) ipv6a) == 0) &&
          (interval_test_ipv6 (data->ip6string, (gchar *) data->array[i][j][3], value) == 1)) {
        find_country_name (cname, (gchar *) data->array[i][j][1], data);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) data->array[i][j][5], 4 * sizeof (gchar));
        year = atoi (temp);
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 4), 2 * sizeof (gchar));
        month = atoi (temp) - 1;
        memset (temp, 0, 5 * sizeof (gchar));
        memcpy (temp, (gchar *) (data->array[i][j][5] + 6), 2 * sizeof (gchar));
        day = atoi (temp);
        if (year != 0) {  // Date is known if not equal to zero.
          sprintf (text, "%s/%i %s to %s (%s) by %s on %s %i, %i.\n", (gchar *) data->array[i][j][3], value, (gchar *) data->array[i][j][6], cname, (gchar *) data->array[i][j][1], rir_www[i], mon_name[month], day, year);
        } else {
          sprintf (text, "%s/%i %s to %s (%s) by %s (date unavailable).\n", (gchar *) data->array[i][j][3], value, (gchar *) data->array[i][j][6], cname, (gchar *) data->array[i][j][1], rir_www[i]);
        }
        fail = 0;
        break;
      }
    }
    if (fail == 0) break;
  }

  if (fail == 1) {
    memset (text, 0, 5000000 * sizeof (gchar));
    sprintf (text, "Failed to find a delegation for IPv6 address %s.\n", data->ip6string);
  }

  post_ip6revsearch_results (text, data);

  // Clear ip_asn_searching flag.
  G_LOCK (ip_asn_searching);
  ip_asn_searching = 0;
  G_UNLOCK (ip_asn_searching);

  g_free (text);

  return (TRUE);
}

// Start idle function which posts IPv6 address delegation search results.
gboolean
post_ip6revsearch_results (gchar *message, SPSData *data)
{
  void *tmp;

  // Allocate memory for string; Idle function will free it.
  tmp = (gchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    data->ip6_results = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'data->ip6_results' in post_ip6revsearch_results().\n");
    exit (EXIT_FAILURE);
  }
  strcpy (data->ip6_results, message);  // Copy message.

  g_idle_add ((GSourceFunc) revip6_results, data);
  sleep (1);

  return (TRUE);
}

// Idle function to show IPv6 address delegation search results in textview.
gboolean
revip6_results (SPSData *data)
{
  GtkTextBuffer *textbuffer20;

  textbuffer20 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview20));
  gtk_text_buffer_set_text (textbuffer20, data->ip6_results, -1);

  // Free memory allocated for message string.
  g_free (data->ip6_results);

  return (FALSE);
}

// Parse character string into 4-byte IPv4 address, and
// determine CIDR (slash) notation or IPv4 range based upon "value"
// given in delegation file.
int
parse_ipv4 (gchar *ip4string, unsigned int val, gchar *answer)
{
  int i, j, p, flag;
  unsigned int ip4[4], inc[4];
  gchar temp[4];

  // Parse character string into 4-byte IPv4 address.
  p = 0;
  for (i=3; i>=0; i--) {
    j = 0;
    memset (temp, 0, 4 * sizeof (gchar));
    while ((ip4string[p] != '.') && (ip4string[p] != 0)) {
      temp[j] = ip4string[p];
      p++;
      j++;
    }
    ip4[i] = atoi (temp);
    p++;  // Skip the '.'
  }

  // Can we use slash notation to express range of addresses?
  flag = 0;
  for (i=0; i<31; i++) {
    if (val == (1u << i)) {
      flag++;
      break;
    }
  }
  if (flag == 1) {
    sprintf (answer, "%s%i.%i.%i.%i/%i", answer, ip4[3], ip4[2], ip4[1], ip4[0], 32 - i);

  // No, we cannot use slash notation. We'll show range instead.
  } else {
    for (i=3; i>0; i--) {
      sprintf (answer, "%s%i.", answer, ip4[i]);  // Starting address
    }
    sprintf (answer, "%s%i", answer, ip4[0]);

    // Calculate increments.
    j = 24;
    for (i=3; i>=0; i--) {
      inc[i] = (val >> j) & 255u;
      j -= 8;
    }

    // Calculate final address by adding increments with carry.
    for (i=0; i<3; i++) {
      if ((ip4[i] + inc[i]) > 255u) {
        inc[i+1]++;
        ip4[i] += inc[i];
        ip4[i] = ip4[i] & 255u;
      } else {
        ip4[i] += inc[i];
      }
    }
    if ((ip4[3] + inc[3]) > 255u) {
      printf ("ERROR: This number of IPv4 addresses exceeds total IPv4 space.\n");
      printf ("       IPv4 address and value: %s %i\n", ip4string, val);
      exit (EXIT_FAILURE);
    } else {
      ip4[3] += inc[3];
    }

    sprintf (answer, "%s to %i.%i.%i.%i", answer, ip4[3], ip4[2], ip4[1], ip4[0]);
  }

  return (EXIT_SUCCESS);
}

// Parse character string into 16-byte IPv6 address.
int
parse_ipv6 (char *ip6string, unsigned int *ip6)
{
  int i, j, k, p, left, right, nright, tempi, flag;
  char temp[5];

  memset (ip6, 0, 16 * sizeof (unsigned int));

  // Determine if a double-colon exists, and if so, where.
  flag = 0;
  p = 0;
  while (ip6string[p] != 0) {
    if ((ip6string[p] == ':') && (ip6string[p+1] == ':')) {
      flag = 1;
      left = p;  // Index of left colon in string.
      right = p + 1;  // Index of right colon in string.
      break;
    }
    p++;
  }

  // We have a double-colon.
  if (flag == 1) {
    // Work on the left side first.
    k = 15;
    p = 0;
    while (p < left) {
      i = 0;
      memset (temp, 0, 5 * sizeof (char));
      while (ip6string[p] != ':') {
        temp[i] = ip6string[p];
        i++;
        p++;
      }
      p++;  // Skip ':'
      sscanf (temp, "%x", &tempi);
      if (tempi <= 255) {
        k--;
        ip6[k] = tempi;
        k--;
      } else {
        ip6[k] = (int) (tempi / 256);
        k--;
        ip6[k] = tempi % 256;
        k--;
      }
    }

    // Now work on right side.
    // First count how many groups on right side of double-colon.
    // Recall that 'right' is string index to right colon of double-colon.
    if (ip6string[right+1] != 0) {  // There's numbers to the right of the right colon.
      i = 0;
      p = right + 1;
      while (ip6string[p] != 0) {
        if (ip6string[p] == ':') i++;
        p++;
      }
      i++;
      nright = i;

      k = (i * 2) - 1;
      p = right + 1;
      for (j=0; j<nright; j++) {
        i = 0;
        memset (temp, 0, 5 * sizeof (char));
        while ((ip6string[p] != ':') && (ip6string[p] != 0)) {
          temp[i] = ip6string[p];
          i++;
          p++;
        }
        p++;  // Skip ':'
        sscanf (temp, "%x", &tempi);
        if (tempi <= 255) {
          k--;
          ip6[k] = tempi;
          k--;
        } else {
          ip6[k] = (int) (tempi / 256);
          k--;
          ip6[k] = tempi % 256;
          k--;
        }
      }
    }

  // We don't have a double-colon.
  } else {
    k = 15;
    p = 0;
    while (ip6string[p] != 0) {
      i = 0;
      memset (temp, 0, 5 * sizeof (char));
      while ((ip6string[p] != ':') && (ip6string[p] != 0)) {
        temp[i] = ip6string[p];
        i++;
        p++;
      }
      p++;  // Skip ':'
      sscanf (temp, "%x", &tempi);
      if (tempi <= 255) {
        k--;
        ip6[k] = tempi;
        k--;
      } else {
        ip6[k] = (int) (tempi / 256);
        k--;
        ip6[k] = tempi % 256;
        k--;
      }
    }
  }

  return (EXIT_SUCCESS);
}

// Test if an IPv4 address is within a range of IPv4 addresses.
int
interval_test_ipv4 (char *teststring, char *ip4string, unsigned int val)
{
  int i, j, p, flag;
  unsigned int test[4], ip4_start[4], inc[4], ip4_end[4];
  int t[4], s[4], e[4];
  char temp[4];

  // Extract the test IPv4 address from string.
  p = 0;
  for (i=3; i>=0; i--) {
    j = 0;
    memset (temp, 0, 4 * sizeof (char));
    while ((teststring[p] != '.') && (teststring[p] != 0)) {
      temp[j] = teststring[p];
      p++;
      j++;
    }
    test[i] = atoi (temp);
    p++;  // Skip the '.'
  }

  // Determine the starting IPv4 address of the range.
  p = 0;
  for (i=3; i>=0; i--) {
    j = 0;
    memset (temp, 0, 4 * sizeof (char));
    while ((ip4string[p] != '.') && (ip4string[p] != 0)) {
      temp[j] = ip4string[p];
      p++;
      j++;
    }
    ip4_start[i] = atoi (temp);
    p++;  // Skip the '.'
  }

  // Calculate increments.
  val -= 1u;
  j = 24;
  for (i=3; i>=0; i--) {
    inc[i] = (val >> j) & 255u;
    j -= 8;
  }

  // Determine final IPv4 address by adding increments with carry.
  for (i=0; i<3; i++) {
    if ((ip4_start[i] + inc[i]) > 255u) {
      inc[i+1]++;
      ip4_end[i] = ip4_start[i] + inc[i];
      ip4_end[i] = ip4_end[i] & 255u;
    } else {
      ip4_end[i] = ip4_start[i] + inc[i];
    }
  }
  if ((ip4_start[3] + inc[3]) > 255u) {
    printf ("ERROR: This number of IPv4 addresses exceeds total IPv4 space.\n");
    exit (EXIT_FAILURE);
  } else {
    ip4_end[3] = ip4_start[3] + inc[3];
  }

  // Test to see if IPv4 address falls within start-end range.
  // To do this, we do two subtractions with carry:
  // 1)  (test - start) > 0 
  // 2)  (end - test) > 0
  // If BOTH (1) and (2) are true, then we are within range.

  // Copy arrays so we don't corrupt originals.
  // Cast as int so we can get negative subtraction results.
  for (i=0; i<4; i++) {
    t[i] = (int) test[i];
    s[i] = (int) ip4_start[i];
  }

  flag = 1;  // Initially assume it does fall within range.

  // Is this IPv4 address less than the starting IP address of the range.
  // Let's check by subtraction, recursively carrying as necessary.
  for (i=0; i<4; i++) {              // Loop through bytes of address from LSB.
    if ((t[i] - s[i]) < 0) {         // Need to carry from next higher order byte.
      if ((i == 3) || (carry (4, i, t, s) == 0)) {
        flag = 0;
        break;
      }
    }
  }

  // Copy arrays so we don't corrupt originals.
  // Cast as int so we can get negative subtraction results.
  for (i=0; i<4; i++) {
    t[i] = (int) test[i];
    e[i] = (int) ip4_end[i];
  }

  // Is this IPv4 address greater than the end IP address of the range.
  // Let's check by subtraction, recursively carrying as necessary
  // Copy arrays so we don't corrupt originals.
  for (i=0; i<4; i++) {              // Loop through bytes of address from LSB.
    if ((e[i] - t[i]) < 0) {         // Need to carry from next higher order byte.
      if ((i == 3) || (carry (4, i, e, t) == 0)) {
        flag = 0;
        break;
      }
    }
  }

  return (flag);
}

// Test if an IPv6 address is within a range of IPv6 addresses.
int
interval_test_ipv6 (char *teststring, char *ip6string, unsigned int val)
{
  int i, flag, hostbits, hostbytes, bit_remainder;
  unsigned int test[16], ip6_start[16], ip6_end[16];
  int t[16], s[16], e[16];

  // Extract the test IPv6 address from string.
  parse_ipv6 (teststring, test);

  // Determine the starting IPv6 address of the range.
  parse_ipv6 (ip6string, ip6_start);

  // Determine final IPv6 address of the range.
  // Begin with start IP address.
  parse_ipv6 (ip6string, ip6_end);
  hostbits = 128 - val;  // Hostbits = 128 - prefix bits
  hostbytes = (int) hostbits / 8;
  bit_remainder = hostbits - (hostbytes * 8);
  for (i=0; i<hostbytes; i++) {
    ip6_end[i] = 255u;
  }
  if ((hostbits % 8) != 0) {
    for (i=0; i<bit_remainder; i++) {
      ip6_end[hostbytes] = ip6_end[hostbytes] | (1u << i);
    }
  }

  // Test to see if IPv6 address falls within start-end range.
  // To do this, we do two subtractions with carry:
  // 1)  (test - start) > 0 
  // 2)  (end - test) > 0
  // If BOTH (1) and (2) are true, then we are within range.

  // Copy arrays so we don't corrupt originals.
  // Cast as int so we can get negative subtraction results.
  for (i=0; i<16; i++) {
    t[i] = (int) test[i];
    s[i] = (int) ip6_start[i];
  }

  flag = 1;  // Initially assume it does fall within range.

  // Is this IPv6 address less than the starting IP address of the range.
  // Let's check by subtraction, recursively carrying as necessary.
  for (i=0; i<16; i++) {             // Loop through bytes of address from LSB.
    if ((t[i] - s[i]) < 0) {         // Need to carry from next higher order byte.
      if ((i == 15) || (carry (16, i, t, s) == 0)) {
        flag = 0;
        break;
      }
    }
  }

  // Copy arrays so we don't corrupt originals.
  // Cast as int so we can get negative subtraction results.
  for (i=0; i<16; i++) {
    t[i] = (int) test[i];
    e[i] = (int) ip6_end[i];
  }

  // Is this IPv6 address greater than the end IP address of the range.
  // Let's check by subtraction, recursively carrying as necessary
  // Copy arrays so we don't corrupt originals.
  for (i=0; i<16; i++) {             // Loop through bytes of address from LSB.
    if ((e[i] - t[i]) < 0) {         // Need to carry from next higher order byte.
      if ((i == 15) || (carry (16, i, e, t) == 0)) {
        flag = 0;
        break;
      }
    }
  }

  return (flag);
}

// Recursive carry for subtracting arbitrarily long numbers.
int
carry (int n, int i, int *t, int *s)
{
  // For IPv4 (4 bytes) n = 4. For IPv6 (16 bytes) n = 16.
  if (i == n) return 0;  // Can't carry from any higher order bytes.

  if (t[i+1] != 0) {
    t[i+1]--;
  } else {
    t[i+1] = 255;
    if (carry (n, i+1, t, s) == 0) return 0;  // Can't carry from any higher order bytes.
  }

  return (1);
}

// Compare two unsigned char strings.
int
compare (unsigned char *string1, unsigned char *string2)
{
  int i, len1, len2;

  len1 = 0;
  while ((string1[len1] != 0) && (len1 < 1024)) len1++;
  len2 = 0;
  while ((string2[len2] != 0) && (len2 < 1024)) len2++;

  if (len1 != len2) {
    return (1);
  } else {
    for (i=0; i<len1; i++) {
      if (string1[i] != string2[i]) {
        return (1);
      }
    }
  }
  return (0);
}

// Get sockaddr (IPv4 or IPv6).
// This excellent bit of code is taken from
// Brian “Beej Jorgensen” Hall (beej@beej.us).
void *
get_in_addr (struct sockaddr *sa)
{
  if (sa->sa_family == AF_INET) {
    return &(((struct sockaddr_in*)sa)->sin_addr);
  }

  return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

// Search country names and extract country code.
gboolean
on_entry500_activate (GtkWidget *entry500, SPSData *data)
{
  int i;
  const gchar *entry500_text;
  gchar *text, *temp, *ccode;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (3 * sizeof (gchar));
  if (tmp != NULL) {
    ccode = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ccode' in on_entry500_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    temp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp' in on_entry500_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry500_activate().\n");
    exit (EXIT_FAILURE);
  }

  if (files_available (data) == 0) {
    //  0 = no files available
    //  1 = only country code file available
    //  2 = all RIR delegation files and country code file available
    sprintf (text, "Please download the country code file first.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (ccode);
    g_free (temp);
    g_free (text);
    return (FALSE);
  }

  // Grab country name from ui.
  entry500_text = gtk_entry_get_text (GTK_ENTRY (entry500));

  // Convert any lower case UTF-8/ASCII to upper case UTF-8/ASCII.
  memset (temp, 0, 1024 * sizeof (gchar));
  strcpy (temp, entry500_text);
  for (i=0; i<strlen (temp); i++) {
    if (((unsigned char)temp[i] > 96) && ((unsigned char)temp[i] < 123)) {
      temp[i] -= 32;
    } else if ((unsigned char)temp[i] == 195) {
      i++;
      if ((((unsigned char)temp[i] > 159) && ((unsigned char)temp[i] < 183)) ||
          (((unsigned char)temp[i] > 183) && ((unsigned char)temp[i] < 191))) {
        temp[i] -= 32;
      }
    }
  }

  find_country_code (temp, ccode, data);
  gtk_entry_set_text (GTK_ENTRY (data->entry500), temp);
  gtk_entry_set_text (GTK_ENTRY (data->entry501), ccode);

  g_free (ccode);
  g_free (temp);
  g_free (text);

  return (TRUE);
}

// Search country codes and extract country name.
gboolean
on_entry501_activate (GtkWidget *entry501, SPSData *data)
{
  int i;
  const gchar *entry501_text;
  gchar *text, *temp, *cname;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    cname = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'cname' in on_entry501_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    temp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp' in on_entry501_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry501_activate().\n");
    exit (EXIT_FAILURE);
  }

  if (files_available (data) == 0) {
    //  0 = no files available
    //  1 = only country code file available
    //  2 = all RIR delegation files and country code file available
    sprintf (text, "Please download the country code file first.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (cname);
    g_free (temp);
    g_free (text);
    return (FALSE);
  }

  // Grab country code from ui.
  entry501_text = gtk_entry_get_text (GTK_ENTRY (entry501));

  // Convert any lower case UTF-8/ASCII to upper case UTF-8/ASCII.
  memset (temp, 0, 1024 * sizeof (gchar));
  strcpy (temp, entry501_text);
  for (i=0; i<strlen (temp); i++) {
    if (((unsigned char)temp[i] > 96) && ((unsigned char)temp[i] < 123)) {
      temp[i] -= 32;
    } else if ((unsigned char)temp[i] == 195) {
      i++;
      if ((((unsigned char)temp[i] > 159) && ((unsigned char)temp[i] < 183)) ||
          (((unsigned char)temp[i] > 183) && ((unsigned char)temp[i] < 191))) {
        temp[i] -= 32;
      }
    }
  }

  find_country_name (cname, temp, data);
  gtk_entry_set_text (GTK_ENTRY (data->entry500), cname);

  g_free (cname);
  g_free (temp);
  g_free (text);

  return (TRUE);
}

// Search for autonomous system (AS) number delegations by country name.
gboolean
on_entry502_activate (GtkWidget *entry502, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  int i;
  const gchar *entry502_text;
  gchar *temp, *ccode, *text;
  GtkTextBuffer *textbuffer;
  GThread *thread;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (3 * sizeof (gchar));
  if (tmp != NULL) {
    ccode = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ccode' in on_entry502_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    temp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp' in on_entry502_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry502_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textviews and say we are searching.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview16));
  gtk_text_buffer_set_text (textbuffer, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (ccode);
    g_free (temp);
    g_free (text);
    return (FALSE);
  }

  // Grab country name from ui.
  entry502_text = gtk_entry_get_text (GTK_ENTRY (entry502));

  // Convert any lower case UTF-8/ASCII to upper case UTF-8/ASCII.
  memset (temp, 0, 1024 * sizeof (gchar));
  strcpy (temp, entry502_text);
  for (i=0; i<strlen (temp); i++) {
    if (((unsigned char)temp[i] > 96) && ((unsigned char)temp[i] < 123)) {
      temp[i] -= 32;
    } else if ((unsigned char)temp[i] == 195) {
      i++;
      if ((((unsigned char)temp[i] > 159) && ((unsigned char)temp[i] < 183)) ||
          (((unsigned char)temp[i] > 183) && ((unsigned char)temp[i] < 191))) {
        temp[i] -= 32;
      }
    }
  }

  find_country_code (temp, ccode, data);
  strcpy (data->cname, temp);
  strcpy (data->ccode, ccode);

  // Spawn thread to perform IPv4 address delegation search.
  thread = g_thread_create ((GThreadFunc) search_asn_by_ccode, data, FALSE, &error);
  if (! thread) {
    g_print ("Error: Unable to create new thread to search ASN delegations by country name.%s\n", error->message);
    g_free (ccode);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry502), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry503), data->ccode);

  g_free (ccode);
  g_free (temp);
  g_free (text);

  return (TRUE);
}

// Search for autonomous system (AS) number delegations by country code.
gboolean
on_entry503_activate (GtkWidget *entry503, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  int i;
  const gchar *entry503_text;
  gchar *temp, *cname, *text;
  GtkTextBuffer *textbuffer;
  GThread *thread;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    cname = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'cname' in on_entry503_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    temp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp' in on_entry503_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry503_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textviews and say we are searching.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview16));
  gtk_text_buffer_set_text (textbuffer, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (cname);
    g_free (temp);
    g_free (text);
    return (FALSE);
  }

  // Grab country code from ui.
  entry503_text = gtk_entry_get_text (GTK_ENTRY (entry503));

  // Convert any lower case UTF-8/ASCII to upper case UTF-8/ASCII.
  memset (temp, 0, 1024 * sizeof (gchar));
  strcpy (temp, entry503_text);
  for (i=0; i<strlen (temp); i++) {
    if (((unsigned char)temp[i] > 96) && ((unsigned char)temp[i] < 123)) {
      temp[i] -= 32;
    } else if ((unsigned char)temp[i] == 195) {
      i++;
      if ((((unsigned char)temp[i] > 159) && ((unsigned char)temp[i] < 183)) ||
          (((unsigned char)temp[i] > 183) && ((unsigned char)temp[i] < 191))) {
        temp[i] -= 32;
      }
    }
  }

  find_country_name (cname, temp, data);
  strcpy (data->cname, cname);
  strcpy (data->ccode, temp);

  // Spawn thread to perform ASN delegation search.
  thread = g_thread_create ((GThreadFunc) search_asn_by_ccode, data, FALSE, &error);
  if (! thread) {
    g_print ("Error: Unable to create new thread to search ASN delegations by country name.%s\n", error->message);
    g_free (cname);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry502), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry503), data->ccode);

  g_free (cname);
  g_free (temp);
  g_free (text);

  return (TRUE);
}

// Reverse search for country by autonomous system (AS) number delegations.
gboolean
on_entry504_activate (GtkWidget *entry504, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  const gchar *entry504_text;
  gchar *text;
  GtkTextBuffer *textbuffer;
  GThread *thread;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry504_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textview and say we are searching.
  textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview17));
  gtk_text_buffer_set_text (textbuffer, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    return (FALSE);
  }

  // Grab country code from ui.
  entry504_text = gtk_entry_get_text (GTK_ENTRY (entry504));

  data->asn = atoi (entry504_text);

  // Spawn thread to perform ASN delegation search.
  thread = g_thread_create ((GThreadFunc) rev_search_asn, data, FALSE, &error);
  if (! thread) {
    g_print ("Error: Unable to create new thread to reverse-search ASN delegations by ASN.%s\n", error->message);
    gtk_text_buffer_set_text (textbuffer, "", -1);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  g_free (text);

  return (TRUE);
}

// Search for IP address delegations by country name.
gboolean
on_entry505_activate (GtkWidget *entry505, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  int i;
  const gchar *entry505_text;
  gchar *temp, *ccode, *text;
  GtkTextBuffer *textbuffer18;
  GtkTextBuffer *textbuffer19;
  GThread *thread4, *thread6;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (guchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry505_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (gchar *) g_malloc0 (3 * sizeof (gchar));
  if (tmp != NULL) {
    ccode = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'ccode' in on_entry505_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    temp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp' in on_entry505_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textviews and say we are searching.
  textbuffer18 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview18));
  gtk_text_buffer_set_text (textbuffer18, "Searching the database...", -1);
  textbuffer19 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview19));
  gtk_text_buffer_set_text (textbuffer19, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (ccode);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  // Grab country name from ui.
  entry505_text = gtk_entry_get_text (GTK_ENTRY (entry505));

  // Convert any lower case UTF-8/ASCII to upper case UTF-8/ASCII.
  memset (temp, 0, 1024 * sizeof (gchar));
  strcpy (temp, entry505_text);
  for (i=0; i<strlen (temp); i++) {
    if (((unsigned char)temp[i] > 96) && ((unsigned char)temp[i] < 123)) {
      temp[i] -= 32;
    } else if ((unsigned char)temp[i] == 195) {
      i++;
      if ((((unsigned char)temp[i] > 159) && ((unsigned char)temp[i] < 183)) ||
          (((unsigned char)temp[i] > 183) && ((unsigned char)temp[i] < 191))) {
        temp[i] -= 32;
      }
    }
  }

  find_country_code (temp, ccode, data);
  strcpy (data->cname, temp);
  strcpy (data->ccode, ccode);

  // Spawn thread to perform IPv4 address delegation search.
  thread4 = g_thread_create ((GThreadFunc) search_ipv4_by_ccode, data, FALSE, &error);
  if (! thread4) {
    g_print ("Error: Unable to create new thread to search IPv4 address delegations by country name.%s\n", error->message);
    g_free (ccode);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  // Spawn thread to perform IPv6 address delegation search.
  thread6 = g_thread_create ((GThreadFunc) search_ipv6_by_ccode, data, FALSE, &error);
  if (! thread6) {
    g_print ("Error: Unable to create new thread to search IPv6 address delegations by country name.%s\n", error->message);
    g_free (ccode);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry505), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry506), data->ccode);

  g_free (ccode);
  g_free (temp);
  g_free (text);

  return (TRUE);
}

// Search for IP address delegations by country code.
gboolean
on_entry506_activate (GtkWidget *entry506, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  int i;
  const gchar *entry506_text;
  gchar *temp, *cname, *text;
  GtkTextBuffer *textbuffer18;
  GtkTextBuffer *textbuffer19;
  GThread *thread4, *thread6;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (gchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    cname = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'cname' in on_entry506_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (1024 * sizeof (gchar));
  if (tmp != NULL) {
    temp = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'temp' in on_entry506_activate().\n");
    exit (EXIT_FAILURE);
  }

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry506_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textviews and say we are searching.
  textbuffer18 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview18));
  gtk_text_buffer_set_text (textbuffer18, "Searching the database...", -1);
  textbuffer19 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview19));
  gtk_text_buffer_set_text (textbuffer19, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (cname);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  // Grab country code from ui.
  entry506_text = gtk_entry_get_text (GTK_ENTRY (entry506));

  // Convert any lower case UTF-8/ASCII to upper case UTF-8/ASCII.
  memset (temp, 0, 1024 * sizeof (gchar));
  strcpy (temp, entry506_text);
  for (i=0; i<strlen (temp); i++) {
    if (((unsigned char)temp[i] > 96) && ((unsigned char)temp[i] < 123)) {
      temp[i] -= 32;
    } else if ((unsigned char)temp[i] == 195) {
      i++;
      if ((((unsigned char)temp[i] > 159) && ((unsigned char)temp[i] < 183)) ||
          (((unsigned char)temp[i] > 183) && ((unsigned char)temp[i] < 191))) {
        temp[i] -= 32;
      }
    }
  }

  find_country_name (cname, temp, data);
  strcpy (data->cname, cname);
  strcpy (data->ccode, temp);

  // Spawn thread to perform IPv4 address delegation search.
  thread4 = g_thread_create ((GThreadFunc) search_ipv4_by_ccode, data, FALSE, &error);
  if (! thread4) {
    g_print ("Error: Unable to create new thread to search IPv4 address delegations by country name.%s\n", error->message);
    g_free (cname);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  // Spawn thread to perform IPv6 address delegation search.
  thread6 = g_thread_create ((GThreadFunc) search_ipv6_by_ccode, data, FALSE, &error);
  if (! thread6) {
    g_print ("Error: Unable to create new thread to search IPv6 address delegations by country name.%s\n", error->message);
    g_free (cname);
    g_free (temp);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  gtk_entry_set_text (GTK_ENTRY (data->entry505), data->cname);
  gtk_entry_set_text (GTK_ENTRY (data->entry506), data->ccode);

  g_free (cname);
  g_free (temp);
  g_free (text);

  return (TRUE);
}

// Reverse-search for IPv4 address delegations by IPv4 address.
gboolean
on_entry507_activate (GtkWidget *entry507, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  const gchar *entry507_text;
  gchar *text;
  GtkTextBuffer *textbuffer20;
  GThread *thread;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry507_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textview and say we are searching.
  textbuffer20 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview20));
  gtk_text_buffer_set_text (textbuffer20, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  // Grab IPv4 address from ui.
  entry507_text = gtk_entry_get_text (GTK_ENTRY (entry507));

  // Check to see if address is a valid IPv4 address.
  if (is_valid_ip4 (entry507_text) != TRUE) {
    sprintf (text, "In callbacks.c: is_valid_ip4() failed in on_entry507_activate().\nAppears to be an invalid IPv4 address..");
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  strcpy (data->ip4string, entry507_text);

  // Spawn thread to perform IPv4 address delegation search.
  thread = g_thread_create ((GThreadFunc) rev_search_ipv4, data, FALSE, &error);
  if (! thread) {
    g_print ("Error: Unable to create new thread to reverse-search IPv4 address delegations by IPv4 address.%s\n", error->message);
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  g_free (text);

  return (TRUE);
}

// Reverse-search for IPv6 address delegations by IPv6 address.
gboolean
on_entry508_activate (GtkWidget *entry508, SPSData *data)
{
  // If a search thread is already active, then do nothing and return (FALSE).
  if (ip_asn_searching == 1) {
    return (FALSE);
  } else {
    ip_asn_searching = 1;
  }

  const gchar *entry508_text;
  gchar *text;
  GtkTextBuffer *textbuffer20;
  struct sockaddr_in6 sa;
  GThread *thread;
  GError *error = NULL;
  void *tmp;

  // Allocate memory for various arrays.

  tmp = (guchar *) g_malloc0 (5000000 * sizeof (gchar));
  if (tmp != NULL) {
    text = tmp;
  } else {
    fprintf (stderr, "ERROR: Cannot allocate memory for array 'text' in on_entry508_activate().\n");
    exit (EXIT_FAILURE);
  }

  // Clear any previous results in the textview and say we are searching.
  textbuffer20 = gtk_text_view_get_buffer (GTK_TEXT_VIEW (data->textview20));
  gtk_text_buffer_set_text (textbuffer20, "Searching the database...", -1);

  // If RIR files and country code file not available, return with no action.
  if (files_available (data) != 2) {
    sprintf (text, "You must download the RIR files and country code file before performing searches.");
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  // Grab IPv6 address from ui.
  entry508_text = gtk_entry_get_text (GTK_ENTRY (entry508));

  // Check to see if address is a valid IPv6 address.
  if (inet_pton (AF_INET6, entry508_text, &sa.sin6_addr) < 1) {
    sprintf (text, "In callbacks.c: inet_pton() failed in on_entry508_activate().\nAppears to be an invalid IPv6 address..");
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    data->parent = data->main_window;
    data->message = text;
    report_error (data);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  strcpy (data->ip6string, entry508_text);

  // Spawn thread to perform IPv4 address delegation search.
  thread = g_thread_create ((GThreadFunc) rev_search_ipv6, data, FALSE, &error);
  if (! thread) {
    g_print ("Error: Unable to create new thread to reverse-search IPv6 address delegations by IPv6 address.%s\n", error->message);
    gtk_text_buffer_set_text (textbuffer20, "", -1);
    g_free (text);
    ip_asn_searching = 0;
    return (FALSE);
  }

  g_free (text);

  return (TRUE);
}
