/*  Simple Packet Sender (SPS) - a linux packet crafting network
    tool using GTK+ graphical interface.

    Copyright (C) 2011  Hohlraum (h0h1r4um@yahoo.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#define _GNU_SOURCE           // setresuid()
#include <unistd.h>           // close(), setresuid()
#include <string.h>           // strcpy, memset(), and memcpy()
#include <time.h>             // localtime(), time()
#include <inttypes.h>         // int32_t

#include <netdb.h>            // struct addrinfo
#include <sys/types.h>        // needed for socket()
#include <sys/socket.h>       // needed for socket()
#include <netinet/in_systm.h>
#include <netinet/in.h>       // IPPROTO_IP, IPPROTO_ICMP, IPPROTO_TCP, IPPROTO_UDP, IPPROTO_IPV6, IPPROTO_ICMPV6
#include <netinet/ip.h>       // struct ip and IP_MAXPACKET (which is 65535), IPPROTO_IPV6, IPPROTO_ICMPV6
#include <netinet/ip6.h>      // struct ip6_hdr
#define __FAVOR_BSD           // Use BSD format of TCP and UDP header
#include <netinet/tcp.h>      // struct tcphdr
#include <netinet/ip_icmp.h>  // struct icmp, ICMP_ECHO
#include <netinet/icmp6.h>    // struct icmp6_hdr, ICMP6_ECHO_REQUEST, ICMP6_TIME_EXCEEDED, struct nd_neighbor_solicit, ND_NEIGHBOR_SOLICIT
#include <netinet/udp.h>      // struct udphdr
#include <arpa/inet.h>        // inet_pton() and inet_ntop()
#include <sys/ioctl.h>        // macro ioctl is defined
#include <bits/ioctls.h>      // defines values for argument "request" of ioctl. Here, we need SIOCGIFHWADDR
#include <net/if.h>           // struct ifreq
#include <bits/socket.h>      // structs msghdr and cmsghdr
#include <linux/if_ether.h>   // ETH_P_IP = 0x0800, ETH_P_IPV6 = 0x86DDD
#include <linux/if_packet.h>  // struct sockaddr_ll (see man 7 packet)
#include <net/ethernet.h>
#include <sys/time.h>         // gettimeofday()

#include <errno.h>            // errno, perror()

// Taken from <linux/ipv6.h>, also in <netinet/in.h>
struct in6_pktinfo {
        struct in6_addr ipi6_addr;
        int             ipi6_ifindex;
};

// Define an struct for ARP header
typedef struct _arp_hdr arp_hdr;
struct _arp_hdr {
  unsigned short htype;
  unsigned short ptype;
  unsigned char hlen;
  unsigned char plen;
  unsigned short opcode;
  unsigned char sender_mac[6];
  unsigned char sender_ip[4];
  unsigned char target_mac[6];
  unsigned char target_ip[4];
};

// Struct to hold ethernet header values
  typedef struct _ethhdr ethhdr;
  struct _ethhdr {
    unsigned char dst_mac[6];
    unsigned char src_mac[6];
    unsigned short int type_code;
  };

// Define some constants
#define ETH_HDRLEN 14      // Ethernet header length
#define IP4_HDRLEN 20      // IPv4 header length
#define IP6_HDRLEN 40      // IPv6 header length
#define TCP_HDRLEN 20      // TCP header length, excludes options data
#define ICMP_HDRLEN 8      // ICMP header length for echo request, excludes data
#define UDP_HDRLEN  8      // UDP header length, excludes data
#define ARP_HDRLEN 28      // ARP header length
#define ARPOP_REQUEST 1    // Taken from <linux/if_arp.h>
#define ARPOP_REPLY 2      // Taken from <linux/if_arp.h>
#define FTP_TIMEOUT 10     // FTP listening timeout
#define FTP_PORT "21"      // FTP port number
#define HTTP_TIMEOUT 10    // HTTP listening timeout
#define HTTP_PORT "80"     // HTTP port number

// www.iso.org's country code filename and directory (they change it once in a while)
#define COUNTRY_DIR "/iso/home/standards/country_codes/"
#define COUNTRY_FILE "country_names_and_code_elements_txt"

// Data objects for Regional Internet Registry names, URLs, directories, and filenames.
// Values are entered in main.c.
gchar **rir_name, **rir_www, **rir_url, **rir_dir, **rir_file;

// Some flags.
gint downloading;         // Flag to indicate a file download thread is active.
gint message_available;   // A message is available to be displayed in a dialog by thread_listen().
gint show_dl_messages;    // Flag to show sent and received messages on stdout as files download.
gint tracing;             // Flag to indicate a tracroute thread is active.
gint stop_thread_listen;  // Flag to cause timeout function thread_listen() to stop.
gint ip_asn_searching;    // Flag to indicate an IP delegation search or ASN delegation search thread is active.

// Main data structure definition
typedef struct _SPSData SPSData;
struct _SPSData {
  gchar *error_text;  // String to hold error messages.

  struct ip ip4hdr_tcp, ip4hdr_tcp_tr, sixto4ip4hdr_tcp, sixto4ip4hdr_tcp_tr;  // _tr is for traceroute
  struct ip6_hdr ip6hdr_tcp, ip6hdr_tcp_tr;
  struct ip ip4hdr_icmp, ip4hdr_icmp_tr, sixto4ip4hdr_icmp, sixto4ip4hdr_icmp_tr;
  struct ip6_hdr ip6hdr_icmp, ip6hdr_icmp_tr;
  struct ip ip4hdr_udp, ip4hdr_udp_tr, sixto4ip4hdr_udp, sixto4ip4hdr_udp_tr;
  struct ip6_hdr ip6hdr_udp, ip6hdr_udp_tr;
  struct tcphdr tcp4hdr, tcp4hdr_tr, tcp6hdr, tcp6hdr_tr;
  struct icmp icmp4hdr, icmp4hdr_tr;
  struct icmp6_hdr icmp6hdr, icmp6hdr_tr;
  struct udphdr udp4hdr, udp4hdr_tr, udp6hdr, udp6hdr_tr;
  GtkWidget *main_window;
  GtkWidget *menubar1;
  GtkWidget *view;
  GtkWidget *parent;  // Pointer to parent that spawned an error. e.g., data->menubar1, data->main_window, or data->traceroute_window

// Send page

// Select Packet Type
  GtkToggleButton *radiobutton1;  // TCP (IPv4) packet type
  GtkToggleButton *radiobutton2;  // ICMP (IPv4) packet type
  GtkToggleButton *radiobutton3;  // UDP (IPv4) packet type
  GtkToggleButton *radiobutton4;  // All (IPv4) packet types (i.e., cycle)
  GtkToggleButton *radiobutton5;  // TCP (IPv6) packet type
  GtkToggleButton *radiobutton6;  // ICMP (IPv6) packet type
  GtkToggleButton *radiobutton7;  // UDP (IPv6) packet type
  GtkToggleButton *radiobutton8;  // All (IPv6) packet types (i.e., cycle)
  unsigned char packet_type;      // Packet type
// Number of Packets to Send
  GtkWidget *entry18;             // Number of packets to send
  gint npackets;                  // Number of packets to send
  GtkToggleButton *checkbutton11; // Flood
// Activity Log
  GtkWidget *textview1;           // Activity log
  gchar *message;                 // Activity message
// IPv6 over IPv4 (6to4)
  GtkToggleButton *checkbutton12; // Tunnel IPv6 over IPv4
  unsigned char sixto4_flag;      // Flag for state of Tunnel IPv6 over IPv4 checkbutton
  GtkWidget *entry125;            // TCP source IPv4 address for 6to4 packets
  GtkWidget *entry160;            // ICMP source IPv4 address for 6to4 packets
  GtkWidget *entry161;            // UDP source IPv4 address for 6to4 packets
  GtkToggleButton *checkbutton13; // TCP source IPv4 "Randomize" for 6to4 packets
  GtkToggleButton *checkbutton26; // ICMP source IPv4 "Randomize" for 6to4 packets
  GtkToggleButton *checkbutton27; // UDP source IPv4 "Randomize" for 6to4 packets
  unsigned char ran_tcp6to4_sourceip;  // Flag for state of TCP source IPv4 "Randomize" for 6to4 packets
  unsigned char ran_icmp6to4_sourceip;  // Flag for state of ICMP source IPv4 "Randomize" for 6to4 packets
  unsigned char ran_udp6to4_sourceip;  // Flag for state of UDP source IPv4 "Randomize" for 6to4 packets
  GtkToggleButton *checkbutton14; // Specify IPv4 ethernet header
  unsigned char sixto4_specify_ether;  // Flag to indicate if 6to4 ethernet header is specified by user
  GtkWidget *entry156;            // Destination link-layer (MAC) address
  GtkWidget *entry157;            // Source link-layer (MAC) address
  GtkWidget *entry158;            // Source interface name
  gchar *sixto4_interfacename;    // 6to4 interface name
  GtkWidget *entry159;            // Ethernet type code
  ethhdr sixto4ethhdr;            // Struct to hold ethernet header values
// Send / Stop
  GtkButton *button1;             // Send
  GtkButton *button5;             // Stop Flood

// TCP page (IPv4)

// Ethernet header
  GtkToggleButton *checkbutton18;  // Specify ethernet header
  unsigned char tcp4_specify_ether;  // Flag to indicate if ethernet header is specified by user
  GtkWidget *entry144;             // Destination link-layer (MAC) address
  GtkWidget *entry145;             // Source link-layer (MAC) address
  GtkWidget *entry150;             // Source interface name
  gchar *tcp4_interfacename;       // Source interface name
  GtkWidget *entry146;             // Ethernet type code;
  ethhdr tcp4ethhdr;               // Struct to hold ethernet header values
// IPv4 header
  GtkWidget *entry29;  // IP header length
  GtkWidget *entry30;  // Protocol version
  GtkWidget *entry31;  // Type of service
  GtkWidget *entry32;  // Total length of datagram
  GtkWidget *entry33;  // ID sequence number
  GtkWidget *entry122; // Zero
  GtkWidget *entry37;  // Do not fragment flag
  GtkWidget *entry38;  // More fragments following flag
  GtkWidget *entry39;  // Fragmentation offset
  GtkWidget *entry40;  // Time-to-live
  GtkWidget *entry74;  // Transport layer protocol
  GtkWidget *entry35;  // Source IP address
  GtkToggleButton *checkbutton5;  // Source IP "Randomize"
  unsigned char ran_tcp4_sourceip;  // Flag for state of source IPv4 "Randomize" checkbutton
  GtkWidget *entry36;  // Destination IP address
  GtkWidget *entry34;  // IP checksum
// TCP header (IPv4)
  GtkWidget *entry41;  // Source port number
  GtkToggleButton *checkbutton7;  // Source port "Randomize"
  unsigned char ran_tcp4_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *entry42;  // Destination port number
  GtkWidget *entry43;  // Sequence number
  GtkWidget *entry44;  // Ack number
  GtkWidget *entry45;  // Reserved
  GtkWidget *entry46;  // Data offset
  GtkWidget *entry47;  // FIN flag
  GtkWidget *entry48;  // SYN flag
  GtkWidget *entry49;  // RST flag
  GtkWidget *entry50;  // PSH flag
  GtkWidget *entry51;  // ACK flag
  GtkWidget *entry52;  // URG flag
  GtkWidget *entry53;  // ECE flag
  GtkWidget *entry54;  // CWR flag
  GtkWidget *entry55;  // Window size
  GtkWidget *entry56;  // TCP checksum
  GtkWidget *entry77;  // Urgent pointer
  GtkToggleButton *radiobutton17;  // ASCII data entry
  GtkToggleButton *radiobutton18;  // Hexadecimal data entry
  unsigned char ascii_hex_tcp4;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry162; // TCP payload data from keyboard
  GtkButton *button20; // TCP payload data from file
  guchar *tcp4_payload;  // TCP payload for IPv4
  gint tcp4_payloadlen;  // TCP payload length for IPv4

  GtkButton *button3;  // View Packet
  GtkButton *button11;  // Restore default settings

// TCP page (IPv6)

// Ethernet header
  GtkWidget *entry131;        // Destination link-layer (MAC) address
  GtkWidget *entry132;        // Source link-layer (MAC) address
  GtkWidget *entry154;        // Source interface name
  gchar *tcp6_interfacename;  // Source interface name
  GtkWidget *entry133;        // Ethernet type code;
  ethhdr tcp6ethhdr;          // Struct to hold ethernet header values
// IPv6 header
  GtkWidget *entry22;  // IP version
  GtkWidget *entry23;  // Traffic class
  GtkWidget *entry24;  // Flow label
  GtkWidget *entry25;  // Payload length
  GtkWidget *entry26;  // Next header
  GtkWidget *entry27;  // Hop limit
  GtkWidget *entry28;  // Source IP address
  GtkToggleButton *checkbutton2;  // Source IPv6 address "Randomize"
  unsigned char ran_tcp6_sourceip;  // Flag for state of source IPv6 "Randomize" checkbutton
  GtkWidget *entry78;  // Destination IP address
// TCP header (IPv6)
  GtkWidget *entry79;  // Source port number
  GtkToggleButton *checkbutton4;  // Source port "Randomize"
  unsigned char ran_tcp6_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *entry80;  // Destination port number
  GtkWidget *entry81;  // Sequence number
  GtkWidget *entry82;  // Ack number
  GtkWidget *entry83;  // Reserved
  GtkWidget *entry84;  // Data offset
  GtkWidget *entry85;  // FIN flag
  GtkWidget *entry86;  // SYN flag
  GtkWidget *entry87;  // RST flag
  GtkWidget *entry88;  // PSH flag
  GtkWidget *entry89;  // ACK flag
  GtkWidget *entry90;  // URG flag
  GtkWidget *entry91;  // ECE flag
  GtkWidget *entry92;  // CWR flag
  GtkWidget *entry93;  // Window size
  GtkWidget *entry94;  // TCP checksum
  GtkWidget *entry95;  // Urgent pointer
  GtkToggleButton *radiobutton19;  // ASCII data entry
  GtkToggleButton *radiobutton20;  // Hexadecimal data entry
  unsigned char ascii_hex_tcp6;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry163; // TCP payload data from keyboard
  GtkButton *button22; // TCP payload data from file
  guchar *tcp6_payload;  // TCP payload for IPv6
  gint tcp6_payloadlen;  // TCP payload length for IPv6

  GtkButton *button6;  // View Packet
  GtkButton *button16;  // Restore default settings

// ICMP page (IPv4)

// Ethernet header
  GtkToggleButton *checkbutton17;  // Specify ethernet header
  unsigned char icmp4_specify_ether;  // Flag to indicate if ethernet header is specified by user
  GtkWidget *entry141;             // Destination link-layer (MAC) address
  GtkWidget *entry142;             // Source link-layer (MAC) address
  GtkWidget *entry151;             // Source interface name
  gchar *icmp4_interfacename;      // Source interface name
  GtkWidget *entry143;             // Ethernet type code;
  ethhdr icmp4ethhdr;              // Struct to hold ethernet header values
// IPv4 header
  GtkWidget *entry57;  // IP header length
  GtkWidget *entry58;  // Protocol version
  GtkWidget *entry59;  // Type of service
  GtkWidget *entry60;  // Total length of datagram
  GtkWidget *entry61;  // ID sequence number
  GtkWidget *entry123; // Zero
  GtkWidget *entry65;  // Do not fragment flag
  GtkWidget *entry66;  // More fragments following flag
  GtkWidget *entry67;  // Fragmentation offset
  GtkWidget *entry68;  // Time-to-live
  GtkWidget *entry75;  // Transport layer protocol
  GtkWidget *entry63;  // Source IP address
  GtkToggleButton *checkbutton9;  // Source IP "Randomize"
  unsigned char ran_icmp4_sourceip;  // Flag for state of source IPv4 "Randomize" checkbutton
  GtkWidget *entry64;  // Destination IP address
  GtkWidget *entry62;  // IP checksum
// ICMP header (IPv4)
  GtkWidget *entry69;  // Message type
  GtkWidget *entry70;  // Message code
  GtkWidget *entry71;  // ICMP checksum
  GtkWidget *entry72;  // Identifier
  GtkWidget *entry73;  // Sequence number
  GtkToggleButton *radiobutton9;   // ASCII data entry
  GtkToggleButton *radiobutton10;  // Hexadecimal data entry
  unsigned char ascii_hex_icmp4;   // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry126; // ICMP payload data from keyboard
  GtkButton *button7;  // ICMP payload data from file
  guchar *icmp4_payload;  // ICMP payload for IPv4
  gint icmp4_payloadlen;  // ICMP payload length for IPv4

  GtkButton *button2;  // View Packet
  GtkButton *button12;  // Restore default settings

// ICMP page (IPv6)

// Ethernet header
  GtkWidget *entry128;         // Destination link-layer (MAC) address
  GtkWidget *entry129;         // Source link-layer (MAC) address
  GtkWidget *entry153;         // Source interface name
  gchar *icmp6_interfacename;  // Source interface name
  GtkWidget *entry130;         // Ethernet type code;
  ethhdr icmp6ethhdr;          // Struct to hold ethernet header values
// IPv6 header
  GtkWidget *entry110;  // IP version
  GtkWidget *entry111;  // Traffic class
  GtkWidget *entry112;  // Flow label
  GtkWidget *entry113;  // Payload length
  GtkWidget *entry114;  // Next header
  GtkWidget *entry115;  // Hop limit
  GtkWidget *entry109;  // Source IP address
  GtkToggleButton *checkbutton10;  // Source IPv6 address "Randomize"
  unsigned char ran_icmp6_sourceip;  // Flag for state of source IPv6 "Randomize" checkbutton
  GtkWidget *entry116;  // Destination IP address
// ICMP header (IPv6)
  GtkWidget *entry117;  // Message type
  GtkWidget *entry118;  // Message code
  GtkWidget *entry119;  // ICMP checksum
  GtkWidget *entry120;  // Identifier
  GtkWidget *entry121;  // Sequence number
  GtkToggleButton *radiobutton11;  // ASCII data entry
  GtkToggleButton *radiobutton12;  // Hexadecimal data entry
  unsigned char ascii_hex_icmp6;   // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry127;  // ICMP payload data from keyboard
  GtkButton *button9;   // ICMP payload data from file
  guchar *icmp6_payload;  // ICMP payload for IPv6
  gint icmp6_payloadlen;  // ICMP payload length for IPv6

  GtkButton *button17;  // View Packet
  GtkButton *button18;  // Restore default settings

// UDP page (IPv4)

// Ethernet header
  GtkToggleButton *checkbutton19;  // Specify ethernet header
  unsigned char udp4_specify_ether;  // Flag to indicate if ethernet header is specified by user
  GtkWidget *entry147;             // Destination link-layer (MAC) address
  GtkWidget *entry148;             // Source link-layer (MAC) address
  GtkWidget *entry152;             // Source interface name
  gchar *udp4_interfacename;       // Source interface name
  GtkWidget *entry149;             // Ethernet type code;
  ethhdr udp4ethhdr;               // Struct to hold ethernet header values
// IPv4 header
  GtkWidget *entry1;   // IP header length
  GtkWidget *entry2;   // Protocol version
  GtkWidget *entry3;   // Type of service
  GtkWidget *entry4;   // Total length of datagram
  GtkWidget *entry5;   // ID sequence number
  GtkWidget *entry124; // Zero
  GtkWidget *entry9;   // Do not fragment flag
  GtkWidget *entry10;  // More fragments following flag
  GtkWidget *entry11;  // Fragmentation offset
  GtkWidget *entry12;  // Time-to-live
  GtkWidget *entry76;  // Transport layer protocol
  GtkWidget *entry7;   // Source IP address
  GtkToggleButton *checkbutton1;  // Source IP "Randomize"
  unsigned char ran_udp4_sourceip;  // Flag for state of source IPv4 "Randomize" checkbutton
  GtkWidget *entry8;   // Destination IP address
  GtkWidget *entry6;   // IP checksum
// UDP header (IPv4)
  GtkWidget *entry13;   // Source port
  GtkToggleButton *checkbutton3;  // Source port "Randomize"
  unsigned char ran_udp4_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *entry14;   // Destination port
  GtkWidget *entry15;   // Length of UDP datagram
  GtkWidget *entry16;   // UDP checksum
  GtkToggleButton *radiobutton13;  // ASCII data entry
  GtkToggleButton *radiobutton14;  // Hexadecimal data entry
  unsigned char ascii_hex_udp4;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry17;   // UDP payload data from keyboard
  GtkButton *button8;   // UDP payload data from file
  guchar *udp4_payload;  // UDP payload for IPv4
  gint udp4_payloadlen;  // UDP payload length for IPv4

  GtkButton *button4;  // View Packet
  GtkButton *button15;  // Restore default settings

// UDP page (IPv6)

// Ethernet header
  GtkWidget *entry134;        // Destination link-layer (MAC) address
  GtkWidget *entry135;        // Source link-layer (MAC) address
  GtkWidget *entry155;        // Source interface name
  gchar *udp6_interfacename;  // Source interface name
  GtkWidget *entry136;        // Ethernet type code;
  ethhdr udp6ethhdr;          // Struct to hold ethernet header values
// IPv6 header
  GtkWidget *entry97;   // IP version
  GtkWidget *entry98;   // Traffic class
  GtkWidget *entry99;   // Flow label
  GtkWidget *entry100;  // Payload length
  GtkWidget *entry101;  // Next header
  GtkWidget *entry102;  // Hop limit
  GtkWidget *entry96;   // Source IP address
  GtkToggleButton *checkbutton6;  // Source IPv6 address "Randomize"
  unsigned char ran_udp6_sourceip;  // Flag for state of source IPv6 "Randomize" checkbutton
  GtkWidget *entry103;  // Destination IP address
// UDP header (IPv6)
  GtkWidget *entry104;   // Source port
  GtkToggleButton *checkbutton8;  // Source port "Randomize"
  unsigned char ran_udp6_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *entry105;   // Destination port
  GtkWidget *entry106;   // Length of UDP datagram
  GtkWidget *entry107;   // UDP checksum
  GtkToggleButton *radiobutton15;  // ASCII data entry
  GtkToggleButton *radiobutton16;  // Hexadecimal data entry
  unsigned char ascii_hex_udp6;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry108;   // UDP payload data from keyboard
  GtkButton *button10;   // UDP payload data from file
  guchar *udp6_payload;  // UDP payload for IPv6
  gint udp6_payloadlen;  // UDP payload length for IPv6

  GtkButton *button13;  // View Packet
  GtkButton *button19;  // Restore default settings

// Host/IP page
  GtkWidget *entry19;    // Host-to-IP: hostname
  GtkWidget *textview3;  // Host-to-IPv4 results
  GtkWidget *textview6;  // Host-to-IPv6 results

  GtkWidget *entry20;    // IP-to-Host IPv4 address
  GtkWidget *entry21;    // IP-to-Host IPv6 address
  GtkWidget *textview2;  // IP-to-Host hostname result

  GtkWidget *textview4;  // This Host hostname
  GtkWidget *textview5;  // This Host IP address

  GtkWidget *entry139;    // This host interface name
  GtkWidget *textview10;  // This host MAC address results

  GtkWidget *entry140;    // IP-to-MAC source interface name
  GtkWidget *spinbutton1;  // Timeout for ARP/Neighbor discovery
  gint timeout;            // Timeout value (s) for ARP/Neighbor discovery
  GtkWidget *entry137;    // IP-to-MAC IPv4 address
  GtkWidget *entry138;    // IP-to-MAC IPv6 address
  GtkWidget *textview11;  // IP-to-MAC results

// Load/Save page

// Load
  GtkButton *button21;             // Load file ...
// Save
  // Send Packet settings
  GtkToggleButton *checkbutton25;  // Packet type, Number of packets to send, Flood mode
  unsigned char settings_save;              // Flag to save Send Packet settings
  // IPv6 over IPv4 (6to4)
  GtkToggleButton *checkbutton15;  // IPv4 Ethernet header and IPv4 header
  unsigned char sixto4_save;                // Flag to save 6to4 information
  // IPv4
  GtkToggleButton *checkbutton16;  // TCP packet (ethernet header, IP header, and TCP header)
  unsigned char ipv4_tcp_save;              // Flag to save IPv4 TCP information
  GtkToggleButton *checkbutton20;  // ICMP packet (ethernet header, IP header, ICMP header, and ICMP data)
  unsigned char ipv4_icmp_save;             // Flag to save IPv4 ICMP information
  GtkToggleButton *checkbutton21;  // UDP packet (ethernet header, IP header, UDP header, and UDP data)
  unsigned char ipv4_udp_save;              // Flag to save IPv4 UDP information
  // IPv6
  GtkToggleButton *checkbutton22;  // TCP packet (ethernet header, IP header, and TCP header)
  unsigned char ipv6_tcp_save;              // Flag to save IPv6 TCP information
  GtkToggleButton *checkbutton23;  // ICMP packet (ethernet header, IP header, ICMP header, and ICMP data)
  unsigned char ipv6_icmp_save;             // Flag to save IPv6 ICMP information
  GtkToggleButton *checkbutton24;  // UDP packet (ethernet header, IP header, UDP header, and UDP data)
  unsigned char ipv6_udp_save;              // Flag to save IPv6 UDP information
  // Save file
  GtkButton *button14;             // Save file ...

// Traceroute page (sps.ui)

  GtkWidget *traceroute_window;  // Traceroute edit packet window
  gint traceroute_flag;          // Flag indicating traceroute edit packet window is activated

// Probe Protocol
  GtkToggleButton *radiobutton21;  // TCP (IPv4) probe packet
  GtkToggleButton *radiobutton22;  // ICMP (IPv4) probe packet
  GtkToggleButton *radiobutton23;  // UDP (IPv4) probe packet
  GtkToggleButton *radiobutton26;  // TCP (IPv6) probe packet
  GtkToggleButton *radiobutton27;  // ICMP (IPv6) probe packet
  GtkToggleButton *radiobutton28;  // UDP (IPv6) probe packet
  GtkToggleButton *checkbutton29;  // Tunnel IPv6 over IPv4
  unsigned char packet_type_tr;    // Packet type
  unsigned char sixto4_tr_flag;    // Flag for state of Tunnel IPv6 over IPv4 checkbutton
// Number of Probes per Hop
  GtkWidget *spinbutton2;  // Number of probes per hop
  gint num_probes;  // Number of probes per hop
// Timeout for Reply
  GtkWidget *spinbutton4;  // Timeout for reply
  gint timeout_tr;         // Timeout for reply
// Maximum number of hops
  GtkWidget *spinbutton3;  // Maximum number of hops
  gint maxhops;  // Maximum number of hops
// Resolve Node Hostnames
  GtkToggleButton *checkbutton34;  // Yes
  int resolve_tr;
// Edit, Start, Stop buttons
  GtkButton *button23;  // Edit packets
  GtkButton *button24;  // Start
  GtkButton *button25;  // Stop
// Activity Log
  GtkWidget *textview7;  // Traceroute activity log
  gchar *tr_message;     // String to contain traceroute results

// Traceroute IPv6 over IPv4 (6to4) (traceroute.ui)

  GtkToggleButton *checkbutton32; // Specify IPv4 ethernet header
  unsigned char sixto4_tr_specify_ether;  // Flag to indicate if 6to4 ethernet header is specified by user
  GtkWidget *entry312;            // Destination link-layer (MAC) address
  GtkWidget *entry313;            // Source link-layer (MAC) address
  GtkWidget *entry314;            // Source interface name
  gchar *sixto4_tr_interfacename; // 6to4 interface name
  GtkWidget *entry315;            // Ethernet type code
  ethhdr sixto4ethhdr_tr;         // Struct to hold ethernet header values
  GtkWidget *entry316;            // TCP source IPv4 address for 6to4 packets
  GtkWidget *entry317;            // ICMP source IPv4 address for 6to4 packets
  GtkWidget *entry318;            // UDP source IPv4 address for 6to4 packets


// Traceroute TCP page (IPv4) (traceroute.ui)

// Ethernet header
  GtkToggleButton *checkbutton28;  // Specify ethernet header
  unsigned char tcp4_tr_specify_ether;  // Flag to indicate if ethernet header is specified by user
  GtkWidget *entry164;             // Destination link-layer (MAC) address
  GtkWidget *entry165;             // Source link-layer (MAC) address
  GtkWidget *entry166;             // Source interface name
  gchar *tcp4_tr_interfacename;    // Source interface name
  GtkWidget *entry167;             // Ethernet type code;
  ethhdr tcp4ethhdr_tr;            // Struct to hold ethernet header values
// IPv4 header
  GtkWidget *entry168;  // IP header length
  GtkWidget *entry169;  // Protocol version
  GtkWidget *entry170;  // Type of service
  GtkWidget *entry171;  // Total length of datagram
  GtkWidget *entry172;  // ID sequence number
  GtkWidget *entry173;  // Zero
  GtkWidget *entry174;  // Do not fragment flag
  GtkWidget *entry175;  // More fragments following flag
  GtkWidget *entry176;  // Fragmentation offset
  GtkWidget *entry177;  // Time-to-live
  GtkWidget *entry178;  // Transport layer protocol
  GtkWidget *entry179;  // Source IP address
  GtkWidget *entry180;  // Destination IP address
  GtkWidget *entry181;  // IP checksum
// TCP header (IPv4)
  GtkWidget *entry182;  // Source port number
  GtkToggleButton *checkbutton30;  // Source port "Randomize"
  unsigned char ran_tcp4_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *entry183;  // Destination port number
  GtkWidget *entry184;  // Sequence number
  GtkWidget *entry185;  // Ack number
  GtkWidget *entry186;  // Reserved
  GtkWidget *entry187;  // Data offset
  GtkWidget *entry188;  // FIN flag
  GtkWidget *entry189;  // SYN flag
  GtkWidget *entry190;  // RST flag
  GtkWidget *entry191;  // PSH flag
  GtkWidget *entry192;  // ACK flag
  GtkWidget *entry193;  // URG flag
  GtkWidget *entry194;  // ECE flag
  GtkWidget *entry195;  // CWR flag
  GtkWidget *entry196;  // Window size
  GtkWidget *entry197;  // TCP checksum
  GtkWidget *entry198;  // Urgent pointer
  GtkToggleButton *radiobutton29;  // ASCII data entry
  GtkToggleButton *radiobutton30;  // Hexadecimal data entry
  unsigned char ascii_hex_tcp4_tr;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry199; // TCP payload data from keyboard
  GtkButton *button29; // TCP payload data from file
  guchar *tcp4_tr_payload;  // TCP payload for IPv4
  gint tcp4_tr_payloadlen;  // TCP payload length for IPv4

  GtkButton *button28;  // View Packet
  GtkButton *button30;  // Restore default settings

// Traceroute TCP page (IPv6) (traceroute.ui)

// Ethernet header
  GtkWidget *entry247;  // Destination link-layer (MAC) address
  GtkWidget *entry248;  // Source link-layer (MAC) address
  GtkWidget *entry249;  // Source interface name
  gchar *tcp6_tr_interfacename;  // Source interface name
  GtkWidget *entry250;  // Ethernet type code;
  ethhdr tcp6ethhdr_tr;  // Struct to hold ethernet header values
// IPv6 header
  GtkWidget *entry251;  // IP version
  GtkWidget *entry252;  // Traffic class
  GtkWidget *entry253;  // Flow label
  GtkWidget *entry254;  // Payload length
  GtkWidget *entry255;  // Next header
  GtkWidget *entry256;  // Hop limit
  GtkWidget *entry257;  // Source IP address
  GtkWidget *entry258;  // Destination IP address
// TCP header (IPv6)
  GtkWidget *entry259;  // Source port number
  GtkToggleButton *checkbutton37;  // Source port "Randomize"
  unsigned char ran_tcp6_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *entry260;  // Destination port number
  GtkWidget *entry261;  // Sequence number
  GtkWidget *entry262;  // Ack number
  GtkWidget *entry263;  // Reserved
  GtkWidget *entry264;  // Data offset
  GtkWidget *entry265;  // FIN flag
  GtkWidget *entry266;  // SYN flag
  GtkWidget *entry267;  // RST flag
  GtkWidget *entry268;  // PSH flag
  GtkWidget *entry269;  // ACK flag
  GtkWidget *entry270;  // URG flag
  GtkWidget *entry271;  // ECE flag
  GtkWidget *entry272;  // CWR flag
  GtkWidget *entry273;  // Window size
  GtkWidget *entry274;  // TCP checksum
  GtkWidget *entry275;  // Urgent pointer
  GtkToggleButton *radiobutton35;  // ASCII data entry
  GtkToggleButton *radiobutton36;  // Hexadecimal data entry
  unsigned char ascii_hex_tcp6_tr;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry276; // TCP payload data from keyboard
  GtkButton *button38; // TCP payload data from file
  guchar *tcp6_tr_payload;  // TCP payload for IPv6
  gint tcp6_tr_payloadlen;  // TCP payload length for IPv6

  GtkButton *button37;  // View Packet
  GtkButton *button39;  // Restore default settings

// Traceroute ICMP page (IPv4) (traceroute.ui)

// Ethernet header
  GtkToggleButton *checkbutton31;  // Specify ethernet header
  unsigned char icmp4_tr_specify_ether;  // Flag to indicate if ethernet header is specified by user
  GtkWidget *entry200;             // Destination link-layer (MAC) address
  GtkWidget *entry201;             // Source link-layer (MAC) address
  GtkWidget *entry202;             // Source interface name
  gchar *icmp4_tr_interfacename;   // Source interface name
  GtkWidget *entry203;             // Ethernet type code;
  ethhdr icmp4ethhdr_tr;           // Struct to hold ethernet header values
// IPv4 header
  GtkWidget *entry204;  // IP header length
  GtkWidget *entry205;  // Protocol version
  GtkWidget *entry206;  // Type of service
  GtkWidget *entry207;  // Total length of datagram
  GtkWidget *entry208;  // ID sequence number
  GtkWidget *entry209;  // Zero
  GtkWidget *entry210;  // Do not fragment flag
  GtkWidget *entry211;  // More fragments following flag
  GtkWidget *entry212;  // Fragmentation offset
  GtkWidget *entry213;  // Time-to-live
  GtkWidget *entry214;  // Transport layer protocol
  GtkWidget *entry215;  // Source IP address
  GtkWidget *entry216;  // Destination IP address
  GtkWidget *entry217;  // IP checksum
// ICMP header (IPv4)
  GtkWidget *entry218;  // Message type
  GtkWidget *entry219;  // Message code
  GtkWidget *entry220;  // ICMP checksum
  GtkWidget *entry221;  // Identifier
  GtkWidget *entry222;  // Sequence number
  GtkToggleButton *radiobutton31;  // ASCII data entry
  GtkToggleButton *radiobutton32;  // Hexadecimal data entry
  unsigned char ascii_hex_icmp4_tr;   // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry223;  // ICMP payload data from keyboard
  GtkButton *button32;  // ICMP payload data from file
  guchar *icmp4_tr_payload;  // ICMP payload for IPv4
  gint icmp4_tr_payloadlen;  // ICMP payload length for IPv4

  GtkButton *button31;  // View Packet
  GtkButton *button33;  // Restore default settings

// Traceroute ICMP page (IPv6) (traceroute.ui)

// Ethernet header
  GtkWidget *entry277;  // Destination link-layer (MAC) address
  GtkWidget *entry278;  // Source link-layer (MAC) address
  GtkWidget *entry279;  // Source interface name
  gchar *icmp6_tr_interfacename;  // Source interface name
  GtkWidget *entry280;  // Ethernet type code;
  ethhdr icmp6ethhdr_tr;  // Struct to hold ethernet header values
// IPv6 header
  GtkWidget *entry281;  // IP version
  GtkWidget *entry282;  // Traffic class
  GtkWidget *entry283;  // Flow label
  GtkWidget *entry284;  // Payload length
  GtkWidget *entry285;  // Next header
  GtkWidget *entry286;  // Hop limit
  GtkWidget *entry287;  // Source IP address
  GtkWidget *entry288;  // Destination IP address
// ICMP header (IPv6)
  GtkWidget *entry289;  // Message type
  GtkWidget *entry290;  // Message code
  GtkWidget *entry291;  // ICMP checksum
  GtkWidget *entry292;  // Identifier
  GtkWidget *entry293;  // Sequence number
  GtkToggleButton *radiobutton37;  // ASCII data entry
  GtkToggleButton *radiobutton38;  // Hexadecimal data entry
  unsigned char ascii_hex_icmp6_tr;  // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry294;  // ICMP payload data from keyboard
  GtkButton *button41;  // ICMP payload data from file
  guchar *icmp6_tr_payload;  // ICMP payload for IPv6
  gint icmp6_tr_payloadlen;  // ICMP payload length for IPv6

  GtkButton *button40;  // View Packet
  GtkButton *button42;  // Restore default settings

// Traceroute UDP page (IPv4) (traceroute.ui)

// Ethernet header
  GtkToggleButton *checkbutton33;  // Specify ethernet header
  unsigned char udp4_tr_specify_ether;  // Flag to indicate if ethernet header is specified by user
  GtkWidget *entry224;             // Destination link-layer (MAC) address
  GtkWidget *entry225;             // Source link-layer (MAC) address
  GtkWidget *entry226;             // Source interface name
  gchar *udp4_tr_interfacename;    // Source interface name
  GtkWidget *entry227;             // Ethernet type code;
  ethhdr udp4ethhdr_tr;            // Struct to hold ethernet header values
// IPv4 header
  GtkWidget *entry228;  // IP header length
  GtkWidget *entry229;  // Protocol version
  GtkWidget *entry230;  // Type of service
  GtkWidget *entry231;  // Total length of datagram
  GtkWidget *entry232;  // ID sequence number
  GtkWidget *entry233;  // Zero
  GtkWidget *entry234;  // Do not fragment flag
  GtkWidget *entry235;  // More fragments following flag
  GtkWidget *entry236;  // Fragmentation offset
  GtkWidget *entry237;  // Time-to-live
  GtkWidget *entry238;  // Transport layer protocol
  GtkWidget *entry239;  // Source IP address
  GtkWidget *entry240;  // Destination IP address
  GtkWidget *entry241;  // IP checksum
// UDP header (IPv4)
  GtkWidget *entry242;  // Source port
  GtkToggleButton *checkbutton35;  // Source port "Randomize"
  unsigned char ran_udp4_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv4)
  GtkWidget *entry243;  // Destination port
  GtkWidget *entry244;  // Length of UDP datagram
  GtkWidget *entry245;  // UDP checksum
  GtkToggleButton *radiobutton33;  // ASCII data entry
  GtkToggleButton *radiobutton34;  // Hexadecimal data entry
  unsigned char ascii_hex_udp4_tr;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry246;  // UDP payload data from keyboard
  GtkButton *button35;  // UDP payload data from file
  guchar *udp4_tr_payload;  // UDP payload for IPv4
  gint udp4_tr_payloadlen;  // UDP payload length for IPv4

  GtkButton *button34;  // View Packet
  GtkButton *button36;  // Restore default settings

// Traceroute UDP page (IPv6) (traceroute.ui)

// Ethernet header
  GtkWidget *entry295;  // Destination link-layer (MAC) address
  GtkWidget *entry296;  // Source link-layer (MAC) address
  GtkWidget *entry297;  // Source interface name
  gchar *udp6_tr_interfacename;  // Source interface name
  GtkWidget *entry298;  // Ethernet type code;
  ethhdr udp6ethhdr_tr;  // Struct to hold ethernet header values
// IPv6 header
  GtkWidget *entry299;  // IP version
  GtkWidget *entry300;  // Traffic class
  GtkWidget *entry301;  // Flow label
  GtkWidget *entry302;  // Payload length
  GtkWidget *entry303;  // Next header
  GtkWidget *entry304;  // Hop limit
  GtkWidget *entry305;  // Source IP address
  GtkWidget *entry306;  // Destination IP address
// UDP header (IPv6)
  GtkWidget *entry307;  // Source port
  GtkToggleButton *checkbutton40;  // Source port "Randomize"
  unsigned char ran_udp6_tr_sourceport;  // Flag for state of source port "Randomize" checkbutton (IPv6)
  GtkWidget *entry308;  // Destination port
  GtkWidget *entry309;  // Length of UDP datagram
  GtkWidget *entry310;  // UDP checksum
  GtkToggleButton *radiobutton39;  // ASCII data entry
  GtkToggleButton *radiobutton40;  // Hexadecimal data entry
  unsigned char ascii_hex_udp6_tr;    // Flag to indicate ASCII (0) or hexadecimal (1) data entry
  GtkWidget *entry311;  // UDP payload data from keyboard
  GtkButton *button44;  // UDP payload data from file
  guchar *udp6_tr_payload;  // UDP payload for IPv6
  gint udp6_tr_payloadlen;  // UDP payload length for IPv6

  GtkButton *button43;  // View Packet
  GtkButton *button45;  // Restore default settings

// IP / ASN delegation page
  gint *rir_files_have;  // Flags indicating which RIR files are loaded.
  GtkWidget *textview8;  // Status of ARIN delegation file.
  GtkButton *button50;   // Download ARIN delegation file.
  GtkWidget *textview9;  // Status of RIPE delegation file.
  GtkButton *button51;   // Download RIPE delegation file.
  GtkWidget *textview12; // Status of AFRINIC delegation file.
  GtkButton *button52;   // Download AFRINIC delegation file.
  GtkWidget *textview13; // Status of APNIC delegation file.
  GtkButton *button53;   // Download APNIC delegation file.
  GtkWidget *textview14; // Status of LACNIC delegation file.
  GtkButton *button54;   // Download LACNIC delegation file.
  GtkWidget *textview15; // Status of ISO 3166-1 country code file.
  GtkButton *button55;   // Download ISO 3166-1 country code file.
  GtkButton *button56;   // View ISO 3166-1 country code file.
  GtkButton *button57;   // Download all RIR files and country code file.
  GtkToggleButton *checkbutton38;  // Flag to show sent and received messages during file downloads
  gint countries_have;   // Flag indicating if country code file is loaded.
  int ftp_family;        // Address family for FTP connection to RIRs.
  int sd_data;           // Socket descriptor for FTP data connection.
  gchar *ftp_data;       // Array to hold FTP data received from remote host.
  gint rir_index;        // Index to strings containing RIR names, URLs, directories, and file names.

  gint nrecords[5];      // Number of records in each RIR's delegation file.
  gint ncountry;         // Number of countries in www.iso.org's country code list.
  unsigned char ****array;  // Array containing all RIR delegation files.
  unsigned char ***cc;   // Array containing www.iso.org's country code list.
  GtkWidget *entry500;   // Country name for country name/code search.
  GtkWidget *entry501;   // Country code for country name/code search.
  GtkWidget *entry502;   // Country name for ASN search.
  GtkWidget *entry503;   // Country code for ASN search.
  GtkWidget *textview16; // ASN delegation results for ASN delegation search.
  gchar *asn_results;    // ASN delegation results for ASN delegation search.
  GtkWidget *entry504;   // ASN for reverse-search for delegation of ASN.
  gint asn;              // ASN for reverse-search for delegation of ASN.
  GtkWidget *textview17; // Results of reverse-search for delegation of ASN.
  GtkWidget *entry505;   // Country name for IP delegation search.
  gchar *cname;          // Country name for IP delegation search.
  GtkWidget *entry506;   // Country code for IP delegation search.
  gchar *ccode;          // Country code for IP delegation search.
  GtkWidget *textview18; // IPv4 address delegation results for IP delegation search.
  gchar *ip4_results;    // IPv4 address delegation results for IP delegation search.
  GtkWidget *textview19; // IPv6 address delegation results for IP delegation search.
  gchar *ip6_results;    // IPv6 address delegation results for IP delegation search.
  GtkWidget *entry507;   // IPv4 address for IP delegation reverse-search.
  gchar *ip4string;      // IPv4 address for IP delegation reverse_search.
  GtkWidget *entry508;   // IPv6 address for IP delegation reverse-search.
  gchar *ip6string;      // IPv6 address for IP delegation reverse_search.
  GtkWidget *textview20; // IP address delegation results for IP delegation reverse-search.

};

// Function prototypes

// Send page 

// Select Packet Type
gboolean on_radiobutton1_toggled (GtkButton *, SPSData *);  // TCP (IPv4) packet type
gboolean on_radiobutton2_toggled (GtkButton *, SPSData *);  // ICMP (IPv4) packet type
gboolean on_radiobutton3_toggled (GtkButton *, SPSData *);  // UDP (IPv4) packet type
gboolean on_radiobutton4_toggled (GtkButton *, SPSData *);  // All (IPv4) packet types (cycle)
gboolean on_radiobutton5_toggled (GtkButton *, SPSData *);  // TCP (IPv6) packet type
gboolean on_radiobutton6_toggled (GtkButton *, SPSData *);  // ICMP (IPv6) packet type
gboolean on_radiobutton7_toggled (GtkButton *, SPSData *);  // UDP (IPv6) packet type
gboolean on_radiobutton8_toggled (GtkButton *, SPSData *);  // All (IPv6) packet types (cycle)
// Number of Packets to Send
gboolean on_entry18_activate (GtkWidget *, SPSData *);      // Number of packets to send
gboolean on_checkbutton11_toggled (GtkButton *, SPSData *); // Flood mode
// Activity Log
void log_activity (SPSData *);  // Activity logger
// IPv6 over IPv4 (6to4)
gboolean on_checkbutton12_toggled (GtkButton *, SPSData *); // Tunnel IPv6 over IPv4
gboolean on_entry125_activate (GtkWidget *, SPSData *);     // TCP IPv4 Source Address for 6to4 tunnel
gboolean on_entry160_activate (GtkWidget *, SPSData *);     // ICMP IPv4 Source Address for 6to4 tunnel
gboolean on_entry161_activate (GtkWidget *, SPSData *);     // UDP IPv4 Source Address for 6to4 tunnel
gboolean on_checkbutton13_toggled (GtkButton *, SPSData *); // Randomize TCP source IPv4 address for 6to4
gboolean on_checkbutton26_toggled (GtkButton *, SPSData *); // Randomize ICMP source IPv4 address for 6to4
gboolean on_checkbutton27_toggled (GtkButton *, SPSData *); // Randomize UDP source IPv4 address for 6to4
gboolean on_checkbutton14_toggled (GtkButton *, SPSData *); // Specify IPv4 ethernet header
gboolean on_entry156_activate (GtkWidget *, SPSData *);     // Destination link-layer (MAC) address
gboolean on_entry157_activate (GtkWidget *, SPSData *);     // Source link-layer (MAC) address
gboolean on_entry158_activate (GtkWidget *, SPSData *);     // Source interface name
gboolean on_entry159_activate (GtkWidget *, SPSData *);     // Ethernet type code
// Send / Stop
gboolean on_button1_clicked (GtkButton *, SPSData *);       // Send packet(s)
gboolean on_button5_clicked (GtkButton *, SPSData *);       // Stop flood
// Send packet functions
gboolean ipv4_send (SPSData *);   // Send IPv4 packet(s)
gboolean ipv6_send (SPSData *);   // Send IPv6 packet(s)
gboolean sixto4_send (SPSData *); // Send IPv6 packet(s) over IPv4 (6to4)

// TCP page (IPv4)

// Ethernet header
gboolean on_checkbutton18_toggled (GtkButton *, SPSData *);  // Specify ethernet header
gboolean on_entry144_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry145_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry150_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry146_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
gboolean on_entry29_activate (GtkWidget *, SPSData *);  // IPv4 header length
gboolean on_entry30_activate (GtkWidget *, SPSData *);  // Protocol version
gboolean on_entry31_activate (GtkWidget *, SPSData *);  // Type-of-service
gboolean on_entry32_activate (GtkWidget *, SPSData *);  // Total length of datagram
gboolean on_entry33_activate (GtkWidget *, SPSData *);  // ID sequence number
gboolean on_entry122_activate (GtkWidget *, SPSData *); // Zero
gboolean on_entry37_activate (GtkWidget *, SPSData *);  // Do not fragment flag
gboolean on_entry38_activate (GtkWidget *, SPSData *);  // More fragments following flag
gboolean on_entry39_activate (GtkWidget *, SPSData *);  // Fragmentation offset
gboolean on_entry40_activate (GtkWidget *, SPSData *);  // Time-to-Live
gboolean on_entry74_activate (GtkWidget *, SPSData *);  // Transport layer protocol
gboolean on_entry35_activate (GtkWidget *, SPSData *);  // Source IPv4 address
gboolean on_checkbutton5_toggled (GtkButton *, SPSData *);  // Randomize source IPv4 address
gboolean on_entry36_activate (GtkWidget *, SPSData *);  // Destination IPv4 address
gboolean on_entry34_activate (GtkWidget *, SPSData *);  // IPv4 Checksum
// TCP header (IPv4)
gboolean on_entry41_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton7_toggled (GtkButton *, SPSData *);  // Randomize source port
gboolean on_entry42_activate (GtkWidget *, SPSData *);  // Destination port
gboolean on_entry43_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_entry44_activate (GtkWidget *, SPSData *);  // Acknowledgement number
gboolean on_entry45_activate (GtkWidget *, SPSData *);  // Reserved 4 bits
gboolean on_entry46_activate (GtkWidget *, SPSData *);  // Data offset
gboolean on_entry47_activate (GtkWidget *, SPSData *);  // FIN flag
gboolean on_entry48_activate (GtkWidget *, SPSData *);  // SYN flag
gboolean on_entry49_activate (GtkWidget *, SPSData *);  // RST flag
gboolean on_entry50_activate (GtkWidget *, SPSData *);  // PSH flag
gboolean on_entry51_activate (GtkWidget *, SPSData *);  // ACK flag
gboolean on_entry52_activate (GtkWidget *, SPSData *);  // URG flag
gboolean on_entry53_activate (GtkWidget *, SPSData *);  // ECE flag
gboolean on_entry54_activate (GtkWidget *, SPSData *);  // CWR flag
gboolean on_entry55_activate (GtkWidget *, SPSData *);  // Window size
gboolean on_entry56_activate (GtkWidget *, SPSData *);  // TCP checksum
gboolean on_entry77_activate (GtkWidget *, SPSData *);  // Urgent pointer
gboolean on_radiobutton17_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton18_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry162_activate (GtkWidget *, SPSData *); // TCP data from keyboard
gboolean on_button20_clicked (GtkButton *, SPSData *);  // TCP data from file

gboolean on_button3_clicked (GtkButton *, SPSData *);   // View IPv4 TCP packet
gboolean on_button11_clicked (GtkButton *, SPSData *);  // Restore IPv4 TCP default values
gboolean tcp4_default (SPSData *);

// TCP page (IPv6)

// Ethernet header
gboolean on_entry131_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry132_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry154_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry133_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
gboolean on_entry22_activate (GtkWidget *, SPSData *);  // IP version
gboolean on_entry23_activate (GtkWidget *, SPSData *);  // Traffic class
gboolean on_entry24_activate (GtkWidget *, SPSData *);  // Flow label
gboolean on_entry25_activate (GtkWidget *, SPSData *);  // Payload length
gboolean on_entry26_activate (GtkWidget *, SPSData *);  // Next header
gboolean on_entry27_activate (GtkWidget *, SPSData *);  // Hop limit
gboolean on_entry28_activate (GtkWidget *, SPSData *);  // Source IPv6 address
gboolean on_checkbutton2_toggled (GtkButton *, SPSData *);  // Randomize Source IPv6 Address
gboolean on_entry78_activate (GtkWidget *, SPSData *);  // Destination IPv6 address
// TCP header (IPv6)
gboolean on_entry79_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton4_toggled (GtkButton *, SPSData *);  // Randomize source port
gboolean on_entry80_activate (GtkWidget *, SPSData *);  // Destination port
gboolean on_entry81_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_entry82_activate (GtkWidget *, SPSData *);  // Acknowledgement number
gboolean on_entry83_activate (GtkWidget *, SPSData *);  // Reserved 4 bits
gboolean on_entry84_activate (GtkWidget *, SPSData *);  // Data offset
gboolean on_entry85_activate (GtkWidget *, SPSData *);  // FIN flag
gboolean on_entry86_activate (GtkWidget *, SPSData *);  // SYN flag
gboolean on_entry87_activate (GtkWidget *, SPSData *);  // RST flag
gboolean on_entry88_activate (GtkWidget *, SPSData *);  // PSH flag
gboolean on_entry89_activate (GtkWidget *, SPSData *);  // ACK flag
gboolean on_entry90_activate (GtkWidget *, SPSData *);  // URG flag
gboolean on_entry91_activate (GtkWidget *, SPSData *);  // ECE flag
gboolean on_entry92_activate (GtkWidget *, SPSData *);  // CWR flag
gboolean on_entry93_activate (GtkWidget *, SPSData *);  // Window size
gboolean on_entry94_activate (GtkWidget *, SPSData *);  // TCP checksum
gboolean on_entry95_activate (GtkWidget *, SPSData *);  // Urgent pointer
gboolean on_radiobutton19_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton20_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry163_activate (GtkWidget *, SPSData *); // TCP data from keyboard
gboolean on_button22_clicked (GtkButton *, SPSData *);  // TCP data from file

gboolean on_button6_clicked (GtkButton *, SPSData *);  // View IPv6 TCP packet
gboolean on_button16_clicked (GtkButton *, SPSData *);  // Restore IPv6 TCP default values
gboolean tcp6_default (SPSData *);

// ICMP page (IPv4)

// Ethernet header
gboolean on_checkbutton17_toggled (GtkButton *, SPSData *);  // Specify ethernet header
gboolean on_entry141_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry142_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry151_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry143_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
gboolean on_entry57_activate (GtkWidget *, SPSData *);  // IPv4 Header Length
gboolean on_entry58_activate (GtkWidget *, SPSData *);  // Protocol Version
gboolean on_entry59_activate (GtkWidget *, SPSData *);  // Type of Service
gboolean on_entry60_activate (GtkWidget *, SPSData *);  // Total Length of Datagram
gboolean on_entry61_activate (GtkWidget *, SPSData *);  // ID Sequence Number
gboolean on_entry123_activate (GtkWidget *, SPSData *); // Zero
gboolean on_entry65_activate (GtkWidget *, SPSData *);  // Do Not Fragment Flag
gboolean on_entry66_activate (GtkWidget *, SPSData *);  // More Fragments Following
gboolean on_entry67_activate (GtkWidget *, SPSData *);  // Fragmentation Offset
gboolean on_entry68_activate (GtkWidget *, SPSData *);  // Time-to-Live
gboolean on_entry75_activate (GtkWidget *, SPSData *);  // Transport Layer Protocol
gboolean on_entry63_activate (GtkWidget *, SPSData *);  // Source IPv4 Address
gboolean on_checkbutton9_toggled (GtkButton *, SPSData *);  // Randomize Source IPv4 Address
gboolean on_entry64_activate (GtkWidget *, SPSData *);  // Destination IPv4 Address
gboolean on_entry62_activate (GtkWidget *, SPSData *);  // IPv4 Checksum
// ICMP message (IPv4)
gboolean on_entry69_activate (GtkWidget *, SPSData *);  // Message type
gboolean on_entry70_activate (GtkWidget *, SPSData *);  // Message code
gboolean on_entry71_activate (GtkWidget *, SPSData *);  // ICMP checksum
gboolean on_entry72_activate (GtkWidget *, SPSData *);  // ID
gboolean on_entry73_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_radiobutton9_toggled (GtkButton *, SPSData *);   // ASCII data entry
gboolean on_radiobutton10_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry126_activate (GtkWidget *, SPSData *); // ICMP data from keyboard
gboolean on_button7_clicked (GtkButton *, SPSData *);   // ICMP data from file

gboolean on_button2_clicked (GtkButton *, SPSData *);   // View IPv4 ICMP packet
gboolean on_button12_clicked (GtkButton *, SPSData *);  // Restore IPv4 ICMP default values
gboolean icmp4_default (SPSData *);

// ICMP page (IPv6)

// Ethernet header
gboolean on_entry128_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry129_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry153_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry130_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
gboolean on_entry110_activate (GtkWidget *, SPSData *);  // IP version
gboolean on_entry111_activate (GtkWidget *, SPSData *);  // Traffic class
gboolean on_entry112_activate (GtkWidget *, SPSData *);  // Flow label
gboolean on_entry113_activate (GtkWidget *, SPSData *);  // Payload length
gboolean on_entry114_activate (GtkWidget *, SPSData *);  // Next header
gboolean on_entry115_activate (GtkWidget *, SPSData *);  // Hop limit
gboolean on_entry109_activate (GtkWidget *, SPSData *);  // Source IPv6 address
gboolean on_checkbutton10_toggled (GtkButton *, SPSData *);  // Randomize Source IPv6 Address
gboolean on_entry116_activate (GtkWidget *, SPSData *);  // Destination IPv6 address
// ICMP message (IPv6)
gboolean on_entry117_activate (GtkWidget *, SPSData *);  // Message type
gboolean on_entry118_activate (GtkWidget *, SPSData *);  // Message code
gboolean on_entry119_activate (GtkWidget *, SPSData *);  // ICMP checksum
gboolean on_entry120_activate (GtkWidget *, SPSData *);  // ID
gboolean on_entry121_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_radiobutton11_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton12_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry127_activate (GtkWidget *, SPSData *);  // ICMP data from keyboard
gboolean on_button9_clicked (GtkButton *, SPSData *);   // ICMP data from file

gboolean on_button17_clicked (GtkButton *, SPSData *);  // View IPv6 ICMP packet
gboolean on_button18_clicked (GtkButton *, SPSData *);  // Restore IPv6 ICMP default values
gboolean icmp6_default (SPSData *);

// UDP page (IPv4)

// Ethernet header
gboolean on_checkbutton19_toggled (GtkButton *, SPSData *);  // Specify ethernet header
gboolean on_entry147_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry148_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry152_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry149_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
gboolean on_entry1_activate (GtkWidget *, SPSData *);   // IPv4 Header Length
gboolean on_entry2_activate (GtkWidget *, SPSData *);   // Protocol Version
gboolean on_entry3_activate (GtkWidget *, SPSData *);   // Type of Service
gboolean on_entry4_activate (GtkWidget *, SPSData *);   // Total Length of Datagram
gboolean on_entry5_activate (GtkWidget *, SPSData *);   // ID Sequence Number
gboolean on_entry124_activate (GtkWidget *, SPSData *); // Zero
gboolean on_entry9_activate (GtkWidget *, SPSData *);   // Do Not Fragment Flag
gboolean on_entry10_activate (GtkWidget *, SPSData *);  // More Fragments Following
gboolean on_entry11_activate (GtkWidget *, SPSData *);  // Fragmentation Offset
gboolean on_entry12_activate (GtkWidget *, SPSData *);  // Time-to-Live
gboolean on_entry76_activate (GtkWidget *, SPSData *);  // Transport Layer Protocol
gboolean on_entry7_activate (GtkWidget *, SPSData *);   // Source IPv4 Address
gboolean on_checkbutton1_toggled (GtkButton *, SPSData *);  // Randomize Source IPv4 Address
gboolean on_entry8_activate (GtkWidget *, SPSData *);   // Destination IPv4 Address
gboolean on_entry6_activate (GtkWidget *, SPSData *);   // IPv4 Checksum
// UDP header (IPv4)
gboolean on_entry13_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton3_toggled (GtkButton *, SPSData *);  // Randomize Source Port
gboolean on_entry14_activate (GtkWidget *, SPSData *);  // Destination Port
gboolean on_entry15_activate (GtkWidget *, SPSData *);  // Length of UDP Datagram
gboolean on_entry16_activate (GtkWidget *, SPSData *);  // UDP Checksum
gboolean on_radiobutton13_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton14_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry17_activate (GtkWidget *, SPSData *);  // UDP Data from keyboard
gboolean on_button8_clicked (GtkButton *, SPSData *);   // UDP data from file

gboolean on_button4_clicked (GtkButton *, SPSData *);   // View IPv4 UDP packet
gboolean on_button15_clicked (GtkButton *, SPSData *);  // Restore IPv4 UDP default values
gboolean udp4_default (SPSData *);

// UDP page (IPv6)

// Ethernet header
gboolean on_entry134_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry135_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry155_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry136_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
gboolean on_entry97_activate (GtkWidget *, SPSData *);   // IP version
gboolean on_entry98_activate (GtkWidget *, SPSData *);   // Traffic class
gboolean on_entry99_activate (GtkWidget *, SPSData *);   // Flow label
gboolean on_entry100_activate (GtkWidget *, SPSData *);  // Payload length
gboolean on_entry101_activate (GtkWidget *, SPSData *);  // Next header
gboolean on_entry102_activate (GtkWidget *, SPSData *);  // Hop limit
gboolean on_entry96_activate (GtkWidget *, SPSData *);   // Source IPv6 address
gboolean on_checkbutton6_toggled (GtkButton *, SPSData *);  // Randomize Source IPv6 Address
gboolean on_entry103_activate (GtkWidget *, SPSData *);  // Destination IPv6 address
// UDP header (IPv6)
gboolean on_entry104_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton8_toggled (GtkButton *, SPSData *);  // Randomize Source Port
gboolean on_entry105_activate (GtkWidget *, SPSData *);  // Destination Port
gboolean on_entry106_activate (GtkWidget *, SPSData *);  // Length of UDP Datagram
gboolean on_entry107_activate (GtkWidget *, SPSData *);  // UDP Checksum
gboolean on_radiobutton15_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton16_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry108_activate (GtkWidget *, SPSData *);  // UDP Data from keyboard
gboolean on_button10_clicked (GtkButton *, SPSData *);   // UDP data from file

gboolean on_button13_clicked (GtkButton *, SPSData *);   // View IPv6 UDP packet
gboolean on_button19_clicked (GtkButton *, SPSData *);  // Restore IPv6 UDP default values
gboolean udp6_default (SPSData *);

// Host/IP page
gboolean on_entry19_activate (GtkWidget *, SPSData *);   // Hostname to IP Address
gboolean on_entry20_activate (GtkWidget *, SPSData *);   // IPv4 Address to Hostname
gboolean on_entry21_activate (GtkWidget *, SPSData *);   // IPv6 Address to Hostname
gboolean on_entry139_activate (GtkWidget *, SPSData *);  // This host's interface name to be converted to link-layer (MAC) address.
gboolean on_entry140_activate (GtkWidget *, SPSData *);  // Interface name for IP to link-layer conversion
gboolean on_spinbutton1_value_changed (GtkWidget *, SPSData *);  // Timeout for ARP/Neighbor discovery
gboolean on_entry137_activate (GtkWidget *, SPSData *);  // IPv4 Address to link-layer (MAC) address
gboolean on_entry138_activate (GtkWidget *, SPSData *);  // IPv6 Address to link-layer (MAC) address

// Load/Save page

// Load
gboolean on_button21_clicked (GtkButton *, SPSData *);   // Load file ...

// Save
// Send Packet settings
gboolean on_checkbutton25_toggled (GtkButton *, SPSData *);  // Packet type, Number of packets to send, Flood mode
// IPv6 over IPv4 (6to4)
gboolean on_checkbutton15_toggled (GtkButton *, SPSData *);  // IPv4 Ethernet header and IPv4 header
// IPv4
gboolean on_checkbutton16_toggled (GtkButton *, SPSData *);  // TCP packet
gboolean on_checkbutton20_toggled (GtkButton *, SPSData *);  // ICMP packet
gboolean on_checkbutton21_toggled (GtkButton *, SPSData *);  // UDP packet
// IPv6
gboolean on_checkbutton22_toggled (GtkButton *, SPSData *);  // TCP packet
gboolean on_checkbutton23_toggled (GtkButton *, SPSData *);  // ICMP packet
gboolean on_checkbutton24_toggled (GtkButton *, SPSData *);  // UDP packet
// Save file
gboolean on_button14_clicked (GtkButton *, SPSData *);       // Save file ...

// Traceroute page (sps.ui)

// Probe Protocol
gboolean on_radiobutton21_toggled (GtkButton *, SPSData *);  // TCP (IPv4) probe packet
gboolean on_radiobutton22_toggled (GtkButton *, SPSData *);  // ICMP (IPv4) probe packet
gboolean on_radiobutton23_toggled (GtkButton *, SPSData *);  // UDP (IPv4) probe packet
gboolean on_radiobutton26_toggled (GtkButton *, SPSData *);  // TCP (IPv6) probe packet
gboolean on_radiobutton27_toggled (GtkButton *, SPSData *);  // ICMP (IPv6) probe packet
gboolean on_radiobutton28_toggled (GtkButton *, SPSData *);  // UDP (IPv6) probe packet
gboolean on_checkbutton29_toggled (GtkButton *, SPSData *);  // Tunnel IPv6 over IPv4

gboolean on_spinbutton2_value_changed (GtkWidget *, SPSData *);  // Number of probes per hop
gboolean on_spinbutton4_value_changed (GtkWidget *, SPSData *);  // Timeout for reply
gboolean on_spinbutton3_value_changed (GtkWidget *, SPSData *);  // Maximum number of hops
// Resolve Node Hostnames
gboolean on_checkbutton34_toggled (GtkButton *, SPSData *);  // Yes
// Edit, Start, Stop buttons
gboolean on_button23_clicked (GtkButton *, SPSData *);  // Edit packets
gboolean on_button24_clicked (GtkButton *, SPSData *);  // Start
gboolean on_button25_clicked (GtkButton *, SPSData *);  // Stop

gboolean post_tr_message (gchar *, SPSData *);  // Start idle function tr_message()
gboolean tr_message (SPSData *);                // Idle function to show traceroute results in textview7


// Traceroute IPv6 over IPv4 (6to4) page (traceroute.ui)

// IPv6 over IPv4 (6to4)
gboolean on_checkbutton32_toggled (GtkButton *, SPSData *); // Specify IPv4 ethernet header
gboolean on_entry312_activate (GtkWidget *, SPSData *);     // Destination link-layer (MAC) address
gboolean on_entry313_activate (GtkWidget *, SPSData *);     // Source link-layer (MAC) address
gboolean on_entry314_activate (GtkWidget *, SPSData *);     // Source interface name
gboolean on_entry315_activate (GtkWidget *, SPSData *);     // Ethernet type code
gboolean on_entry316_activate (GtkWidget *, SPSData *);     // TCP IPv4 Source Address for 6to4 tunnel
gboolean on_entry317_activate (GtkWidget *, SPSData *);     // ICMP IPv4 Source Address for 6to4 tunnel
gboolean on_entry318_activate (GtkWidget *, SPSData *);     // UDP IPv4 Source Address for 6to4 tunnel
gboolean sixto4_tr_show (SPSData *);


// TCP page (IPv4) (traceroute.ui)

// Ethernet header
gboolean on_checkbutton28_toggled (GtkButton *, SPSData *);  // Specify ethernet header
gboolean on_entry164_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry165_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry166_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry167_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
gboolean on_entry168_activate (GtkWidget *, SPSData *);  // IPv4 header length
gboolean on_entry169_activate (GtkWidget *, SPSData *);  // Protocol version
gboolean on_entry170_activate (GtkWidget *, SPSData *);  // Type-of-service
gboolean on_entry171_activate (GtkWidget *, SPSData *);  // Total length of datagram
gboolean on_entry172_activate (GtkWidget *, SPSData *);  // ID sequence number
gboolean on_entry173_activate (GtkWidget *, SPSData *); // Zero
gboolean on_entry174_activate (GtkWidget *, SPSData *);  // Do not fragment flag
gboolean on_entry175_activate (GtkWidget *, SPSData *);  // More fragments following flag
gboolean on_entry176_activate (GtkWidget *, SPSData *);  // Fragmentation offset
gboolean on_entry177_activate (GtkWidget *, SPSData *);  // Time-to-Live
gboolean on_entry178_activate (GtkWidget *, SPSData *);  // Transport layer protocol
gboolean on_entry179_activate (GtkWidget *, SPSData *);  // Source IPv4 address
gboolean on_entry180_activate (GtkWidget *, SPSData *);  // Destination IPv4 address
gboolean on_entry181_activate (GtkWidget *, SPSData *);  // IPv4 Checksum
// TCP header (IPv4)
gboolean on_entry182_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton30_toggled (GtkButton *, SPSData *);  // Randomize source port
gboolean on_entry183_activate (GtkWidget *, SPSData *);  // Destination port
gboolean on_entry184_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_entry185_activate (GtkWidget *, SPSData *);  // Acknowledgement number
gboolean on_entry186_activate (GtkWidget *, SPSData *);  // Reserved 4 bits
gboolean on_entry187_activate (GtkWidget *, SPSData *);  // Data offset
gboolean on_entry188_activate (GtkWidget *, SPSData *);  // FIN flag
gboolean on_entry189_activate (GtkWidget *, SPSData *);  // SYN flag
gboolean on_entry190_activate (GtkWidget *, SPSData *);  // RST flag
gboolean on_entry191_activate (GtkWidget *, SPSData *);  // PSH flag
gboolean on_entry192_activate (GtkWidget *, SPSData *);  // ACK flag
gboolean on_entry193_activate (GtkWidget *, SPSData *);  // URG flag
gboolean on_entry194_activate (GtkWidget *, SPSData *);  // ECE flag
gboolean on_entry195_activate (GtkWidget *, SPSData *);  // CWR flag
gboolean on_entry196_activate (GtkWidget *, SPSData *);  // Window size
gboolean on_entry197_activate (GtkWidget *, SPSData *);  // TCP checksum
gboolean on_entry198_activate (GtkWidget *, SPSData *);  // Urgent pointer
gboolean on_radiobutton29_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton30_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry199_activate (GtkWidget *, SPSData *); // TCP data from keyboard
gboolean on_button29_clicked (GtkButton *, SPSData *);  // TCP data from file

gboolean on_button28_clicked (GtkButton *, SPSData *);  // View IPv4 TCP packet
gboolean on_button30_clicked (GtkButton *, SPSData *);  // Restore IPv4 TCP default values
gboolean tcp4_tr_default (SPSData *);
gboolean tcp4_tr_show (SPSData *);

// TCP page (IPv6) (traceroute.ui)

// Ethernet header
gboolean on_entry247_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry248_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry249_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry250_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
gboolean on_entry251_activate (GtkWidget *, SPSData *);  // IP version
gboolean on_entry252_activate (GtkWidget *, SPSData *);  // Traffic class
gboolean on_entry253_activate (GtkWidget *, SPSData *);  // Flow label
gboolean on_entry254_activate (GtkWidget *, SPSData *);  // Payload length
gboolean on_entry255_activate (GtkWidget *, SPSData *);  // Next header
gboolean on_entry256_activate (GtkWidget *, SPSData *);  // Hop limit
gboolean on_entry257_activate (GtkWidget *, SPSData *);  // Source IPv6 address
gboolean on_entry258_activate (GtkWidget *, SPSData *);  // Destination IPv6 address
// TCP header (IPv6)
gboolean on_entry259_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton37_toggled (GtkButton *, SPSData *);  // Randomize source port
gboolean on_entry260_activate (GtkWidget *, SPSData *);  // Destination port
gboolean on_entry261_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_entry262_activate (GtkWidget *, SPSData *);  // Acknowledgement number
gboolean on_entry263_activate (GtkWidget *, SPSData *);  // Reserved 4 bits
gboolean on_entry264_activate (GtkWidget *, SPSData *);  // Data offset
gboolean on_entry265_activate (GtkWidget *, SPSData *);  // FIN flag
gboolean on_entry266_activate (GtkWidget *, SPSData *);  // SYN flag
gboolean on_entry267_activate (GtkWidget *, SPSData *);  // RST flag
gboolean on_entry268_activate (GtkWidget *, SPSData *);  // PSH flag
gboolean on_entry269_activate (GtkWidget *, SPSData *);  // ACK flag
gboolean on_entry270_activate (GtkWidget *, SPSData *);  // URG flag
gboolean on_entry271_activate (GtkWidget *, SPSData *);  // ECE flag
gboolean on_entry272_activate (GtkWidget *, SPSData *);  // CWR flag
gboolean on_entry273_activate (GtkWidget *, SPSData *);  // Window size
gboolean on_entry274_activate (GtkWidget *, SPSData *);  // TCP checksum
gboolean on_entry275_activate (GtkWidget *, SPSData *);  // Urgent pointer
gboolean on_radiobutton35_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton36_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry276_activate (GtkWidget *, SPSData *); // TCP data from keyboard
gboolean on_button38_clicked (GtkButton *, SPSData *);  // TCP data from file

gboolean on_button37_clicked (GtkButton *, SPSData *);  // View IPv6 TCP packet
gboolean on_button39_clicked (GtkButton *, SPSData *);  // Restore IPv6 TCP default values
gboolean tcp6_tr_default (SPSData *);
gboolean tcp6_tr_show (SPSData *);

// ICMP page (IPv4) (traceroute.ui)

// Ethernet header
gboolean on_checkbutton31_toggled (GtkButton *, SPSData *);  // Specify ethernet header
gboolean on_entry200_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry201_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry202_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry203_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
gboolean on_entry204_activate (GtkWidget *, SPSData *);  // IPv4 Header Length
gboolean on_entry205_activate (GtkWidget *, SPSData *);  // Protocol Version
gboolean on_entry206_activate (GtkWidget *, SPSData *);  // Type of Service
gboolean on_entry207_activate (GtkWidget *, SPSData *);  // Total Length of Datagram
gboolean on_entry208_activate (GtkWidget *, SPSData *);  // ID Sequence Number
gboolean on_entry209_activate (GtkWidget *, SPSData *); // Zero
gboolean on_entry210_activate (GtkWidget *, SPSData *);  // Do Not Fragment Flag
gboolean on_entry211_activate (GtkWidget *, SPSData *);  // More Fragments Following
gboolean on_entry212_activate (GtkWidget *, SPSData *);  // Fragmentation Offset
gboolean on_entry213_activate (GtkWidget *, SPSData *);  // Time-to-Live
gboolean on_entry214_activate (GtkWidget *, SPSData *);  // Transport Layer Protocol
gboolean on_entry215_activate (GtkWidget *, SPSData *);  // Source IPv4 Address
gboolean on_entry216_activate (GtkWidget *, SPSData *);  // Destination IPv4 Address
gboolean on_entry217_activate (GtkWidget *, SPSData *);  // IPv4 Checksum
// ICMP message (IPv4)
gboolean on_entry218_activate (GtkWidget *, SPSData *);  // Message type
gboolean on_entry219_activate (GtkWidget *, SPSData *);  // Message code
gboolean on_entry220_activate (GtkWidget *, SPSData *);  // ICMP checksum
gboolean on_entry221_activate (GtkWidget *, SPSData *);  // ID
gboolean on_entry222_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_radiobutton31_toggled (GtkButton *, SPSData *);   // ASCII data entry
gboolean on_radiobutton32_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry223_activate (GtkWidget *, SPSData *); // ICMP data from keyboard
gboolean on_button32_clicked (GtkButton *, SPSData *);   // ICMP data from file

gboolean on_button31_clicked (GtkButton *, SPSData *);   // View IPv4 ICMP packet
gboolean on_button33_clicked (GtkButton *, SPSData *);  // Restore IPv4 ICMP default values
gboolean icmp4_tr_default (SPSData *);
gboolean icmp4_tr_show (SPSData *);

// ICMP page (IPv6) (traceroute.ui)

// Ethernet header
gboolean on_entry277_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry278_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry279_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry280_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
gboolean on_entry281_activate (GtkWidget *, SPSData *);  // IP version
gboolean on_entry282_activate (GtkWidget *, SPSData *);  // Traffic class
gboolean on_entry283_activate (GtkWidget *, SPSData *);  // Flow label
gboolean on_entry284_activate (GtkWidget *, SPSData *);  // Payload length
gboolean on_entry285_activate (GtkWidget *, SPSData *);  // Next header
gboolean on_entry286_activate (GtkWidget *, SPSData *);  // Hop limit
gboolean on_entry287_activate (GtkWidget *, SPSData *);  // Source IPv6 address
gboolean on_entry288_activate (GtkWidget *, SPSData *);  // Destination IPv6 address
// ICMP message (IPv6)
gboolean on_entry289_activate (GtkWidget *, SPSData *);  // Message type
gboolean on_entry290_activate (GtkWidget *, SPSData *);  // Message code
gboolean on_entry291_activate (GtkWidget *, SPSData *);  // ICMP checksum
gboolean on_entry292_activate (GtkWidget *, SPSData *);  // ID
gboolean on_entry293_activate (GtkWidget *, SPSData *);  // Sequence number
gboolean on_radiobutton37_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton38_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry294_activate (GtkWidget *, SPSData *);  // ICMP data from keyboard
gboolean on_button41_clicked (GtkButton *, SPSData *);  // ICMP data from file

gboolean on_button40_clicked (GtkButton *, SPSData *);  // View IPv6 ICMP packet
gboolean on_button42_clicked (GtkButton *, SPSData *);  // Restore IPv6 ICMP default values
gboolean icmp6_tr_default (SPSData *);
gboolean icmp6_tr_show (SPSData *);

// UDP page (IPv4) (traceroute.ui)

// Ethernet header
gboolean on_checkbutton33_toggled (GtkButton *, SPSData *);  // Specify ethernet header
gboolean on_entry224_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry225_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry226_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry227_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv4 header
gboolean on_entry228_activate (GtkWidget *, SPSData *);  // IPv4 Header Length
gboolean on_entry229_activate (GtkWidget *, SPSData *);  // Protocol Version
gboolean on_entry230_activate (GtkWidget *, SPSData *);  // Type of Service
gboolean on_entry231_activate (GtkWidget *, SPSData *);  // Total Length of Datagram
gboolean on_entry232_activate (GtkWidget *, SPSData *);  // ID Sequence Number
gboolean on_entry233_activate (GtkWidget *, SPSData *);  // Zero
gboolean on_entry234_activate (GtkWidget *, SPSData *);  // Do Not Fragment Flag
gboolean on_entry235_activate (GtkWidget *, SPSData *);  // More Fragments Following
gboolean on_entry236_activate (GtkWidget *, SPSData *);  // Fragmentation Offset
gboolean on_entry237_activate (GtkWidget *, SPSData *);  // Time-to-Live
gboolean on_entry238_activate (GtkWidget *, SPSData *);  // Transport Layer Protocol
gboolean on_entry239_activate (GtkWidget *, SPSData *);  // Source IPv4 Address
gboolean on_entry240_activate (GtkWidget *, SPSData *);  // Destination IPv4 Address
gboolean on_entry241_activate (GtkWidget *, SPSData *);  // IPv4 Checksum
// UDP header (IPv4)
gboolean on_entry242_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton35_toggled (GtkButton *, SPSData *);  // Randomize Source Port
gboolean on_entry243_activate (GtkWidget *, SPSData *);  // Destination Port
gboolean on_entry244_activate (GtkWidget *, SPSData *);  // Length of UDP Datagram
gboolean on_entry245_activate (GtkWidget *, SPSData *);  // UDP Checksum
gboolean on_radiobutton33_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton34_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry246_activate (GtkWidget *, SPSData *);  // UDP Data from keyboard
gboolean on_button35_clicked (GtkButton *, SPSData *);  // UDP data from file

gboolean on_button34_clicked (GtkButton *, SPSData *);  // View IPv4 UDP packet
gboolean on_button36_clicked (GtkButton *, SPSData *);  // Restore IPv4 UDP default values
gboolean udp4_tr_default (SPSData *);
gboolean udp4_tr_show (SPSData *);

// UDP page (IPv6) (traceroute.ui)

// Ethernet header
gboolean on_entry295_activate (GtkWidget *, SPSData *);  // Destination link-layer (MAC) address
gboolean on_entry296_activate (GtkWidget *, SPSData *);  // Source link-layer (MAC) address
gboolean on_entry297_activate (GtkWidget *, SPSData *);  // Source interface name
gboolean on_entry298_activate (GtkWidget *, SPSData *);  // Ethernet type code
// IPv6 header
gboolean on_entry299_activate (GtkWidget *, SPSData *);  // IP version
gboolean on_entry300_activate (GtkWidget *, SPSData *);  // Traffic class
gboolean on_entry301_activate (GtkWidget *, SPSData *);  // Flow label
gboolean on_entry302_activate (GtkWidget *, SPSData *);  // Payload length
gboolean on_entry303_activate (GtkWidget *, SPSData *);  // Next header
gboolean on_entry304_activate (GtkWidget *, SPSData *);  // Hop limit
gboolean on_entry305_activate (GtkWidget *, SPSData *);  // Source IPv6 address
gboolean on_entry306_activate (GtkWidget *, SPSData *);  // Destination IPv6 address
// UDP header (IPv6)
gboolean on_entry307_activate (GtkWidget *, SPSData *);  // Source Port
gboolean on_checkbutton40_toggled (GtkButton *, SPSData *);  // Randomize Source Port
gboolean on_entry308_activate (GtkWidget *, SPSData *);  // Destination Port
gboolean on_entry309_activate (GtkWidget *, SPSData *);  // Length of UDP Datagram
gboolean on_entry310_activate (GtkWidget *, SPSData *);  // UDP Checksum
gboolean on_radiobutton39_toggled (GtkButton *, SPSData *);  // ASCII data entry
gboolean on_radiobutton40_toggled (GtkButton *, SPSData *);  // Hexadecimal data entry
gboolean on_entry311_activate (GtkWidget *, SPSData *);  // UDP Data from keyboard
gboolean on_button44_clicked (GtkButton *, SPSData *);  // UDP data from file

gboolean on_button43_clicked (GtkButton *, SPSData *);  // View IPv6 UDP packet
gboolean on_button45_clicked (GtkButton *, SPSData *);  // Restore IPv6 UDP default values
gboolean udp6_tr_default (SPSData *);
gboolean udp6_tr_show (SPSData *);

// Send functions for traceroute
gboolean ipv4_tr_send (SPSData *);  // Perform IPv4 traceroute
gboolean free_ipv4_tr_mem (char *, unsigned char *, unsigned char *, char *, char *, gchar *);  // Free allocated memory used in ipv4_tr_send().
gboolean create_tcp4_frame (unsigned char *, int, SPSData *);  // Create IPv4 TCP ethernet frame for traceroute
gboolean create_icmp4_frame (unsigned char *, int, SPSData *);  // Create IPv4 ICMP ethernet frame for traceroute
gboolean create_udp4_frame (unsigned char *, int, SPSData *);  // Create IPv4 UDP ethernet frame for traceroute
gboolean ipv6_tr_send (SPSData *);  // Perform IPv6 traceroute
gboolean free_ipv6_tr_mem (char *, unsigned char *, unsigned char *, char *, char *, gchar *);  // Free allocated memory used in ipv6_tr_send().
gboolean sixto4_tr_send (SPSData *);  // Perform IPv6 over IPv4 (6to4) traceroute
gboolean free_sixto4_tr_mem (char *, unsigned char *, unsigned char *, char *, char *, char *, gchar *);  // Free allocated memory used in sixto4_tr_send().
gboolean create_tcp6_frame (unsigned char *, int, SPSData *);  // Create IPv6 TCP ethernet frame for traceroute
gboolean create_icmp6_frame (unsigned char *, int, SPSData *);  // Create IPv6 ICMP ethernet frame for traceroute
gboolean create_udp6_frame (unsigned char *, int, SPSData *);  // Create IPv6 UDP ethernet frame for traceroute

// FTP functions
gboolean ftp_rir (SPSData *);  // FTP routine to retrieve RIR delegation files.
gboolean free_rir_mem (gchar *, gchar *, gchar *, SPSData *);
gboolean download_all (SPSData *);  // Download all RIR delegation files and country code file.
int cntl_sock (gchar *, gchar *, SPSData *);  // Request a socket descriptor for FTP control socket.
int data_sock (gchar *, gint, SPSData *);  // Request a socket descriptor for FTP data socket.
int cntl_send (int, gchar *, SPSData *);  // Send a command to FTP control socket.
int cntl_recv (int, SPSData *);  // Receive FTP command responses from remote host on the control socket.
int parse_ip4_port (gchar *, int *, SPSData *);  // Extract IPv4 address and port number for creating FTP data socket in passive mode.
int parse_ip6_port (int *, SPSData *);  // Extract port number for creating IPv6 FTP data socket in passive mode.
int data_recv (SPSData *);  // Receive and save data on the FTP data socket.
gboolean on_button50_clicked (GtkButton *, SPSData *);  // Download ARIN delegation file.
gboolean on_button51_clicked (GtkButton *, SPSData *);  // Download RIPE delegation file.
gboolean on_button52_clicked (GtkButton *, SPSData *);  // Download AFRINIC delegation file.
gboolean on_button53_clicked (GtkButton *, SPSData *);  // Download APNIC delegation file.
gboolean on_button54_clicked (GtkButton *, SPSData *);  // Download LACNIC delegation file.
gboolean on_button55_clicked (GtkButton *, SPSData *);  // Download www.iso.org's ISO 3166-1 country code list.
gboolean on_button56_clicked (GtkButton *, SPSData *);  // View www.iso.org's ISO 3166-1 country code list.
gboolean on_button57_clicked (GtkButton *, SPSData *);  // Download all RIR files and www.iso.org's country code list.
gboolean on_entry500_activate (GtkWidget *, SPSData *); // Country name (for country name/code search).
gboolean on_entry501_activate (GtkWidget *, SPSData *); // Country code (for country name/code search).
gboolean on_entry502_activate (GtkWidget *, SPSData *); // Search for autonomous system (AS) number delegations by country name.
gboolean on_entry503_activate (GtkWidget *, SPSData *); // Search for autonomous system (AS) number delegations by country code.
gboolean on_entry504_activate (GtkWidget *, SPSData *); // Reverse search for country by autonomous system (AS) number delegations.
gboolean on_entry505_activate (GtkWidget *, SPSData *); // Search for IP delegations by country name.
gboolean on_entry506_activate (GtkWidget *, SPSData *); // Search for IP delegations by country code.

// HTTP functions
gboolean get_countries (SPSData *);  // Retrieve www.iso.org's ISO 3166-1 country code list
int http_sock (gchar *, gchar *, const gchar *, SPSData *);  // Request socket descriptor for HTTP connection
int send_get (int, gchar *, SPSData *);  // Send HTTP GET
int http_recv (int, gchar *, SPSData *);  // Receive and save data resulting from HTTP GET

gboolean on_checkbutton38_toggled (GtkButton *button, SPSData *);  // Change state of flag to show sent and received messages during file downloads
gboolean rir_loading (SPSData *);  // Change RIR delegation file status to "Loading".
gboolean rir_not_available (SPSData *);  // Change RIR delegation file status to "Not Available" because an error occurred while downloading.
gboolean rir_available (SPSData *);  // Update RIR delegation file status.
gboolean cc_loading (SPSData *);  // Change country code file status to "Loading".
gboolean cc_not_available (SPSData *);  // Change country code file status to "Not Available" because an error occurred while downloading.
gboolean cc_available (SPSData *);  // Update country code file status.

// Functions for reading and searching ASN / IP delegation files and www.iso.org's ISO 3166-1 country code list.
gboolean read_delegations_and_countries (SPSData *);
int grab_line (FILE *, unsigned char *);

gboolean find_country_code (gchar *, gchar *, SPSData *);
gboolean find_country_name (gchar *, gchar *, SPSData *);

gboolean search_asn_by_ccode (SPSData *);
gboolean post_asnsearch_results (gchar *, SPSData *);
gboolean asn_results (SPSData *);

gboolean rev_search_asn (SPSData *);
gboolean post_asnrevsearch_results (gchar *, SPSData *);
gboolean revasn_results (SPSData *);

gboolean search_ipv4_by_ccode (SPSData *);
gboolean post_ip4search_results (gchar *, SPSData *);
gboolean ip4_results (SPSData *);

gboolean rev_search_ipv4 (SPSData *);
gboolean post_ip4revsearch_results (gchar *, SPSData *);
gboolean revip4_results (SPSData *);

gboolean search_ipv6_by_ccode (SPSData *);
gboolean post_ip6search_results (gchar *, SPSData *);
gboolean ip6_results (SPSData *);

gboolean rev_search_ipv6 (SPSData *);
gboolean post_ip6revsearch_results (gchar *, SPSData *);
gboolean revip6_results (SPSData *);

int parse_ipv4 (char *, unsigned int, char *);
int parse_ipv6 (char *, unsigned int *);
int interval_test_ipv4 (char *, char *, unsigned int);
int interval_test_ipv6 (char *, char *, unsigned int);
int compare (unsigned char *, unsigned char *);
int carry (int, int, int *, int *);  // Recursive carry function for numerical subtraction.

// Checksum routines
unsigned short int checksum (unsigned short int *, int);
unsigned short int tcp4_checksum (struct ip, struct tcphdr, unsigned char *, int);
unsigned short int tcp6_checksum (struct ip6_hdr, struct tcphdr, unsigned char *, int);
unsigned short int icmp4_checksum (struct icmp, unsigned char *, int);
unsigned short int icmp6_checksum (struct ip6_hdr, struct icmp6_hdr, unsigned char *, int);
unsigned short int udp4_checksum (struct ip, struct udphdr, unsigned char *, int);
unsigned short int udp6_checksum (struct ip6_hdr, struct udphdr, unsigned char *, int);

// Miscellaneous function prototypes
unsigned char bin8_to_dec (unsigned char *);
gushort ran_val1 ();
gushort ran_val2 ();
gchar *ran4_addr (gchar *);
gchar *ran6_addr (gchar *);
gboolean this_host (SPSData *);
gchar *thishost_ipv4 (char *);
gchar *thishost_ipv6 (char *);
gboolean show_packet (unsigned char *, gint, gchar *, gint, gint);
gboolean is_valid_ip4 (const gchar *);
gboolean report_warning (SPSData *);
gboolean report_error (SPSData *);
gboolean arp (unsigned char *, SPSData *);
gboolean neighbor_discovery (unsigned char *, SPSData *);
gboolean mac_bin2string (unsigned char *, char *);
gboolean mac_string2bin (char *, unsigned char *);
gboolean is_valid_hexdata (const gchar *, SPSData *);
gboolean is_valid_hex (gchar);
gboolean on_traceroute_window_destroy (GtkWidget *, SPSData *);
gint file_exists (gchar *);
gint files_available (SPSData *);
void *get_in_addr (struct sockaddr *);
gboolean thread_listen (SPSData *);  // Timeout function to create dialog pop-up for messages/errors from file download or traceroute threads.

// Menubar functions
gboolean on_about_requested ();
gboolean on_documentation_activate ();
